/*ScianAnimation.c
  Eric Pepke
  21 May 1993

  Advanced keyframe animation for SciAn
*/

#include "Scian.h"
#include "ScianTypes.h"
#include "ScianIDs.h"
#include "ScianNames.h"
#include "ScianWindows.h"
#include "ScianObjWindows.h"
#include "ScianDialogs.h"
#include "ScianScripts.h"
#include "ScianErrors.h"
#include "ScianSymbols.h"
#include "ScianEvents.h"
#include "ScianLists.h"
#include "ScianArrays.h"
#include "ScianControls.h"
#include "ScianButtons.h"
#include "ScianTextBoxes.h"
#include "ScianTitleBoxes.h"
#include "ScianIcons.h"
#include "ScianStyle.h"
#include "ScianObjFunctions.h"
#include "ScianColors.h"
#include "ScianObjFunctions.h"
#include "ScianTemplates.h"
#include "ScianTemplateHelper.h"
#include "ScianSnap.h"
#include "ScianAnimation.h"
#include "ScianTrackControls.h"
#include "ScianDraw.h"
#include "ScianMethods.h"
#include "ScianFixedClasses.h"
#include "ScianProjects.h"
#include "ScianWindowFunctions.h"
#include "ScianVisWindows.h"
#include "ScianRecorders.h"

#define EqualTime(x, y)  (ABS((x) - (y)) < 0.01)

#define NFASTFRAMES	3	/*Frame step for fast motion*/

#define ANIMBUTTONWIDTH		SMALLICONBUTTONSIZE
#define ANIMBUTTONHEIGHT	SMALLICONBUTTONSIZE

#define P2T(p)	((((p) - l) * (maxTime - minTime) / ((real) r - l)) + minTime)
#define T2P(t)	floor(((((t) - minTime) * ((real) r - l) / (maxTime - minTime)) + l) + 0.5)

ObjPtr sequenceClass;		/*Class of a sequence*/
ObjPtr trackClass;		/*Class of a track within a sequence*/
ObjPtr subTrackClass;		/*Class of a subtrack within a track*/
ObjPtr keyFrameClass;		/*Class of a keyframe within a subtrack*/

char **betterVarNames = 0;	/*Better variable names*/
int nBetterVarNames = 0;	/*Number of better var names*/

#ifdef HAVE_PROTOTYPES
static ObjPtr NewTrack(ObjPtr class, ObjPtr repObj)
#else
static ObjPtr NewTrack(class, repObj)
ObjPtr class;
ObjPtr repObj;
#endif
/*Creates a new track to control repObj*/
{
    ObjPtr retVal;

    retVal = NewObject(class ? class : trackClass, 0L);
    SetVar(retVal, REPOBJ, repObj);
    MakeVar(repObj, NAME);
    SetVar(retVal, NAME, GetVar(repObj, NAME));
    return retVal;
}

#ifdef HAVE_PROTOTYPES
static ObjPtr NewSubTrack(ObjPtr class, ObjPtr repObj, char *name)
#else
static ObjPtr NewSubTrack(class, repObj, name)
ObjPtr class;
ObjPtr repObj;
char *name;
#endif
/*Creates a new subtrack to control repObj's variable group named name*/
{
    ObjPtr retVal;

    retVal = NewObject(class ? class : subTrackClass, 0L);
    SetVar(retVal, REPOBJ, repObj);
    SetVar(retVal, NAME, NewString(name));
    return retVal;
}

#ifdef HAVE_PROTOTYPES
static ObjPtr NewKeyFrame(real time, ObjPtr snapshot)
#else
static ObjPtr NewKeyFrame(time, snapshot)
real time;
ObjPtr snapshot;
#endif
/*Makes a new keyframe at time using snapshot.*/
{
    ObjPtr retVal;

    retVal = NewObject(keyFrameClass, 0L);
    SetVar(retVal, TIME, NewReal(time));
    SetVar(retVal, SNAPSHOT, snapshot);

    return retVal;
}

ObjPtr DeleteKeyFrame(keyFrame)
ObjPtr keyFrame;
/*Deletes a key frame from the current window.*/
{
    ObjPtr track;
    int index;
    int k;
    ObjPtr *elements;
    ObjPtr keyFrames;
    ObjPtr var;
    long nKeyFrames;

    track = GetVar(keyFrame, PARENT);

    keyFrames = GetVar(track, KEYFRAMES);
    if (!keyFrames) return ObjFalse;

    elements = ELEMENTS(keyFrames);
    nKeyFrames = DIMS(keyFrames)[0];

    index = -1;
    for (k = 0; k < nKeyFrames; ++k)
    {
	if (elements[k] == keyFrame) index = k;
    }

    if (index >= 0)
    {
	SaveVarSnapshotForUndo(track, KEYFRAMES);
	Select(keyFrame, false);
	var = DeleteFromArray(keyFrames, index);
	SetVar(track, KEYFRAMES, var);
	ImInvalid(track);
    }
    else
    {
	ReportError("DeleteKeyframe", "Internal error--key frame not in array");
	return ObjFalse;
    }
    return ObjFalse;
}

ObjPtr DeleteTrack(track)
ObjPtr track;
/*Deletes a track from the current window.*/
{
    ObjPtr sequence;
    int index;
    int k;
    ObjPtr *elements;
    ObjPtr tracks;
    ObjPtr var;
    long nTracks;

    if (!selWinInfo) return ObjFalse;

    sequence = GetVar((ObjPtr) track, PARENT);

    tracks = GetVar(sequence, TRACKS);
    if (!tracks) return ObjFalse;

    elements = ELEMENTS(tracks);
    nTracks = DIMS(tracks)[0];

    index = -1;
    for (k = 0; k < nTracks; ++k)
    {
	if (elements[k] == track) index = k;
    }

    if (index >= 0)
    {
	SaveVarSnapshotForUndo(sequence, TRACKS);
	Select(track, false);
	var = DeleteFromArray(tracks, index);
	SetVar(sequence, TRACKS, var);
	ImInvalid(sequence);
    }
    else
    {
	ReportError("DeleteTrack", "Internal error--key frame not in array");
	return ObjFalse;
    }
    return ObjFalse;
}

ObjPtr DeleteSubTrack(track)
ObjPtr track;
/*Deletes a track from the current window.*/
{
    ObjPtr superTrack;
    int index;
    int k;
    ObjPtr *elements;
    ObjPtr tracks;
    ObjPtr var;
    long nTracks;

    if (!selWinInfo) return ObjFalse;

    superTrack = GetVar((ObjPtr) track, PARENT);

    tracks = GetVar(superTrack, SUBTRACKS);
    if (!tracks) return ObjFalse;

    elements = ELEMENTS(tracks);
    nTracks = DIMS(tracks)[0];

    index = -1;
    for (k = 0; k < nTracks; ++k)
    {
	if (elements[k] == track) index = k;
    }

    if (index >= 0)
    {
	SaveVarSnapshotForUndo(superTrack, SUBTRACKS);
	Select(track, false);
	var = DeleteFromArray(tracks, index);
	SetVar(superTrack, SUBTRACKS, var);
	ImInvalid(superTrack);
	var = GetVar(superTrack, PARENT);
	ImInvalid(var);
    }
    else
    {
	ReportError("DeleteSubTrack", "Internal error--key frame not in array");
	return ObjFalse;
    }
    return ObjFalse;
}

#ifdef HAVE_PROTOTYPES
void InsertKeySnapshot(ObjPtr track, real time, ObjPtr snapshot)
#else
void InsertKeySnapshot(track, time, snapshot)
ObjPtr track, snapshot;
real time;
#endif
/*Inserts a key snapshot at time into track, or just updates 
  what's there.*/
{
    ObjPtr keyframes;
    ThingListPtr runner;
    long index;
    long dim;
    ObjPtr *elements;
    ObjPtr var;

    keyframes = GetVar(track, KEYFRAMES);
    if (!keyframes)
    {
	dim = 1;
	keyframes = NewArray(AT_OBJECT, 1, &dim);
	index = 0;

	elements = ELEMENTS(keyframes);
	elements[0] = NewKeyFrame(time, snapshot);
	SetVar(elements[0], PARENT, track);
	SetVar(track, KEYFRAMES, keyframes);
    }
    else
    {
	/*Search to see if there is already a keyframe there*/
	index = SearchRealVar(keyframes, TIME, time);

	elements = ELEMENTS(keyframes);

	if (index > 0)
	{
	    /*See if the element just before is the one*/
	    MakeVar(elements[index - 1], TIME);
	    var = GetVar(elements[index - 1], TIME);
	    if (var && IsReal(var) && EqualTime(GetReal(var), time))
	    {
		/*Found it!  Just change it's snapshot, touch the keyframe, 
		  and return*/

		SetVar(elements[index - 1], SNAPSHOT, snapshot);
		SetVar(track, KEYFRAMES, keyframes);
		return;
	    }
	}

	if (index < DIMS(keyframes)[0])
	{
	    /*See if the element here is the one*/
	    MakeVar(elements[index], TIME);
	    var = GetVar(elements[index], TIME);
	    if (var && IsReal(var) && EqualTime(GetReal(var), time))
	    {
		/*Found it!  Just change it's snapshot, touch the keyframe, 
		  and return*/

		SetVar(elements[index], SNAPSHOT, snapshot);
		SetVar(track, KEYFRAMES, keyframes);
		return;
	    }
	}

	/*Insert keyframe where it belongs*/
	var = NewKeyFrame(time, snapshot);
	SetVar(var, PARENT, track);
	keyframes = InsertInArray(keyframes,
		var, index);
	SetVar(track, KEYFRAMES, keyframes);
    }
}

#ifdef HAVE_PROTOTYPES
ObjPtr TrackObjectWithinSequence(ObjPtr sequence, ObjPtr object, ObjPtr trackClass)
#else
ObjPtr TrackObjectWithinSequence(sequence, object, trackClass)
ObjPtr sequence;
ObjPtr object;
ObjPtr trackClass;
#endif
/*Returns the track for object within sequence or adds one*/
{
    ObjPtr tracks, track, var;
    ObjPtr *elements;
    long k, dim;

    /*First find out if there is a track with that object*/
    tracks = GetVar(sequence, TRACKS);
    if (tracks)
    {
	for (k = 0; k < DIMS(tracks)[0]; ++k)
	{
	    var = GetObjectElement(tracks, &k);
	    if (GetVar(var, REPOBJ) == object)
	    {
		return var;
	    }
	}
    }

    /*Fell through, object not found.  Create a new track*/
    track = NewTrack(trackClass, object);
    SetVar(trackClass, PARENT, sequence);

    if (tracks)
    {
	/*Insert track in sorted place*/
	MakeVar(object, NAME);
	var = GetStringVar("TrackObjectWithinSequence", object, NAME);
	if (!var) return NULLOBJ;
	k = SearchStringVar(tracks, NAME, GetString(var));
	if (k < 0) return NULLOBJ;
	tracks = InsertInArray(tracks, track, k);
    }
    else
    {
	/*Create single array to hold tracks*/
	dim = 1;
	tracks = NewArray(AT_OBJECT, 1, &dim);
	elements = (ObjPtr *) ELEMENTS(tracks);
	elements[0] = track;
    }
    SetVar(track, PARENT, sequence);

    SetVar(sequence, TRACKS, tracks);
    return track;
}

#ifdef HAVE_PROTOTYPES
ObjPtr GetTrackedObjectWithinSequence(ObjPtr sequence, ObjPtr object)
#else
ObjPtr GetTrackedObjectWithinSequence(sequence, object)
ObjPtr sequence;
ObjPtr object;
#endif
/*Returns the track for object within sequence or returns NULLOBJ*/
{
    ObjPtr tracks, track, var;
    ObjPtr *elements;
    long k, dim;

    /*First find out if there is a track with that object*/
    tracks = GetVar(sequence, TRACKS);
    if (tracks)
    {
	for (k = 0; k < DIMS(tracks)[0]; ++k)
	{
	    var = GetObjectElement(tracks, &k);
	    if (GetVar(var, REPOBJ) == object)
	    {
		return var;
	    }
	}
    }

    /*Fell through, object not found.*/
    return NULLOBJ;
}

#ifdef HAVE_PROTOTYPES
ObjPtr TrackVarGroupWithinSequence(ObjPtr sequence, ObjPtr object, char *group, ObjPtr trackClass)
#else
ObjPtr TrackVarGroupWithinSequence(sequence, object, group, trackClass)
ObjPtr sequence;
ObjPtr object;
char *group;
ObjPtr trackClass;
#endif
/*Returns the track for vargroup group of object within sequence
  or adds one*/
{
    ObjPtr track, var, name;
    ObjPtr subTracks, subTrack;
    ObjPtr *elements;
    long k, dim;

    track = TrackObjectWithinSequence(sequence, object, NULLOBJ);

    /*First find out if there is a subtrack for that group*/
    
    subTracks = GetVar(track, SUBTRACKS);
    if (subTracks)
    {
	for (k = 0; k < DIMS(subTracks)[0]; ++k)
	{
	    var = GetObjectElement(subTracks, &k);
	    MakeVar(var, NAME);
	    name = GetVar(var, NAME);
	    if (name && (0 == strcmp2(GetString(name), group)))
	    {
		return var;
	    }
	}
    }

    /*Fell through, object not found.  Create a new track*/
    subTrack = NewSubTrack(trackClass, object, group);

    if (IsSelected(track)) Select(subTrack, true);

    if (subTracks)
    {
	/*Insert track in sorted place*/
	k = SearchStringVar(subTracks, NAME, group);
	if (k < 0) return NULLOBJ;
	subTracks = InsertInArray(subTracks, subTrack, k);
    }
    else
    {
	/*Create single array to hold tracks*/
	dim = 1;
	subTracks = NewArray(AT_OBJECT, 1, &dim);
	elements = (ObjPtr *) ELEMENTS(subTracks);
	elements[0] = subTrack;
    }
    SetVar(subTrack, PARENT, track);
    SetVar(track, SUBTRACKS, subTracks);

    /*Also need to touch TRACKS of sequence*/
    SetVar(sequence, TRACKS, GetVar(sequence, TRACKS));
    return subTrack;
}

#ifdef HAVE_PROTOTYPES
ObjPtr GetTrackedVarGroupWithinSequence(ObjPtr sequence, ObjPtr object, char *group)
#else
ObjPtr GetTrackedVarGroupWithinSequence(sequence, object, group)
ObjPtr sequence;
ObjPtr object;
char *group;
#endif
/*Returns the track for vargroup group of object within sequence
  or NULLOBJ*/
{
    ObjPtr track, var, name;
    ObjPtr subTracks, subTrack;
    ObjPtr *elements;
    long k, dim;

    track = GetTrackedObjectWithinSequence(sequence, object);
    if (!track) return NULLOBJ;

    /*First find out if there is a subtrack for that group*/
    
    subTracks = GetVar(track, SUBTRACKS);
    if (subTracks)
    {
	for (k = 0; k < DIMS(subTracks)[0]; ++k)
	{
	    var = GetObjectElement(subTracks, &k);
	    MakeVar(var, NAME);
	    name = GetVar(var, NAME);
	    if (name && (0 == strcmp2(GetString(name), group)))
	    {
		return var;
	    }
	}
    }

    /*Fell through, object not found.*/
    return NULLOBJ;
}

#if 1
static void toado(ObjPtr sequence, char *name)
{
    ObjPtr temp;
    temp = NewObject(NULLOBJ, 0L);
    SetVar(temp, NAME, NewString(name));
    TrackVarGroupWithinSequence(sequence, temp, "Guano", NULLOBJ);
    TrackVarGroupWithinSequence(sequence, temp, "Bat", NULLOBJ);
}
#endif

#ifdef HAVE_PROTOTYPES
ObjPtr NewSequence(char *name)
#else
ObjPtr NewSequence(name)
char *name;
#endif
/*Returns a new sequence with name, also shows controls*/
{
    ObjPtr retVal;

    /*UPDATE log command to make sequence*/
    retVal = NewObject(sequenceClass, 0L);
    SetVar(retVal, NAME, NewString(name));

    if (!CurProject()) NewProject();

    AddSequenceToProject(retVal);

#if 0
    /*DEBUG give it some tracks by default*/
    {
	toado(retVal, "And");
	toado(retVal, "Did");
	toado(retVal, "Those");
	toado(retVal, "Teeth");
	toado(retVal, "In");
	toado(retVal, "Ancient");
	toado(retVal, "Times");
	toado(retVal, "Walk");
	toado(retVal, "Upon");
	toado(retVal, "England's");
	toado(retVal, "Mountains");
	toado(retVal, "Green?");
    }
#endif

#if 0
    DeferMessage(retVal, SHOWCONTROLS);
#endif

    return retVal;
}

int seqSerialNum = 1;

#ifdef HAVE_PROTOTYPES
void DoNewSequence(void)
#else
void DoNewSequence()
#endif
/*Asks the user for a name and makes the new sequence*/
{
    sprintf(tempStr, "Sequence %d", seqSerialNum++);
    NewSequence(tempStr);
}

#ifdef HAVE_PROTOTYPES
void ExpressTrack(ObjPtr track, real time)
#else
void ExpressTrack(track, time)
ObjPtr track;
real time;
#endif
/*Expresses a track by adjusting all the variables it controls to time*/
{
    ObjPtr var;
    ObjPtr *elements;
    long k;
    FuncTyp method;

    
    var = GetVar(track, SUBTRACKS);
    if (var)
    {
	elements = ELEMENTS(var);
	for (k = 0; k < DIMS(var)[0]; ++k)
	{
	    method = GetMethod(elements[k], EXPRESS);
	    if (method)
	    {
		(*method)(elements[k], time);
	    }
	}
    }
}

static ObjPtr ExpressDefaultSubTrack(track, time)
ObjPtr track;
real time;
/*Default method to express a variable group subtrack at time*/
{
    ObjPtr *elements;
    long index;
    ObjPtr keyFrames;
    ObjPtr var;
    ObjPtr frame1, frame2;	/*Frames for interpolation*/
    real weight;
    real n1, n2;
    ObjPtr newSnap;

    /*Get the keyframes*/
    keyFrames = GetVar(track, KEYFRAMES);
    if (!keyFrames) return ObjFalse;

    /*Find the index for this frame*/
    elements = ELEMENTS(keyFrames);
    index = SearchRealVar(keyFrames, TIME, time);
    if (index <= 0)
    {
	frame1 = elements[0];
	frame2 = 0;
    }
    else if (index >= DIMS(keyFrames)[0])
    {
	frame1 = elements[DIMS(keyFrames)[0] - 1];
	frame2 = 0;
    }
    else
    {
	frame1 = elements[index - 1];
	frame2 = elements[index];
    }

    if (frame2)
    {
	/*Calculate weight*/
	var = GetVar(frame1, TIME);
	if (!var) return ObjFalse;
	n1 = GetReal(var);
	var = GetVar(frame2, TIME);
	if (!var) return ObjFalse;
	n2 = GetReal(var);
	weight = (time - n1) / (n2 - n1);
    
	newSnap = InterpSnapshots(GetVar(frame1, SNAPSHOT),
			    GetVar(frame2, SNAPSHOT), weight);
	ApplySnapshot(newSnap);
    }
    else
    {
	ApplySnapshot(GetVar(frame1, SNAPSHOT));
    }

    return ObjTrue;
}

#ifdef HAVE_PROTOTYPES
void ExpressSequence(ObjPtr sequence)
#else
void ExpressSequence(sequence)
ObjPtr sequence;
#endif
/*Expresses a sequence by adjusting all the variables it controls to 
  the current time*/
{
    ObjPtr var;
    real time;
    ObjPtr *elements;
    long k;

    var = GetRealVar("ExpressSequence", sequence, TIME);
    if (!var) return;
    time = GetReal(var);

    InhibitLogging(true);
    var = GetVar(sequence, TRACKS);
    if (var)
    {
	elements = ELEMENTS(var);
	for (k = 0; k < DIMS(var)[0]; ++k)
	{
	    ExpressTrack(elements[k], time);
	}
    }
    InhibitLogging(false);
}

static ObjPtr ChangeSequenceTime(control)
ObjPtr control;
/*Changes the time in a sequence in response to control*/
{
    ObjPtr repObj;

    repObj = GetObjectVar("ChangedSequenceTime", control, REPOBJ);
    if (repObj)
    {
	SetVar(repObj, TIME, GetValue(control));
	if (GetPredicate(repObj, ENABLEEXPRESSION))
	{
	    ExpressSequence(repObj);
	}
    }

    return ObjTrue;
}

static ObjPtr MakeAnimationControlHiValue(object)
ObjPtr object;
/*Makes an animation control's hi value from it's repobj's duration*/
{
    ObjPtr repObj;

    repObj = GetVar(object, REPOBJ);
    if (repObj)
    {
	SetVar(object, HIVALUE, GetVar(repObj, DURATION));
    }

    return ObjTrue;
}

static ObjPtr MakeAnimationControlValue(control)
ObjPtr control;
/*Makes the value of an animation control*/
{
    ObjPtr repObj;

    repObj = GetVar(control, REPOBJ);
    if (repObj)
    {
	SetVar(control, VALUE, GetVar(repObj, TIME));
	AutoScroll(control);
	return ObjTrue;
    }
    else
    {
	SetVar(control, VALUE, NewReal(0.0));
	AutoScroll(control);
	return ObjTrue;
    }
}

static ObjPtr ShowSequenceControls(object, windowName)
ObjPtr object;
char *windowName;
/*Makes a new control window to control sequence object*/
{
    WinInfoPtr controlWindow;
    ObjPtr var;
    ObjPtr panel;
    ObjPtr contents;
    ObjPtr control, trackControl;
    ObjPtr radio, button, textBox;
    int left, right, bottom, top, width;
    WinInfoPtr dialogExists;
    int l, r, b, t;
    real duration;

    GetTemplateBounds(SequenceTemplate, "Panel", &left, &right, &bottom, &top);

    dialogExists = DialogExists((WinInfoPtr) object, NewString("Controls"));
    controlWindow = GetDialog((WinInfoPtr) object, NewString("Controls"), windowName, 
	(right - left), (top - bottom), 1280, 1024, WINUI);
    
    if (!dialogExists)
    {
	SetVar((ObjPtr) controlWindow, REPOBJ, object);

	/*Put in a help string*/
	SetVar((ObjPtr) controlWindow, HELPSTRING,
	    NewString("This window shows controls for a an animation sequence.  Using the controls in this \
window, you can define keyframes and produce mini-movies."));

	/*Add in a panel*/
	panel = TemplatePanel(SequenceTemplate, "Panel");
	if (!panel)
	{
	    return ObjFalse;
	}
	contents = GetVar((ObjPtr) controlWindow, CONTENTS);
	PrefixList(contents, panel);
	SetVar(panel, PARENT, (ObjPtr) controlWindow);

	contents = GetVar(panel, CONTENTS);

	/*Make a new track control to hold the tracks*/
	var = GetVar(object, DURATION);
	if (var)
	{
	    duration = GetReal(var);
	}
	else
	{
	    duration = 10.0;
	}
	trackControl = TemplateTrackControl(SequenceTemplate, "Animation Tracks",
				  0.0, duration);
	SetVar(trackControl, PARENT, panel);
	SetVar(trackControl, REPOBJ, object);
	PrefixList(contents, trackControl);
	SetVar(trackControl, HELPSTRING,
	    NewString("This control allows you to do lots of neat stuff with \
animation."));
	SetVar(trackControl, STICKINESS, NewInt(STICKYLEFT + STICKYRIGHT + STICKYBOTTOM + STICKYTOP));
	SetMethod(trackControl, CHANGEDVALUE, ChangeSequenceTime);
	DeclareIndirectDependency(trackControl, HIVALUE, REPOBJ, DURATION);
	SetMethod(trackControl, HIVALUE, MakeAnimationControlHiValue);
	DeclareIndirectDependency(trackControl, VALUE, REPOBJ, TIME);
	SetVar(trackControl, WHICHVAR, NewSymbol(TIME));
	SetMethod(trackControl, VALUE, MakeAnimationControlValue);
	SetVar(trackControl, INHIBITRECORDING, ObjTrue);

	/*Create the auto edit check box*/
	control = TemplateCheckBox(SequenceTemplate, "Enable Auto Edit", false);
	if (control)
	{
	    SetVar(control, PARENT, panel);
	    PrefixList(contents, control);
	    SetVar(control, HELPSTRING,
		NewString("When this check box is on, anything you do to any window or visualization \
object will automatically add new keyframes at the time given by the sequence cursor.  This is \
the main way that keyframes are placed into an animation.  With this check box on, move the time \
cursor to the time you wish to change.  Then change the value of any control in any window.  The changed \
will be updated automatically in the animation control."));
	    SetVar(control, INHIBITRECORDING, ObjTrue);
	    AssocDirectControlWithVar(control, object, RECORDING);
	    SetVar(control, STICKINESS, NewInt(STICKYLEFT + STICKYBOTTOM));
	}

	/*Create the auto express check box*/
	control = TemplateCheckBox(SequenceTemplate, "Enable Playback", false);
	if (control)
	{
	    SetVar(control, PARENT, panel);
	    PrefixList(contents, control);
	    SetVar(control, HELPSTRING,
		NewString("When this check box is on, whenever you move the sequence cursor, \
the current state of the animation will be played back into all objects.  The control is normally on, \
so that you can see the effects of what you do on the screen.  However, it is sometimes \
useful to turn it off temporarily in order to make a batch of changes to the animation control \
without having to wait for the windows to be redrawn."));
	    SetVar(control, INHIBITRECORDING, ObjTrue);
	    AssocDirectControlWithVar(control, object, ENABLEEXPRESSION);
	    SetVar(control, STICKINESS, NewInt(STICKYLEFT + STICKYBOTTOM));
	}

	/*Create the constrain check box*/
	control = TemplateCheckBox(SequenceTemplate, "Constrain to Keyframes", false);
	if (control)
	{
	    SetVar(control, PARENT, panel);
	    PrefixList(contents, control);
	    SetVar(control, HELPSTRING,
		NewString("When this check box is on, actions within the sequence control \
will be constrained to lie on keyframes.  When the check box is off, no constrainst will be \
applied."));
	    SetVar(control, INHIBITRECORDING, ObjTrue);
	    AssocDirectControlWithVar(control, object, CONSTRAINTOKEY);
	    SetVar(control, STICKINESS, NewInt(STICKYLEFT + STICKYBOTTOM));
	}

	/*Create the frame rate boxes*/
	textBox = TemplateTextBox(SequenceTemplate, "Frame Rate Text", 0, "Frame Rate:");
	SetVar(textBox, PARENT, panel);
	SetTextAlign(textBox, RIGHTALIGN);
	PrefixList(contents, textBox);

	textBox = TemplateTextBox(SequenceTemplate, "Frame Rate", EDITABLE + WITH_PIT + ONE_LINE, "");
	SetVar(textBox, PARENT, panel);
	SetTextAlign(textBox, RIGHTALIGN);
	PrefixList(contents, textBox);
	SetVar(textBox, INHIBITRECORDING, ObjTrue);
	AssocTextRealControlWithVar(textBox, object, FRAMERATE, 0.0, 1000.0, TR_NE_BOTTOM);

	textBox = TemplateTextBox(SequenceTemplate, "fps", 0, "fps");
	SetVar(textBox, PARENT, panel);
	PrefixList(contents, textBox);

	/*Create the duration boxes*/
	textBox = TemplateTextBox(SequenceTemplate, "Duration Text", 0, "Duration:");
	SetVar(textBox, PARENT, panel);
	SetTextAlign(textBox, RIGHTALIGN);
	PrefixList(contents, textBox);

	textBox = TemplateTextBox(SequenceTemplate, "Duration", EDITABLE + WITH_PIT + ONE_LINE, "");
	SetVar(textBox, PARENT, panel);
	SetTextAlign(textBox, RIGHTALIGN);
	PrefixList(contents, textBox);
	SetVar(textBox, INHIBITRECORDING, ObjTrue);
	AssocTextRealControlWithVar(textBox, object, DURATION, 0.0, 100000.0, TR_NE_BOTTOM);

	textBox = TemplateTextBox(SequenceTemplate, "seconds", 0, "seconds");
	SetVar(textBox, PARENT, panel);
	PrefixList(contents, textBox);

	/*Create the icon buttons for the animation control*/
	radio = NewRadioButtonGroup("Animation Control");
	SetVar(radio, INHIBITRECORDING, ObjTrue);
	SetVar(radio, PARENT, panel);
	SetVar(radio, REPOBJ, object);
	SetVar(radio, HELPSTRING,
		NewString("These radio buttons control the recording and playback \
of this animation sequence."));
	PrefixList(contents, radio); 
	SetVar(radio, STICKINESS, NewInt(STICKYLEFT + STICKYBOTTOM));

	b = MINORBORDER;
	l = MINORBORDER;
	t = b + ANIMBUTTONHEIGHT;

	r = l + ANIMBUTTONWIDTH;
	button = NewIconButton(l, r, b, t,
			ICONREWIND, UIGREEN, "Rewind", BS_PLAIN);
	AddRadioButton(radio, button);
	SetVar(button, STICKINESS, NewInt(STICKYLEFT + STICKYBOTTOM));
	SetVar(button, HELPSTRING,
	    NewString("Pressing this button rewinds the sequence to the very beginning."));
	l = r;
	
	r = l + ANIMBUTTONWIDTH;
	button = NewIconButton(l, r, b, t,
			ICONFR, UIGREEN, "Fast\nReverse", BS_PLAIN);
	AddRadioButton(radio, button);
	SetVar(button, HELPSTRING,
	    NewString("Pressing this button plays the sequence in the reverse \
direction at three times normal speed."));
	SetVar(button, STICKINESS, NewInt(STICKYLEFT + STICKYBOTTOM));
	l = r;
	
	r = l + ANIMBUTTONWIDTH;
	button = NewIconButton(l, r, b, t,
			ICONREV, UIGREEN, "Reverse", BS_PLAIN);
	AddRadioButton(radio, button);
	SetVar(button, HELPSTRING,
	    NewString("Pressing this button plays the sequence in the reverse \
direction."));
	SetVar(button, STICKINESS, NewInt(STICKYLEFT + STICKYBOTTOM));
	l = r;
	
	r = l + ANIMBUTTONWIDTH;
	button = NewIconButton(l, r, b, t,
			ICONSTOP, UIGREEN, "Stop", BS_PLAIN);
	AddRadioButton(radio, button);
	SetVar(button, HELPSTRING,
	    NewString("Pressing this button stops playback and recording."));
	SetVar(button, STICKINESS, NewInt(STICKYLEFT + STICKYBOTTOM));
	l = r;
	
	r = l + ANIMBUTTONWIDTH;
	button = NewIconButton(l, r, b, t,
			ICONPLAY, UIGREEN, "Play", BS_PLAIN);
	AddRadioButton(radio, button);
	SetVar(button, HELPSTRING,
	    NewString("Pressing this button plays the sequence in the forward \
direction."));
	SetVar(button, STICKINESS, NewInt(STICKYLEFT + STICKYBOTTOM));
	l = r;
	
	r = l + ANIMBUTTONWIDTH;
	button = NewIconButton(l, r, b, t,
			ICONFF, UIGREEN, "Fast\nForward", BS_PLAIN);
	AddRadioButton(radio, button);
	SetVar(button, HELPSTRING,
	    NewString("Pressing this button plays the sequence in the forward \
direction at three times normal speed."));
	SetVar(button, STICKINESS, NewInt(STICKYLEFT + STICKYBOTTOM));
	l = r;
	
	r = l + ANIMBUTTONWIDTH;
	button = NewIconButton(l, r, b, t,
			ICONTOEND, UIGREEN, "To End", BS_PLAIN);
	AddRadioButton(radio, button);
	SetVar(button, HELPSTRING,
	    NewString("Pressing this button advances the sequence to the very end."));
	SetVar(button, STICKINESS, NewInt(STICKYLEFT + STICKYBOTTOM));
	l = r;
	
	l = r + MAJORBORDER;

	r = l + ANIMBUTTONWIDTH;
	button = NewIconButton(l, r, b, t,
			ICONRECORD, UIRED, "Record", BS_PLAIN);
	AddRadioButton(radio, button);
	SetVar(button, HELPSTRING,
	    NewString("Pressing this button records the sequence to the currently \
selected animation recorder."));
	SetVar(button, STICKINESS, NewInt(STICKYLEFT + STICKYBOTTOM));
	l = r + MINORBORDER;

	AssocDirectControlWithVar(radio, object, ANIMATIONSTATE);
	
    }
    return (ObjPtr) controlWindow;
}

static ObjPtr DrawTrackName(track, l, r, b, t)
ObjPtr track;
int l, r, b, t;
/*Draws a track name within l, r, b, t*/
{
    int texty;
    ObjPtr name;
    ObjPtr subTracks;
    int k;
    ObjPtr *elements;

    MakeVar(track, APPEARANCE);

    texty = t - TC_CELLHEIGHT / 2 - TCFONTSIZE / 2 + TC_TEXTBUP;

    FillUIRect(l, r, t - TC_CELLHEIGHT, t, IsSelected(track) ? UIYELLOW : UIGRAY62);

    SetUIColor(UIBLACK);
    MakeVar(track, NAME);
    name = GetVar(track,  NAME);
    if (name)
    {
	DrawAString(LEFTALIGN, l + 1 + TCTEXTLOFF,
		    texty, GetString(name)); 
    }
    else
    {
	DrawAString(LEFTALIGN, l + 1 + TCTEXTLOFF,
		    texty, "?");
    }

    if (GetPredicate(track, EXPANDEDP))
    {
	/*Draw the subtracks*/
	ObjPtr var;
	FuncTyp method;

	
	subTracks = GetVar(track, SUBTRACKS);
	if (subTracks)
	{
	    t -= TC_CELLHEIGHT;
	    elements = ELEMENTS(subTracks);
	    for (k = 0; k < DIMS(subTracks)[0]; ++k)
	    {

		MakeVar(elements[k], TRACKHEIGHT);
		var = GetVar(elements[k], TRACKHEIGHT);
		if (var)
		{
		    b = t - GetInt(var);

		    method = GetMethod(elements[k], DRAWNAME);
		    if (method)
		    {
			(*method)(elements[k], l, r, b, t);
		    }
		    t = b;
		}
	    }
	}
    }

    return ObjTrue;
}

#ifdef HAVE_PROTOTYPES
static void DrawSubTrackMarker(int x, int y, Bool highlight)
#else
static void DrawSubTrackMarker(x, y, Bool highlight)
int x, y;
#endif
/*Draws a subtrack keyframe marker at x, y*/
{
    int left, right, bottom, top;

    left = x - TC_STMARKHWIDTH;
    right = x + TC_STMARKHWIDTH + 1;
    bottom = y - TC_STMARKHHEIGHT;
    top = y + TC_STMARKHHEIGHT + 1;


    /* face */
    FillUIRect(left, right, bottom, top - 1, highlight ? UIHIBACKGROUND : UIBACKGROUND);

    /* outside */
    FrameUIRect(left, right, bottom, top, UIBLACK);

#if 0
    /*face*/
    FillUIRect(left + TC_STMARKEDGE, right - TC_STMARKEDGE, 
		bottom + TC_STMARKEDGE, top - TC_STMARKEDGE, UIBACKGROUND);

    /* bottom */
    SetUIColor(UIBOTTOMEDGE);
    FillIntQuad( left, bottom,
		 right, bottom,
		 right - TC_STMARKEDGE, bottom + TC_STMARKEDGE,
		 left + TC_STMARKEDGE, bottom + TC_STMARKEDGE);

    /* left */
    SetUIColor(UILEFTEDGE);
    FillIntQuad( left,  bottom,
		 left + TC_STMARKEDGE, bottom + TC_STMARKEDGE,
		 left + TC_STMARKEDGE, top - TC_STMARKEDGE,
		 left, top);

    /* right */
    SetUIColor(UIRIGHTEDGE);
    FillIntQuad( right - TC_STMARKEDGE, bottom + TC_STMARKEDGE,
		 right, bottom,
		 right, top,
		 right - TC_STMARKEDGE, top - TC_STMARKEDGE);

    /* top */
    SetUIColor(UITOPEDGE);
    FillIntQuad( left + TC_STMARKEDGE, top - TC_STMARKEDGE,
		 right - TC_STMARKEDGE, top - TC_STMARKEDGE,
		 right, top,
		 left, top);
#endif
}

#ifdef HAVE_PROTOTYPES
static void DrawTrackMarker(int x, int y)
#else
static void DrawTrackMarker(x, y)
int x, y;
#endif
/*Draws a keyframe marker at x, y*/
{
    int left, right, bottom, top;

    left = x - TC_TMARKWIDTH;
    right = x + TC_TMARKWIDTH + 1;
    bottom = y - TC_TMARKHEIGHT;
    top = y + TC_TMARKHEIGHT + 1;

    /* face */
    FillUIRect(left, right, bottom, top - 1, UIBLACK);

#if 0
    /* outside */
    FrameUIRect(left, right, bottom, top, UIBLACK);
#endif

#if 0
    /* top */
    DrawUILine(left, top - 1, right, top - 1, UISHARPTOPEDGE);

    /* right */
    DrawUILine(right, top - 1, right, bottom, UISHARPRIGHTEDGE);

    /* left */
    DrawUILine(left - 1, bottom, left - 1, top - 1, UISHARPLEFTEDGE);

    /* bottom */
    DrawUILine(left, bottom, right, bottom, UISHARPBOTTOMEDGE);

    /* line */
    DrawUILine(x, bottom + 1, x, top - 2, UIBLACK);
#endif
}

static ObjPtr DrawTrackLine(track, l, r, b, t, minTime, maxTime)
ObjPtr track;
int l, r, b, t;
real minTime, maxTime;
/*Draws a track line within l, r, b, t between minTime and maxTime*/
{
    int middle;
    ObjPtr name, sequence;
    ObjPtr *subTracks, *keyframes;
    int k;
    ObjPtr var, keyframesVar;
    int nSubtracks, nKeyframes;
    int pixel;
    real testTime, bestTime;
    real curTime;
    int index;
    Bool anyToDraw;
    real time;

    MakeVar(track, APPEARANCE);

    middle = t - TC_CELLHEIGHT / 2;

#if 0
    DrawUILine(l, middle + 1, r, middle + 1, UIYELLOW);
    DrawUILine(l, middle, r, middle, UIYELLOW);
#endif

    /*Get the sequence that owns this track -> sequence*/
    sequence = GetObjectVar("DrawTrackLine", track, PARENT);
    if (!sequence) return ObjFalse;

    /*Get the subtracks*/
    
    var = GetVar(track, SUBTRACKS);
    if (var)
    {
	subTracks = ELEMENTS(var);
	nSubtracks = DIMS(var)[0];
    }
    else
    {
	subTracks = (ObjPtr *) 0;
    }


    /*Get starting pixel*/
    pixel = l - TC_STMARKHWIDTH;

    /*Get starting frame*/
    curTime = P2T(pixel);
    if (curTime < 0.) curTime = 0.;

    if (subTracks)
    {
	/*Draw the track summary*/

	/*Go through all the keyframes, setting up indices*/
	for (k = 0; k < nSubtracks; ++k)
	{
	    keyframesVar = GetVar(subTracks[k], KEYFRAMES);
	    if (keyframesVar)
	    {
		keyframes = ELEMENTS(keyframesVar);
		nKeyframes = DIMS(keyframesVar)[0];
	
		/*Get the index of the first keyframe to draw*/
		index = SearchRealVar(keyframesVar, TIME, curTime);
		if (index > 0) --index;

		/*Stick it into the keyframe*/
		SetVar(keyframesVar, INDEX, NewInt(index));
	    }
	}

	do
	{
	    /*Find keyframe which has lowest TIME greater than or
	      equal to curTime, automatically advancing anything less
	      than curTime*/

	    bestTime = 999999999.0;
	    anyToDraw = false;

	    for (k = 0; k < nSubtracks; ++k)
	    {
		keyframesVar = GetVar(subTracks[k], KEYFRAMES);
		if (keyframesVar)
		{
		    keyframes = ELEMENTS(keyframesVar);
		    nKeyframes = DIMS(keyframesVar)[0];

		    var = GetIntVar("DrawTrackLine", keyframesVar, INDEX);
		    if (!var) continue;
		    index = GetInt(var);

		    if (index >= nKeyframes) continue;

		    var = GetRealVar("DrawTrackLine", keyframes[index], TIME);
		    if (!var) continue;

		    while (index < nKeyframes && (testTime = GetReal(var)) < curTime && !EqualTime(testTime, curTime))
		    {
			++index;
			SetVar(keyframesVar, INDEX, NewInt(index));
			if (index < nKeyframes)
			{
			    var = GetRealVar("DrawTrackLine", keyframes[index], TIME);
			}
		    }
		    if (index >= nKeyframes) continue;

		    bestTime = MIN(bestTime, testTime);
		    anyToDraw = true;
		}
	    }

	    /*Now we should have a bestFrame*/
	    if (anyToDraw)
	    {
		/*Print out the bestFrame*/
		time = bestTime;
		pixel = T2P(time);
		DrawTrackMarker(pixel, middle);

		/*Go to next frame*/
		curTime = bestTime;

		/*Advance the indices*/
		for (k = 0; k < nSubtracks; ++k)
		{
		    keyframesVar = GetVar(subTracks[k], KEYFRAMES);
		    if (keyframesVar)
		    {
			keyframes = ELEMENTS(keyframesVar);
			nKeyframes = DIMS(keyframesVar)[0];
    
			var = GetIntVar("DrawTrackLine", keyframesVar, INDEX);
			if (!var) continue;
			index = GetInt(var);
    
			if (index >= nKeyframes) continue;
    
			var = GetRealVar("DrawTrackLine", keyframes[index], TIME);
			if (!var) continue;

			testTime = GetReal(var);

			if (testTime < curTime || EqualTime(testTime, curTime))
			{
			    SetVar(keyframesVar, INDEX, NewInt(index + 1));
			}
		    }
		}
	    }

	} while (anyToDraw && pixel < r + TC_STMARKHWIDTH);
    }

    if (GetPredicate(track, EXPANDEDP))
    {
	/*Draw the subtracks*/
	FuncTyp method;

	if (subTracks)
	{
	    t -= TC_CELLHEIGHT;
	    for (k = 0; k < nSubtracks; ++k)
	    {

		MakeVar(subTracks[k], TRACKHEIGHT);
		var = GetVar(subTracks[k], TRACKHEIGHT);
		if (var)
		{
		    b = t - GetInt(var);

		    method = GetMethod(subTracks[k], DRAWTRACK);
		    if (method)
		    {
			(*method)(subTracks[k], l, r, b, t, minTime, maxTime);
		    }
		    t = b;
		}
	    }
	}
    }

    return ObjTrue;
}

static ObjPtr DrawSubTrackLine(track, l, r, b, t, minTime, maxTime)
ObjPtr track;
int l, r, b, t;
real minTime, maxTime;
/*Draws a subtrack within l, r, b, t between minTime and maxTime*/
{
    int middle;
    ObjPtr subTracks;
    ObjPtr keyframes;
    ObjPtr sequence;
    ObjPtr *elements;
    int pixel;
    real testTime;
    real curTime, beginTime, endTime;
    long index;
    long firstKeyframe;
    int startPixel;
    ObjPtr var;
    ObjPtr snap1, snap2;

    /*Get the elements to draw*/
    keyframes = GetVar(track, KEYFRAMES);
    if (!keyframes) return ObjTrue;
    elements = ELEMENTS(keyframes);

    /*Get middle of cell for drawing things*/
    middle = t - TC_CELLHEIGHT / 2;

    /*Get the sequence that owns this  subtrack -> track -> sequence*/
    sequence = GetObjectVar("DrawSubTrackLine", track, PARENT);
    if (!sequence) return ObjFalse;
    sequence = GetObjectVar("DrawSubTrackLine", sequence, PARENT);
    if (!sequence) return ObjFalse;

    /*Get starting pixel*/
    pixel = l - TC_STMARKHWIDTH;

    /*Get starting time*/
    curTime = P2T(pixel);

    /*Get ending pixel*/
    pixel = r + TC_STMARKHWIDTH;

    /*Get ending time*/
    endTime = P2T(pixel);

    /*Get the index of the first keyframe to draw*/
    firstKeyframe = SearchRealVar(keyframes, TIME, curTime);
    if (firstKeyframe > 0) --firstKeyframe;

    /*Get the time of the first keyframe*/
    var = GetRealVar("DrawSubTrackLine", elements[firstKeyframe], TIME);
    if (!var)
    {
	return ObjFalse;
    }
    beginTime = GetReal(var);

    /*Print out the highlights*/
    index = firstKeyframe;
    curTime = beginTime;
    while (curTime <= endTime && index < DIMS(keyframes)[0])
    {
	/*Print out the curFrame*/
	pixel = T2P(curTime);

	if (IsSelected(elements[index]))
	{
	    FillUIRect(pixel - TC_KEYHIBORDER, pixel + TC_KEYHIBORDER + 1,
		   b, t, UIYELLOW);
	    if (index < DIMS(keyframes)[0] - 1)
	    {
		/*Draw the line to the next frame*/
		startPixel = pixel + TC_STMARKHWIDTH + 1;
		var = GetRealVar("DrawSubTrackLine", elements[index + 1], TIME);
		if (!var)
		{
		    return ObjFalse;
		}
		testTime = GetReal(var);
		pixel = T2P(testTime);
	
		snap1 = GetVar(elements[index], SNAPSHOT);
		snap2 = GetVar(elements[index + 1], SNAPSHOT);
	
		if (IsSelected(elements[index + 1]) && startPixel < pixel - TC_STMARKHWIDTH)
		{
		    /*Draw a continuation line*/
		    FillUIRect(startPixel, pixel - TC_STMARKHWIDTH,
				b, t,
				UIYELLOW);
		}
	    }
	}

	/*Advance index and curTime*/
	++index;
	if (index < DIMS(keyframes)[0])
	{
	    var = GetRealVar("DrawSubTrackLine", elements[index], TIME);
	    if (!var)
	    {
		return ObjFalse;
	    }
	    curTime = GetReal(var);
	}
    }

    /*Print out the keyframes*/
    index = firstKeyframe;
    curTime = beginTime;
    while (curTime <= endTime && index < DIMS(keyframes)[0])
    {
	/*Print out the curFrame*/
	pixel = T2P(curTime);
	DrawSubTrackMarker(pixel, middle, GetPredicate(elements[index], HIGHLIGHTED));

	if (index < DIMS(keyframes)[0] - 1)
	{
	    /*Draw the line to the next frame*/
	    startPixel = pixel + TC_STMARKHWIDTH + 1;
	    var = GetRealVar("DrawSubTrackLine", elements[index + 1], TIME);
	    if (!var)
	    {
		return ObjFalse;
	    }
	    testTime = GetReal(var);
	    pixel = T2P(testTime);
    
	    snap1 = GetVar(elements[index], SNAPSHOT);
	    snap2 = GetVar(elements[index + 1], SNAPSHOT);
    
	    if (startPixel < pixel - TC_STMARKHWIDTH)
	    {
		/*Draw a continuation line*/
		FillUIRect(startPixel, pixel - TC_STMARKHWIDTH,
			    middle - 1, middle + 1,
			    EqualSnapshots(snap1, snap2) ? UIGRAY50 : UIBLUE);
	    }
	}

	/*Advance index and curTime*/
	++index;
	if (index < DIMS(keyframes)[0])
	{
	    var = GetRealVar("DrawSubTrackLine", elements[index], TIME);
	    if (!var)
	    {
		return ObjFalse;
	    }
	    curTime = GetReal(var);
	}
    }

    return ObjTrue;
}

static ObjPtr DrawSubTrackName(track, l, r, b, t)
ObjPtr track;
int l, r, b, t;
/*Draws a subtrack name within l, r, b, t*/
{
    int texty;
    ObjPtr name;
    ObjPtr subTracks;
    int k;
    ObjPtr *elements;

    MakeVar(track, APPEARANCE);

    texty = t - TC_CELLHEIGHT / 2 - TCFONTSIZE / 2;

    FillUIRect(l, r, t - TC_CELLHEIGHT, t, IsSelected(track) ? UIYELLOW : UIGRAY62);

    SetUIColor(UIBLACK);
    MakeVar(track, NAME);
    name = GetVar(track,  NAME);
    if (name)
    {
	DrawAString(LEFTALIGN, l + 1 + TCSUBTEXTLOFF,
		    texty, GetString(name)); 
    }
    else
    {
	DrawAString(LEFTALIGN, l + 1 + TCSUBTEXTLOFF,
		    texty, "?");
    }

    return ObjTrue;
}

static MakeMethod MakeTrackAppearance(track)
ObjPtr track;
/*Makes a track's appearance*/
{
    SetVar(track, APPEARANCE, ObjTrue);
    ImInvalid(track);
    return ObjTrue;
}

static MakeMethod MakeSubTrackAppearance(track)
ObjPtr track;
/*Makes a subtrack's appearance*/
{
    SetVar(track, APPEARANCE, ObjTrue);
    ImInvalid(track);
    return ObjTrue;
}

static ObjPtr PressTrackName(track, x, y, flags)
ObjPtr track;
int x, y;
long flags;
/*Press in a track control name section.  y is negative from top*/
{
    ObjPtr parent;

    if (y > -TC_CELLHEIGHT)
    {
	/*It's a press in the main name*/
	if (TOOL(flags) == T_HELP)
	{
	    ContextHelp(track);
	    return ObjTrue;
	}

	if (TOOL(flags) == T_ROTATE)
	{
	    flags |= F_EXTEND;
	}

	if (flags & F_EXTEND)
	{
	    Select(track, IsSelected(track) ? false : true);
	}
	else
	{
	    if (!IsSelected(track))
	    {
		DeselectAll();
		Select(track, true);
	    }
	    if (flags & F_DOUBLECLICK)
	    {
		/*Expand or contract*/
		SetVar(track, EXPANDEDP, GetPredicate(track, EXPANDEDP) ? ObjFalse : ObjTrue);

		/*Have to touch parent's TRACKS*/
		parent = GetObjectVar("PressTrackName", track, PARENT);
		if (parent)
		{
		    SetVar(parent, TRACKS, GetVar(parent, TRACKS));
		}

		/*DIKEO log*/
	    }
	}
    }
    else if (GetPredicate(track, EXPANDEDP))
    {
	/*It may be a press in a subtrack*/
	int ty;
	ObjPtr var;
	ObjPtr subTracks;
	int height;
	long k;
	ObjPtr *elements;

	ty = - TC_CELLHEIGHT;

	subTracks = GetVar(track, SUBTRACKS);
	if (subTracks)
	{
	    elements = ELEMENTS(subTracks);
	    for (k = 0; k < DIMS(subTracks)[0]; ++k)
	    {
		MakeVar(elements[k], TRACKHEIGHT);
		var = GetVar(elements[k], TRACKHEIGHT);
		if (var)
		{
		    height = GetInt(var);
		    if ((ty - height) < y &&
			 (ty >= y)) break;
		    ty -= height;
		}
	    }

	    if (k < DIMS(subTracks)[0])
	    {
		/*It is a press in a subtrack*/
		FuncTyp method;

		method = GetMethod(elements[k], PRESSNAME);
		if (method)
		{
		    (*method)(elements[k], x, y - ty, flags);
		}
	    }
	}
    }

    return ObjTrue;
}

static ObjPtr PressTrack(track, x, y, l, r, minTime, maxTime, flags)
ObjPtr track;
int x, y;
int l, r;
real minTime, maxTime;
long flags;
/*Press in a track control track section.  y is negative from top*/
{
    ObjPtr parent;

    if (y > -TC_CELLHEIGHT)
    {
	/*It's a press in the main name*/
	if (TOOL(flags) == T_HELP)
	{
	    ContextHelp(track);
	    return ObjTrue;
	}

	if (TOOL(flags) == T_ROTATE)
	{
	    flags |= F_EXTEND;
	}

#if 0
	if (flags & F_EXTEND)
	{
	    Select(track, IsSelected(track) ? false : true);
	}
	else
	{
	    if (!IsSelected(track))
	    {
		DeselectAll();
		Select(track, true);
	    }
	    if (flags & F_DOUBLECLICK)
	    {
		/*Expand or contract*/
		SetVar(track, EXPANDEDP, GetPredicate(track, EXPANDEDP) ? ObjFalse : ObjTrue);

		/*Have to touch parent's TRACKS*/
		parent = GetObjectVar("PressTrackName", track, PARENT);
		if (parent)
		{
		    SetVar(parent, TRACKS, GetVar(parent, TRACKS));
		}

		/*DIKEO log*/
	    }
	}
#endif
    }
    else if (GetPredicate(track, EXPANDEDP))
    {
	/*It may be a press in a subtrack*/
	int ty;
	ObjPtr var;
	ObjPtr subTracks;
	int height;
	long k;
	ObjPtr *elements;

	ty = - TC_CELLHEIGHT;

	subTracks = GetVar(track, SUBTRACKS);
	if (subTracks)
	{
	    elements = ELEMENTS(subTracks);
	    for (k = 0; k < DIMS(subTracks)[0]; ++k)
	    {
		MakeVar(elements[k], TRACKHEIGHT);
		var = GetVar(elements[k], TRACKHEIGHT);
		if (var)
		{
		    height = GetInt(var);
		    if ((ty - height) < y &&
			 (ty >= y)) break;
		    ty -= height;
		}
	    }

	    if (k < DIMS(subTracks)[0])
	    {
		/*It is a press in a subtrack*/
		FuncTyp method;

		method = GetMethod(elements[k], PRESSTRACK);
		if (method)
		{
		    (*method)(elements[k], x, y - ty, l, r, minTime, maxTime, flags);
		}
	    }
	}
    }

    return ObjTrue;
}

static ObjPtr PressSubTrackName(track, x, y, flags)
ObjPtr track;
int x, y;
long flags;
/*Press in a subtrack control name section.  y is negative from top*/
{
    ObjPtr parent;

    if (TOOL(flags) == T_HELP)
    {
	ContextHelp(track);
	return ObjTrue;
    }

    if (TOOL(flags) == T_ROTATE)
    {
	flags |= F_EXTEND;
    }

    if (flags & F_EXTEND)
    {
	Select(track, IsSelected(track) ? false : true);
    }
    else
    {
	if (!IsSelected(track))
	{
	    DeselectAll();
	    Select(track, true);
	}
    }

    /*Make its parent invalid*/
    parent = GetVar(track, PARENT);
    SetVar(parent, APPEARANCE, ObjTrue);

    return ObjTrue;
}

static ObjPtr PressSubTrack(track, x, y, l, r, minTime, maxTime, flags)
ObjPtr track;
int x, y;
int l, r;
real minTime, maxTime;
long flags;
/*Press in a subtrack control track section.  y is negative from top*/
{
    ObjPtr parent;
    ObjPtr keyframes;
    ObjPtr *elements;
    int nKeyframes;
    int k;
    int newX, newY;
    int startX;
    real curTime, startTime, lastTime, testTime;
    int pixel;
    int tp1, tp2;
    int index;
    ObjPtr var;
    ObjPtr selList;
    ThingListPtr runner;
    ObjPtr sequence;

    if (TOOL(flags) == T_HELP)
    {
	ContextHelp(track);
	return ObjTrue;
    }

    if (TOOL(flags) == T_ROTATE)
    {
	flags |= F_EXTEND;
    }

    /*Look for a keyframe*/
    keyframes = GetVar(track, KEYFRAMES);
    if (!keyframes) return ObjTrue;

    elements = ELEMENTS(keyframes);
    nKeyframes = DIMS(keyframes)[0];

    /*Get starting pixel*/
    pixel = l - TC_STMARKHWIDTH;

    /*Get starting frame*/
    curTime = P2T(pixel);

    /*Get the index of the first keyframe to test*/
    index = SearchRealVar(keyframes, TIME, curTime);
    if (index > 0) --index;

    if (0 == (flags & F_EXTEND))
    {
	DeselectAll();
    }

    /*Accumulate selected keyframes*/
    selList = NewList();

    do
    {
	MakeVar(elements[index], TIME);
	var = GetRealVar("PressSubTrack", elements[index], TIME);
	if (var)
	{
	    tp1 = T2P(GetReal(var)) - TC_STMARKHWIDTH;
	    tp2 = tp1 + 2 * TC_STMARKHWIDTH;

	    if (tp1 < x && x <= tp2)
	    {
		if (flags & F_EXTEND)
		{
		    Select(elements[index], IsSelected(elements[index]) ? false : true);
		}
		else
		{
		    Select(elements[index], true);
		    PostfixList(selList, elements[index]);
		}
	    }
	}
	++index;
    } while (index < nKeyframes && tp1 < x);

    parent = GetVar(track, PARENT);
    sequence = GetVar(parent, PARENT);

    /*Do the drag*/
    if (0 == (flags & F_EXTEND))
    {
	Bool constrain;
	real frameWidth;
	real loValue, hiValue;

	var = GetVar(sequence, DURATION);
	if (var) hiValue = GetReal(var);
	loValue = 0.0;

	MakeVar(sequence, FRAMERATE);
	var = GetVar(sequence, FRAMERATE);
	if (!var)
	{
	    return ObjTrue;
	}
	frameWidth = 1.0 / GetReal(var);

	constrain = GetPredicate(sequence, CONSTRAINTOKEY);
	
	/*Set original times*/
	for (runner = LISTOF(selList); runner; runner = runner -> next)
	{
	    SetVar(runner -> thing, HIGHLIGHTED, ObjTrue);
	    SetVar(runner -> thing, ORIGTIME, GetVar(runner -> thing, TIME));
	}

	startX = x;
	startTime = P2T(startX);
	lastTime = startTime;

	DrawMe(track);
	DrawMe(parent);

	while (Mouse(&newX, &newY))
	{
	    curTime = P2T(newX) - startTime;

	    if (curTime != lastTime)
	    {
		for (runner = LISTOF(selList); runner; runner = runner -> next)
		{
		    var = GetVar(runner -> thing, ORIGTIME);
		    if (var)
		    {
			testTime = GetReal(var) + curTime;
			if (testTime < loValue - frameWidth * 0.5)
			{
			    testTime = loValue;
			}
			if (testTime > hiValue - frameWidth * 0.5)
			{
			    testTime = hiValue - frameWidth;
			}
			if (constrain)
			{
			    testTime = floor(testTime / frameWidth + 0.5) * frameWidth;
			}
			SetVar(runner -> thing, TIME, NewReal(testTime));
		    }
		}

		lastTime = curTime;
    
		DrawMe(track);
		DrawMe(parent);
	    }
	}
	for (runner = LISTOF(selList); runner; runner = runner -> next)
	{
	    SetVar(runner -> thing, HIGHLIGHTED, ObjFalse);
	}

	/*Just in case stuff has changed, sort the list*/
	keyframes = SortArrayByRealVar(keyframes, TIME);
	SetVar(track, KEYFRAMES, keyframes);

	elements = ELEMENTS(keyframes);
	nKeyframes = DIMS(keyframes)[0];

	for (runner = LISTOF(selList); runner; runner = runner -> next)
	{
	    SetVar(runner -> thing, HIGHLIGHTED, ObjFalse);
	}
	DrawMe(track);
	DrawMe(parent);
    }

#if 0
    if (flags & F_EXTEND)
    {
	Select(track, IsSelected(track) ? false : true);
    }
    else
    {
	if (!IsSelected(track))
	{
	    DeselectAll();
	    Select(track, true);
	}
    }
#endif

    /*Make its parent invalid*/
    SetVar(parent, APPEARANCE, ObjTrue);

    return ObjTrue;
}
static ObjPtr MakeTrackHeight(track)
ObjPtr track;
/*Makes a track's TRACKHEIGHT*/
{
    int height;

    height = TC_CELLHEIGHT;

    if (GetPredicate(track, EXPANDEDP))
    {
	ObjPtr subTracks;

	subTracks = GetVar(track, SUBTRACKS);
	if (subTracks)
	{
	    ObjPtr *elements;
	    ObjPtr var;
	    int k;

	    elements = ELEMENTS(subTracks);

	    for (k = 0; k < DIMS(subTracks)[0]; ++k)
	    {
		MakeVar(elements[k], TRACKHEIGHT);
		var = GetVar(elements[k], TRACKHEIGHT);
		if (var)
		{
		    height += GetInt(var);
		}
	    }
	}
    }

    SetVar(track, TRACKHEIGHT, NewInt(height));

    return ObjTrue;
}

static ObjPtr MakeSubTrackHeight(track)
ObjPtr track;
/*Makes a subtrack's TRACKHEIGHT*/
{
    SetVar(track, TRACKHEIGHT, NewInt(TC_CELLHEIGHT));
    return ObjTrue;
}

static ObjPtr SelectTrack(track, whether)
ObjPtr track;
Bool whether;
/*Extra stuff for selecting a track; selects all its subtracks*/
{
    ObjPtr subTracks;
    ObjPtr *elements;
    long k;

    subTracks = GetVar(track, SUBTRACKS);
    if (subTracks)
    {
	elements = ELEMENTS(subTracks);
	for (k = 0; k < DIMS(subTracks)[0]; ++k)
	{
	    Select(elements[k], whether);
	}
    }
}

#ifdef HAVE_PROTOTYPES
void DefineBetterVarName(NameTyp var, char *name)
#else
void DefineBetterVarName(var, name)
NameTyp var;
char *name;
#endif
/*Defines a better var name for variable var as name*/
{
    int k;

    if (var >= nBetterVarNames)
    {
	if (betterVarNames)
	{
	    betterVarNames = (char **) Realloc(betterVarNames, (var + 1) * sizeof(char *));
	}
	else
	{
	    betterVarNames = (char **) Alloc((var + 1) * sizeof(char *));
	}
	for (k = nBetterVarNames; k <= var; ++k)
	{
	    betterVarNames[k] = NULL;
	}
	nBetterVarNames = var + 1;
    }
    SAFEFREE(betterVarNames[var]);
    betterVarNames[var] = (char *) Alloc(strlen(name) + 1);
    strcpy(betterVarNames[var], name);
}

#ifdef HAVE_PROTOTYPES
char *GetGoodVarName(NameTyp var)
#else
char *GetGoodVarName(var)
NameTyp var;
#endif
/*Gets a good var name for variable var and returns it.  Return value is
volatile and disappears after next call to GetGoodVarName.*/
{
    char *name;
    static char processedName[300];
    int k;

    if (var < nBetterVarNames && betterVarNames[var])
    {
	strcpy(processedName, betterVarNames[var]);
    }
    else
    {
    name = GetInternalString(var);
    strcpy(processedName, name);

    for (k = 1; processedName[k]; ++k)
    {
	processedName[k] = tolower(processedName[k]);
    }
    }
    return processedName;
}

#ifdef HAVE_PROTOTYPES
void RecordObjectVar(ObjPtr object, NameTyp whichVar)
#else
void RecordObjectVar(object, whichVar)
ObjPtr object;
NameTyp whichVar;
#endif
/*Records the current value of an object var in recording sequences*/
{
    ObjPtr result;

    result = AllProjectSequences();
    if (result && LISTOF(result))
    {
	ThingListPtr runner;
	char *name;

	for (runner = LISTOF(result); runner; runner = runner -> next)
	{
	    ObjPtr track, snapshot;
	    real time;
	    ObjPtr var;

	    if (!GetPredicate(runner -> thing, RECORDING)) continue;

	    name = GetGoodVarName(whichVar);

	    track = TrackVarGroupWithinSequence(runner -> thing, object, name, NULLOBJ);
	    if (track)
	    {
		var = GetVar(runner -> thing, TIME);
		if (var)
		{
		    time = GetReal(var);
		}
		else
		{
		    time = 0.0;
		}

		snapshot = TakeSingleVarSnapshot(object, whichVar);
		InsertKeySnapshot(track, time, snapshot);
	    }
	}
    }
}

#ifdef HAVE_PROTOTYPES
ObjPtr TrackBoundsInvalid(ObjPtr object, unsigned long changeCount)
#else
ObjPtr TrackBoundsInvalid(object, changeCount)
ObjPtr object;
unsigned long changeCount;
#endif
/*For an track, tests to see if it needs drawing.  Returns
  NULLOBJ	if it does not
  ObjTrue	it it needs to be redrawn but does not know where

  In order for an object to be declared invalid, its changed bounds
  must have been changed more recently than changeCount
*/
{
    ObjPtr subTracks;

    MakeVar(object, APPEARANCE);

    if (GetVarChangeCount(object, CHANGEDBOUNDS) > changeCount)
    {
	/*Object is not good, so return true*/
	return ObjTrue;
    }

    subTracks = GetVar(object, SUBTRACKS);
    if (subTracks && IsObjArray(subTracks))
    {
	/*Still, maybe some of the contents need to be drawn*/
	ObjPtr *elements;
	long k;

	elements = ELEMENTS(subTracks);

	for (k = 0; k < DIMS(subTracks)[0]; ++k)
	{
	    if (BoundsInvalid(elements[k], changeCount))
	    {
		return ObjTrue;
	    }
	}
    }

    return NULLOBJ;
}

void SetupWindowForSequence(win)
WinInfoPtr win;
/*Sets up a vis window for sequence*/
{
    PopWindow(win);
    DoWindowFunction(WF_HIDEPANEL, win);
    DoWindowFunction(WF_HIDEFRAME, win);
}

void RecoverWindowFromRecording(win)
WinInfoPtr win;
/*Recovers up a vis window from sequence*/
{
    PopWindow(win);
    DoWindowFunction(WF_SHOWPANEL, win);
    DoWindowFunction(WF_SHOWFRAME, win);
}

#ifdef HAVE_PROTOTYPES
void IdleAnimation(void)
#else
void IdleAnimation()
#endif
/*Idles the animation*/
{
    ObjPtr sequences;
    ObjPtr sequence;
    ThingListPtr runner;
    ObjPtr var;
    int animationState;
    int recordState;

    sequences = AllProjectSequences();
    if (sequences)
    {
	for (runner = LISTOF(sequences); runner; runner = runner -> next)
	{
	    sequence = runner -> thing;
	    var = GetVar(sequence, ANIMATIONSTATE);
	    if (var)
	    {
		animationState = GetInt(var);

		var = GetVar(sequence, RECORDSTATE);
		if (var)
		{
		    recordState = GetInt(var);
		}
		else
		{
		    recordState = RS_NULL;
		}

		if (animationState != AC_RECORD)
		{
		    if (recordState == RS_RUNNING || recordState == RS_STOP)
		    {
			StopRecording();
			DisconnectRecorder();
			SetVar(sequence, recordState, NewInt(RS_RECOVER));
		    }
		}

		switch(animationState)
		{
		    case AC_REWIND:
			/*Rewind to beginning*/
			SetVar(sequence, TIME, NewReal(0.0));
			SetVar(sequence, ANIMATIONSTATE, NewInt(AC_STOP));
			SetVar(sequence, RECORDSTATE, NewInt(RS_NULL));
			if (GetPredicate(sequence, ENABLEEXPRESSION))
			{
			    ExpressSequence(sequence);
			}
			break;
		    case AC_FREVERSE:
			/*Step back a number of frames*/
			{
			    real frameRate;
			    long frame;
			    real time;

			    SetVar(sequence, RECORDSTATE, NewInt(RS_NULL));

			    MakeVar(sequence, FRAMERATE);
			    var = GetRealVar("IdleSequence", sequence, FRAMERATE);
			    if (!var) break;
			    frameRate = GetReal(var);

			    MakeVar(sequence, TIME);
			    var = GetRealVar("IdleSequence", sequence, TIME);
			    if (!var) break;
			    time = GetReal(var);
			    
			    if (EqualTime(time, 0.0) || time <= 0.0)
			    {
				/*At end; stop*/
				SetVar(sequence, ANIMATIONSTATE, NewInt(AC_STOP));
			    }
			    else
			    {
				/*Retard*/
				frame = floor(time * frameRate + 0.5);
				frame -= NFASTFRAMES;
				if (frame < 0) frame = 0;
				time = frame / frameRate;
    
				SetVar(sequence, TIME, NewReal(time));
				ExpressSequence(sequence);
			    }
			}
			break;
		    case AC_REVERSE:
			/*Go to previous frame*/
			{
			    real frameRate;
			    long frame;
			    real time;

			    SetVar(sequence, RECORDSTATE, NewInt(RS_NULL));

			    MakeVar(sequence, FRAMERATE);
			    var = GetRealVar("IdleSequence", sequence, FRAMERATE);
			    if (!var) break;
			    frameRate = GetReal(var);

			    MakeVar(sequence, TIME);
			    var = GetRealVar("IdleSequence", sequence, TIME);
			    if (!var) break;
			    time = GetReal(var);
			    
			    if (EqualTime(time, 0.0) || time <= 0.0)
			    {
				/*At end; stop*/
				SetVar(sequence, ANIMATIONSTATE, NewInt(AC_STOP));
			    }
			    else
			    {
				/*Retard*/
				frame = floor(time * frameRate + 0.5);
				--frame;
				time = frame / frameRate;
    
				SetVar(sequence, TIME, NewReal(time));
				ExpressSequence(sequence);
			    }
			}
			break;
		    case AC_STOP:
			/*Stopped; do nothing*/
			SetVar(sequence, RECORDSTATE, NewInt(RS_NULL));
			break;
		    case AC_PLAY:
			/*Go to next frame*/
			{
			    real frameRate;
			    long frame;
			    real time, lastTime;

			    SetVar(sequence, RECORDSTATE, NewInt(RS_NULL));

			    MakeVar(sequence, FRAMERATE);
			    var = GetRealVar("IdleSequence", sequence, FRAMERATE);
			    if (!var) break;
			    frameRate = GetReal(var);

			    MakeVar(sequence, TIME);
			    var = GetRealVar("IdleSequence", sequence, TIME);
			    if (!var) break;
			    time = GetReal(var);
			    
			    var = GetRealVar("IdleSequence", sequence, DURATION);
			    if (!var) break;
			    lastTime = GetReal(var);
			    lastTime -= 1.0 / frameRate;

			    if (EqualTime(time, lastTime) || time >= lastTime)
			    {
				/*At end; stop*/
				SetVar(sequence, ANIMATIONSTATE, NewInt(AC_STOP));
			    }
			    else
			    {
				/*Advance*/
				frame = floor(time * frameRate + 0.5);
				++frame;
				time = frame / frameRate;
    
				SetVar(sequence, TIME, NewReal(time));
				ExpressSequence(sequence);
			    }
			}
			break;
		    case AC_RECORD:
			/*Record something*/
			{
			    ObjPtr var;
			    real nSeconds;
			    int nFrames;
			    int ore;

			    ore = recordEnabled;
			    recordEnabled = true;

			    var = GetIntVar("IdleAnimation", sequence, RECORDSTATE);
			    if (!var)
			    {
				SetVar(sequence, ANIMATIONSTATE, NewInt(AC_STOP));
				recordEnabled = ore;
				break;
			    }

			    switch (recordState)
			    {
				case RS_NULL:
				    /*Nothing has happened yet, just go to the setup*/
				    SetVar(sequence, RECORDSTATE, NewInt(RS_SETUP));
				    break;

				case RS_SETUP:
				    /*Do the setup*/
				    InhibitLogging(true);
				    ForAllVisWindows(SetupWindowForSequence);
				    InhibitLogging(false);
				    SetVar(sequence, RECORDSTATE, NewInt(RS_PREPARE));
				    break;
				case RS_PREPARE:
				    /*Do the preparation to record*/
				    var = GetRealVar("IdleSequence", sequence, DURATION);
				    if (!var)
				    {
					SetVar(sequence, ANIMATIONSTATE, NewInt(AC_STOP));
					break;
				    }
				    nSeconds = GetReal(var);

				    MakeVar(sequence, FRAMERATE);
				    var = GetRealVar("IdleSequence", sequence, FRAMERATE);
				    if (!var)
				    {
					SetVar(sequence, ANIMATIONSTATE, NewInt(AC_STOP));
					break;
				    }

				    nFrames = (int) (nSeconds * GetReal(var)) + 1;
				    if (AdjustToVisWindows(curRecorder))
				    {
					if (ConnectRecorder())
					{
					    if (!PrepareToRecord(nFrames))
					    {
						WarnUser(CW_NOSPACETORECORD);
						SetVar(sequence, ANIMATIONSTATE, NewInt(AC_STOP));
						break;
					    }
					}
					else
					{
					    WarnUser(CW_CANNOTCONNECTTOREC);
					    SetVar(sequence, ANIMATIONSTATE, NewInt(AC_STOP));
					    break;
					}
				    }
				    else
				    {
					WarnUser(CW_NOVISWINDOWS);
					SetVar(sequence, ANIMATIONSTATE, NewInt(AC_STOP));
					break;
				    }

				    SetVar(sequence, RECORDSTATE, NewInt(RS_RUNNING));
				    break;

				case RS_RUNNING:
				    /*Running the recording; snap another frame*/
				    {
					real frameRate;
					long frame;
					real time, lastTime;

					MakeVar(sequence, FRAMERATE);
					var = GetRealVar("IdleSequence", sequence, FRAMERATE);
					if (!var) break;
					frameRate = GetReal(var);
		
					MakeVar(sequence, TIME);
					var = GetRealVar("IdleSequence", sequence, TIME);
					if (!var) break;
					time = GetReal(var);
					
					var = GetRealVar("IdleSequence", sequence, DURATION);
					if (!var) break;
					lastTime = GetReal(var);
					lastTime -= 1.0 / frameRate;
		
					if (EqualTime(time, lastTime) || time >= lastTime)
					{
					    /*At end; stop*/
					    SetVar(sequence, RECORDSTATE, NewInt(RS_STOP));
					}
					else
					{
					    SnapOneFrame();

					    /*Advance*/
					    frame = floor(time * frameRate + 0.5);
					    ++frame;
					    time = frame / frameRate;
		
					    SetVar(sequence, TIME, NewReal(time));
					    if (GetPredicate(sequence, ENABLEEXPRESSION))
					    {
						ExpressSequence(sequence);
					    }
					}
				    }
				    break;

				case RS_STOP:
				    /*Stop the recording.*/
				    StopRecording();
				    DisconnectRecorder();
				    SetVar(sequence, RECORDSTATE, NewInt(RS_RECOVER));
				    break;

				case RS_RECOVER:
				    /*Recover the vis windows*/
				    ForAllVisWindows(RecoverWindowFromRecording);
				    SetVar(sequence, RECORDSTATE, NewInt(RS_NULL));
				    SetVar(sequence, ANIMATIONSTATE, NewInt(AC_STOP));
				    break;
			    }
			    recordEnabled = ore;
			}
			break;
		    case AC_FFORWARD:
			/*Skip some frames*/
			{
			    real frameRate;
			    long frame;
			    real time, lastTime;

			    SetVar(sequence, RECORDSTATE, NewInt(RS_NULL));

			    MakeVar(sequence, FRAMERATE);
			    var = GetRealVar("IdleSequence", sequence, FRAMERATE);
			    if (!var) break;
			    frameRate = GetReal(var);

			    MakeVar(sequence, TIME);
			    var = GetRealVar("IdleSequence", sequence, TIME);
			    if (!var) break;
			    time = GetReal(var);
			    
			    var = GetRealVar("IdleSequence", sequence, DURATION);
			    if (!var) break;
			    lastTime = GetReal(var);
			    lastTime -= 1.0 / frameRate;

			    if (EqualTime(time, lastTime) || time >= lastTime)
			    {
				/*At end; stop*/
				SetVar(sequence, ANIMATIONSTATE, NewInt(AC_STOP));
			    }
			    else
			    {
				/*Advance*/
				frame = floor(time * frameRate + 0.5);
				frame += NFASTFRAMES;
				time = frame / frameRate;
				if (time > lastTime) time = lastTime;
    
				SetVar(sequence, TIME, NewReal(time));
				ExpressSequence(sequence);
			    }
			}
			break;
		    case AC_TOEND:
			/*Go to the end*/
			SetVar(sequence, TIME, GetVar(sequence, DURATION));
			SetVar(sequence, ANIMATIONSTATE, NewInt(AC_STOP));
			SetVar(sequence, RECORDSTATE, NewInt(RS_NULL));
			if (GetPredicate(sequence, ENABLEEXPRESSION))
			{
			    ExpressSequence(sequence);
			}
			break;
		}
	    }
	}
    }
}

#ifdef HAVE_PROTOTYPES
void InitAnimation(void)
#else
void InitAnimation()
#endif
/*Initializes the animations*/
{
    sequenceClass = NewObject(NULLOBJ, 0L);
    SetMethod(sequenceClass, SHOWCONTROLS, NewControlWindow);
    SetMethod(sequenceClass, NEWCTLWINDOW, ShowSequenceControls);
    SetVar(sequenceClass, RECORDING, ObjTrue);
    SetVar(sequenceClass, ENABLEEXPRESSION, ObjTrue);
    SetVar(sequenceClass, TIME, NewReal(0.0));
    SetVar(sequenceClass, FRAMERATE, NewReal(30.0));
    SetVar(sequenceClass, DURATION, NewReal(15.0));
    AddToReferenceList(sequenceClass);
    SetVar(sequenceClass, DOUBLECLICK, NewString(OF_SHOW_CONTROLS));
    SetVar(sequenceClass, ANIMATIONSTATE, NewInt(AC_STOP));
    SetVar(sequenceClass, RECORDSTATE, NewInt(RS_NULL));
    SetVar(sequenceClass, CONSTRAINTOKEY, ObjTrue);
    AddSnapVar(sequenceClass, RECORDING);
    AddSnapVar(sequenceClass, FRAMERATE);
    AddSnapVar(sequenceClass, DURATION);
    AddSnapVar(sequenceClass, TIME);
    AddSnapVar(sequenceClass, ENABLEEXPRESSION);
    AddSnapVar(sequenceClass, ANIMATIONSTATE);
    AddSnapVar(sequenceClass, CONSTRAINTOKEY);
    DefineFixedClass(sequenceClass, CLASS_SEQUENCE, "sequence");

    trackClass = NewObject(NULLOBJ, 0L);
    AddToReferenceList(trackClass);
    DeclareDependency(trackClass, TRACKHEIGHT, EXPANDEDP);
    DeclareDependency(trackClass, TRACKHEIGHT, CHANGED);
    DeclareDependency(trackClass, TRACKHEIGHT, SUBTRACKS);
    SetMethod(trackClass, TRACKHEIGHT, MakeTrackHeight);
    SetMethod(trackClass, DRAWNAME, DrawTrackName);
    SetMethod(trackClass, DRAWTRACK, DrawTrackLine);
    DeclareDependency(trackClass, APPEARANCE, SELECTED);
    DeclareDependency(trackClass, APPEARANCE, FRAMERATE);
    DeclareDependency(trackClass, APPEARANCE, TRACKHEIGHT);
    DeclareDependency(trackClass, APPEARANCE, CHANGED);
    DeclareDependency(trackClass, APPEARANCE, EXPANDEDP);
    SetMethod(trackClass, APPEARANCE, MakeTrackAppearance);
    SetMethod(trackClass, PRESSNAME, PressTrackName);
    SetMethod(trackClass, PRESSTRACK, PressTrack);
    SetMethod(trackClass, SELECT, SelectTrack);
    SetMethod(trackClass, BOUNDSINVALID, TrackBoundsInvalid);
    DeclareIndirectDependency(trackClass, APPEARANCE, REPOBJ, SELECTED);
    SetMethod(trackClass, DELETE, DeleteTrack);
    SetVar(trackClass, NAME, NewString("Track"));
    DefineFixedClass(trackClass, CLASS_TRACK, "track");

    subTrackClass = NewObject(NULLOBJ, 0L);
    AddToReferenceList(subTrackClass);
    SetMethod(subTrackClass, TRACKHEIGHT, MakeSubTrackHeight);
    SetMethod(subTrackClass, DRAWNAME, DrawSubTrackName);
    SetMethod(subTrackClass, PRESSNAME, PressSubTrackName);
    SetMethod(subTrackClass, PRESSTRACK, PressSubTrack);
    DeclareDependency(subTrackClass, APPEARANCE, KEYFRAMES);
    DeclareDependency(subTrackClass, APPEARANCE, CHANGED);
    SetMethod(subTrackClass, APPEARANCE, MakeSubTrackAppearance);
    SetMethod(subTrackClass, DRAWTRACK, DrawSubTrackLine);
    SetMethod(subTrackClass, EXPRESS, ExpressDefaultSubTrack);
    SetMethod(subTrackClass, DELETE, DeleteSubTrack);
    SetVar(subTrackClass, NAME, NewString("SubTrack"));
    DefineFixedClass(subTrackClass, CLASS_SUBTRACK, "subTrack");

    keyFrameClass = NewObject(NULLOBJ, 0L);
    AddToReferenceList(keyFrameClass);
    SetMethod(keyFrameClass, DELETE, DeleteKeyFrame);
    SetVar(keyFrameClass, NAME, NewString("Keyframe"));
    DefineFixedClass(keyFrameClass, CLASS_KEYFRAME, "keyframe");

    DefineBetterVarName(ROTQUAT, "Rotation");
    DefineBetterVarName(FOCUSDIST, "Focus Distance");
    DefineBetterVarName(ISOVAL, "Isovalue");
}

#ifdef HAVE_PROTOTYPES
void KillAnimation(void)
#else
void KillAnimation()
#endif
/*Kills animation*/
{
    int k;
    AddToReferenceList(keyFrameClass);
    RemoveFromReferenceList(subTrackClass);
    RemoveFromReferenceList(trackClass);
    RemoveFromReferenceList(sequenceClass);
    for (k = 0; k < nBetterVarNames; ++k)
    {
	SAFEFREE(betterVarNames[k]);
    }
    SAFEFREE(betterVarNames);
}
