/*ScianDepend.c
  Eric Pepke
  May 22, 1991
  Dependencies system for SciAn
*/

#include "Scian.h"
#include "ScianTypes.h"
#include "ScianIDs.h"
#include "ScianErrors.h"
#include "ScianDepend.h"

char localTempStr[500];

Bool mustMake = false;

static void CallBack(obj, destVar, indVar, sourceVar)
ObjPtr obj;
NameTyp destVar, indVar, sourceVar;
{
    ObjPtr indObject;
    if (indVar)
    {
	SubMakeVar(obj, indVar);
	indObject = GetVar(obj, indVar);
	if (!indObject) return;
    }
    else
    {
	indObject = obj;
    }
    SubMakeVar(indObject, sourceVar);

    if (CompareVarChangeCounts(obj, destVar, indObject, sourceVar))
    {
	/*Yes, it needs to be made*/

#ifdef DEBUG
	    ObjPtr name;
	    char out[256];
	    name = GetVar(obj, NAME);

	    sprintf(out, "Need to make object %lx ", obj);
	    if (name)
	    {
		sprintf(localTempStr, "(%s) ", GetString(name));
		strcat(out, localTempStr);
	    }
	    sprintf(localTempStr, "variable %s ", IDName(destVar));
	    strcat(out, localTempStr);

	    sprintf(localTempStr, "due to change in %s", IDName(sourceVar));
	    strcat(out, localTempStr);

	    puts(out);
#endif
	mustMake = true;
    }
}

int testomatic (double);

#ifndef RELEASE
int nMainMakeVars = 0;
int nSubMakeVars = 0;
#endif

#ifdef HAVE_PROTOTYPES
Bool BasicMakeVar(ObjPtr origObj, NameTyp destVar)
#else
Bool BasicMakeVar(origObj, destVar)
ObjPtr origObj;
int destVar;
#endif
/*Makes destVar within origObj.  Returns true iff it had to make it.*/
{
    ObjPtr class, var;
    long initialCount;
    FuncTyp method;
    Bool retVal = false;
    Bool oldMustMake;

    oldMustMake = mustMake;
    mustMake = false;

    initialCount = GetVarChangeCount(origObj, destVar);
    GetDependencies(origObj, destVar, CallBack);
    retVal = initialCount == GetVarChangeCount(origObj, destVar) ? false : true;

    if (GetVarChangeCount(origObj, destVar) <= 0)
    {
#ifdef DEBUG
	    ObjPtr name;
	    char out[256];
	    name = GetVar(origObj, NAME);

	    sprintf(out, "Need to make object %lx ", origObj);
	    if (name)
	    {
		sprintf(localTempStr, "(%s) ", GetString(name));
		strcat(out, localTempStr);
	    }
	    sprintf(localTempStr, "variable %s ", IDName(destVar));
	    strcat(out, localTempStr);

	    sprintf(localTempStr, "first time");
	    strcat(out, localTempStr);

	    puts(out);
#endif
	mustMake = true;

    }

    if (mustMake)
    {
	/*Yes, it needs to be made*/
	FuncTyp method;

#ifdef DEBUG
	    ObjPtr name;
	    char out[256];
	    name = GetVar(origObj, NAME);

	    sprintf(out, "Making object %lx ", origObj);
	    if (name)
	    {
		sprintf(localTempStr, "(%s) ", GetString(name));
		strcat(out, localTempStr);
	    }
	    sprintf(localTempStr, "variable %s", IDName(destVar));
	    strcat(out, localTempStr);

	    puts(out);
#endif

	method = GetMethod(origObj, destVar);
	if (method)
	{
	    (*method)(origObj);
	}
	else
	{
	    SetVar(origObj, destVar, GetVar(origObj, destVar));
	    /*ReportError("MakeVar", "No make method found");*/
	}
    }

    mustMake = oldMustMake;

    return retVal;
}

#ifdef HAVE_PROTOTYPES
Bool MakeVar(ObjPtr origObj, NameTyp destVar)
#else
Bool MakeVar(origObj, destVar)
ObjPtr origObj;
int destVar;
#endif
/*Makes destVar within origObj.  Returns true iff it had to make it.*/
{
#ifndef RELEASE
    ++nMainMakeVars;
#endif

    BasicMakeVar(origObj, destVar);
}

#ifdef HAVE_PROTOTYPES
Bool SubMakeVar(ObjPtr origObj, NameTyp destVar)
#else
Bool SubMakeVar(origObj, destVar)
ObjPtr origObj;
int destVar;
#endif
/*Makes destVar within origObj.  Returns true iff it had to make it.*/
{
#ifndef RELEASE
    ++nSubMakeVars;
#endif

    BasicMakeVar(origObj, destVar);
}

#ifdef HAVE_PROTOTYPES
void DeclareDependency(ObjPtr object, NameTyp destVar, NameTyp sourceVar)
#else
void DeclareDependency(object, destVar, sourceVar)
ObjPtr object;
NameTyp destVar, sourceVar;
#endif
/*Declares a variable dependency within object of destVar on sourceVar*/
{
    SetDependency(object, destVar, sourceVar);
}

#ifdef HAVE_PROTOTYPES
void DeclareIndirectDependency(ObjPtr object, NameTyp destVar, NameTyp indirectObject, NameTyp sourceVar)
#else
void DeclareIndirectDependency(object, destVar, indirectObject, sourceVar)
ObjPtr object;
NameTyp destVar, indirectObject, sourceVar;
#endif
/*Declares a variable dependency within object of destVar on sourceVar
  within object indirectObject*/
{
    SetIndirectDependency(object, destVar, indirectObject, sourceVar);
}
