/*ScianFixedClasses.c
  Eric Pepke
  March 3, 1992

  Fixed classes for SciAn
*/

#include "Scian.h"
#include "ScianTypes.h"
#include "ScianIDs.h"

char *defaultClassNames[N_CLASS_IDS] = 
    {
	    "icon",
	    "corral",
	    "visWindow",
	    "light",
	    "clock",
	    "observer",
	    "renderer",
	    "timedObj",
	    "visObj",
	    "panel",
	    "dataset",
	    "file",
	    "textBox",
	    "space",
	    "dialog",
	    "fileReader",
	    "recorder",
	    "snapshot",
	    "sequence",
	    "menu",
	    "action",
	    "folder",
	    "fileType",
	    "unixFile",
	    "unixDir",
	    "palette"
    };

/*Classes, must be set by initialization routines*/
ObjPtr classClasses[N_CLASS_IDS];
char *classNames[N_CLASS_IDS];

#ifdef HAVE_PROTOTYPES
char *ClassIDToName(int ID)
#else
char *ClassIDToName(ID)
int ID;
#endif
/*Returns a name for ID or null*/
{
    if (ID < 0 || ID >= N_CLASS_IDS) return NULL;

    return classNames[ID];
}

#ifdef HAVE_PROTOTYPES
int ClassNameToID(char *name)
#else
int ClassNameToID(name)
char *name;
#endif
/*Returns an ID for name or -1*/
{
    int k;
    for (k = 0; k < N_CLASS_IDS; ++k)
    {
	if (classNames[k] && (0 == strcmp2(name, classNames[k])))
	{
	    return k;
	}
    }
    return -1;
}

#ifdef HAVE_PROTOTYPES
void DefineFixedClass(ObjPtr fixedClass, int classID, char *name)
#else
void DefineFixedClass(fixedClass, classID, name)
ObjPtr fixedClass;
int classID;
char *name;
#endif
/*Defines a fixed class with ID classID and name*/
{
    classClasses[classID] = fixedClass;

    SetVar(fixedClass, CLASSID, NewInt(classID));

    if (name)
    {
	SAFEFREE(classNames[classID]);
        classNames[classID] = Alloc(strlen(name) + 1);
	strcpy(classNames[classID], name);
    }
}

#ifdef HAVE_PROTOTYPES
ObjPtr GetFixedClassObject(int classID)
#else
ObjPtr GetFixedClassObject(classID)
int classID;
#endif
/*Return the class for a fixed class ID*/
{
    return classClasses[classID];
}

#ifdef HAVE_PROTOTYPES
void InitFixedClasses(void)
#else
void InitFixedClasses()
#endif
/*Initializes the fixed classes*/
{
    int k;
    for (k = 0; k < N_CLASS_IDS; ++k)
    {
	classClasses[k] = NULLOBJ;
	if (k < N_DEFAULT_CLASS_IDS)
	{
            classNames[k] = Alloc(strlen(defaultClassNames[k]) + 1);
	    strcpy(classNames[k], defaultClassNames[k]);
	}
	else
	{
	    classNames[k] = NULL;
	}
    }
}

#ifdef HAVE_PROTOTYPES
void KillFixedClasses(void)
#else
void KillFixedClasses()
#endif
/*Initializes the fixed classes*/
{
    int k;
    for (k = 0; k < N_DEFAULT_CLASS_IDS; ++k)
    {
	classClasses[k] = NULLOBJ;
        SAFEFREE(classNames[k]);
    }
}

