/*ScianLights.c
  Lights
  Eric Pepke
  May 21, 1991
*/

#include "Scian.h"
#include "ScianTypes.h"
#include "ScianColors.h"
#include "ScianLists.h"
#include "ScianArrays.h"
#include "ScianWindows.h"
#include "ScianSpaces.h"
#include "ScianDraw.h"
#include "ScianObjWindows.h"
#include "ScianVisWindows.h"
#include "ScianVisObjects.h"
#include "ScianIcons.h"
#include "ScianIDs.h"
#include "ScianControls.h"
#include "ScianButtons.h"
#include "ScianErrors.h"
#include "ScianDialogs.h"
#include "ScianSliders.h"
#include "ScianTextBoxes.h"
#include "ScianTitleBoxes.h"
#include "ScianStyle.h"
#include "ScianComplexControls.h"
#include "ScianLights.h"
#include "ScianMethods.h"
#include "ScianScripts.h"
#include "ScianEvents.h"
#include "ScianObjFunctions.h"
#include "ScianSymbols.h"
#include "ScianFontSystem.h"
#include "ScianColorControls.h"
#include "ScianTemplates.h"
#include "ScianTemplateHelper.h"
#include "ScianFixedClasses.h"

ObjPtr lightClass;			/*Class of individual light*/
int lightsNum = 0;
ObjPtr commonLights = NULLOBJ;	
Bool lastTwoSided = false;		/*Last change to twoSided*/
int nTotalLights = 0;

#define TOTALLIGHTS	8		/*Total number of possible lights*/
#define NLIGHTS		2		/*Number of lights available by default*/

lightSpec defLights[] =
    {
	{(M_PI_4 * 1.2), M_PI_4, {1.0, 1.0, 1.0}, 1.0, false},
	/*{M_PI_2, M_PI_4, {0.7, 1.0, 1.0}, 0.5, false},*/
	{-M_PI_4, - 0.7 * M_PI_4, {0.7, 1.0, 1.0}, 0.5, false},
	{M_PI, M_PI_4, {1.0, 0.0, 0.5}, 0.2, true},
	{0.0, M_PI_2, {1.0, 1.0, 0.6}, 0.3, true}
    };

#ifdef GRAPHICS
/*Don't change this without checking references*/
float lightModel[] = {
			AMBIENT, 0.2, 0.2, 0.2,
#ifdef TWOSIDE
			TWOSIDE, 0.0,
#endif
			LMNULL};
#endif

#ifdef HAVE_PROTOTYPES
void TwoSided(Bool whether)
#else
void TwoSided(whether)
Bool whether;
#endif
/*Goes to two sided lighting on whether*/
{
#ifdef GRAPHICS
    {
	if (whether != lastTwoSided)
	{
	    int k;
	    lastTwoSided = whether;
	    if (whether)
	    {
		/*Bind all the reverse lights*/
		for (k = 0; k < nTotalLights; ++k)
		{
		    if (rgbp)
		    {
			lmbind(LIGHT0 + k + nTotalLights, k * 2 + 2);
		    }
		}
	    }
	    else
	    {
		/*Unbind all the reverse lights*/
		for (k = 0; k < nTotalLights; ++k)
		{
		    if (rgbp)
		    {
		    	lmbind(LIGHT0 + k + nTotalLights, 0);
		    }
		}
	    }
	}
    }
#endif
}

#ifdef HAVE_PROTOTYPES
void StartLights(ObjPtr space, ObjPtr lights, Bool twoSided)
#else
void StartLights(space, lights, twoSided)
ObjPtr space, lights;
Bool twoSided;
#endif
/*Starts the lights*/
{
    ThingListPtr runner;
    int k;

#ifdef GRAPHICS
    if (rgbp)
    {
	lmbind(LMODEL, 1);
    }
#endif

    runner = LISTOF(lights);
    k = 0;
    while (runner)
    {
	ObjPtr light;
	ObjPtr var;
	real brightness;
	real baseColor[3];
	real location[3];
#ifdef GRAPHICS
	float tempLight[50];
	float *lightPtr;
#endif

	light = runner -> thing;

	if (k < TOTALLIGHTS && !GetPredicate(runner -> thing, HIDDEN))
	{
	
	    var = GetVar(light, BRIGHTNESS);
	    if (var && IsReal(var))
	    {
		brightness = GetReal(var);
	    }
	    else
	    {
		brightness = 1.0;
	    }
	    var = GetVar(light, BASECOLOR);
	    if (var)
	    {
		Array2CArray(baseColor, var);
	    }
	    else
	    {
		baseColor[0] = 1.0;
		baseColor[1] = 1.0;
		baseColor[2] = 1.0;
	    }
	    var = GetFixedArrayVar("StartLights", light, LOCATION, 1, 3L);
	    if (!var)
	    {
		return;
	    }
	    Array2CArray(location, var);

	    if (GetPredicate(light, LIGHTBYOBSERVER))
	    {
		ObjPtr observer;
		ObjPtr var;

		MakeVar(space, OBSERVER);
		observer = GetVar(space, OBSERVER);
		if (observer)
		{
		    var = GetVar(observer, ROTQUAT);
		    if (var)
		    {
			Quaternion qRot;
			 
			Array2CArray(qRot, var);
			QRot(location, qRot, location);
		    }
		}
	    }
			

#ifdef GRAPHICS
	    lightPtr = tempLight;
	    *lightPtr++ = POSITION;
	    *lightPtr++ = location[0];
	    *lightPtr++ = location[1];
	    *lightPtr++ = location[2];
	    *lightPtr++ = 0.0;
	    *lightPtr++ = LCOLOR;
	    *lightPtr++ = baseColor[0] * brightness;
	    *lightPtr++ = baseColor[1] * brightness;
	    *lightPtr++ = baseColor[2] * brightness;
	    *lightPtr++ = LMNULL; 

	    if (rgbp)
	    {
	        lmdef(DEFLIGHT, k * 2 + 1, 0, tempLight);

		lmbind(LIGHT0 + k, k * 2 + 1);
	    }

	    {
		/*Make the reverse light*/
		lightPtr = tempLight;
		*lightPtr++ = POSITION;
		*lightPtr++ = -location[0];
		*lightPtr++ = -location[1];
		*lightPtr++ = -location[2];
		*lightPtr++ = 0.0;
		*lightPtr++ = LCOLOR;
		*lightPtr++ = baseColor[0] * brightness;
		*lightPtr++ = baseColor[1] * brightness;
		*lightPtr++ = baseColor[2] * brightness;
		*lightPtr++ = LMNULL; 
		if (rgbp)
		{
		    lmdef(DEFLIGHT, k * 2 + 2, 0, tempLight);
		}
	    }
#endif
	    ++k;
	}
	runner = runner -> next;
    }
    nTotalLights = k;
}

void StopLights(space)
ObjPtr space;
/*Stops lights in space*/
{
    int k;

#ifdef GRAPHICS
    for (k = 0; k < TOTALLIGHTS; ++k)
    {
	if (rgbp)
	{
	    lmbind(LIGHT0 + k, 0);
	}
    }

    if (rgbp)
    {
	lmbind(LMODEL, 0);
    }
#endif
}

ObjPtr NewLight(k)
int k;
/*Returns a new light with default definition k*/
{
    ObjPtr light;
    ObjPtr baseColor;
    ObjPtr name;
    ObjPtr loc;
    real *locPtr;

    light = NewObject(lightClass, 0);

    /*Set light position*/
    loc = NewRealArray(1, 3L);
    locPtr = ArrayMeat(loc);
    *locPtr++ = rcos(defLights[k] . theta) * rcos(defLights[k] . phi);
    *locPtr++ = rsin(defLights[k] . theta) * rcos(defLights[k] . phi);
    *locPtr++ = rsin(defLights[k] . phi);
    SetVar(light, LOCATION, loc);

    /*Set light color*/
    baseColor = NewRealArray(1, 3L);
    CArray2Array(baseColor, defLights[k] . baseColor);
    SetVar(light, BASECOLOR, baseColor);

    /*Set light brightness*/
    SetVar(light, BRIGHTNESS, NewReal(defLights[k] . brightness));

    /*Set whether the light is enabled or not*/
    SetVar(light, HIDDEN, NewInt(defLights[k] . enabled));
    
    return light;
}

static ObjPtr MakeLightName(light)
ObjPtr light;
/*Makes the NAME of a light*/
{
    ObjPtr name;

    sprintf(tempStr, "Light %d", ++lightsNum);

    name = NewString(tempStr);
    SetVar(light, NAME, name);

    return ObjTrue;
}



ObjPtr NewLights()
/*Returns a new list of lights*/
{
    ObjPtr lights;
    int k;

    if (oneLights)
    {
	if (commonLights) return commonLights;
    }

    lights = NewList();
    for (k = 0; k < NLIGHTS; ++k)
    {
	ObjPtr light;

	light = NewLight(k);

	/*Set the light's parent*/
	SetVar(light, PARENT, lights);

	PostfixList(lights, light);
    }

    return lights;
}

static ObjPtr CloneLight(light)
ObjPtr light;
/*Clones a light light*/
{
    ObjPtr newLight;

    newLight = NewLight(0);
    CopyVar(newLight, light, LOCATION);
    CopyVar(newLight, light, BASECOLOR);
    CopyVar(newLight, light, BRIGHTNESS);
    CopyVar(newLight, light, HIDDEN);
    CopyVar(newLight, light, LIGHTBYOBSERVER);

    return newLight;
}

static ObjPtr BindLightToSpace(light, space)
ObjPtr light, space;
/*Makes space know about a light.*/
{
    ObjPtr lightsList;

    lightsList = GetVar(space, LIGHTS);
    if (!lightsList)
    {
	lightsList = NewList();
	SetVar(space, LIGHTS, lightsList);
    }
	
    PrefixList(lightsList, light);
    SetVar(light, PARENT, space);

    if (ListCount(lightsList) > TOTALLIGHTS)
    {
	char errmes[256];
	sprintf(errmes, "Only %d lights can be shown.\n", TOTALLIGHTS);
	ReportError("BindLightToSpace", errmes); 
    }
    return ObjTrue;
}

static ObjPtr HideLight(light)
ObjPtr light;
/*Hides a light*/
{
    SetVar(light, HIDDEN, ObjTrue);
    ImInvalid(light);
    return ObjTrue;
}

static ObjPtr ShowLight(light)
ObjPtr light;
/*Shows a light*/
{
    SetVar(light, HIDDEN, ObjFalse);
    ImInvalid(light);
    return ObjTrue;
}


static ObjPtr lightForRGB;

static void LightRGB(visWindow)
WinInfoPtr visWindow;
/*Sends a deferred message to visWindow if it has lightForRGB*/
{
    ObjPtr space;
    ObjPtr lights;

    space = FindSpace(visWindow);
    if (space)
    {
	MakeVar(space, LIGHTS);
	lights = GetVar(space, LIGHTS);
	if (lights)
	{
	    ThingListPtr runner;

	    runner = LISTOF(lights);
	    while (runner)
	    {
		if (runner -> thing == lightForRGB)
		{
		    DeferMessage((ObjPtr) visWindow, SETRGBMESSAGE);
		    return;
		}
		runner = runner -> next;
	    }
	}
    }
}

static ObjPtr ShowLightControls(object, windowName)
ObjPtr object;
char *windowName;
/*Makes a new control window to control light*/
{
    WinInfoPtr controlWindow;
    ObjPtr var;
    ObjPtr panel;
    ObjPtr contents;
    ObjPtr lights;
    ThingListPtr runner;
    ObjPtr firstIcon = 0;
    WinInfoPtr dialogExists;
    int k;
    int left, right, bottom, top, l, r, b, t;

    GetTemplateBounds(LightTemplate, "Panel", &left, &right, &bottom, &top);

    dialogExists = DialogExists((WinInfoPtr) object, NewString("Controls"));
    controlWindow = GetDialog((WinInfoPtr) object, NewString("Controls"), windowName, 
	right - left, top - bottom, right - left, top - bottom, WINUI + WINFIXEDSIZE);
    SetVar((ObjPtr) controlWindow, HELPSTRING,
	NewString("This window shows controls for a light source at infinity."));

#ifdef GRAPHICS
    if (hasRGB)
    {
	lightForRGB = object;
	ForAllVisWindows(LightRGB);
    }
#endif

    if (!dialogExists)
    {
	ObjPtr control, radio, swatch, titleBox, checkBox, corral, sw, textBox, slider, icon, colorObj;
	ObjPtr var, button;

	SetVar((ObjPtr) controlWindow, REPOBJ, object);

	/*Add in a panel*/
	panel = NewPanel(greyPanelClass, 0, right - left, 0, top - bottom);
	if (!panel)
	{
	    return ObjFalse;
	}
	contents = GetVar((ObjPtr) controlWindow, CONTENTS);
	PrefixList(contents, panel);
	SetVar(panel, PARENT, (ObjPtr) controlWindow);

	contents = NewList();
	SetVar(panel, CONTENTS, contents);

	left += MINORBORDER;
	right -= MINORBORDER;
	bottom += MINORBORDER;
	top -= MINORBORDER;

	l = left;
	t = top;
	r = l + 2 * MINORBORDER + MAJORBORDER + COLORWHEELWIDTH + SLIDERWIDTH;
	b = t - TITLEBOXTOP - COLORWHEELWIDTH - TEXTBOXSEP - TEXTBOXHEIGHT - SWATCHHEIGHT - 3 * MINORBORDER;

	/*Make the title box*/

	titleBox = NewTitleBox(l, r, b, t, "Light Color");
	PrefixList(contents, titleBox);
	SetVar(titleBox, PARENT, contents);
    
	b = t - TITLEBOXTOP;
    
	t = b - MINORBORDER;
	b = t - SWATCHHEIGHT;
	l = l + MINORBORDER;
	r = r - MINORBORDER;
     
   
	swatch = NewDirectSwatch(l, r, b, t, "Light Color Swatch", object, BASECOLOR);
	PrefixList(contents, swatch);
	SetVar(swatch, PARENT, contents);

	/*Create the color control*/
	r = l + COLORWHEELWIDTH;
	t = b - MINORBORDER;;
	b = t - COLORWHEELWIDTH;
 	control = NewColorWheel(l, r, b, t, "Color");
	SetVar(control, HELPSTRING, NewString("This color wheel controls the base color of a light.  \
It will only have an effect if the window containing the light is in full color mode."));
	PrefixList(contents, control);
	SetVar(control, PARENT, panel);
	AssocColorControlWithVar(control, object, BASECOLOR);

	var = GetFixedArrayVar("ShowLightControls", object, BASECOLOR, 1, 3L);
	if (!var)
	{
	    real *baseColor;

	    var = NewRealArray(1, 3L);
	    baseColor = ELEMENTS(var);

	    baseColor[0] = 1.0;
	    baseColor[1] = 1.0;
	    baseColor[2] = 1.0;
	    SetVar(object, BASECOLOR, var);
	}

        l = r + MAJORBORDER;
	r = l + SLIDERWIDTH;

	/*Create the brightness slider*/
	var = GetRealVar("ShowLightControls", object, BRIGHTNESS);
	if (!var)
	{
	    SetVar(object, BRIGHTNESS, NewReal(1.0));
	}
	slider = NewSlider(l, r, b, t,
		       PLAIN, "Brightness");
	SetVar(slider, HELPSTRING, NewString("This slider controls the brightness \
of the light.  It will only have an effect if the window containing the light is in full color mode."));
	if (!slider)
	{
	    return ObjFalse;
	}
	PrefixList(contents, slider);
	SetVar(slider, PARENT, panel);
	SetSliderRange(slider, 1.0, 0.0, 0.0);
	AssocDirectControlWithVar(slider, object, BRIGHTNESS);

	/*Create the text box*/
	t = b - TEXTBOXSEP;
	b = t - TEXTBOXHEIGHT;
    
	l =  left + MINORBORDER;
	r = l + COLORWHEELWIDTH;
    
	textBox = NewTextBox(l, r, b, t,
			 0, "Color Text", "Color");
	PrefixList(contents, textBox);
	SetVar(textBox, PARENT, panel);
	SetTextAlign(textBox, CENTERALIGN);

	l = r + MAJORBORDER;
	r = l + SLIDERWIDTH;
     
	/*Create the brightness text box*/
	textBox = NewTextBox(l - 40, r + 40, b, t,
			 0, "Brightness Text", "Brightness");
	PrefixList(contents, textBox);
	SetVar(textBox, PARENT, panel);
	SetTextAlign(textBox, CENTERALIGN);

	/*Create the radio buttons*/
	radio = NewRadioButtonGroup("Anchor Radio");
	SetVar(radio, PARENT, panel);
	PrefixList(contents, radio);

	button = TemplateRadioButton(LightTemplate, "Anchor to Space");
	AddRadioButton(radio, button);
	button = TemplateRadioButton(LightTemplate, "Anchor to Observer");
	AddRadioButton(radio, button);

	AssocDirectControlWithVar(radio, object, LIGHTBYOBSERVER);
    }

    return (ObjPtr) controlWindow;
}

#define LIGHTDSPSTEP 0.1  /*Step to display outward for lights*/
static real globalLightRadius;

ObjPtr DrawLight(light)
ObjPtr light;
/*Draws a light in a space*/
{
    float lightrad = 1.0;
    if (IsSelected(light))
    {
	ObjPtr var;
	real brightness;
	real baseColor[3];
	real location[3];
	float clr[3];
	real azimuth;
	real xyrange;
	real elevation;
    
	lightrad += LIGHTDSPSTEP;

	var = GetVar(light, BRIGHTNESS);
	if (var && IsReal(var))
	{
	    brightness = GetReal(var);
	}
	else
	{
	    brightness = 1.0;
	}
	var = GetVar(light, BASECOLOR);
	if (var)
	{
	    Array2CArray(baseColor, var);
	}
	else
	{
	    baseColor[0] = 1.0;
	    baseColor[1] = 1.0;
	    baseColor[2] = 1.0;
	}
	var = GetFixedArrayVar("DrawLights", light, LOCATION, 1, 3L);
	if (!var)
	{
	    return;
	}
	Array2CArray(location, var);


	if (GetPredicate(light, LIGHTBYOBSERVER))
	{
	    ObjPtr observer;
	    ObjPtr var;

	    MakeVar(curSpace, OBSERVER);
	    observer = GetVar(curSpace, OBSERVER);
	    if (observer)
	    {
		var = GetVar(observer, ROTQUAT);
		if (var)
		{
		    Quaternion qRot;
		     
		    Array2CArray(qRot, var);
		    QRot(location, qRot, location);
		}
	    }
	}
			
	/*Calculate the azimuth*/
	if (location[0] == 0.0 && location[1] == 0.0)
	{
	    azimuth = 0.0;
	}
	else
	{
	    azimuth = atan2(location[1], location[0]);
	}

	/*Calculate the xyrange and elevation*/
	xyrange = (float) sqrt(location[0] * location[0] + location[1] * location[1]);
	elevation = atan2(location[2], xyrange);
	if (elevation >= M_PI) elevation -= 2 * M_PI;

	SetLineWidth(2);
	SetUIColor(UIGRAY50);

	/*Draw the horizon circle*/
	{
	    NudgeFarther();
	    NudgeFarther();
#ifdef GRAPHICS
	    circ(0.0, 0.0, globalLightRadius);
#endif
	    NudgeCloser();
	    NudgeCloser();
	}

	/*Draw the X space line*/
	SetUIColor(UIWHITE);
	DrawSpaceLine(0.0, 0.0, 0.0, lightrad * globalLightRadius, 0.0, 0.0);

	/*Choose a color for subsequent stuff*/
#ifdef GRAPHICS
	clr[0] = baseColor[0] * (brightness * 0.8 + 0.2);
	clr[1] = baseColor[1] * (brightness * 0.8 + 0.2);
	clr[2] = baseColor[2] * (brightness * 0.8 + 0.2);
	if (rgbp)
	{
	    c3f(clr);
	}
	else
	{
	    color(uiColorIndex[ClosestUIColor(clr)]);
	}
#endif

	/*Draw the azimuth measurement*/
	sprintf(tempStr, "%4.0f deg.", (float) azimuth * 180.0 / M_PI);
	SetUIFont(UIBOLDLARGEFONT);
	DrawSpaceString((lightrad + 0.05) * globalLightRadius * rcos(azimuth * 0.5),
			(lightrad + 0.05) * globalLightRadius * rsin(azimuth * 0.5),
			0.0,
			tempStr);

#ifdef GRAPHICS
	/*Draw the azimuth circle*/
	if (azimuth < 0.0)
	{
	    arc(0.0, 0.0, lightrad * globalLightRadius, azimuth * 1800.0 / M_PI, 0);
	}
	else
	{
	    arc(0.0, 0.0, lightrad * globalLightRadius, 0, azimuth * 1800.0 / M_PI);
	}
#endif

	/*Draw the azimuth line*/
	DrawSpaceLine(0.0, 0.0, 0.0, lightrad * globalLightRadius * rcos(azimuth), lightrad * globalLightRadius * sin(azimuth), 0.0);

#ifdef GRAPHICS
	/*Draw the elevation arc*/
	pushmatrix();
	rotate(900, 'x');
	rot(azimuth * 180.0 / M_PI, 'y');
	if (elevation < 0.0)
	{
	    arc(0.0, 0.0, lightrad * globalLightRadius, elevation * 1800.0 / M_PI, 0);
	}
	else
	{
	    arc(0.0, 0.0, lightrad * globalLightRadius, 0, elevation * 1800.0 / M_PI);
	}
#endif

	/*Draw the elevation measurement*/
	sprintf(tempStr, "%4.0f deg.", (float) elevation * 180.0 / M_PI);
	SetUIFont(UIBOLDLARGEFONT);
	DrawSpaceString((lightrad + 0.05) * globalLightRadius * rcos(elevation * 0.5),
			(lightrad + 0.05) * globalLightRadius * sin(elevation * 0.5),
			0.0,
			tempStr);

#ifdef GRAPHICS
	popmatrix();
#endif

	SetLineWidth(3);

	DrawSpaceLine(0.0, 0.0, 0.0, 30.0 * location[0],
	     30.0 * location[1],
	     30.0 * location[2]);
	SetLineWidth(1);
    }
}	

#ifdef HAVE_PROTOTYPES
void DrawLights(ObjPtr space, ObjPtr lights, int action, real radius)
#else
void DrawLights(space, lights, action, radius)
ObjPtr space, lights;
int action;
real radius;
#endif
/*Draw lights in space if they want to be displayed
  Action is DRAWSPACE, PICKSPACE, etc.
  Radius is radius of sphere in center area*/
{
#ifdef GRAPHICS
    ThingListPtr runner;
 
    zbuffer(FALSE);
    globalLightRadius = radius;
    runner = LISTOF(lights);
    while (runner)
    {
	ObjPtr light;

	DrawObject(runner -> thing);
	runner = runner -> next;
    }

    zbuffer(TRUE);
#endif
}

static ObjPtr DeleteLight(object)
ObjPtr object;
/*Deletes an object from the current window.*/
{
    ObjPtr deleteObject;

    if (!selWinInfo || !object) return ObjFalse;
    deleteObject = ObjectWhichRepresents(selWinInfo, object);
    Select(object, false);
    if (deleteObject)
    {
	ObjPtr space;
	ObjPtr lights;
	ObjPtr parent, newList;
	ObjPtr contentsList;

	space = FindSpace(selWinInfo);
	if (space)
	{
	    MakeVar(space, LIGHTS);
	    lights = GetListVar("DeleteLights", space, LIGHTS);
	    if (lights)
	    {
		SaveVarSnapshotForUndo(space, LIGHTS);
		newList = RemoveFromList(lights, object);
		SetVar(newList, PARENT, GetVar(lights, PARENT));
		SetVar(space, LIGHTS, newList);
	    }
	}

	object = deleteObject;

	contentsList = GetVar(object, PARENT);
	if (!contentsList || !IsList(contentsList))
	{
	    contentsList = GetVar(contentsList, CONTENTS);
	}
	if (contentsList && IsList(contentsList))
	{

	    parent = GetVar(contentsList, PARENT);
	    if (parent)
	    {
		SaveVarSnapshotForUndo(parent, CONTENTS);
		newList = RemoveFromList(contentsList, object);
		SetVar(newList, PARENT, parent);
		SetVar(parent, CONTENTS, newList);
	    }
	    else
	    {
		ReportError("DeleteObject", "No parent");
		DeleteFromList(contentsList, object);
	    }
	    return ObjTrue;
	}
	else
	{
	    ReportError("DeleteObject","Internal error: cannot find contents list");
	}
    }
    return ObjFalse;
}

static MakeMethod MakeLightAppearance(light)
ObjPtr light;
/*Dummy make for light appearance*/
{
    SetVar(light, APPEARANCE, ObjTrue);
    ImInvalid(light);
    return ObjTrue;
}

void InitLights()
{
    ObjPtr icon;
    ObjPtr list;
#ifdef GRAPHICS
    Object o;
    float v[3], n[3];
    short c[3];
#endif

#ifdef GRAPHICS
    lmdef(DEFLMODEL, 1, 0, lightModel);
#endif

    /*Create class of lights*/
    lightClass = NewObject(controllerClass, 0);
    AddToReferenceList(lightClass);
    SetMethod(lightClass, NAME, MakeLightName);
    SetMethod(lightClass, BINDTOSPACE, BindLightToSpace);
    SetMethod(lightClass, NEWCTLWINDOW, ShowLightControls);
    SetVar(lightClass, DOUBLECLICK, NewString(OF_SHOW_CONTROLS));
    SetMethod(lightClass, HIDE, HideLight);
    SetMethod(lightClass, SHOW, ShowLight);
    list = NewList();
    PrefixList(list, NewSymbol(HIDDEN));
    SetVar(lightClass, SNAPVARS, list);
    SetMethod(lightClass, CLONE, CloneLight);
    SetMethod(lightClass, DRAW, DrawLight);
    DefineFixedClass(lightClass, CLASS_LIGHT, "light");
    DeclareDependency(lightClass, APPEARANCE, SELECTED);
    SetMethod(lightClass, APPEARANCE, MakeLightAppearance);
    SetVar(lightClass, LIGHTBYOBSERVER, NewInt(0));
    AddSnapVar(lightClass, LIGHTBYOBSERVER);
    AddSnapVar(lightClass, BRIGHTNESS);
    AddSnapVar(lightClass, BASECOLOR);
    AddSnapVar(lightClass, LOCATION);

    icon = NewIcon(0, 0, ICONLIGHTS, "Light");
#ifdef GRAPHICS
#include "ScianGeoLight.h"
    SetVar(icon, CACHEGRAPHICS, NewInt(o));
#endif

    SetVar(icon, HELPSTRING,
	NewString("This icon represents a light.  When this icon is selected, \
a ray showing the direction and color of the light will appear in the space.  \
You can rotate the light by pressing and dragging in the space with the center \
mouse button.  You can show the controls for the light, such as the brightness and \
color, by selecting it and choosing the Show Controls item in the Object menu.\n\
\nYou can drag the light icon to the corral of another visualization \
window to have it illuminate that space as well.  A space can hold up to eight lights.  \
The easiest way to add additional lights is to select a light and choose Duplicate from \
the Object menu.")); 
    SetMethod(lightClass, DUPLICATE, DuplicateSpaceObject);
    SetMethod(lightClass, DELETE, DeleteLight);
    SetVar(lightClass, DEFAULTICON, icon);
    SetVar(lightClass, LIGHTBYOBSERVER, NewInt(0));
}

void KillLights()
{
    DeleteThing(lightClass);
}
