/*ScianPostScript.c
  Eric Pepke, with Jim Lyons as "Jim"
  June 4, 1993
  PostScript pseudorecorder*/

#include "Scian.h"
#include "ScianTypes.h"
#include "ScianArrays.h"
#include "ScianFontSystem.h"
#include "ScianLists.h"
#include "ScianWindows.h"
#include "ScianObjWindows.h"
#include "ScianIcons.h"
#include "ScianControls.h"
#include "ScianButtons.h"
#include "ScianDraw.h"
#include "ScianIDs.h"
#include "ScianPostScript.h"
#include "ScianRecorders.h"
#include "ScianStyle.h"
#include "ScianTemplates.h"
#include "ScianTemplateHelper.h"
#include "ScianTextBoxes.h"

extern short LinToGamma[];

static long curFrame = 0;

/*** IDs change ***/
#ifndef PAGEWIDTH
#define PAGEWIDTH   IMGWIDTH
#define PAGEHEIGHT  IMGHEIGHT
#endif

/* globals used for PostScript drawing */
FILE *psFile;		/* pointer to temp file */
int iconsUsed[32];	/* maximum number of *different* icons in PS file */
int numIconsUsed;	/* number used so far = next index to use */

static ObjPtr ReturnTrue(object)
ObjPtr object;
/*Returns true*/
{
    return ObjTrue;
}

static ObjPtr PrepareToRecordPostScript(object, nFrames)
ObjPtr object;
long nFrames;
/*Prepares to record nframes*/
{
    curFrame = 0;
    return ObjTrue;
}

static ObjPtr SnapOneFramePostScript(object)
ObjPtr object;
/*Snap a single PostScript frame*/
{
    int l, b, r, t,  w, h;
    ObjPtr var;
    ObjPtr windows;
    ThingListPtr runner;
    Bool eps, saveWindowFrame;

    saveWindowFrame = GetPredicate(object, SAVEWINDOWFRAME);

    var = GetVar(object, PSFORMAT);
    if (var)
    {
	eps = (GetInt(var) == PS_EPS);
    }
    else
    {
	eps = false;
    }
    
    var = GetVar(object, FRAMEOX);
    if (var)
    {
	l = GetInt(var);
    }
    else
    {
	l = 0;
    }

    var = GetVar(object, FRAMEOY);
    if (var)
    {
	b = GetInt(var);
    }
    else
    {
	b = 0;
    }

    var = GetVar(object, FRAMEWIDTH);
    if (var)
    {
	w = GetInt(var);
    }
    else
    {
	w = 0;
    }

    var = GetVar(object, FRAMEHEIGHT);
    if (var)
    {
	h = GetInt(var);
    }
    else
    {
	h = 0;
    }

    r = l + w - 1;
    t = b + h - 1;

    if (saveWindowFrame)
    {
		l -= WINBL;
		r += WINBR;
		b -= WINBB;
		t += WINBT;
    }

    do
    {
	++curFrame;
	sprintf(tempStr, eps ? "screen%04d.eps" : "screen%04d.ps" , curFrame);
    } while (fopen(tempStr, "r"));

    SetVar(object, FILENAME, NewString(tempStr));

    drawFile = fopen(tempStr, "w");
    if (drawFile)
    {
	if (saveWindowFrame)
	{
	    l -= WINFRAMEWID;
	    r += WINFRAMEWID;
	    b -= WINFRAMEWID;
	    t += WINFRAMEWID + WINTITLEHT;
	}

	drawingMode = DRAW_POSTSCRIPT;
	BeginDrawing(l, r, b, t);

	windows = GetVar(object, RECORDWINDOWS);
	if (windows)
	{
	    runner = LISTOF(windows);
	    while(runner)
	    {
		WinInfoPtr window;
		int ox, oy;
		int wl, wr, wb, wt;
		int sx, sy;

		window = (WinInfoPtr) (runner -> thing);

		SelWindow(window);

		GetWindowOrigin(&ox, &oy);
		GetWindowBounds(&wl, &wr, &wb, &wt);

		SetDrawingWindow(window);
		sx = ox - l;
		sy = oy - b;

		BeginDrawSubscreen(sx, sy);
		DrawWindowInfo(window);
		EndDrawSubscreen(sx, sy);

		RestoreDrawingWindow();

		if (saveWindowFrame)
		{
		    BeginDrawSubscreen(sx, sy);
		    DrawPSFrame(window -> winTitle, 0, wr - wl, 0, wt - wb);
		    EndDrawSubscreen(sx, sy);
		}

		runner = runner -> next;
	    }
	}
	EndDrawing(!eps);
	drawingMode = DRAW_SCREEN;
	fclose(drawFile);
    }

    return ObjTrue;
}

static ObjPtr AddPostScriptControls(object, panelContents)
ObjPtr object, panelContents;
/*Adds controls appropriate to a PostScript object to panelContents*/
{
    ObjPtr titleBox, button, radio, var;
    ObjPtr textBox, checkBox;

/* Image Quality */
    /*Make title box for quality*/
    titleBox = TemplateTitleBox(PostScriptTemplate, "Quality");
    PrefixList(panelContents, titleBox);
    SetVar(titleBox, PARENT, panelContents);

    /*Make quality radio button*/
    radio = NewRadioButtonGroup("Image Quality");
    PrefixList(panelContents, radio);
    SetVar(radio, PARENT, panelContents);
    SetVar(radio, HELPSTRING, NewString("This radio button group sets the quality of \
the image portions of the PostScript file.  Lower numbers give lower quality with a substantial \
savings in file size."));

    /*Make the radios*/
    button = TemplateRadioButton(PostScriptTemplate, "Full Quality");
    AddRadioButton(radio, button);
    SetVar(button, HELPSTRING, NewString("When this radio button is on, the \
image portions of the PostScript file are saved at full quality."));

    button = TemplateRadioButton(PostScriptTemplate, "1/2 Quality");
    AddRadioButton(radio, button);
    SetVar(button, HELPSTRING, NewString("When this radio button is on, the \
image portions of the PostScript file are saved at half quality, resulting in file \
1/4 as large."));

    button = TemplateRadioButton(PostScriptTemplate, "1/3 Quality");
    AddRadioButton(radio, button);
    SetVar(button, HELPSTRING, NewString("When this radio button is on, the \
image portions of the PostScript file are saved at a third quality, resulting in file \
1/9 as large."));

    button = TemplateRadioButton(PostScriptTemplate, "1/4 Quality");
    AddRadioButton(radio, button);
    SetVar(button, HELPSTRING, NewString("When this radio button is on, the \
image portions of the PostScript file are saved at a quarter quality, resulting in file \
1/16 as large."));

    AssocDirectControlWithVar(radio, object, IMGQUAL);

/* Image Color */
    /* Title Box */
    titleBox = TemplateTitleBox(PostScriptTemplate, "Color");
    PrefixList(panelContents, titleBox);
    SetVar(titleBox, PARENT, panelContents);

    /*Make image type radio button*/
    radio = NewRadioButtonGroup("Image Color");
    PrefixList(panelContents, radio);
    SetVar(radio, PARENT, panelContents);
    SetVar(radio, HELPSTRING, NewString("This radio button group sets the type of \
image file generated."));

    /*Make the radios*/
    button = TemplateRadioButton(PostScriptTemplate, "8-bit Gray");
    AddRadioButton(radio, button);
    SetVar(button, HELPSTRING, NewString("When this radio button is on, the \
picture generated is an 8-bit grayscale image."));

    button = TemplateRadioButton(PostScriptTemplate, "12-bit Color");
    AddRadioButton(radio, button);
    SetVar(button, HELPSTRING, NewString("When this radio button is on, the \
picture generated is a color image with 4 bits per component."));

    button = TemplateRadioButton(PostScriptTemplate, "24-bit Color");
    AddRadioButton(radio, button);
    SetVar(button, HELPSTRING, NewString("When this radio button is on, the \
picture generated is a color image with 8 bits per component."));

    AssocDirectControlWithVar(radio, object, IMGTYPE); 

/* Scaling */
    /*Make image scaling textbox*/
    textBox = TemplateTextBox(PostScriptTemplate, "Scale image by",  ONE_LINE,
		"Scale image by");
    PrefixList(panelContents, textBox);
    SetVar(textBox, PARENT, panelContents);
    SetTextAlign(textBox, RIGHTALIGN);

    textBox = TemplateTextBox(PostScriptTemplate, "Scaling",  EDITABLE + WITH_PIT + ONE_LINE,
		"");
    PrefixList(panelContents, textBox);
    SetVar(textBox, PARENT, panelContents);
    SetTextAlign(textBox, RIGHTALIGN);
    AssocTextRealControlWithVar(textBox, object, IMGSCALING, 1.0, 1000.0, 0);
    SetVar(textBox, HELPSTRING, NewString("Enter into this text box the amount \
to enlarge or reduce the image.  100% is full size."));

    textBox = TemplateTextBox(PostScriptTemplate, "%", ONE_LINE, "%");
    PrefixList(panelContents, textBox);
    SetVar(textBox, PARENT, panelContents);

/* Format */
    /* Title Box */
    titleBox = TemplateTitleBox(PostScriptTemplate, "Format");
    PrefixList(panelContents, titleBox);
    SetVar(titleBox, PARENT, panelContents);

    /*Make format radio group*/
    radio = NewRadioButtonGroup("Image Format");
    PrefixList(panelContents, radio);
    SetVar(radio, PARENT, panelContents);
    SetVar(radio, HELPSTRING, NewString("This radio button group sets the format \
of the generated file to EPS or PostScript."));

    /*Make the radios*/
    button = TemplateRadioButton(PostScriptTemplate, "EPS");
    AddRadioButton(radio, button);
    SetVar(button, HELPSTRING, NewString("When this radio button is on, the \
file generated is an Encapsulated PostScript file. Use this setting when the \
file is to be embedded in another PostScript document."));
    
    button = TemplateRadioButton(PostScriptTemplate, "PostScript ");
    AddRadioButton(radio, button);
    SetVar(button, HELPSTRING, NewString("When this radio button is on, the \
file generated is a PostScript file. Use this setting when the \
file is to be printed directly."));

    AssocDirectControlWithVar(radio, object, PSFORMAT);

/* Orientation */
    /* Title Box */
    titleBox = TemplateTitleBox(PostScriptTemplate, "Orientation");
    PrefixList(panelContents, titleBox);
    SetVar(titleBox, PARENT, panelContents);

    /*Make orientation radio group*/
    radio = NewRadioButtonGroup("Image Orientation");
    PrefixList(panelContents, radio);
    SetVar(radio, PARENT, panelContents);
    SetVar(radio, HELPSTRING, NewString("This radio button group determines \
the image's orientation on the page. This setting only applies if Format is \
set to 'PostScript'."));

    /*Make the radios*/
    button = TemplateRadioButton(PostScriptTemplate, "Portrait");
    AddRadioButton(radio, button);
    SetVar(button, HELPSTRING, NewString("When this radio button is on, the \
image will be printed without rotation."));
    
    button = TemplateRadioButton(PostScriptTemplate, "Landscape");
    AddRadioButton(radio, button);
    SetVar(button, HELPSTRING, NewString("When this radio button is on, the \
image will be rotated 90 degrees counterclockwise before printing."));

    AssocDirectControlWithVar(radio, object, ORIENTATION);

/* Placement */
    /* Title Box */
    titleBox = TemplateTitleBox(PostScriptTemplate, "Placement");
    PrefixList(panelContents, titleBox);
    SetVar(titleBox, PARENT, panelContents);

    /*Make placement radio group*/
    radio = NewRadioButtonGroup("Image Placement");
    PrefixList(panelContents, radio);
    SetVar(radio, PARENT, panelContents);
    SetVar(radio, HELPSTRING, NewString("This radio button group determines \
the image's placement on the page. This setting only applies if Format is set \
to 'PostScript'."));

    /*Make the radios*/
    button = TemplateRadioButton(PostScriptTemplate, "Center on Page");
    AddRadioButton(radio, button);
    SetVar(button, HELPSTRING, NewString("When this radio button is on, the \
image will be printed centered on the page. Enter the page dimensions in \
the text boxes."));
    
    button = TemplateRadioButton(PostScriptTemplate, "Offset");
    AddRadioButton(radio, button);
    SetVar(button, HELPSTRING, NewString("When this radio button is on, the \
image will be offset from the bottom left of the page by the amounts \
entered in the text boxes."));

    AssocDirectControlWithVar(radio, object, PLACEMENT);

    /* Make the text boxes */
    textBox = TemplateTextBox(PostScriptTemplate, "Width",
	EDITABLE + WITH_PIT + ONE_LINE, "8.5");
    PrefixList(panelContents, textBox);
    SetVar(textBox, PARENT, panelContents);
    SetTextAlign(textBox, RIGHTALIGN);
    SetVar(textBox, HELPSTRING, NewString("Enter into this box the width of \
the page on which the image will be centered. Dimensions must be in inches in \
this version of SciAn."));

    AssocTextRealControlWithVar(textBox, object, PAGEWIDTH, 0.0, 24.0, 0);

    textBox = TemplateTextBox(PostScriptTemplate, "W", ONE_LINE, "W");
    PrefixList(panelContents, textBox);
    SetVar(textBox, PARENT, panelContents);
    
    textBox = TemplateTextBox(PostScriptTemplate, "Height",
	EDITABLE + WITH_PIT + ONE_LINE, "11");
    PrefixList(panelContents, textBox);
    SetVar(textBox, PARENT, panelContents);
    SetTextAlign(textBox, RIGHTALIGN);
    SetVar(textBox, HELPSTRING, NewString("Enter into this box the height of \
the page on which the image will be centered. Dimensions must be in inches in \
this version of SciAn."));
    
    AssocTextRealControlWithVar(textBox, object, PAGEHEIGHT, 0.0, 72.0, 0);

    textBox = TemplateTextBox(PostScriptTemplate, "H", ONE_LINE, "H");
    PrefixList(panelContents, textBox);
    SetVar(textBox, PARENT, panelContents);
    
    textBox = TemplateTextBox(PostScriptTemplate, "Left Offset",
	EDITABLE + WITH_PIT + ONE_LINE, "1");
    PrefixList(panelContents, textBox);
    SetVar(textBox, PARENT, panelContents);
    SetTextAlign(textBox, RIGHTALIGN);
    SetVar(textBox, HELPSTRING, NewString("Enter into this box the offset from \
the left edge of the page to the image. Dimensions must be in inches in \
this version of SciAn."));
    
    AssocTextRealControlWithVar(textBox, object, LEFTOFFSET, 0.0, 24.0, 0);

    textBox = TemplateTextBox(PostScriptTemplate, "L", ONE_LINE, "from Left");
    PrefixList(panelContents, textBox);
    SetVar(textBox, PARENT, panelContents);
    
    textBox = TemplateTextBox(PostScriptTemplate, "Bottom Offset",
	EDITABLE + WITH_PIT + ONE_LINE, "2");
    PrefixList(panelContents, textBox);
    SetVar(textBox, PARENT, panelContents);
    SetTextAlign(textBox, RIGHTALIGN);
    SetVar(textBox, HELPSTRING, NewString("Enter into this box the offset from \
the bottom of the page to the image. Dimensions must be in inches in \
this version of SciAn."));
    
    AssocTextRealControlWithVar(textBox, object, BOTTOMOFFSET, 0.0, 72.0, 0);

    textBox = TemplateTextBox(PostScriptTemplate, "B", ONE_LINE, "from Bottom");
    PrefixList(panelContents, textBox);
    SetVar(textBox, PARENT, panelContents);
    
    return ObjTrue;
}

#ifdef HAVE_PROTOTYPES
void InitPostScript(void)
#else
void InitPostScript()
#endif
/*Create a new PostScript ersatz recorder*/
{
    ObjPtr recorder, icon;

    recorder = NewRecorder(NULLOBJ, "PostScript", "Adobe");
    SetMethod(recorder, CONNECT, ReturnTrue);
    SetMethod(recorder, DISCONNECT, ReturnTrue);
    SetMethod(recorder, PREPARETORECORD, PrepareToRecordPostScript);
    SetMethod(recorder, STOPRECORDING, ReturnTrue);
    SetMethod(recorder, SNAPONEFRAME, SnapOneFramePostScript);
    SetVar(recorder, FRAMESOURCE, NewInt(FS_WINDOW));
    SetVar(recorder, ENABLEDSOURCES, NewInt(FB_WINDOW));
    SetVar(recorder, IMGQUAL, NewInt(0));
    AddSnapVar(recorder, IMGQUAL);
    SetVar(recorder, IMGTYPE, NewInt(PS_GRAYSCALE));
    AddSnapVar(recorder, IMGTYPE);
    SetVar(recorder, PSFORMAT, NewInt(PS_EPS));
    AddSnapVar(recorder, PSFORMAT);
    SetVar(recorder, ORIENTATION, NewInt(PS_PORTRAIT));
    AddSnapVar(recorder, ORIENTATION);
    SetVar(recorder, PLACEMENT, NewInt(PS_CENTERED));
    AddSnapVar(recorder, PLACEMENT);
    SetVar(recorder, LEFTOFFSET, NewReal(1.0));
    AddSnapVar(recorder, LEFTOFFSET);
    SetVar(recorder, BOTTOMOFFSET, NewReal(2.0));
    AddSnapVar(recorder, BOTTOMOFFSET);
    SetVar(recorder, PAGEWIDTH, NewReal(8.5));
    AddSnapVar(recorder, PAGEWIDTH);
    SetVar(recorder, PAGEHEIGHT, NewReal(11.0));
    AddSnapVar(recorder, PAGEHEIGHT);
    SetVar(recorder, IMGSCALING, NewReal(50.0));
    AddSnapVar(recorder, IMGSCALING);

    SetMethod(recorder, ADDCONTROLS, AddPostScriptControls);
    icon = NewIcon(0, 0, ICONRECORDER, "PostScript");
    SetVar(recorder, CONTROLICON, icon);
    SetVar(icon, PANELHELP, NewString("This control panel lets you edit \
the parameters of the PostScript driver,  such as file format and image size and \
orientation. These parameters apply to files generated by the pseudorecorder as \
well as files generated by the Screen Save function when 'PostScript' is chosen \
as the Screen Save format in the Preferences window."));

    RegisterRecorder(recorder);
}

#ifdef HAVE_PROTOTYPES
void KillPostScript(void)
#else
void KillPostScript()
#endif
{
}

#ifdef HAVE_PROTOTYPES
void SavePSImage(FILE *psFile, RGBAPixel *imageBuffer, int left, int right, int bottom, int top)
#else
void SavePSImage(psFile, imageBuffer, left, right, bottom, top)
FILE *psFile;
RGBAPixel *imageBuffer;
int left;
int right;
int bottom;
int top;
#endif
/*Saves an image in imageBuffer within left, right, bottom, top into psFile*/
{
    real y, dummy;
    int iVal;
    long k;
    int pixWrap;
    long nPixels;
    long i, j, ip, jp;
    long width, height;
    int subSample, imgType;
    ObjPtr var;

    var = GetVar(curRecorder, IMGQUAL);
    if (var)
    {
	subSample = GetInt(var) + 1;
    }
    else
    {
	subSample = 1;
    }
    
    var = GetVar(curRecorder, IMGTYPE);
    if (var)
    {
	imgType = GetInt(var);
    }
    else
    {
	imgType = PS_GRAYSCALE;
    }
    
    width = right - left;
    height = top - bottom;

#define PIXPERLINE	10

    if (imgType != PS_GRAYSCALE) /*** need to impl 12-bit color, color subsampling ***/
    {
	/*Color image.  Do stuff before pixels*/
	fprintf(psFile, "%% Color Space image within %d %d %d %d\n",
		    left, right, bottom, top);
	fprintf(psFile, "%d %d moveto\n", left, bottom);
	fprintf(psFile, "/imgstr %d string def\n", 3 * width);
	fprintf(psFile, "%d %d %d matrix \n", right - left, top - bottom, 8);
    
	/*Do pixels*/
	nPixels = width * height;
	pixWrap = 0;
	fputs("{currentfile imgstr readhexstring pop} false 3 colorimage\n", psFile);
	for (k = 0; k < nPixels; ++k)
	{
	    fprintf(psFile, "%02X%02X%02X%s",
		    LinToGamma[imageBuffer[k] . red],
		    LinToGamma[imageBuffer[k] . green],
		    LinToGamma[imageBuffer[k] . blue],
		    (++pixWrap) % PIXPERLINE ? " " : "\n");
	}
    }
    else
    {
	/*B/W image.  Do stuff before pixels*/
	fprintf(psFile, "%% B/W Space image within %d %d %d %d\n",
		    left, right, bottom, top);
	fprintf(psFile, "/imgstr %d string def\n", width / subSample);
	fprintf(psFile, "%d %d %d\n", width / subSample, height / subSample, 8);
	fprintf(psFile, "%d %d matrix translate\n",
		    -left, -bottom);
	fprintf(psFile, "%lg %lg matrix scale\n",
		    1.0 / (double) subSample, 1.0 / (double) subSample);
	fputs(" matrix concatmatrix\n", psFile);
    
	/*Do pixels*/
	nPixels = (right - left) * (top - bottom);
	pixWrap = 0;
	fputs("{currentfile imgstr readhexstring pop} image\n", psFile);
    
    
	for (j = 0; j < height; j += subSample)
	{
	    for (i = 0; i < width; i += subSample)
	    {
		real avgPixel;
    
		avgPixel = 0.0;
    
		/*Calculate an average pixel*/
		for (jp = j; jp < j + subSample; ++jp)
		{
		    for (ip = i; ip < i + subSample; ++ip)
		    {
			k = ip + jp * width;
    
			RGB2YIQ(&y, &dummy, &dummy,
			    imageBuffer[k] . red / 255.0,
			    imageBuffer[k] . green / 255.0,
			    imageBuffer[k] . blue / 255.0);
    
			avgPixel += y;
		    }
		}
		avgPixel /= (real) subSample;
		avgPixel /= (real) subSample;
    
		iVal = avgPixel * 255.0;
		if (iVal < 0) iVal = 0;
		if (iVal > 255) iVal = 255;
		fprintf(psFile, "%02X%s", LinToGamma[iVal],
		    (++pixWrap) % PIXPERLINE ? " " : "\n");
	    }
	}
    }
    fputs("\n\n", psFile);

#undef PIXPERLINE

}

/*
 * Icons in PostScript files 10/28/92
 * In order to make the PostScript file of a window self contained, a bitmap
 * font of just the icons used in that window is defined in the file. In PS
 * drawing mode, RegisterIconUsed keeps a list of the icon numbers used.
 * It returns the index of the icon in the list. EndDrawing uses the list
 * to create the font definition. If there are any icons in the window, the
 * file must be rewritten with the font definition in the header. The font
 * definition is generated by GenIconDef() in ScianIcons.c.
 */

#ifdef HAVE_PROTOTYPES
void DrawPSFrame(char *name, int left, int right, int bottom, int top)
#else
void DrawPSFrame(name, left, right, bottom, top)
char *name;
int left;
int right;
int bottom;
int top;
#endif
/*Draws a PostScript frame around left, right, bottom, top*/
{
    left -= WINFRAMEWID;
    right += WINFRAMEWID;
    bottom -= WINFRAMEWID;
    top += WINFRAMEWID + WINTITLEHT;
    /* draw a generic window frame with title only */
    fprintf(psFile,
"newpath %d %d moveto %d %d lineto %d %d lineto %d %d lineto closepath stroke\n", 
	left, bottom, left, top, right, top, right, bottom);
    fprintf(psFile, 
"newpath %d %d moveto %d %d lineto %d %d lineto %d %d lineto closepath stroke\n", 
	left + WINFRAMEWID - 1, bottom + WINFRAMEWID - 1,
	left + WINFRAMEWID - 1, top - WINFRAMEWID + 1,
	right - WINFRAMEWID + 1, top - WINFRAMEWID + 1,
	right - WINFRAMEWID + 1, bottom + WINFRAMEWID - 1);
    fprintf(psFile, "newpath %d %d moveto %d %d lineto stroke\n", 
	left + WINFRAMEWID - 1, top - WINFRAMEWID - WINTITLEHT + 1, 
	right - WINFRAMEWID + 1, top - WINFRAMEWID - WINTITLEHT + 1);
    SetupFont(WINTITLEFONT, 12);
    fprintf(psFile, "%d %d moveto (%s) show\n", 
	left + WINFRAMEWID + 20, top - WINFRAMEWID - WINTITLEHT + 5,
	PSShow(name));
    RegisterFontUsed(WINTITLEFONT);
}

#ifdef HAVE_PROTOTYPES
void BeginDrawSubscreen(int ox, int oy)
#else
void BeginDrawSubscreen(ox, oy)
int ox, oy;
#endif
/*Begins a drawing segment offset by ox, oy*/
{
    if (drawingMode == DRAW_POSTSCRIPT && psFile)
    {
	if (ox != 0 || oy != 0)
	    fprintf(psFile, "%d %d translate\n", ox, oy);
    }
}

#ifdef HAVE_PROTOTYPES
void EndDrawSubscreen(int ox, int oy)
#else
void EndDrawSubscreen(ox, oy)
int ox, oy;
#endif
/*Ends a drawing segment offset by ox, oy*/
{
    if (drawingMode == DRAW_POSTSCRIPT && psFile)
    {
	if (ox != 0 || oy != 0)
	    fprintf(psFile, "%d %d translate\n", -ox,  -oy);
    }
}

#ifdef HAVE_PROTOTYPES
void BeginDrawing(int left, int right, int bottom, int top)
#else
void BeginDrawing(left, right, bottom, top)
int left, right, bottom, top;	/*Coordinates of window INTERIOR*/
#endif
{
#ifdef GRAPHICS
    real scale;
    Bool eps;
    int imgOrient, imgPlace, xTran, yTran, imgWid, imgHt;
    real pageWidth, pageHeight, leftOffset, bottomOffset;
    ObjPtr var;
    ObjPtr windows;
    struct timeval date;

    psFile = tmpfile();
    if (!(drawingMode == DRAW_POSTSCRIPT && psFile)) return;
    gettimeofday(&date, (struct timezone *)0);

/* build list of window titles in tempStr */
    windows = GetVar(curRecorder, RECORDWINDOWS);
    if (windows)
    {
	ThingListPtr runner;
	
	*tempStr = '\0';	/* clear temp string */
	runner = LISTOF(windows);
	while(runner)
	{
	    WinInfoPtr window;

	    window = (WinInfoPtr) (runner -> thing);
	    strcat(tempStr, window -> winTitle);
	    runner = runner -> next;
	    if(runner) strcat(tempStr, ", ");
	}
    }
    
    var = GetVar(curRecorder, IMGSCALING);
    if (var)
    {
	scale = GetReal(var) * 0.01;	/* scale is in per cent */
    }
    else
    {
	scale = 0.5;
    }
    
    var = GetVar(curRecorder, PSFORMAT);
    if (var)
    {
	eps = (GetInt(var) == PS_EPS);
    }
    else
    {
	eps = false;
    }

    if (eps)
    {
	/* just normalize */
	right -= left;
	right = (int) (scale * (real) (right + 1));
	top -= bottom;
	top = (int) (scale * (real) (top + 1));
	left = bottom = 0;
    }
    else
    {
	/* compute bounding box and translation */

	var = GetVar(curRecorder, ORIENTATION);
	if (var)
	{
	    imgOrient = GetInt(var);
	}
	else
	{
	    imgOrient = PS_PORTRAIT;
	}
	
	var = GetVar(curRecorder, PLACEMENT);
	if (var)
	{
	    imgPlace = GetInt(var);
	}
	else
	{
	    imgPlace = PS_CENTERED;
	}
	if (imgPlace == PS_CENTERED)
	{
/*** units assumed to be inches (sorry Euro-pals) until generalized units handling ***/
	    var = GetVar(curRecorder, PAGEWIDTH);
	    if (var)
	    {
		pageWidth = 72*GetReal(var);
	    }
	    else
	    {
		pageWidth = 72*8.5;
	    }
	    var = GetVar(curRecorder, PAGEHEIGHT);
	    if (var)
	    {
		pageHeight = 72*GetReal(var);
	    }
	    else
	    {
		pageHeight = 72*11;
	    }
	}
	else
	{
	    var = GetVar(curRecorder, LEFTOFFSET);
	    if (var)
	    {
		leftOffset = 72*GetReal(var);
	    }
	    else
	    {
		leftOffset = 72*1;
	    }
	    var = GetVar(curRecorder, BOTTOMOFFSET);
	    if (var)
	    {
		bottomOffset = 72*GetReal(var);
	    }
	    else
	    {
		bottomOffset = 72*2;
	    }
	}

	imgWid = scale*(right - left + 1);
	imgHt = scale*(top - bottom + 1);
	if (imgOrient == PS_PORTRAIT)
	{
	    if (imgPlace == PS_CENTERED)
	    {
		xTran = (pageWidth - imgWid)/2;
		yTran = (pageHeight - imgHt)/2;
	    }
	    else
	    {
		xTran = leftOffset;
		yTran = bottomOffset;
	    }
	    left = xTran;
	    right = xTran + imgWid;
	    bottom = yTran;
	    top = yTran + imgHt;
	}
	else /* landscape */
	{
	    if (imgPlace == PS_CENTERED)
	    {
		xTran = (pageWidth + imgHt)/2;
		yTran = (pageHeight - imgWid)/2;
	    }
	    else
	    {
		xTran = leftOffset + imgHt;
		yTran = bottomOffset;
	    }
	    left = xTran - imgHt;
	    right = xTran;
	    bottom = yTran;
	    top = yTran + imgWid;
	}
    }

    /* initialize counters */
    numFontsUsed = 0;
    numIconsUsed = 0;

/* now write PostScript prologue */

    fputs("%!PS-Adobe-2.0", psFile);
    if (eps) fputs(" EPSF-2.0", psFile);
    fprintf(psFile, "\n\
%%%%Title: %s\n\
%%%%Creator: SciAn %s\n\
%%%%CreationDate: %s\
%%%%For: %s\n\
%%%%DocumentFonts: (atend)\n\
%%%%BoundingBox: %d %d %d %d\n\
%%%%EndComments\n\
save\n\n",	/* NOTE: icon font def is inserted after first blank line of file */
	tempStr, SCIANVERSION, ctime(&date.tv_sec), getenv("USER"),
	left, bottom, right, top);

    fputs("/resetclip {matrix currentmatrix currentfont currentlinewidth\n\
grestore gsave setlinewidth setfont setmatrix } def\n\n", psFile);

#ifdef PS_SCREEN
    fputs("currentscreen 3 1 roll pop pop 75 0 3 -1 roll setscreen\n", psFile);
#endif
    
    fputs("gsave\n1 setlinewidth\n", psFile);
    if (!eps)
    {
	fprintf(psFile, "%d %d translate\n", xTran, yTran);
	if (imgOrient == PS_LANDSCAPE) fputs("90 rotate\n", psFile);
    }
    if (scale != 1.0) fprintf(psFile, "%g %g scale\n\n\n", scale, scale);
    else fputs("\n\n", psFile);

#endif
}

#ifdef HAVE_PROTOTYPES
void EndDrawing(Bool showpagep)
#else
void EndDrawing(showpagep)
Bool showpagep;
#endif
{
#ifdef GRAPHICS
    if (drawingMode == DRAW_POSTSCRIPT && psFile && drawFile)
    {
	char buf[256];
        int i;
	
	/* copy header from temp file (copies to first blank line) */
	rewind(psFile);
	while (fgets(buf, 256, psFile))
	{
	    fputs(buf, drawFile);
	    if (*buf == '\n') break;
	}
	
	/* generate font of icons used, if needed */
	GenIconDef(drawFile);
	
	/* copy the rest of the temp file */
	while (fgets(buf, 256, psFile))
	{
	    fputs(buf, drawFile);
	}
	fclose(psFile);

	if (showpagep)
	{
	    fputs("\nshowpage\n", drawFile);
	}

        /* write PostScript trailer -- list fonts used */
        fputs("\nrestore\n\n%%Trailer\n%%DocumentFonts:", drawFile);
	for (i=0; i<numFontsUsed; ++i)
        {
            fprintf(drawFile, " %s", fontsUsed[i]);

	    /* free the fontsUsed string space */
            Free(fontsUsed[i]);
            fontsUsed[i] = NIL;
        }
        fputs("\n", drawFile);
    }
#endif
}

#ifdef HAVE_PROTOTYPES
int RegisterIconUsed(int icon)
#else
int RegisterIconUsed(icon)
int icon;
#endif
{
#ifdef GRAPHICS
    int i;

    /* check list of icons used */
    for (i = 0; i < numIconsUsed; ++i)
    {
        if (icon == iconsUsed[i]) return i; /* already got it */
    }
    if (numIconsUsed >= 32)
    {
        ReportError("RegisterIconUsed","Too many different icons!");
        return -1;
    }
    iconsUsed[numIconsUsed] = icon;
    return numIconsUsed++;
#else
    return 0;
#endif
}
