/*ScianProjects.c
  Eric Pepke
  May 11, 1994

  Projects system in SciAn
*/

#include "Scian.h"
#include "ScianTypes.h"
#include "ScianIDs.h"
#include "ScianNames.h"
#include "ScianWindows.h"
#include "ScianObjWindows.h"
#include "ScianDialogs.h"
#include "ScianScripts.h"
#include "ScianErrors.h"
#include "ScianSymbols.h"
#include "ScianEvents.h"
#include "ScianLists.h"
#include "ScianArrays.h"
#include "ScianControls.h"
#include "ScianButtons.h"
#include "ScianTextBoxes.h"
#include "ScianTitleBoxes.h"
#include "ScianIcons.h"
#include "ScianStyle.h"
#include "ScianObjFunctions.h"
#include "ScianDatabase.h"
#include "ScianColors.h"
#include "ScianObjFunctions.h"
#include "ScianTemplates.h"
#include "ScianTemplateHelper.h"
#include "ScianSnap.h"
#include "ScianAnimation.h"
#include "ScianTrackControls.h"
#include "ScianDraw.h"
#include "ScianMethods.h"
#include "ScianFixedClasses.h"
#include "ScianProjects.h"

ObjPtr projectClass = NULLOBJ;

#ifdef HAVE_PROTOTYPES
ObjPtr AllProjectSequences(void)
#else
ObjPtr AllProjectSequences()
#endif
/*Returns a list of all the project sequences*/
{
    ObjPtr project;

    project = CurProject();
    if (project)
    {
	return GetVar(project, SEQUENCES);
    }
    else
    {
	return NULLOBJ;
    }
}

#ifdef HAVE_PROTOTYPES
ObjPtr CurProject(void)
#else
ObjPtr CurProject()
#endif
/*Returns the current project or NULLOBJ*/
{
    ObjPtr keyList, result;

    keyList = NewList();
    PostfixList(keyList, NewSymbol(CLASSID));
    PostfixList(keyList, NewInt(CLASS_PROJECT));
    result = SearchDatabase(keyList);
    if (result && LISTOF(result))
    {
	return LISTOF(result) -> thing;
    }
    else
    {
	return NULLOBJ;
    }
}

static MakeMethod MakeVisWinCorralAppearance(corral)
ObjPtr corral;
/*Makes the appearance of a vis window corral*/
{
    ObjPtr repObj;
    ObjPtr windows;
    ThingListPtr runner1, runner2;
    ObjPtr contents;
    ObjPtr objectsToAdd;
    ObjPtr icon, var;

    repObj = GetObjectVar("MakeVisWinCorralAppearance", corral, REPOBJ);
    if (!repObj) return ObjFalse;

    windows = GetListVar("MakeVisWinCorralAppearance", repObj, VISWINDOWS);
    if (!windows) return ObjFalse;

    contents = GetListVar("MakeVisWinCorralAppearance", corral, CONTENTS);
    if (!contents) return ObjFalse;

    objectsToAdd = NewList();
    for (runner1 = LISTOF(windows); runner1; runner1 = runner1 -> next)
    {
	for (runner2 = LISTOF(contents); runner2; runner2 = runner2 -> next)
	{
	    if (GetVar(runner2 -> thing, REPOBJ) == runner1 -> thing) break;
	}
	if (!runner2)
	{
	    PostfixList(objectsToAdd, runner1 -> thing);
	}
    }

    if (LISTOF(objectsToAdd))
    {
	ImInvalid(corral);

	for (runner1 = LISTOF(objectsToAdd); runner1; runner1 = runner1 -> next)
	{
	    var = GetVar(runner1 -> thing, NAME);
	    icon = NewIcon(0, 0, ICONWINDOW, var ? GetString(var) : "?");
	    SetVar(icon, REPOBJ, runner1 -> thing);
	    SetVar(icon, ICONLOC, NULLOBJ);
	    DropIconInCorral(corral, icon);
	}
    }

    SetVar(corral, APPEARANCE, ObjTrue);

    return ObjTrue;
}

static MakeMethod MakeSequenceCorralAppearance(corral)
ObjPtr corral;
/*Makes the appearance of a sequence corral*/
{
    ObjPtr repObj;
    ObjPtr sequences;
    ThingListPtr runner1, runner2;
    ObjPtr contents;
    ObjPtr objectsToAdd;
    ObjPtr icon, var;

    repObj = GetObjectVar("MakeSequenceAppearance", corral, REPOBJ);
    if (!repObj) return ObjFalse;

    sequences = GetListVar("MakeSequenceAppearance", repObj, SEQUENCES);
    if (!sequences) return ObjFalse;

    contents = GetListVar("MakeSequenceAppearance", corral, CONTENTS);
    if (!contents) return ObjFalse;

    objectsToAdd = NewList();
    for (runner1 = LISTOF(sequences); runner1; runner1 = runner1 -> next)
    {
	for (runner2 = LISTOF(contents); runner2; runner2 = runner2 -> next)
	{
	    if (GetVar(runner2 -> thing, REPOBJ) == runner1 -> thing) break;
	}
	if (!runner2)
	{
	    PostfixList(objectsToAdd, runner1 -> thing);
	}
    }

    if (LISTOF(objectsToAdd))
    {
	ImInvalid(corral);

	for (runner1 = LISTOF(objectsToAdd); runner1; runner1 = runner1 -> next)
	{
	    var = GetVar(runner1 -> thing, NAME);
	    icon = NewIcon(0, 0, ICONSEQUENCE, var ? GetString(var) : "?");
	    SetVar(icon, REPOBJ, runner1 -> thing);
	    SetVar(icon, ICONLOC, NULLOBJ);
	    DropIconInCorral(corral, icon);
	}
    }

    SetVar(corral, APPEARANCE, ObjTrue);

    return ObjTrue;
}

void AddSequenceToProject(sequence)
ObjPtr sequence;
/*Adds sequence to the current project*/
{
    ObjPtr project, sequences;

    project = CurProject();
    if (project)
    {
	sequences = GetListVar("AddSeqWindowToProject", project, SEQUENCES);
	if (sequences)
	{
	    PostfixList(sequences, sequence);
	    SetVar(project, SEQUENCES, sequences);
	}
    }
}

void MaybeAddSequenceToProject(window)
ObjPtr window;
/*Adds window to the current project*/
{
    ObjPtr repObj, var;

    repObj = GetVar(window, REPOBJ);
    if (repObj)
    {
	var = GetVar(repObj, CLASSID);
	if (var && (GetInt(var) == CLASS_SEQUENCE))
	{
	    AddSequenceToProject(repObj);
	}
    }
}

void AddVisWindowToProject(window)
ObjPtr window;
/*Adds window to the current project*/
{
    ObjPtr project, windows;

    project = CurProject();
    if (project)
    {
	windows = GetListVar("AddVisWindowToProject", project, VISWINDOWS);
	if (windows)
	{
	    PostfixList(windows, window);
	    SetVar(project, VISWINDOWS, windows);
	}
    }
}

#ifdef HAVE_PROTOTYPES
void NewProject(void)
#else
void NewProject()
#endif
/*Create a new project*/
{
    if (CurProject())
    {
	WarnUser(CW_NOTWOPROJECTS);
    }
    else
    {
	WinInfoPtr win;
	ObjPtr panel, contents, project, corral, titleBox;
	int l, r, b, t;

	/*Create the window*/
	win = NewObjWindow(NULLOBJ, "Project", WINUI, PROJWINWIDTH, PROJWINHEIGHT, SCRWIDTH, SCRHEIGHT);
    
	/*Add a help string*/
	SetVar((ObjPtr) win, HELPSTRING, 
	    NewString("This window shows the currently open project."));
    
	/*Put in a panel*/
	panel = NewPanel(greyPanelClass, 0, PROJWINWIDTH, 0, PROJWINHEIGHT);
	SetVar(panel, STICKINESS, NewInt(STICKYLEFT + STICKYRIGHT +
					 STICKYBOTTOM + STICKYTOP));
    
	contents = GetVar((ObjPtr) win, CONTENTS);
	PrefixList(contents, panel);
	SetVar(panel, PARENT, (ObjPtr) win);
	contents = GetVar(panel, CONTENTS);

	/*Make the new project*/
	project = NewObject(projectClass, 0L);
	SetVar(project, VISWINDOWS, NewList());
	SetVar(project, SEQUENCES, NewList());

	/*Stick project on to window*/
	SetVar((ObjPtr) win, REPOBJ, project);

	/*Add it to the database*/
	AddVolObjToDatabase(project);

	/*Fill the window with controls*/
	l = MINORBORDER;
	r = PROJWINWIDTH - MINORBORDER;
	t = PROJWINHEIGHT - MINORBORDER;
	b = t - 2 * MINORBORDER - TITLEBOXTOP - ONECORRALHEIGHT - BARWIDTH - CORRALBARBORDER;

	titleBox = NewTitleBox(l, r, b, t, "Visualization Windows");
	PrefixList(contents, titleBox);
	SetVar(titleBox, PARENT, panel);
	SetVar(titleBox, STICKINESS, NewInt(STICKYLEFT + STICKYRIGHT + STICKYTOP));

	l += MINORBORDER;
	r -= MINORBORDER;
	b += MINORBORDER;
	t -= MINORBORDER + TITLEBOXTOP;

	corral = NewIconCorral(NULLOBJ, l, r, b, t, BARRIGHT + BARBOTTOM);
	PrefixList(contents, corral);
	SetVar(corral, PARENT, panel);
	SetVar(corral, STICKINESS, NewInt(STICKYLEFT + STICKYRIGHT + STICKYTOP));
	DeclareIndirectDependency(corral, APPEARANCE, REPOBJ, VISWINDOWS);
	SetMethod(corral, APPEARANCE, MakeVisWinCorralAppearance);
	SetVar(corral, REPOBJ, project);
	SetVar(corral, ICONOFFSET, NewInt(-5));

	l -= MINORBORDER;
	r += MINORBORDER;
	t = b - 2 * MINORBORDER;
	b = t - 2 * MINORBORDER - TITLEBOXTOP - ONECORRALHEIGHT - BARWIDTH - CORRALBARBORDER;

	titleBox = NewTitleBox(l, r, b, t, "Sequences");
	PrefixList(contents, titleBox);
	SetVar(titleBox, PARENT, panel);
	SetVar(titleBox, STICKINESS, NewInt(STICKYLEFT + STICKYRIGHT + STICKYTOP));

	l += MINORBORDER;
	r -= MINORBORDER;
	b += MINORBORDER;
	t -= MINORBORDER + TITLEBOXTOP;

	corral = NewIconCorral(NULLOBJ, l, r, b, t, BARBOTTOM);
	PrefixList(contents, corral);
	SetVar(corral, PARENT, panel);
	SetVar(corral, STICKINESS, NewInt(STICKYLEFT + STICKYRIGHT + STICKYTOP));
	DeclareIndirectDependency(corral, APPEARANCE, REPOBJ, SEQUENCES);
	SetMethod(corral, APPEARANCE, MakeSequenceCorralAppearance);
	SetVar(corral, REPOBJ, project);
	SetVar(corral, ICONOFFSET, NewInt(-5));

	ForAllVisWindows(AddVisWindowToProject);
	ForAllWindows(MaybeAddSequenceToProject);

	DoNewSequence();
    }
}

#ifdef HAVE_PROTOTYPES
void DoNewProject(void)
#else
void DoNewProject()
#endif
/*Makes a new project*/
{
    NewProject();
}


static ObjPtr RemoveProjFromDatabase(project)
ObjPtr project;
/*Removes the project from the database*/
{
    DeleteObjFromDatabase(project);

    return ObjTrue;
}

ObjPtr LogProjectDefinition(object)
ObjPtr object;
/*Logs a project object definition*/
{
    char *retVar;
    ObjPtr v;
    ObjPtr var;
    int classID;
    ObjPtr depList;
    ObjPtr varList;
    ThingListPtr runner, runner2, runner3;

    InhibitWindow(true);

    var = GetVar(object, VISWINDOWS);
    if (var)
    {
	for (runner = LISTOF(var); runner; runner = runner -> next)
	{
	    FuncTyp method;
	    method = GetMethod(runner -> thing, SAVEALLCONTROLS);
	    if (method)
	    {
		(*method)(runner -> thing);
	    }
	}
    }

    /*Determine all the dependent vars*/
    depList = NewList();
    PostfixList(depList, object);
    depList = AddExternalSnapVarsToSet(depList, object);

    /*Make the variables*/
    varList = NewList();
    for (runner = LISTOF(depList); runner; runner = runner -> next)
    {
	retVar = LogScriptVarForObject(runner -> thing);
	/*If it's in the database, emit a setter*/
	if (IsUniqueInDatabase(runner -> thing))
	{
	    v = GetVar(runner -> thing, CLASSID);
	    if (v)
	    {
		classID = GetInt(v);
		MakeVar(runner -> thing, NAME);
		v = GetVar(runner -> thing, NAME);
		if (v)
		{
		    char name[200];
		    Log("try $");
		    Log(retVar);
		    Log(" = ");
		    Log(ClassIDToName(classID));
		    Log("@");
		    MakeObjectSimpleName(name, runner -> thing);
		    Log(name);
		    Log("\n");
		}
	    }
	}
	PostfixList(varList, NewString(retVar));
    }

    Log("\n");

    /*Make the definition*/
    runner2 = LISTOF(varList);
    for (runner = LISTOF(depList); runner; runner = runner -> next)
    {
	/*Log header*/
	Log("begin definition ");
	Log("$");
	Log(GetString(runner2 -> thing));
	Log("\n");

	/*Log history*/
	var = GetVar(runner -> thing, HISTORY);
	if (var)
	{
	    /*There's a specified history*/
	    for (runner3 = LISTOF(var); runner3; runner3 = runner3 -> next)
	    {
		LogObject(runner3 -> thing);
	    }

	    /*If it's in the database, emit a setter*/
	    if (IsUniqueInDatabase(runner -> thing))
	    {
		v = GetVar(runner -> thing, CLASSID);
		if (v)
		{
		    classID = GetInt(v);
		    MakeVar(runner -> thing, NAME);
		    v = GetVar(runner -> thing, NAME);
		    if (v)
		    {
			char name[200];
			Log("    $");
			Log(GetString(runner2 -> thing));
			Log(" = ");
			Log(ClassIDToName(classID));
			Log("@");
			MakeObjectSimpleName(name, runner -> thing);
			Log(name);
			Log("\n");
		    }
		}
	    }
	}
	else
	{
	    /*Make default history, ab initio*/

	    v = GetVar(runner -> thing, CLASSID);
	    if (v)
	    {
		classID = GetInt(v);
		MakeVar(runner -> thing, NAME);
		v = GetVar(runner -> thing, NAME);
		if (v)
		{
		    char name[200];
		    Log("    $");
		    Log(GetString(runner2 -> thing));
		    Log(" = new(");
		    Log(ClassIDToName(classID));
		    Log("@");
		    MakeObjectSimpleName(name, runner -> thing);
		    Log(name);
		    Log(")\n");
		}
	    }
	}

	/*Log current state*/
	LogSnapVars(runner -> thing);

	/*Log end*/
	Log("end definition\n\n");
	runner2 = runner2 -> next;
    }

    InhibitWindow(false);

    return ObjTrue;
}

ObjPtr SaveProjectFile(object, name)
ObjPtr object;
char *name;
/*Saves a object as name*/
{
    SaveObject(object, name);
    return ObjTrue;
}

#ifdef HAVE_PROTOTYPES
void InitProjects(void)
#else
void InitProjects()
#endif
/*Initialize the project system*/
{
    projectClass = NewObject(NULLOBJ, 0L);
    AddToReferenceList(projectClass);
    DefineFixedClass(projectClass, CLASS_PROJECT, "project");
    SetVar(projectClass, NAME, NewString("Current Project"));
    SetMethod(projectClass, CLEANUP, RemoveProjFromDatabase);
    AddSnapVar(projectClass, SEQUENCES);
    SetMethod(projectClass, SAVECPANEL, SaveSnapshotControls);
    SetMethod(projectClass, SAVEALLCONTROLS, LogProjectDefinition);
    SetMethod(projectClass, SAVEFILE, SaveProjectFile);
}

#ifdef HAVE_PROTOTYPES
void KillProjects(void)
#else
void KillProjects()
#endif
/*Kill the project system*/
{
    RemoveFromReferenceList(projectClass);
}
