/*ScianScripts.c
  Stuff for scripts in scian
  Eric Pepke
*/

#include "Scian.h"
#include "ScianTypes.h"
#include "ScianArrays.h"
#include "ScianIDs.h"
#include "ScianEvents.h"
#include "ScianRecorders.h"
#include "ScianColors.h"
#include "ScianWindows.h"
#include "ScianObjWindows.h"
#include "ScianDialogs.h"
#include "ScianVisWindows.h"
#include "ScianSliders.h"
#include "ScianTimers.h"
#include "ScianMethods.h"
#include "ScianLists.h"
#include "ScianErrors.h"
#include "ScianGarbageMan.h"
#include "ScianSpaces.h"
#include "ScianIcons.h"
#include "ScianSockets.h"
#include "ScianObjFunctions.h"
#include "ScianGlobalFunctions.h"
#include "ScianCurrentFunctions.h"
#include "ScianWindowFunctions.h"
#include "ScianRepobjFunctions.h"
#include "ScianHelp.h"
#include "ScianStyle.h"
#include "ScianNames.h"
#include "ScianTextFiles.h"
#include "ScianDatabase.h"
#include "ScianSymbols.h"
#include "ScianControls.h"
#include "ScianScripts.h"
#include "ScianFixedClasses.h"
#include "ScianParser.h"
#include "ScianCalculator.h"

Bool runningScript = false;		/*True iff running script*/
Bool settingUp = false;			/*True iff setting up*/
WinInfoPtr scriptWindow = 0;		/*The current window of the script*/
extern ObjPtr perspecControlClass;	/*Class of perspecive controls*/
long framesLeft = 0;			/*Frames left to record*/
Bool abortScript = false;		/*Abort script*/
TextFilePtr curScript = 0;		/*Current script being read, if any*/
char *scriptLine = 0;			/*Current line of the script*/
Bool lineAccepted = true;		/*True iff last script line was accepted*/
int inhibitRecording = 0;		/*Inhibits recording*/
int inhibitWindow = 0;			/*Inhibits window saving*/
ObjPtr scriptDummy = NULLOBJ;		/*Dummy object for script stuff marking*/
ObjPtr compiledCode = NULLOBJ;		/*Code that we're compiling, else null for immediate*/

TextFilePtr savedScripts[10];		/*Stack for saving scripts*/
int curScriptIndex = 0;

/*IDs for running tasks*/
#define RT_NONE		0		/*No running task*/
#define RT_RECORDING	1		/*Recording some time*/
int runningTask = RT_NONE;		/*Current running task*/

#define MAXSTACKDEPTH	100		/*Maximum depth of block stack*/

/*Block in a stack*/
typedef struct
    {
	ScriptVarPtr vars;		/*Variables within this block*/
	int command;			/*The command*/
	ObjPtr object;			/*The (optional) object this is about*/
	ScriptVarPtr cVar;		/*The object this block is defining*/
	ObjPtr compiledCode;		/*Pointer to compiled code*/
    } Block;

Block commandStack[MAXSTACKDEPTH];	/*Stack of commands*/
int nextStackElement = 0;		/*Next element on the stack*/

typedef struct				/*Element of file logging stack*/
    {
	TextFilePtr logFile;		/*Text file for logging*/
	int logInhibit;			/*Counter to inhibit logging*/
	WinInfoPtr logWindow;		/*Current window set in log*/
	ScriptVarPtr vars;		/*Variables in the script*/
    } LogElement;

#define MAXLOGFILES	50		/*Maximum number of log files open*/

LogElement logStack[MAXLOGFILES];	/*Stack of log files*/
int nextLogElement = 0;			/*Next element on the stack*/

#define KWBEGIN		0	/*Begin a block*/
#define KWEND		1	/*End current modal command*/
#define KWRECORDER	2	/*(set) recorder*/
#define KWDESELECT	3	/*Deselect an icon*/
#define KWSELECT	4	/*Select an icon*/
#define KWSHELL		5	/*Shell escape*/
#define KWALIGNMENT	6	/*(Set) alignment OBSELETE*/
#define KWLOCATION	7	/*(Set) location*/
#define KWVIDEOSCREEN	8	/*Resize current window to video screen*/
#define KWFULLSCREEN	9	/*Resize current window to full screen*/
#define KWROTATE	10	/*Rotate space*/
#define KWSHOW		11	/*Show something*/
#define KWPANEL		12	/*(Show/Hide) the panel of icons*/
#define KWRECORDING	13	/*(Begin/End) recording*/
#define KWSNAP		14	/*Snap a single frame*/
#define KWWINDOW	15	/*Select a new window*/
#define KWSET		16	/*Set something*/
#define KWCONTROLS	17	/*(Show) controls for a vis object*/
#define KWPUSHWINDOW	18	/*Push a window behind all others*/
#define KWQUIT		19	/*Quit*/
#define KWSHEAR		20	/*Shear the matrix*/
#define KWFPS		21	/*(Set) fps*/
#define KWEYEPOSN	22	/*Eye position*/
#define KWROLL		23	/*Roll of eyeball*/
#define KWPITCH		24	/*Pitch of eyeball*/
#define KWYAW		25	/*Yaw of eyeball*/
#define KWRECTANGLE	26	/*Rectangle*/
#define KWHIDE		27	/*Hide the frame of a window*/
#define KWEXIT		28	/*Exit on next loop*/
#define KWVALUE		29	/*(Set) value of a something*/
#define KWFONT		30	/*(Set) the font of something OBSELETE*/
#define KWSIZE		31	/*(Set) the size of something's font OBSELETE*/
#define KWCLOSE		32	/*Closes a window*/
#define KWENDPOINTS	33	/*(Set) endpoints*/
#define KWDROP		34	/*Drops whatever is in the drag buffer*/
#define KWBOUNDS	35	/*(Set) the bounds of an object*/
#define KWTILE		36	/*Tile to a certain width and height*/
#define KWANNOTATION	37	/*Add annotation*/
#define KWFRAME		38	/*(Show/Hide) the frame of a window*/
#define KWLOCATE	39	/*Locate a window*/
#define KWDELETE	40	/*Delete stuff*/
#define KWSCRSAVE	41	/*Save the screen*/
#define KWRECORD	42	/*Record some number of seconds*/
#define KWSCREEN	43	/*(Set) screen*/
#define KWCOLOR		44	/*(Set) color*/
#define KWPREFERENCES	45	/*(Show) preferences*/
#define KWTIME		46	/*Time*/
#define KWOFF		47	/*(Turn) off*/
#define KWTURN		48	/*Turn (on/off)*/
#define KWON		49	/*(Turn) on*/
#define KWLINE		50	/*Make a new line*/
#define KWFUNCTIONBOX	51	/*(Set) Function box*/
#define KWFILESWINDOW	52	/*(Show) FilesWindow*/
#define KWSELECTALL	53	/*Selectall*/
#define KWDESELECTALL	54	/*Deselectall*/
#define KWHELP		55	/*(Show) help*/
#define KWFILEREADERS   56	/*(Show) fileReaders*/
#define KWDATASETS	57	/*(Show) datasets*/
#define KWROTATION	58	/*(Set) rotation*/
#define KWVERSION	59	/*Version*/
#define KWFRONT		60	/*(Show) front (panel controls)*/
#define KWBACK		61	/*(Show) back (panel controls)*/
#define KWSETUP		62	/*(Begin/End) setup*/
#define KWSAVE		63	/*Save something*/
#define KWTIMEREADOUT	64	/*Make a time readout*/
#define KWSPACE		65	/*(Show) space (controls)*/
#define KWEFFECT	66	/*Color special effect OBSELETE*/
#define KWKEEP		67	/*Keep changes to palette OBSELETE*/
#define KWREVERT	68	/*Revert to last saved palette OBSELETE*/
#define KWMODEL		69	/*Color model OBSELETE*/
#define KWSNAPSHOT	70	/*(Begin) Snapshot*/
#define KWOBSERVER	71	/*(Begin Snapshot) Observer*/
#define KWVARIABLE	72	/*(set) Variable <value>*/
#define KWOBJECT	73	/*(Begin Snapshot) */
#define KWRECORDERS	74	/*recorders*/
#define KWSCRIPT	75	/*Begin/end script, used for nesting scripts*/
#define KWDEFINITION	76	/*Begin/end definition*/
#define KWREAD		77	/*Read (a file) */
#define KWTRY		78	/*Try a script line*/
#define NKEYWORDS	79	/*Number of keywords*/


char *keywords[NKEYWORDS] =
	{
	    "begin",
	    "end",
	    "recorder",
	    "deselect",
	    "select",
	    "shell",
	    "alignment",
	    "location",
	    "videoscreen",
	    "fullscreen",
	    "rotate",
	    "show",
	    "panel",
	    "recording",
	    "snap",
	    "window",
	    "set",
	    "controls",
	    "pushwindow",
	    "quit",
	    "shear",
	    "fps",
	    "eyeposn",
	    "roll",
	    "pitch",
	    "yaw",
	    "rectangle",
	    "hide",
	    "exit",
	    "value",
	    "font",
	    "size",
	    "close",
	    "endpoints",
	    "drop",
	    "bounds",
	    "tile",
	    "annotation",
	    "frame",
	    "locate",
	    "delete",
	    "scrsave",
	    "record",
	    "screen",
	    "color",
	    "preferences",
	    "time",
	    "off",
	    "turn",
	    "on",
	    "line",
	    "functionbox",
	    "fileswindow",
	    "selectall",
	    "deselectall",
	    "help",
	    "filereaders",
	    "datasets",
	    "rotation",
	    "version",
	    "front",
	    "back",
	    "setup",
	    "save",
	    "timereadout",
	    "space",
	    "effect",
	    "keep",
	    "revert",
	    "model",
	    "snapshot",
	    "observer",
	    "variable",
	    "object",
	    "recorders",
	    "script",
	    "definition",
	    "read",
	    "try"
	};

#ifdef HAVE_PROTOTYPES
void InhibitWindow(Bool whether)
#else
void InhibitWindow(whether)
Bool whether;
#endif
/*Inhibits window saving*/
{
    if (whether) ++inhibitWindow;
    else --inhibitWindow;
}

#ifdef HAVE_PROTOTYPES
Bool BeginCommand(char *);
Bool EndCommand(char *);
#else
Bool BeginCommand();
Bool EndCommand();
#endif

#ifdef HAVE_PROTOTYPES
void MarkVars(ScriptVarPtr vars)
#else
void MarkVars(vars)
ScriptVarPtr vars;
#endif
/*Marks the vars*/
{
    if (!vars) return;
    MarkVars(vars -> left);
    MarkVars(vars -> right);
    if (vars -> value && (vars -> value != ERROBJ))
    {
	MarkObject(vars -> value);
    }
    if (vars -> definition)
    {
	MarkObject(vars -> definition);
    }
}

static ObjPtr MarkScriptStuff(dummy)
ObjPtr dummy;
/*Marks the script stuff*/
{
    int k;

    for (k = 0; k < nextLogElement; ++k)
    {
	MarkVars(logStack[k] . vars);
    }
    for (k = 0; k < nextStackElement; ++k)
    {
	MarkVars(commandStack[k] . vars);
	MarkObject(commandStack[k] . object);
	MarkObject(commandStack[k] . compiledCode);
    }
    

    return ObjTrue;
}

#ifdef HAVE_PROTOTYPES
ScriptVarPtr AddScriptVar(ScriptVarPtr *vars, char *name)
#else
ScriptVarPtr AddScriptVar(vars, name)
ScriptVarPtr *vars;
char *name;
#endif
/*Adds a script var, or returns one if already there*/
{
    if (!*vars)
    {
	*vars = newp(ScriptVar);
	(*vars) -> left = 0;
	(*vars) -> right = 0;
	(*vars) -> name = Alloc(strlen(name) + 1);
	strcpy((*vars) -> name, name);
	(*vars) -> value = ERROBJ;
	(*vars) -> definition = NULLOBJ;
	return *vars;
    }
    else
    {
	int c;
	c = strcmp2(name, (*vars) -> name);
	if (c < 0)
	{
	    return AddScriptVar(&((*vars) -> left), name);
	}
	else if (c > 0)
	{
	    return AddScriptVar(&((*vars) -> right), name);
	}
	else
	{
	    return *vars;
	}
    }
}

#ifdef HAVE_PROTOTYPES
ScriptVarPtr FindScriptVar(ScriptVarPtr vars, char *name)
#else
ScriptVarPtr FindScriptVar(vars, name)
ScriptVarPtr vars;
char *name;
#endif
/*Finds a script var from the name*/
{
    if (!vars)
    {
	return (ScriptVarPtr) 0;
    }
    else
    {
	int c;
	c = strcmp2(name, vars -> name);
	if (c < 0)
	{
	    return FindScriptVar(vars -> left, name);
	}
	else if (c > 0)
	{
	    return FindScriptVar(vars -> right, name);
	}
	else
	{
	    return vars;
	}
    }
}

#ifdef HAVE_PROTOTYPES
ObjPtr GetReadScriptVar(char *name)
#else
ObjPtr GetReadScriptVar(name)
char *name;
#endif
/*Gets a script var while reading a script*/
{
    ScriptVarPtr var;
    int k;

    for (k = nextStackElement - 1; k >= 0; --k)
    {
	var = FindScriptVar(commandStack[k] . vars, name);
	if (var) break;
    }
    
    if (var)
    {
	if (var -> value == ERROBJ)
	{
	    /*Try to evaluate the definition*/
	    Eval(var -> definition);
	}
	return var -> value;
    }
    else
    {
	return ERROBJ;
    }
}

#ifdef HAVE_PROTOTYPES
ScriptVarPtr FindReadScriptVar(char *name)
#else
ScriptVarPtr FindReadScriptVar(name)
char *name;
#endif
/*Finds a read var, returns it or NULL*/
{
    int k;
    ScriptVarPtr var;

    for (k = nextStackElement - 1; k >= 0; --k)
    {
	var = FindScriptVar(commandStack[k] . vars, name);
	if (var) break;
    }
    return var;
}

#ifdef HAVE_PROTOTYPES
ScriptVarPtr SetReadScriptVar(char *name, ObjPtr object)
#else
ScriptVarPtr SetReadScriptVar(name, object)
char *name;
ObjPtr object;
#endif
/*Sets a script var while reading a script*/
{
    ScriptVarPtr var;

    var = FindReadScriptVar(name);
    
    if (var)
    {
	var -> value = object;
	return var;
    }
    else
    {
	return NewReadScriptVar(name, object);
    }
}

#ifdef HAVE_PROTOTYPES
ScriptVarPtr NewReadScriptVar(char *name, ObjPtr object)
#else
ScriptVarPtr NewReadScriptVar(name, object)
char *name;
ObjPtr object;
#endif
/*Adds a script var while reading a script*/
{
    ScriptVarPtr retVal;

    retVal = AddScriptVar(&(commandStack[nextStackElement - 1] . vars), name);

    if (retVal)
    {
	retVal -> value = object;
    }
    return retVal;
}

#ifdef HAVE_PROTOTYPES
ScriptVarPtr FindScriptValue(ScriptVarPtr vars, ObjPtr value)
#else
ScriptVarPtr FindScriptValue(vars, value)
ScriptVarPtr vars;
ObjPtr value;
#endif
/*Finds a script var from the value*/
{
    if (!vars)
    {
	return (ScriptVarPtr) 0;
    }
    else
    {
	ScriptVarPtr test;

	if (vars -> value == value) return vars;
	test = FindScriptValue(vars -> left, value);
	if (test) return test;
	test = FindScriptValue(vars -> right, value);
	return test;	
   }
}

#ifdef HAVE_PROTOTYPES
void FreeScriptVars(ScriptVarPtr vars)
#else
void FreeScriptVars(vars)
ScriptVarPtr vars;
#endif
/*Frees vars*/
{
    if (!vars) return;
    FreeScriptVars(vars -> left);
    vars -> left = (ScriptVarPtr) 0;
    FreeScriptVars(vars -> right);
    vars -> right = (ScriptVarPtr) 0;
    Free(vars -> name);
    Free(vars);
}

TextFilePtr OpenLogFile(name)
char *name;
/*Opens a log file named name in the current directory.  OpenLogFile and
  CloseLogFile may be nested.  If !name, it's stdout*/
{
    struct timeval date;
    char line[200];

    logStack[nextLogElement] . logFile = OpenTextFile(name, TF_WRITE + TF_AUTOWRAP);
    ParseFromFile(logStack[nextLogElement] . logFile);

    if (logStack[nextLogElement] . logFile)
    {
	logStack[nextLogElement] . logInhibit = 0;
	logStack[nextLogElement] . logWindow = 0;
	logStack[nextLogElement] . vars = (ScriptVarPtr) 0;
    }
    else
    {
	WinInfoPtr alertWindow;
	char msg[300];
	getcwd(tempStr, TEMPSTRSIZE);

	sprintf(msg, "File %s cannot be saved in directory %s.",
		name, tempStr);
	alertWindow = AlertUser(UIERRORALERT, (WinInfoPtr) NULLOBJ,
			msg, (FuncTyp) 0, 1, "Oh Well");
	SetVar((ObjPtr) alertWindow, HELPSTRING,
		NewString("A file cannot be saved in the specified directory.  \
This is probably because the permissions on the directory are not set up to allow \
you to write to it.  Check the permissions on the directory and try again."));
	SetVar((ObjPtr) alertWindow, INHIBITLOGGING, ObjTrue);
	return 0;
    }

    /*Write out a version number as first line in file*/
    PutTextString(logStack[nextLogElement] . logFile, SCIANVERSION);
    PutTextChar(logStack[nextLogElement] . logFile, '\n');

    /*And a time stamp*/
    gettimeofday(&date, (struct timezone *)0);

    sprintf(line, "time %s\n", ctime(&date.tv_sec));
    PutTextString(logStack[nextLogElement] . logFile, line);

    return logStack[nextLogElement++] . logFile;
}

void CloseLogFile()
/*Closes a log file opened by OpenLogFile.*/
{
    if (!nextLogElement) return;
    --nextLogElement;
    FreeScriptVars(logStack[nextLogElement] . vars);
    logStack[nextLogElement] . vars = (ScriptVarPtr) 0;
    CloseTextFile(logStack[nextLogElement] . logFile);
}

void CloseAllLogFiles()
/*Closes all log files*/
{
    while (nextLogElement) CloseLogFile();
}

Bool logBeginning = true;	/*True iff it's at the beginning of a log line*/

Bool AnyLogging()
/*Returns true iff any logging is to be done*/
{
    if (!nextLogElement)
    {
	return false;
    }

    if (logStack[nextLogElement - 1] . logInhibit)
    {
	return false;
    }
    return true;
}

#ifdef HAVE_PROTOTYPES
ScriptVarPtr FindLogScriptVar(char *name)
#else
ScriptVarPtr FindLogScriptVar(name)
char *name;
#endif
/*Finds a log script var with name*/
{
    ScriptVarPtr var;
    int k;

    for (k = nextLogElement - 1; k >= 0; --k)
    {
	if (logStack[k] . logInhibit)
	{
	    /*Done; don't do no more.*/
	    return (ScriptVarPtr) 0;
	}

	var = FindScriptVar(logStack[k] . vars, name);
	if (var) return var;
    }

    return (ScriptVarPtr) 0;
}

#ifdef HAVE_PROTOTYPES
ScriptVarPtr FindLogScriptValue(ObjPtr value)
#else
ScriptVarPtr FindLogScriptValue(value)
ObjPtr value;
#endif
/*Finds a log script var with value*/
{
    ScriptVarPtr var;
    int k;

    for (k = nextLogElement - 1; k >= 0; --k)
    {
	if (logStack[k] . logInhibit)
	{
	    /*Done; don't do no more.*/
	    return (ScriptVarPtr) 0;
	}

	var = FindScriptValue(logStack[k] . vars, value);
	if (var) return var;
    }

    return (ScriptVarPtr) 0;
}

#ifdef HAVE_PROTOTYPES
char *LogScriptVarForObject(ObjPtr object)
#else
char *LogScriptVarForObject(object)
ObjPtr object;
#endif
/*Logs a script variable to describe object*/
{
    ObjPtr var;
    char *putNumber;
    char *s, *dest;
    int k;
    ScriptVarPtr newVar;
    static char varStr[AUTOVARMAXLEN];

    newVar = FindLogScriptValue(object);
    if (newVar)
    {
	return newVar -> name;
    }

    /*Determine a name*/
    MakeVar(object, NAME);
    var = GetVar(object, NAME);
    if (var)
    {
	s = GetString(var);
	k = 0;
	dest = varStr;
	if (isdigit(*s))
	{
	    *dest++ = '\\';
	    ++k;
	}
	while (*s && k < AUTOVARMAXLEN - 6)
	{
	    if (*s == ' ')
	    {
		*dest++ = '_';
		++k;
		++s;
	    }
	    else
	    {
		if (!isalpha(*s) && !isdigit(*s))
		{
		    *dest++ = '\\';
		    ++k;
		}
		*dest++ = *s++;
		++k;
	    }
	}
	*dest = 0;
    }

    putNumber = varStr + strlen(varStr);

    k = 0;

    do
    {
	if (k) sprintf(putNumber, "%d", k);
	++k;
    } while (FindLogScriptVar(varStr));

    /*Now there's a good name*/
    for (k = nextLogElement - 1; k >= 0; --k)
    {
	if (logStack[k] . logInhibit)
	{
	    /*Done; don't do no more.*/
	    break;
	}

	newVar = AddScriptVar(&(logStack[k] . vars), varStr);
	newVar -> value = object;
    }

    /*Actually stuff it out there*/
    Log("variable $");
    Log(varStr);
    Log("\n");
    return varStr;
}

#ifdef HAVE_PROTOTYPES
void Log(char *s)
#else
void Log(s)
char *s;
#endif
/*Logs command s to all the open log files*/
{
    int whichLog;
    char *t;

    if (!inhibitWindow)
    {
	if (selWinInfo && GetPredicate((ObjPtr) selWinInfo, INHIBITLOGGING))
	{
	    return;
	}
    }

    if (!nextLogElement) return;

    for (whichLog = nextLogElement - 1; whichLog >= 0; --whichLog)
    {
	if (logStack[whichLog] . logInhibit)
	{
	    /*Done; don't do no more.*/
	    break;
	}
        
	if (!inhibitWindow && logBeginning && selWinInfo && logStack[whichLog] . logWindow != selWinInfo)
	{
	    /*Make a window command*/
	    char windowName[256];
	    int k;

	    /*Copy window title*/
	    strcpy(windowName, "window ");
	    t = selWinInfo -> winTitle;
	    k = strlen(windowName);
	    while (*t)
	    {
		if (*t == ' ')
		{
		    windowName[k++] = '_';
		    ++t;
		}
		else
		{
		    if (!isalpha(*t) && ((k == 0) || !isdigit(*t)))
		    {
			windowName[k++] = '\\';
		    }
		    windowName[k++] = *t++;
		}
	    }
	    windowName[k] = '\n';
	    windowName[k + 1] = '\0';

	    PutTextString(logStack[whichLog] . logFile, windowName);
	    logStack[whichLog] . logWindow = selWinInfo;
	}
	PutTextString(logStack[whichLog] . logFile, s);
    }
    /*Determine if it's in a line*/
    t = s;
    if (*t)
    {
	while (*t) ++t;
	--t;
	if (*t == '\n')
	{
	    logBeginning = true;
	}
	else
	{
	    logBeginning = false;
	}
    }
}

#ifdef HAVE_PROTOTYPES
void InhibitLogging(Bool whether)
#else
void InhibitLogging(whether)
Bool whether;
#endif
/*Inhibits or disinhibits logging based on whether*/
{
    if (nextLogElement)
    {
	if (whether)
	{
	    ++logStack[nextLogElement - 1] . logInhibit;
	}
	else
	{
	    --logStack[nextLogElement - 1] . logInhibit;
	}
    }
	if (whether)
	{
	    ++inhibitRecording;
	}
	else
	{
	    --inhibitRecording;
	}
}

#ifdef HAVE_PROTOTYPES
void LogVariable(ObjPtr object, NameTyp variable)
#else
void LogVariable(object, variable)
ObjPtr object; NameTyp variable;
#endif
/*Logs a variable*/
{
    if (AnyLogging() && !GetPredicate(object, INHIBITLOGGING))
    {
	char cmd[256];
	ObjPtr value;

	value = GetVar(object, variable);

	if (IsInt(value))
	{
	    sprintf(cmd, "set variable %s (int) ", GetInternalString(variable));
	}
	else
	{
	    sprintf(cmd, "set variable %s ", GetInternalString(variable));
	}
	Log(cmd);
	LogObject(value);
	Log("\n");
    }
}

void LogControl(object)
ObjPtr object;
{
    ObjPtr repObj;
    ObjPtr var;
    NameTyp whichVar;

    if (GetPredicate(object, INHIBITLOGGING)) return;

    if (AnyLogging())
    {
	char cmd[256];

	Log("set value ");
	MakeObjectName(cmd, object);
	Log(cmd);
	Log(" ");
	LogObject(GetVar(object, VALUE));
	Log("\n");
    }

    /*Record the object var, if appropriate*/

    if (GetPredicate(object, INHIBITRECORDING)) return;

    if (inhibitRecording) return;

    repObj = GetVar(object, REPOBJ);
    var = GetVar(object, WHICHVAR);
    if (repObj && var)
    {
	whichVar = GetSymbolID(var);
	RecordObjectVar(repObj, whichVar);
    }
}

#ifdef HAVE_PROTOTYPES
void LogReal(real value)
#else
void LogReal(value)
real value;
#endif
/*Logs a real number*/
{
    char s[200];

    if (value == missingData)
    {
	strcpy(s, "# Missing");
    }
    else if (value >= PLUSINF)
    {
	strcpy(s, "# Infinity");
    }
    else if (value <= MINUSINF)
    {
	strcpy(s, "# -Infinity");
    }
    else
    {
	sprintf(s, "%g", (float) value);
    }
    Log(s);
}

#ifdef HAVE_PROTOTYPES
void LogObject(ObjPtr object)
#else
void LogObject(object)
ObjPtr object;
#endif
/*Logs an object*/
{
    FuncTyp method;

    if (!object)
    {
	Log("NULLOBJ");
	return;
    }

    method = GetMethod(object, LOGEXPR);
    if (method)
    {
	(*method)(object);
	return;
    }

    /*Drop through, special case logs*/
    if (IsReal(object))
    {
	LogReal(GetReal(object));
    }
    else if (IsInt(object))
    {
	sprintf(tempStr, "%ld", (long) GetInt(object));
	Log(tempStr);
    }
    else if (IsShortArray(object))
    {
	long *indices;
	long *dims;
	short *elements;
	int rank;
	int k;
	char numStr[100];

	rank = RANK(object);
	dims = DIMS(object);
	elements = ELEMENTS(object);

	indices = Alloc(sizeof(long) * rank);
	for (k = 0; k < rank; ++k)
	{
	    indices[k] = 0;
	}

	Log("(shortarray) ");
	for (k = 0; k < rank; ++k) Log("[");

	for(;;)
	{
	    /*Spit out the next value*/
	    sprintf(numStr, "%d", *elements++);
	    Log(numStr);

	    for (k = rank - 1; k >= 0; --k)
	    {
		++indices[k];
		if (indices[k] >= dims[k])
		{
		    Log("]");
		    indices[k] = 0;
		}
		else
		{
		    Log(" ");
		    break;
		}
	    }

	    if (k < 0) break;
	    while (k < rank - 1)
	    {
		Log("[");
		++k;
	    }
	}
    }
    else if (IsPointerArray(object))
    {
	long *indices;
	long *dims;
	void **elements;
	int rank;
	int k;
	char numStr[100];

	rank = RANK(object);
	dims = DIMS(object);
	elements = ELEMENTS(object);

	indices = Alloc(sizeof(long) * rank);
	for (k = 0; k < rank; ++k)
	{
	    indices[k] = 0;
	}

	Log("(pointerarray) ");
	for (k = 0; k < rank; ++k) Log("[");

	for(;;)
	{
	    /*Spit out the next value*/
	    sprintf(numStr, "%x", *elements++);
	    Log(numStr);

	    for (k = rank - 1; k >= 0; --k)
	    {
		++indices[k];
		if (indices[k] >= dims[k])
		{
		    Log("]");
		    indices[k] = 0;
		}
		else
		{
		    Log(" ");
		    break;
		}
	    }

	    if (k < 0) break;
	    while (k < rank - 1)
	    {
		Log("[");
		++k;
	    }
	}
    }
    else if (IsByteArray(object))
    {
	long *indices;
	long *dims;
	char *elements;
	int rank;
	int k;
	char numStr[100];

	rank = RANK(object);
	dims = DIMS(object);
	elements = ELEMENTS(object);

	indices = Alloc(sizeof(long) * rank);
	for (k = 0; k < rank; ++k)
	{
	    indices[k] = 0;
	}

	Log("(bytearray) ");
	for (k = 0; k < rank; ++k) Log("[");

	for(;;)
	{
	    /*Spit out the next value*/
	    sprintf(numStr, "%d", *elements++);
	    Log(numStr);

	    for (k = rank - 1; k >= 0; --k)
	    {
		++indices[k];
		if (indices[k] >= dims[k])
		{
		    Log("]");
		    indices[k] = 0;
		}
		else
		{
		    Log(" ");
		    break;
		}
	    }

	    if (k < 0) break;
	    while (k < rank - 1)
	    {
		Log("[");
		++k;
	    }
	}
    }
    else if (IsObjArray(object))
    {
	long *indices;
	long *dims;
	ObjPtr *elements;
	int rank;
	int k;
	char numStr[100];

	rank = RANK(object);
	dims = DIMS(object);
	elements = ELEMENTS(object);

	indices = Alloc(sizeof(long) * rank);
	for (k = 0; k < rank; ++k)
	{
	    indices[k] = 0;
	}

	Log("(objectarray) ");
	for (k = 0; k < rank; ++k) Log("[");

	for(;;)
	{
	    /*Spit out the next value*/
	    LogObject(*elements++);
	    Log(numStr);

	    for (k = rank - 1; k >= 0; --k)
	    {
		++indices[k];
		if (indices[k] >= dims[k])
		{
		    Log("]");
		    indices[k] = 0;
		}
		else
		{
		    Log(" ");
		    break;
		}
	    }

	    if (k < 0) break;
	    while (k < rank - 1)
	    {
		Log("[");
		++k;
	    }
	}
    }
    else if (IsRealArray(object))
    {
	long *indices;
	long *dims;
	real *elements;
	int rank;
	int k;
	char numStr[100];

	rank = RANK(object);
	dims = DIMS(object);
	elements = ELEMENTS(object);

	indices = Alloc(sizeof(long) * rank);
	for (k = 0; k < rank; ++k)
	{
	    indices[k] = 0;
	}

	for (k = 0; k < rank; ++k) Log("[");

	for(;;)
	{
	    /*Spit out the next value*/
	    if (*elements == missingData ||
		*elements >= plusInf ||
		*elements <= minusInf)
	    {
		Log("#");
	    }

	    PrintNumber(numStr, *elements++);
	    Log(numStr);

	    for (k = rank - 1; k >= 0; --k)
	    {
		++indices[k];
		if (indices[k] >= dims[k])
		{
		    Log("]");
		    indices[k] = 0;
		}
		else
		{
		    Log(" ");
		    break;
		}
	    }

	    if (k < 0) break;
	    while (k < rank - 1)
	    {
		Log("[");
		++k;
	    }
	}
    }
    else if (IsByteArray(object))
    {
	long *indices;
	long *dims;
	char *elements;
	int rank;
	int k;
	char numStr[100];

	rank = RANK(object);
	dims = DIMS(object);
	elements = ELEMENTS(object);

	indices = Alloc(sizeof(long) * rank);
	for (k = 0; k < rank; ++k)
	{
	    indices[k] = 0;
	}

	Log("(bytearray) ");
	for (k = 0; k < rank; ++k) Log("[");

	for(;;)
	{
	    /*Spit out the next value*/
	    sprintf(numStr, "%d", *elements++);
	    Log(numStr);

	    for (k = rank - 1; k >= 0; --k)
	    {
		++indices[k];
		if (indices[k] >= dims[k])
		{
		    Log("]");
		    indices[k] = 0;
		}
		else
		{
		    Log(" ");
		    break;
		}
	    }

	    if (k < 0) break;
	    while (k < rank - 1)
	    {
		Log("[");
		++k;
	    }
	}
    }
    else if (IsPointerArray(object))
    {
	long *indices;
	long *dims;
	void **elements;
	int rank;
	int k;
	char numStr[100];

	rank = RANK(object);
	dims = DIMS(object);
	elements = ELEMENTS(object);

	indices = Alloc(sizeof(long) * rank);
	for (k = 0; k < rank; ++k)
	{
	    indices[k] = 0;
	}

	Log("(pointerarray) ");
	for (k = 0; k < rank; ++k) Log("[");

	for(;;)
	{
	    /*Spit out the next value*/
	    sprintf(numStr, "%x", *elements++);
	    Log(numStr);

	    for (k = rank - 1; k >= 0; --k)
	    {
		++indices[k];
		if (indices[k] >= dims[k])
		{
		    Log("]");
		    indices[k] = 0;
		}
		else
		{
		    Log(" ");
		    break;
		}
	    }

	    if (k < 0) break;
	    while (k < rank - 1)
	    {
		Log("[");
		++k;
	    }
	}
    }
    else if (IsObjArray(object))
    {
	long *indices;
	long *dims;
	ObjPtr *elements;
	int rank;
	int k;

	rank = RANK(object);
	dims = DIMS(object);
	elements = ELEMENTS(object);

	indices = Alloc(sizeof(long) * rank);
	for (k = 0; k < rank; ++k)
	{
	    indices[k] = 0;
	}

	Log("(objectarray) ");
	for (k = 0; k < rank; ++k) Log("[");

	for(;;)
	{
	    /*Spit out the next value*/
	    LogObject(*elements++);

	    for (k = rank - 1; k >= 0; --k)
	    {
		++indices[k];
		if (indices[k] >= dims[k])
		{
		    Log("]");
		    indices[k] = 0;
		}
		else
		{
		    Log(" ");
		    break;
		}
	    }

	    if (k < 0) break;
	    while (k < rank - 1)
	    {
		Log("[");
		++k;
	    }
	}
    }
    else if (IsString(object))
    {
	/*String*/
	LogString(GetString(object));
    }
    else if (IsList(object))
    {
	/*List*/
	ThingListPtr runner;

	Log("{");
	for (runner = LISTOF(object); runner; runner = runner -> next)
	{
	    LogObject(runner -> thing);
	    if (runner -> next) Log(", ");
	}
	Log("}");
    }
    else
    {
	LogObjectName(object);
    }
}

void LogObjectName(object)
ObjPtr object;
/*Logs the name of an object*/
{
    MakeObjectName(tempStr, object);
    Log(tempStr);
}

#ifdef HAVE_PROTOTYPES
void AbortScriptForError(void)
#else
void AbortScriptForError()
#endif
/*Aborts a script for an error*/
{
    if (abortScriptP)
    {
	abortScript = true;
    }
}

void ScriptError(e)
char *e;
/*Prints out an error e*/
{
    if (curScript)
    {
	CurLineError(curScript, e, begToken, endToken);
    }
    AbortScriptForError();
}

void SelectNamedWindow(name)
char *name;
/*Selects a window by name name*/
{
    WinInfoPtr namedWindow;
    namedWindow = GetWinFromTitle(name);
    if (!namedWindow)
    {
	ScriptError("Cannot find a unique window by that name.");
	return;
    }
    SelWindow(namedWindow);
    scriptWindow = namedWindow;
}

static ObjPtr testObj;
static WinInfoPtr repWindow;
static ObjPtr globalRep;

void TestWindowOnObject(window)
WinInfoPtr window;
/*Tests to see if window has an object which represents testObj*/
{
    ObjPtr rep;
    if (rep = ObjectWhichRepresents(window, testObj))
    {
	repWindow = window;
	globalRep = rep;
    }
}

#ifdef HAVE_PROTOTYPES
void MakeObjectSimpleName(char *dest, ObjPtr object)
#else
void MakeObjectSimpleName(dest, object)
char *dest;
ObjPtr object;
#endif
/*Makes a simple object name from object and puts it into dest.*/
{
    ObjPtr name;
    char *s;

    name = GetStringVar("MakeObjectSimpleName", object, NAME);
    if (!name)
    {
	*dest = 0;
	return;
    }

    /*Copy name*/
    s = GetString(name);
    if (isdigit(*s))
    {
	*dest++ = '\\';
    }
    while (*s)
    {
	if (*s == ' ')
	{
	    *dest++ = '_';
	    ++s;
	}
	else
	{
	    if (!isalpha(*s) && !isdigit(*s))
	    {
		*dest++ = '\\';
	    }
	    *dest++ = *s++;
	}
    }
    *dest = 0;
}

#ifdef HAVE_PROTOTYPES
Bool IsUniqueInDatabase(ObjPtr object)
#else
Bool IsUniqueInDatabase(object)
ObjPtr object;
#endif
/*Returns true iff object exists and is unique in the database*/
{
    ObjPtr keyList, resultList, classID;
    keyList = NewList();
    PostfixList(keyList, NewSymbol(NAME));
    PostfixList(keyList, GetVar(object, NAME));
    PostfixList(keyList, NewSymbol(CLASSID));
    PostfixList(keyList, classID = GetVar(object, CLASSID));
    resultList = SearchDatabase(keyList);
    if (classID && resultList && LISTOF(resultList) && !(LISTOF(resultList) -> next))
    {
	return true;
    }
    else
    {
	return false;
    }
}

#ifdef HAVE_PROTOTYPES
Bool MakeObjectName(char *dest, ObjPtr object)
#else
Bool MakeObjectName(dest, object)
char *dest;
ObjPtr object;
#endif
/*Makes a long object name from object and puts it into dest.  Returns 
  true iff the object name requires the current window.*/
{
    register char *s;
    ObjPtr name;
    ObjPtr representative = NULLOBJ;
    Bool retVal = true;
    Bool repObjP;
    ScriptVarPtr var;

    /*See if it's a variable*/
    var = FindLogScriptValue(object);
    if (var)
    {
	strcpy(dest, "$");
	strcpy(dest + 1, var -> name);
	return false;
    }

    /*See if there's an entry in the database*/
    if (IsUniqueInDatabase(object))
    {
	/*Unique database entry*/
	char *name;
	ObjPtr classID;

	classID = GetVar(object, CLASSID);
	name = ClassIDToName(GetInt(classID));

	if (!name)
	{
	    ReportError("MakeObjectName", "No fixed class name");
	}
	else
	{
	    strcpy(dest, name);
	    while (*dest) ++dest;
	    *dest++ = '@';
	    retVal = false;
	}
    }
    else if (!IsWindow(object) &&
	((!IsValidWindow(selWinInfo)) ||
	 (!(representative = ObjectWhichRepresents(selWinInfo, object)))))
    {
	/*Must find a window that knows about the object*/
	testObj = object;
	repWindow = (WinInfoPtr) 0;

	ForAllWindows(TestWindowOnObject);

	if (repWindow)
	{
	    representative = globalRep;
	    MakeObjectName(dest, (ObjPtr) repWindow);
	    while (*dest) ++dest;
	}
	else
	{
	    *dest++ = '?';
	}
	*dest++ = ':';
    }

    repObjP = false;
    if (representative)
    {
	if (!GetPredicate(representative, REPOBJONLY) && GetVar(representative, REPOBJ) == object)
	{
	    repObjP = true;
	}
	object = representative;
    }

    MakeObjectSimpleName(dest, object);
    if (repObjP)
    {
	strcpy(dest, ".REPOBJ");
    }
    return retVal;
}

int ParseKeyword(s, l, n)
char *s;
char *l[];
int n;
/*Parses keyword s in list l with number of elements n.Returns its number or -1 == not found, -2 ==
  ambiguious.*/
{
    int match = -1;		/*Last found match*/
    int i, k;			/*Counters*/

    /*Seek an exact match*/
    for (k = 0; k < n; ++k)
    {
	for (i = 0; s[i]; ++i)
	{
	    if (tolower(s[i]) != tolower(l[k][i]))
	    {
		break;
	    }
	}
	if (!s[i] && !keywords[k][i])
	{
	    /*It's a match!*/
	    return k;
	}
    }

    /*Seek a partial match*/
    for (k = 0; k < n; ++k)
    {
	for (i = 0; s[i]; ++i)
	{
	    if (tolower(s[i]) != tolower(l[k][i]))
	    {
		break;
	    }
	}
	if (!s[i])
	{
	    /*It's a match!*/
	    if (match >= 0)
	    {
		/*Double match*/
		return -2;
	    }
	    match = k;
	}
    }
    return match;
}

char *PrintScriptString(dest, s)
char *dest, *s;
/*Prints a script string from s into dest*/
{
	*dest++ = '"';
	while (*s)
	{
	    if (*s == '"')
	    {
		*dest++ = '\\';
	    }
	    if (*s == '\n')
	    {
		*dest++ = '\\';
		*dest++ = 'n';
	    }
	    else if (*s == '\t')
	    {
		*dest++ = '\\';
		*dest++ = 't';
	    }
	    else if (*s == '\r')
	    {
		*dest++ = '\\';
		*dest++ = 'r';
	    }
	    else if (*s < ' ')
	    {
		*dest++ = '\\';
		sprintf(tempStr, "%3o", *s);
		*dest++ = tempStr[0];
		*dest++ = tempStr[1];
		*dest++ = tempStr[2];
	    }
	    else
	    {
		*dest++ = *s;
	    }
	    ++s;
	}
	*dest++ = '"';
	*dest = 0;
	return dest;
}

#ifdef HAVE_PROTOTYPES
void LogString(char *s)
#else
void LogString(s)
char *s;
#endif
/*Logs a string value*/
{
    int len;
    char *d;

    len = strlen(s);
    if (len < (TEMPSTRSIZE / 4))
    {
	d = tempStr;
    }
    else
    {
	d = Alloc(len * 2);
    }

    PrintScriptString(d, s);
    Log(d);

    if (len >= (TEMPSTRSIZE / 4))
    {
	Free(d);
    }
}

int FindCommandInStack(command)
int command;
/*Finds the topmost command command in the stack.  Returns it's index or -1
  if not found.*/
{
    int k;
    for (k = nextStackElement - 1; k >= 0; --k)
    {
	if (commandStack[k] . command == command)
	{
	    break;
	}
    }
    return k;
}

Bool BeginCommand(args)
char *args;
/*Begins a begin-end block with command in args.*/
{
    char cmdNum;
    char arg[256];

    if (nextStackElement >= MAXSTACKDEPTH)
    {
	ScriptError("There are too many nested begin-end blocks.");
	return true;
    }

    SHIFTKW(arg, args);

    if (strlen(arg))
    {
	cmdNum = ParseKeyword(arg, keywords, NKEYWORDS);
	if (cmdNum == -1)
	{
	    ScriptError("This command is unknown.");
	    return true;
	}
	else if (cmdNum == -2)
	{
	    ScriptError("This command is ambiguous.");
	    return true;
	}
    }
    else
    {
	/*It's an anonymous block*/
	cmdNum = KWBEGIN;
    }
    commandStack[nextStackElement] . command = cmdNum;
    commandStack[nextStackElement] . object = NULLOBJ;
    commandStack[nextStackElement] . compiledCode = NULLOBJ;
    commandStack[nextStackElement] . vars = (ScriptVarPtr) 0;
    switch (cmdNum)
    {
	case KWBEGIN:
	    Log("begin\n");
	    break;
	case KWRECORDING:
	    Log("begin recording\n");
	    if (FindCommandInStack(KWRECORDING) >= 0)
	    {
		ScriptError("The begin recording command cannot be nested.");
		return true;
	    }
	    else
	    {
		real nSeconds;
		
		SHIFTNUM(arg, args);
		if (1 != sscanf(arg, "%g", &nSeconds))
		{
		    ScriptError("A maximum time in seconds is expected here.");
		    return true;
		}
		if (nSeconds < 0.0) nSeconds = -nSeconds;
		if (AdjustToVisWindows(curRecorder))
		if (ConnectRecorder())
		{
		    int nFrames;
		    ObjPtr var;

		    var = GetVar(curRecorder, FRAMERATE);
		    if (var)
		    {
			nFrames = (int) (nSeconds * GetReal(var)) + 1;
		    }
		    else
		    {
			nFrames = (int) (nSeconds * 30.0) + 1;
		    }
		    PushNonVisWindows();
		    if (!PrepareToRecord(nFrames))
		    {
			ScriptError("There is a problem recording that many frames.");
			return true;
		    }
		}
		else
		{
		    ScriptError("Could not connect to the recorder.");
		    return true;
		}
		else
		{
		    ScriptError("There are no visualization windows to record.");
		    return true;
		}
	    }
	    break;
	case KWSETUP:
	    Log("begin setup\n");
	    settingUp = true;
	    break;
	case KWSNAPSHOT:
	    /*Beginning of an object snapshot*/
	    {
		ObjPtr value;
		int cmdWhere;

		value = ERROBJ;
		SKIPBLANKS(args);
		Log("begin snapshot ");
		Log(args);
		Log("\n");

		cmdWhere = FindCommandInStack(KWDEFINITION);
		if (cmdWhere >= 0)
		{
		    args = ParseExpression(args, &value);
		}
		else
		{
		    args = EvalExpressionArg(args, &value);
		}
		if (value != ERROBJ)
		{
		    commandStack[nextStackElement] . object = value;
		}
		else
		{
		    AbortScriptForError();
		    commandStack[nextStackElement] . object = NULLOBJ;
		}
	    }
	    break;
	case KWSCRIPT:
	    /*Sub script, do nothing*/
	    break;
	case KWDEFINITION:
	    /*Definition of variable*/
	    SKIPBLANKS(args);
	    if (*args == '$') ++args;
	    Log("begin definition $");
	    Log(args);
	    Log("\n");
	    SHIFTKW(arg, args);
	    commandStack[nextStackElement] . cVar = FindReadScriptVar(arg);
	    if (!commandStack[nextStackElement] . cVar)
	    {
		ScriptError("This variable has not been declared");
		return true;
	    }
	    commandStack[nextStackElement] . object = NewScriptVariable(commandStack[nextStackElement] . cVar);
	    commandStack[nextStackElement] . compiledCode = NewCodeSequence();
	    compiledCode = commandStack[nextStackElement] . compiledCode;
	    break;
	default:
	    ScriptError("This command may not start a begin-end block");
	    return true;
    }
    ++nextStackElement;
    return true;
}

Bool EndCommand(args)
char *args;
/*Ends the top command on the stack, or up to command in args*/
{
    char arg[256];
    if (!nextStackElement)
    {
	ScriptError("There are too many ends and not enough begins.");
	return true;
    }

    SHIFTKW(arg, args);
    if (strlen(arg))
    {
	int cmdNum, cmdWhere;
	cmdNum = ParseKeyword(arg, keywords, NKEYWORDS);
	cmdWhere = FindCommandInStack(cmdNum);
	if (cmdWhere < 0)
	{
	    ScriptError("There is no such block in force.");
	    return true;
	}
	else
	{
	    do
	    {
		EndCommand("");
	    } while (commandStack[nextStackElement] . command != cmdNum);
	}
    }
    else
    {
	/*End the top command*/
	--nextStackElement;
	switch(commandStack[nextStackElement] . command)
	{
	    case KWBEGIN:
		Log("end\n");
		break;
	    case KWRECORDING:
		Log("end recording\n");
		StopRecording();
		DisconnectRecorder();
		break;
	    case KWSNAPSHOT:
		{
		    int cmdWhere;

		    Log("end snapshot\n");

		    cmdWhere = FindCommandInStack(KWDEFINITION);
		    if (cmdWhere < 0)
		    {
			ImInvalid(commandStack[nextStackElement] . object);
		    }
		}
		break;
	    case KWSETUP:
		Log("end setup\n");
		settingUp = false;
		break;
	    case KWSCRIPT:
		/*Script, do nothing.*/
		break;
	    case KWDEFINITION:
		/*Definition*/
		{
		    int cmdWhere;
		
		    Log("end definition\n");

		    commandStack[nextStackElement] . cVar -> definition = 
		        commandStack[nextStackElement] . compiledCode;

		    cmdWhere = FindCommandInStack(KWDEFINITION);
		    if (cmdWhere >= 0)
		    {
			compiledCode = commandStack[cmdWhere] . compiledCode;
		    }
		    else
		    {
			compiledCode = NULLOBJ;
		    }
		}
		break;
	}
	FreeScriptVars(commandStack[nextStackElement] . vars);
	commandStack[nextStackElement] . vars = (ScriptVarPtr) 0;
    }
    return true;
}

Bool DoSetCommand(kwn, args)
int kwn;
char *args;
/*Does a set command*/
{
    char arg[256];

    switch(kwn)
    {
	case KWVARIABLE:
	    /*Set a variable within a snapshot*/
	    {
		int element1, element2;

		element1 = FindCommandInStack(KWSNAPSHOT);
		element2 = FindCommandInStack(KWDEFINITION);


		if (element1 == 0 && element2 == 0)
		{
		    ScriptError("A set variable command can only occur within a snapshot or definition");
		}
		else if (element2 >= 0)
		{
		    /*Compile a setvar, just like it says*/
		    ObjPtr stmt;
		    NameTyp internalID;
		    ObjPtr value;

		    if (element1 < element2) element1 = element2;

		    Log(args);
		    SHIFTKW(arg, args);
		    internalID = GetInternalID(arg);
		    
		    value = ERROBJ;
		    args = ParseExpression(args, &value);
	
		    stmt = NewSetObjVar(commandStack[element1] . object, internalID, value);
		    AppendStatementToCode(commandStack[element2] . compiledCode, stmt);
		}
		else
		{
		    /*Do a setvar, just like it says*/
		    NameTyp internalID;
		    ObjPtr value;

		    SHIFTKW(arg, args);
		    internalID = GetInternalID(arg);
		    
		    value = ERROBJ;
		    args = EvalExpressionArg(args, &value);
		    if (value != ERROBJ)
		    {
			if (commandStack[element1] . object)
			{
			    SetVar(commandStack[element1] . object, internalID, value);
			    LogVariable(commandStack[element1] . object, internalID);
			}
		    }
		    else
		    {
			AbortScriptForError();
		    }
		}
	    }
	    break;
	case KWRECORDER:
	    {
		ObjPtr object;
		if (FindCommandInStack(KWRECORDING) >= 0)
		{
		    ScriptError("You can't set the recorder while recording");
		    break;
		}
		object = ERROBJ;
		args = EvalExpressionArg(args, &object);
		if (object != ERROBJ)
		{
		    if (IsString(object))
		    {
			if (!SetRecorder(GetString(object)))
			{
			    ScriptError("This recorder does not exist");
			}
		    }
		    else
		    {
			ScriptError("A quoted name of a recorder is expected here");
		    }
		}
		else
		{
		    AbortScriptForError();
		}
	    }
	    break;
	case KWVALUE:
	    /*Set the value of a control*/
	    {
		ObjPtr object, value;
		object = ERROBJ;

		args = EvalExpressionArg(args, &object);
		if (object != ERROBJ)
		{
		    if (object && IsObject(object))
		    {
			value = ERROBJ;
			args = EvalExpressionArg(args, &value);
			if (value != ERROBJ)
			{
			    ObjPtr var;

			    var = GetVar(object, NAME);
			
			    SetValue(object, value);
			}
			else
			{
			    AbortScriptForError();
			}
		    }
		    else
		    {
			ScriptError("An object name is expected here");
		    }
		}
		else
		{
		    AbortScriptForError();
		}
	    }
	    break;
	case KWROTATION:
	    {
		ObjPtr rotation, axis;
		rotation = ERROBJ;

		args = EvalExpressionArg(args, &rotation);
		if (rotation != ERROBJ)
		{
		    if (rotation && IsReal(rotation))
		    {
			axis = ERROBJ;
			args = EvalExpressionArg(args, &axis);
			if (axis != ERROBJ && axis &&
			    IsRealArray(axis) && RANK(axis) == 1 && DIMS(axis)[0] == 3)
			{
			    real *elements;
			    real degrees;
			    elements = ELEMENTS(axis);
			    degrees = GetReal(rotation);
			    SetRotationMotor(degrees * M_PI / 180.0, elements[0], elements[1], elements[2]);
			}
			else
			{
			    AbortScriptForError();
			}
		    }
		    else
		    {
			ScriptError("A rotation speed is expected here");
		    }
		}
		else
		{
		    AbortScriptForError();
		}
	    }
	    break;
	case KWFUNCTIONBOX:
	    {
		ObjPtr object, functionBox;
		object = ERROBJ;

		args = EvalExpressionArg(args, &object);
		if (object != ERROBJ)
		{
		    if (IsObject(object))
		    {
			functionBox = ERROBJ;
			args = EvalExpressionArg(args, &functionBox);
			if (functionBox != ERROBJ)
			{
			    SetFunctionBox(object, functionBox);
			    InhibitLogging(true);
			    ImposeColorFunction(object);
			    ForAllVisWindows(ImInvalid);
			    InhibitLogging(false);
			}
			else
			{
			    AbortScriptForError();
			}
		    }
		    else
		    {
			ScriptError("An object name is expected here");
		    }
		}
		else
		{
		    AbortScriptForError();
		}
	    }
	    break;
	case KWCOLOR:
	    {
		ObjPtr object;
		int which, r, g, b;

		object = ERROBJ;

		args = EvalExpressionArg(args, &object);
		if (object != ERROBJ)
		{
		    if (IsObject(object))
		    {
			if (4 == sscanf(args, " %d %d %d %d \n", &which, &r, &g, &b))
			{
			    SetColorBarColor(object, which, r, g, b);
			}
			else
			{
			    ScriptError("A color number and components are expected here");
			}
		    }
		    else
		    {
			ScriptError("An object name is expected here");
		    }
		}
		else
		{
		    AbortScriptForError();
		}
	    }
	    break;
	case KWLOCATION:
	    {
		ObjPtr object, value;
		object = ERROBJ;
		args = EvalExpressionArg(args, &object);
		if (object != ERROBJ)
		{
		    if (IsObject(object))
		    {
			real location[3];
			if (3 == sscanf(args, " [%g %g %g]", 
				&(location[0]),
				&(location[1]),
				&(location[2])))
			{
			    ObjPtr var;
			    var = NewRealArray(1, 3L);
			    CArray2Array(var, location);
			    SetVar(object, LOCATION, var);
			    ImInvalid(object);
			    {
				char cmd[300];
				char *s;
				sprintf(cmd, "set location ");

				s = &(cmd[0]);
				while (*s) ++s;
				MakeObjectName(s, object);
				while (*s) ++s;
				sprintf(s, " [%g %g %g]\n",
					location[0],
					location[1],
					location[2]);
				Log(cmd);
			    }
			}
			else
			{
			    ScriptError("A vector location is expected here");
			}
		    }
		    else
		    {
			ScriptError("An object name is expected here");
		    }
		}
		else
		{
		    AbortScriptForError();
		}
	    }
	    break;
	case KWBOUNDS:
	    {
		ObjPtr object;
		ObjPtr bounds;
		char *oldArgs;
		object = ERROBJ;
		oldArgs = args;
		args = EvalExpressionArg(args, &object);
		if (object != ERROBJ)
		{
		    if (IsObject(object))
		    {
			bounds = ERROBJ;
			args = EvalExpressionArg(args, &bounds);
			if (bounds != ERROBJ)
			{
			    if (IsRealArray(bounds) && RANK(bounds) == 1)
			    {
				ObjPtr testBounds;
				testBounds = GetVar(object, BOUNDS);
				if (!testBounds ||
				    (IsRealArray(testBounds) && 
				     RANK(testBounds) == 1 &&
				     DIMS(testBounds)[0] == DIMS(bounds)[0]))
				{
				    {
					char cmd[256];
					sprintf(cmd, "set bounds %s\n", oldArgs);
					Log(cmd);
				    }
			    	    SetVar(object, BOUNDS, bounds);
				    ImInvalid(object);
				}
				else
				{
				    ScriptError("These bounds are of wrong size for the object");
				}
			    }
			    else
			    {
				ScriptError("A vector is expected here");
			    }
			}
			else
			{
			    AbortScriptForError();
			}
		    }
		    else
		    {
			ScriptError("An object is expected here");
		    }
		}
		else
		{
		    AbortScriptForError();
		}
	    }
	    break;
	case KWENDPOINTS:
	    {
		ObjPtr object;
		int x1, y1, x2, y2;
		char *oldArgs;
		object = ERROBJ;
		oldArgs = args;
		args = EvalExpressionArg(args, &object);
		if (object != ERROBJ)
		{
		    if (IsObject(object))
		    {
			if (4 == sscanf(args, " %d %d %d %d", &x1, &y1, &x2, &y2))
			{
			    SetEndpoints(object, x1, y1, x2, y2);
			    ImInvalid(object);
			}
			else
			{
			    ScriptError("Four endpoint coordinates are expected here");
			}
		    }
		    else
		    {
			ScriptError("An object name is expected here");
		    }
		}
		else
		{
		    AbortScriptForError();
		}
	    }
	    break;
	case KWSCREEN:
	    {
		/*Set the size of the recording screen*/
		int x, y;
		if (2 != sscanf(args, " %d %d", &x, &y))
		{
		    ScriptError("A width and height of the recording screen is expected here");
		    break;
		}
		SetScreenSize(x, y);
	    }
	    break;
	case KWFPS:
	    {
		/*Set the frames per second*/
		real fps;
		if (1 != sscanf(args, " %g", &fps))
		{
		    ScriptError("A number of frames per second is expected here");
		    break;
		}
		SetFPS(fps);
	    }
	    break;
	default:
	    ScriptError("This is not a paramater that can be set");
	    break;
    }
    return true;
}


Bool DoSaveCommand(kwn, args)
int kwn;
char *args;
/*Does a save command*/
{
    char arg[256];

    switch(kwn)
    {
	case KWCONTROLS:
	    DoSaveObject();
	    break;
	case KWSCREEN:
	    SaveScreen(scriptWindow, F_OPTIONDOWN);
	    break;
	case KWWINDOW:
	    SaveScreen(scriptWindow, 0);
	    break;
	default:
	    ScriptError("This is not something that can be saved");
	    return;
	    break;
    }
    return true;
}

Bool DoShowHideCommand(showp, kwn, args)
Bool showp;
int kwn;
char *args;
/*Does a show or hide command, showp is true iff it's show*/
{
    char arg[256];

    switch(kwn)
    {
	case KWPANEL:
	    if (showp)
	    {
		DoShowPanel();
	    }
	    else
	    {
		DoHidePanel();
	    }
	    break;
	case KWDATASETS:
	    if (showp)
	    {
		PopDatasetsWindow();
	    }
	    else
	    {
		ScriptError("You can't hide the datasets window this way.  Close the window instead");
	    }
	    break;
	case KWFILEREADERS:
	    if (showp)
	    {
		PopFileReadersWindow();
	    }
	    else
	    {
		ScriptError("You can't hide the file readers window this way.  Close the window instead");
	    }
	    break;
	case KWRECORDERS:
	    if (showp)
	    {
		PopRecorderDriversWindow();
	    }
	    else
	    {
		ScriptError("You can't hide the recorder drivers window this way.  Close the window instead");
	    }
	    break;
	case KWFILESWINDOW:
	    if (showp)
	    {
		DoNewFileWindow();
	    }
	    else
	    {
		ScriptError("You can't hide the file window this way.  Close the window instead");
	    }
	    break;
	case KWFRAME:
	    if (showp)
	    {
		DoShowFrame();
	    }
	    else
	    {
		DoHideFrame();
	    }
	    break;
	case KWFRONT:
	    if (showp)
	    {
		DoShowFrontPanelControls();
	    }
	    else
	    {
		ScriptError("You can't hide the front panel controls this way.  Close the window instead");
	    }
	    break;
	case KWBACK:
	    if (showp)
	    {
		DoShowBackPanelControls();
	    }
	    else
	    {
		ScriptError("You can't hide the back panel controls this way.  Close the window instead");
	    }
	    break;
	case KWSPACE:
	    if (showp)
	    {
		DoShowSpaceControls();
	    }
	    else
	    {
		ScriptError("You can't hide the space controls this way.  Close the window instead");
	    }
	    break;

	default:
	    ScriptError("This keyword is not a settable parameter");
	    break;
    }
    return true;
}

#ifdef HAVE_PROTOTYPES
void ShowScriptControlPanels(Bool whether)
#else
void ShowScriptControlPanels(whether)
Bool whether;
#endif
/*Turns subsequent showing of control panels within a script on or off*/
{
    showControlPanels = whether;
    Log(whether ? "turn show controls on\n" : "turn show controls off");
}

void VersionCommand(args)
char *args;
/*Does a version command*/
{
    char arg[256];
    float scianVersion, scriptVersion;
    SHIFTNUM(arg, args);

    if (1 != sscanf(arg, "%g", &scriptVersion))
    {
	ScriptError("A number is expected here");
	return;
    }
		
    if (1 != sscanf(SCIANVERSION, "Version %g", &scianVersion))
    {
	ReportError("VersionCommand", "Internal error: cannot parse SciAn version");
	return;
    }

    if (scianVersion < scriptVersion)
    {
	ScriptError("The script is newer than this version of SciAn");
	if (abortScriptP)
	{
	    abortScript = true;
	}
    }
    if (scianVersion > scriptVersion)
    {
#if 0
	ScriptError("The script is older than this version of SciAn");
	if (abortScriptP)
	{
	    abortScript = true;
	}
#endif
    }
}

#ifdef HAVE_PROTOTYPES
void SelectCommand(char *args, Bool selectP)
#else
void SelectCommand(args, selectP)
char *args;
Bool selectP;
#endif
/*Does select or deselect depending on selectP*/
{
	ObjPtr object;
	object = ERROBJ;
	args = EvalExpressionArg(args, &object);
 	if (object != ERROBJ)
	{
	    if (IsObject(object))
	    {
		Select(object, selectP);
	    }
	    else
	    {
		ScriptError("An object name is expected here");
	    }
	}
	else
	{
	    AbortScriptForError();
	}
}

#ifdef HAVE_PROTOTYPES
void TurnCommand(char *args)
#else
void TurnCommand(args)
char *args;
#endif
/*Does turn show controls on/off.  Normal turn on/off done by obj
  functions.*/
    {
	char arg[256];
	int kwn;
	SHIFTKW(arg, args);
	kwn = ParseKeyword(arg, keywords, NKEYWORDS);
	if (kwn == KWSHOW)
	{
	    SHIFTKW(arg, args);
	    kwn = ParseKeyword(arg, keywords, NKEYWORDS);
	    if (kwn == KWCONTROLS)
	    {
		SHIFTKW(arg, args);
		kwn = ParseKeyword(arg, keywords, NKEYWORDS);
		
		if (kwn == KWON)
		{
		    ShowScriptControlPanels(true);
		}
		else if (kwn == KWOFF)
		{
		    ShowScriptControlPanels(false);
		}
	    }
	    else
	    {
		ScriptError("'Controls' is expected here");
	    }
	}
	else
	{
	    ScriptError("Either 'on' or 'off' is expected here");
	}
    }


Bool DoScriptCommand(kwn, args)
int kwn;
char *args;
/*Does command kwn with args*/
{
    char arg[256];

    switch(kwn)
    {
	case KWVERSION:
	    VersionCommand(args);
	    break;
	case KWTIME:
	    break;
	case KWDELETE:
	    DoDelete();
	    break;
	case KWTURN:
	    /*Turn show controls on/off.  Turn on/off handled by objfunctions*/
	    TurnCommand(args);
	    break;
	case KWSELECT:
	    SelectCommand(args, true);
	    break;
	case KWDESELECT:
	    SelectCommand(args, false);
	    break;
	case KWSELECTALL:
	    DoSelectAllIcons();
	    break;
	case KWDESELECTALL:
	    DeselectAll();
	    break;
	case KWVARIABLE:
	    SKIPBLANKS(args);
	    if (*args == '$') ++args;
	    SHIFTKW(arg, args);
	    if (strlen(arg))
	    {
		AddScriptVar(&(commandStack[nextStackElement - 1] . vars), arg);
	    }
	    else
	    {
		ScriptError("A variable name is expected here");
	    }
	    break;
	case KWANNOTATION:
	case KWTIMEREADOUT:
	case KWRECTANGLE:
	    {
		char name[256];
		ObjPtr bounds;
		SHIFTKW(name, args);
		if (strlen(name))
		{
		    bounds = ERROBJ;
		    args = EvalExpressionArg(args, &bounds);
		    if (bounds != ERROBJ)
		    {
			if (IsRealArray(bounds) && RANK(bounds) == 1 &&
				DIMS(bounds)[0] == 4)
			{
			    switch(kwn)
			    {
				case KWANNOTATION:
				    AddAnnotation(name, ArrayMeat(bounds));
				    break;
				case KWTIMEREADOUT:
				    AddTimeReadout(name, ArrayMeat(bounds));
				    break;
				case KWRECTANGLE:
				    AddRectangle(name, ArrayMeat(bounds));
				    break;
			    }
			}
			else
			{
			    ScriptError("A 4-vector is expected here");
			}
		    }
		    else
		    {
			AbortScriptForError();
		    }
		}
		else
		{
		    ScriptError("A name is expected here");
		}
	    }
	    break;
	case KWLINE:
	    {
		char name[256];
		int x1, y1, x2, y2;
		SHIFTKW(name, args);
		if (strlen(name))
		{
		    if (4 == sscanf(args, " %d %d %d %d", &x1, &y1, &x2, &y2))
		    {
			AddLine(name, x1, y1, x2, y2);
		    }
		    else
		    {
			ScriptError("Four endpoint coordinates are expected here");
		    }
		}
		else
		{
		    ScriptError("A name is expected here");
		}
	    }
	    break;
	case KWSET:			/*Set something or other*/
	    {
		int k;
		SHIFTKW(arg, args);
		/*Find out what to set*/
		kwn = ParseKeyword(arg, keywords, NKEYWORDS);
		if (kwn == -1)
		{
		    ScriptError("This parameter is unknown.");
		    break;
		}
		else if (kwn == -2)
		{
		    ScriptError("This parameter is ambiguous.");
		    break;
		}
		else
		{
		    return DoSetCommand(kwn, args);
		}
	    }
	    break;
	case KWSAVE:			/*Save something or other*/
	    {
		int k;
		SHIFTKW(arg, args);
		/*Find out what to set*/
		kwn = ParseKeyword(arg, keywords, NKEYWORDS);
		if (kwn == -1)
		{
		    ScriptError("This parameter is unknown.");
		    break;
		}
		else if (kwn == -2)
		{
		    ScriptError("This parameter is ambiguous.");
		    break;
		}
		else
		{
		    return DoSaveCommand(kwn, args);
		}
	    }
	    break;
	case KWSHOW:			/*Show something or other*/
	    {
		int k;
		SHIFTKW(arg, args);
		/*Find out what to set*/
		kwn = ParseKeyword(arg, keywords, NKEYWORDS);
		if (kwn == -1)
		{
		    ScriptError("This parameter is unknown.");
		    break;
		}
		else if (kwn == -2)
		{
		    ScriptError("This parameter is ambiguous.");
		    break;
		}
		else
		{
		    return DoShowHideCommand(true, kwn, args);
		}
	    }
	    break;
	case KWHIDE:			/*Hide something or other*/
	    {
		int k;
		SHIFTKW(arg, args);
		/*Find out what to set*/
		kwn = ParseKeyword(arg, keywords, NKEYWORDS);
		if (kwn == -1)
		{
		    ScriptError("This parameter is unknown.");
		    break;
		}
		else if (kwn == -2)
		{
		    ScriptError("This parameter is ambiguous.");
		    break;
		}
		else
		{
		    return DoShowHideCommand(false, kwn, args);
		}
	    }
	    break;
	case KWDROP:
	    {
		int dropX, dropY;
		
		SHIFTNUM(arg, args);
		if (1 != sscanf(arg, "%d", &dropX))
		{
		    ScriptError("A number is expected here");
		    break;
		}

		SHIFTNUM(arg, args);
		if (1 != sscanf(arg, "%d", &dropY))
		{
		    ScriptError("A number is expected here");
		    break;
		}

		if (scriptWindow && dragBuffer)
		{
		    FuncTyp method;
		    /*Drop*/
		    method = GetMethod((ObjPtr) scriptWindow, DROPOBJECTS);
		    if (method)
		    {
			iconXOff = 0;
			iconYOff = 0;
			{
			    char cmd[256];
			    sprintf(cmd, "drop %d %d\n", dropX, dropY);
			    Log(cmd);
			    InhibitLogging(true);
			}
			BeginUndo(UT_CHANGESNAPSHOT);
			(*method)(scriptWindow, dragBuffer, dropX, dropY);
			EndUndo();
			DeleteThing(dragBuffer);
			dragBuffer = NULLOBJ;
			    InhibitLogging(false);
		    }
		}
		else
		{
		    ScriptError("There is no selected window");
		}
	    }
	    break;
	case KWVIDEOSCREEN:
	    DoVideoScreen();
	    break;
	case KWFULLSCREEN:
	    DoMaxScreen();
	    break;
	case KWTILE:
	    {
		int width, height;

		SHIFTNUM(arg, args);
		if (1 != sscanf(arg, "%d", &width))
		{
		    ScriptError("A width is expected here");
		    break;
		}

		SHIFTNUM(arg, args);
		if (1 != sscanf(arg, "%d", &height))
		{
		    ScriptError("A height is expected here");
		    break;
		}

		Tile(width, height);
	    }
	    break;
	case KWLOCATE:
	    {
		int l, r, b, t;
		if (4 != sscanf(args, " %d %d %d %d", &l, &r, &b, &t))
		{
		    ScriptError("The bounds of the window are expected here.");
		    return true;
		}
		LocateWindow(l, r, b, t);
	    }
	    break;
	case KWEYEPOSN:
	    {
		if (scriptWindow)
		{
		    ObjPtr array, observer /*, observers*/;
		    array = ERROBJ;
    
		    args = EvalExpressionArg(args, &array);
		    if (array != ERROBJ)
		    {
			observer = FindObserver(scriptWindow);
			if (!observer) break;

			if (IsArray(array) && RANK(array) == 1 && DIMS(array)[0] == 3)
			{
			    real *p;
			    SetVar(observer, LOCATION, array);
			    ImInvalid(observer);
			    p = ELEMENTS(array);
			    {
				char cmd[256];
				sprintf(cmd, "eyeposn [%g %g %g]\n",
				    p[0], p[1], p[2]);
				Log(cmd);
			    }
			}
			else
			{
			    ScriptError("A 3-vector is expected here");
			}
		    }
		    else
		    {
			AbortScriptForError();
		    }
		}
	    }
	    break;
	case KWROLL:
	    {
		float value;
		SHIFTKW(arg, args);
		if (strlen(arg) && 1 == sscanf(arg, " %g", &value))
		{
		    if (scriptWindow)
		    {
			ObjPtr observer /*, observers*/;
			observer = FindObserver(scriptWindow);
			if (!observer) break;

			SetVar(observer, ROLL, NewReal(value));
		    }
		}
		else
		{
		    ScriptError("An amount to roll is expected here.");
		}
	    }
	    break;
	case KWPITCH:
	    {
		float value;
		SHIFTKW(arg, args);
		if (strlen(arg) && 1 == sscanf(arg, " %g", &value))
		{
		    if (scriptWindow)
		    {
			ObjPtr observer/*, observers*/;
			observer = FindObserver(scriptWindow);
			if (!observer) break;

			SetVar(observer, PITCH, NewReal(value));
		    }
		}
		else
		{
		    ScriptError("An amount to pitch is expected here.");
		}
	    }
	    break;
	case KWYAW:
	    {
		ScriptError("YAW is no longer supported.");
	    }
	    break;
	case KWROTATE:
	    {
		char axis;
		float amount;

		SHIFTKW(arg, args);
		if (strlen(arg) == 1 && (*arg == 'x' || *arg == 'y' || *arg == 'z'))
		{
		    axis = *arg;	

		    SHIFTNUM(arg, args);
		    if (!strlen(arg) || 1 != sscanf(arg, "%f", &amount))
		    {
			ScriptError("An amount to rotate is expected here.");
			return true;
		    }

		    if (scriptWindow)
		    {
			RotateOrthoWindow(scriptWindow, axis, amount);
		    }
		}
		else
		{
		    ScriptError("Unrecognized keyword.  An axis (x, y, or z) is expected.");
		}
		return true;
	    }
	    break;
	case KWSHEAR:
	    ScriptError("Shearing is no longer supported.");
	    return true;
	    break;
	case KWRECORD:
	    if (FindCommandInStack(KWRECORDING) < 0)
	    {
		ScriptError("This command can only exist within a BEGIN RECORDING block.");
		return true;
	    }
	    else
	    {
		real nSecs;
		if (1 != sscanf(args, " %g", &nSecs))
		{
		    
		    ScriptError("A number of seconds to record is expected here.");
		}
		else
		{
		    runningTask = RT_RECORDING;
		    framesLeft = nSecs * 30 + 0.5;
		}
		return false;
	    }
	case KWBEGIN:
	    return BeginCommand(args);
	case KWEND:
	    return EndCommand(args);
	case KWSNAP:
	    if (FindCommandInStack(KWRECORDING) < 0)
	    {
		ScriptError("This command can only exist within a BEGIN RECORDING block.");
		return true;
	    }
	    PushNonVisWindows();
	    if (!SnapOneFrame())
	    {
		ScriptError("Could not snap a single frame.");
	    }
	    ++TrashDayFlag;
	    break;
	case KWWINDOW:
	    SHIFTKW(arg, args);
	    if (!strlen(arg))
	    {
		ScriptError("The name of a window is expected here");
		return true;
	    }
	    SelectNamedWindow(arg);
	    break;
	case KWEXIT:
	case KWQUIT:
	    DoQuit();
	    break;
	case KWCLOSE:
	    CloseWindow(scriptWindow);
	    SelWindow(0);
	    scriptWindow = 0;
	    break;
	case KWPUSHWINDOW:
	    if (scriptWindow)
	    {
		PushWindow(scriptWindow);
	    }
	    else
	    {
		ScriptError("There is no current window");
	    }
	    break;
	case KWSHELL:			/*Shell escape*/
	    SKIPBLANKS(args);
	    if ((getuid() == geteuid()) &&
		(getgid() == getegid()))
	    {
	        system(args);
	    }
	    break;
	case KWSCRSAVE:
	    {
		char fileName[256];
		int l, r, b, t;
		int n;
		n = sscanf(args, " %s %d %d %d %d", fileName, &l, &r, &b, &t);
		if (n == 5)
		{
		    sprintf(tempStr, "scrsave %s %d %d %d %d", fileName, l, r, b, t);
		    if ((getuid() == geteuid()) &&
			(getgid() == getegid()))
		    {
		        system(tempStr);
		    }
		}
		else if (n == 1)
		{
		    sprintf(tempStr, "scrsave %s", fileName);
		    if ((getuid() == geteuid()) &&
			(getgid() == getegid()))
		    {
		        system(tempStr);
		    }
		}
		else
		{
		    ScriptError("A filename and four sides are expected here");
		    return true;
		}
	    }
	    break;
	case KWREAD:
	    {
		ObjPtr path = ERROBJ, fileName = ERROBJ, fileFormat = ERROBJ;
		ObjPtr stmt;

		args = ParseExpression(args, &path);
		if (path == ERROBJ)
		{
		    ScriptError("A path name is expected here");
		    return true;
		}
		SKIPBLANKS(args);
		if (*args != ',')
		{
		    ScriptError("A comma is expected here");
		    return true;
		}
		++args;

		args = ParseExpression(args, &fileName);
		if (fileName == ERROBJ)
		{
		    ScriptError("A file name is expected here");
		    return true;
		}
		SKIPBLANKS(args);
		if (*args != ',')
		{
		    ScriptError("A comma is expected here");
		    return true;
		}
		++args;

		args = ParseExpression(args, &fileFormat);
		if (fileFormat == ERROBJ)
		{
		    ScriptError("A file format is expected here");
		    return true;
		}

		stmt = NewFileRead(path, fileName, fileFormat);

		if (compiledCode)
		{
		    /*Compile*/
		    AppendStatementToCode(compiledCode, stmt);
		}
		else
		{
		    /*Immediate execution*/
		    Eval(stmt);
		}
	    }
	    break;
	default:
	    ScriptError("This keyword is not a command");
	    break;
    }
    return true;
}

#ifdef HAVE_PROTOTYPES
void BeginScript(char *name)
#else
void BeginScript(name)
char *name;
#endif
/*Begins a new script opened from file name*/
{
    /*Save curScript*/
    savedScripts[curScriptIndex++] = curScript;

    curScript = OpenTextFile(name, TF_READ + TF_CONTINUATION);
    if (curScript)
    {
	runningScript = true;
	abortScript = false;
	ParseFromFile(curScript);
	BeginCommand("script");
    }
    else
    {
	fprintf(stderr, "BeginScript: Cannot open script file %s.\n", name);
    }
}

void EndScript()
/*Ends a script*/
{
    if (curScript)
    {
	EndCommand("script");

	/*Make control panels shown*/
	ShowScriptControlPanels(true);
	CloseTextFile(curScript);
	curScript = savedScripts[--curScriptIndex];
	if (!curScript)
	{
	    runningScript = false;
	}
	MySetCursor(0);
    }
}

Bool ReadScriptLine()
/*Reads and does a script line*/
{
    if (curScript)
    {
	int k, c;

	if (!lineAccepted)
	{
	    lineAccepted = InterpretScriptLine(scriptLine);
	    return true;
	}

	if (abortScript)
	{
	    fprintf(stderr, "Aborting script.\n");
	    EndScript();
	    return false;
	}

	/*Read a script line*/
	scriptLine = GetNextLine(curScript);
	if (!scriptLine)
	{
	    EndScript();
	    return false;
	}
	else
	{
	    lineAccepted = InterpretScriptLine(scriptLine);
	    return true;
	}
    }
    return true;
}

static Bool ScriptTask()
/*Run the current script task*/
{
    switch(runningTask)
    {
	case RT_RECORDING:
	    if (framesLeft <= 0)
	    {
		runningTask = 0;
		return true;
	    }
	    PushNonVisWindows();
	    if (!SnapOneFrame())
	    {
		ScriptError("Could not snap a single frame.");
	    }
	    ++TrashDayFlag;
	    --framesLeft;
	    break;
    }
    return false;
}

#ifdef HAVE_PROTOTYPES
char *DoAssignmentCommand(char *s)
#else
char *DoAssignmentCommand(s)
char *s;
#endif
/*Does an assignment command*/
{
    char varName[256];
    ObjPtr value;
    int cmdWhere;

    Log(s);

    ++s;
    SHIFTKW(varName, s);
    SKIPBLANKS(s);

    if (*s != '=')
    {
	ScriptError("An equal sign (=) is expected here");
	return s;
    }
    ++s;

    value = ERROBJ;

    cmdWhere = FindCommandInStack(KWDEFINITION);
    if (cmdWhere >= 0)
    {
	ScriptVarPtr var;
	
	var = FindReadScriptVar(varName);
	if (var)
	{
	    ObjPtr stmt;

	    s = ParseExpression(s, &value);

	    stmt = NewSetVar(var, value);
	    AppendStatementToCode(commandStack[cmdWhere] . compiledCode, stmt);
	}
    }
    else
    {
	s = EvalExpressionArg(s, &value);
	SKIPBLANKS(s);
	if (*s)
	{
	    ScriptError("There is unexpected stuff after the expression");
	}
    
	if (value != ERROBJ)
	{
	    SetReadScriptVar(varName, value);
	}
    
	return s;
    }
}

Bool InterpretScriptLine(s)
char *s;
/*Interprets a script line in character string s.  Returns true iff the script
  line needs to be thrown away*/
{
    char curKeyword[256];
    int kwn;
    Bool retVal;
    char *t;

    ParseFromFile(curScript);

    if (runningTask != RT_NONE)
    {
	/*There's a running task*/
	return ScriptTask();
    }

    /*Look at first keyword*/
    t = s;
    SKIPBLANKS(s);
    if (s > t)
    {
	strncpy(curKeyword, s, s - t);
	Log(curKeyword);
    }
    if (*s == '#')
    {
	if (logging)
	{
	    Log(s);
	    Log("\n");
	}
	return true;
    }
    if (*s == '$')
    {
	DoAssignmentCommand(s);
	return true;
    }

    if (scriptWindow && IsValidWindow(scriptWindow))
    {
	SelWindow(scriptWindow);
    }
    else
    {
	SelWindow((WinInfoPtr) 0);
    }

    if (ObjFunctionScriptLine(s))
    {
	retVal = true;
    }
    else if (WindowFunctionScriptLine(s))
    {
	retVal = true;
    }
    else if (RepobjFunctionScriptLine(s))
    {
	retVal = true;
    }
    else if (CurrentFunctionScriptLine(s))
    {
	retVal = true;
    }
    else if (GlobalFunctionScriptLine(s))
    {
	retVal = true;
    }
    else
    {
	if (!strlen(s))
	{
	    retVal = true;
	}
	else
	{
	    SHIFTKW(curKeyword, s);
    
	    if (!strlen(curKeyword))
	    {
		ScriptError("This command is unknown.");
		retVal = true;
	    }
	    else
	    {
	    
		/*Find out what it is*/
		kwn = ParseKeyword(curKeyword, keywords, NKEYWORDS);
		if (kwn == -1)
		{
		    ScriptError("This command is unknown.");
		    retVal = true;
		}
		else if (kwn == -2)
		{
		    ScriptError("This command is ambiguous.");
		    retVal = true;
		}
		else if (kwn == KWTRY)
		{
		    InhibitCalcErrors(true);
		    retVal = InterpretScriptLine(s);
		    InhibitCalcErrors(false);
		}
		else
		{
		    retVal = DoScriptCommand(kwn, s);
		}
	    }
	}
    }
    return retVal;
}

#ifdef HAVE_PROTOTYPES
void ReadScriptFile(char *pathName, char *fileName)
#else
void ReadScriptFile(pathName, fileName)
char *pathName;
char *fileName;
#endif
/*Queues up a script file to be read*/
{
    QueueScript(pathName, fileName);
}

#ifdef HAVE_PROTOTYPES
void InitScripts(void)
#else
void InitScripts()
#endif
/*Initializes the script system*/
{
    scriptDummy = NewObject(NULLOBJ, 0L);
    SetMethod(scriptDummy, MARK, MarkScriptStuff);
    AddToReferenceList(scriptDummy);
}

#ifdef HAVE_PROTOTYPES
void KillScripts(void)
#else
void KillScripts()
#endif
/*Kills the script system*/
{
    RemoveFromReferenceList(scriptDummy);
}

