/* ScianSound.c
 * John R. Murray
 * June 1994
 */

#include "Scian.h"
#include "ScianTypes.h"
#include "ScianIDs.h"
#include "ScianGarbageMan.h"
#include "ScianErrors.h"
#include "ScianLists.h"
#include "ScianSound.h"
#include "ScianButtons.h"
#include "ScianEvents.h"
#include "ScianWindows.h"
#include "ScianTimers.h"
#include "ScianDialogs.h"

int soundWorking = 0; /* should be used primarily (exclusively?) in sound interface section */
int soundVetoed = 0; /* should be used primarily (exclusively?) in soundobj section */

int first_time = 1;

ObjPtr soundClass;		/* class of all sounds */
ObjPtr soundBellClass;		/* class of bell sounds */
ObjPtr soundFmBellClass;	/* class of FM synthesis bells */
ObjPtr soundAddBellClass;	/* class of additive synthesis bells */
ObjPtr soundWoodblockClass;	/* class of woodblock sounds */
ObjPtr soundFmToneClass;	/* class of plain FM tones */
ObjPtr soundList;
char *sampleDir = ".";

#ifdef CAVESOUND

#ifdef VROOMSOUND

/* ========== NEW sound stuff */

float fmToneActor = -1.0;

#include "vssClient.h"

void SoundOpen()
{

#ifdef WINDOWSCAVE
    if (CAVEP && CAVEWall != CAVE_FRONT_WALL)
    {
	return;
    }
#endif

    if (!FBgnSoundServer())
    {
	ReportError("SoundOpen", "connection to sound server failed");
	soundWorking = 0;
	return;
    }
    soundWorking = 1;
}

void SoundClose()
{
#ifdef WINDOWSCAVE
    if (CAVEP && CAVEWall != CAVE_FRONT_WALL)
    {
	return;
    }
#endif
    if (soundWorking && !first_time)
    {
	EndSoundServer();
    }
    return;
}

void SoundStop()
{
    soundWorking = 0;
    /* kill all on-the-fly sound objects that were created */
    fprintf(stderr, "sound server error - vetoing sound on this end\n");
}

void SoundRestart()
{
    fprintf(stderr, "SoundRestart will NOT work\n");
    /* no easy way to kill all old sound objects and force recreation */
    SoundOpen();
}

void PlaySampleSound(sname)
char *sname;
{
    float aSampleActor;

    if (first_time)
    {
	SoundOpen();
	first_time = 0;
    }

    if (!soundWorking)
    {
	return;
    }

#ifdef WINDOWSCAVE
    if (CAVEP && CAVEWall != CAVE_FRONT_WALL)
    {
	return;
    }
#endif

    if (-1 == (aSampleActor = createActor(SampleActor)))
    {
	SoundStop();
	return;
    }
    if (-1 == actorMessage(A_SetDirectory, "FS", aSampleActor, sampleDir))
    {
	SoundStop();
	return;
    }
    if (-1 == actorMessage(A_BeginNote, "FS", aSampleActor, sname))
    {
	SoundStop();
	return;
    }
}

#ifdef HAVE_PROTOTYPES
void PlayFmBellSound(float amp, float pitch, float dur)
#else
void PlayFmBellSound(amp, pitch, dur)
float amp, pitch, dur;
#endif
{
    float aFmBell;

    if (first_time)
    {
	SoundOpen();
	first_time = 0;
    }

    if (!soundWorking)
    {
	return;
    }

#ifdef WINDOWSCAVE
    if (CAVEP && CAVEWall != CAVE_FRONT_WALL)
    {
	return;
    }
#endif

    if ((aFmBell = createActor(FmBellActor)) == -1)
    {
	ReportError("PlayFmBellSound", "error creating FmBellActor");
	SoundStop();
	return;
    }
    if (-1 == actorMessage(A_Hit, "FFFF", aFmBell, amp, pitch, dur))
    {
	SoundStop();
	return;
    }
}

#ifdef HAVE_PROTOTYPES
void PlayAddBellSound(float amp, float pitch, float dur)
#else
void PlayAddBellSound(amp, pitch, dur)
float amp, pitch, dur;
#endif
{
    float aAddBell;
    if (first_time)
    {
	SoundOpen();
	first_time = 0;
    }

    if (!soundWorking)
    {
	return;
    }

#ifdef WINDOWSCAVE
    if (CAVEP && CAVEWall != CAVE_FRONT_WALL)
    {
	return;
    }
#endif

    if ((aAddBell = createActor(AddBellActor)) == -1)
    {
	ReportError("PlayAddBellSound", "error creating AddBellActor");
	SoundStop();
	return;
    }
    if (-1 == actorMessage(A_Hit, "FFFF", aAddBell, amp, pitch, dur))
    {
	SoundStop();
	return;
    }
}

#ifdef HAVE_PROTOTYPES
void PlayWoodblockSound(float amp, float pitch, float dur)
#else
void PlayWoodblockSound(amp, pitch, dur)
float amp, pitch, dur;
#endif
{
    float aFmWoodblock;
    if (first_time)
    {
	SoundOpen();
	first_time = 0;
    }

    if (!soundWorking)
    {
	return;
    }

#ifdef WINDOWSCAVE
    if (CAVEP && CAVEWall != CAVE_FRONT_WALL)
    {
	return;
    }
#endif

    if ((aFmWoodblock = createActor(FmWoodblockActor)) == -1)
    {
	ReportError("PlayWoodblockSound", "error creating FmWoodblockActor");
	SoundStop();
	return;
    }
    if (-1 == actorMessage(A_SetCenterFreq, "FF", aFmWoodblock, pitch))
    {
	SoundStop();
	return;
    }
    if (-1 == actorMessage(A_Hit, "FFFF", aFmWoodblock, amp, 0, dur))
    {
	SoundStop();
	return;
    }
/*    if (-1 == actorMessage(A_Hit, "F", aFmWoodblock)) hak hak hak; */
}

float PlayFmToneSound(handle, amp, pitch, dur, modIndex, CMratio)
float *handle, amp, pitch, dur, modIndex, CMratio;
{
float noteHandle;

    if (first_time)
    {
	SoundOpen();
	first_time = 0;
    }

    if (!soundWorking)
    {
	return;
    }

#ifdef WINDOWSCAVE
    if (CAVEP && CAVEWall != CAVE_FRONT_WALL)
    {
	return;
    }
#endif

    if (*handle == -1.0)
    {
	if ((*handle = createActor(FmActor)) == -1)
	{
	    ReportError("PlayFmToneSound", "error creating FmActor");
	    SoundStop();
	    return;
	}
	{
#define ARRLEN 7
	    float envArray[ARRLEN];

	    envArray[0] = 0.0;
	    envArray[1] = 0.01 * dur;
	    envArray[2] = 1.0;
	    envArray[3] = 0.75 * dur;
	    envArray[4] = 1.0;
	    envArray[5] = 0.24 * dur;
	    envArray[6] = 0.0;

	    if (-1 == actorMessage(A_SetDefaultEnv, "FIA", *handle, ARRLEN, envArray))
	    {
		SoundStop();
		return;
	    }
	}

#ifdef DEBUGSOUND
fprintf(stderr, "created new FM actor\n");
#endif

    }
#ifdef DEBUGSOUND
else fprintf(stderr, "actor already existed? handle = %f\n", *handle);
fprintf(stderr, "sending BeginNote, amp = %f, pitch = %f, modIndex = %f, CMratio = %f\n",
		amp, pitch, modIndex, CMratio);
#endif
    noteHandle = actorMessage(A_BeginNote, "FFFFF", *handle, amp, pitch, modIndex, CMratio);
    if (noteHandle == -1)
    {
	ReportError("PlayFmToneSound", "error in note handle return value");
	SoundStop();
    }
    return noteHandle;
}

EndNote(actorHandle, noteHandle)
float actorHandle, noteHandle;
{
#ifdef DEBUGSOUND
fprintf(stderr, "sending EndNote message, actor = %f, note = %f\n", actorHandle, noteHandle);
#endif
    if (-1 == actorMessage(A_EndNote, "FF", actorHandle, noteHandle))
    {
	SoundStop();
	return;
    }
}

#else

/* ========== OLD sound stuff */

/* include file for CAVE sound */
#include "client.h"


INSTR *instPtr = 0;
float noteHandle = hNil;

/* NOTE should be split into ScianObjSound and ScianSound? */

void SoundOpen()
{
    instPtr = Alloc(sizeof(INSTR));
    if (!FReadFile("samp.in", instPtr))
    {
	soundWorking = 0;
fprintf(stderr, "couldn't read instrument file samp.in\n");
	return;
    }

    if (!FBgnSoundServer())
    {
	soundWorking = 0;
fprintf(stderr, "connection to sound server failed\n");
	return;
    }
}

void SoundClose()
{
    EndSoundServer();
    return;
}

int bad_inst = 0;

void PlaySampleSound(sName)
char *sName;
{
    mm mmT;
    if (first_time)
    {
	SoundOpen();
	first_time = 0;
    }

    if (!instPtr && !bad_inst)
    {
fprintf(stderr, "no instrument pointer!\n");
	bad_inst = 1;
	return;
    }

    /* NOTE: may have to keep track of note handles */
    noteHandle = HNoteFromPinstr(instPtr);
fprintf(stderr, "noteHandle = %f\n", noteHandle);


    sprintf(mmT.name, "_%s", szFileSamp);
    mmT.param[0] = noteHandle;
    strcpy((char*)(mmT.param+1), sName);
    Msgsend(NULL, &mmT, strlen(sName)/sizeof(float) + 5);
	/* + 5 is random */

    UnmuteNote(noteHandle);
}


#endif

/* ===== ABOVE - sound interface functions. BELOW - Sound Obj functions ===== */

/* returns a new sound object */
ObjPtr NewSampleSoundType(soundName)
char *soundName;
{
    ObjPtr retVal;

#ifdef DEBUGSOUND
fprintf (stderr, "soundClass = %lx\n", soundClass);
#endif
    retVal = NewObject(soundClass, 0);
    SetVar(retVal, FILENAME, NewString(soundName));
    if (!soundList)
    {
#ifdef DEBUGSOUND
fprintf(stderr, "remade soundlist!\n");
#endif
	soundList = NewList();
	AddToReferenceList(soundList);
    }
    PrefixList (soundList, retVal);
    return retVal;
}

ObjPtr EndObjNote(note)
ObjPtr note;
{
    float actorHandle, noteHandle;
  
    noteHandle = GetReal(GetRealVar("EndObjNote", note, SOUNDNOTEHANDLE));
    actorHandle = GetReal(GetRealVar("EndObjectNote", note, SOUNDACTORHANDLE));
    /* EndNote(actorHandle, noteHandle); */
    SetVar(note, SOUNDON, ObjFalse);
    return ObjTrue;
}

ObjPtr NewNote(sound, noteHandle)
ObjPtr sound;
float noteHandle;
{
    ObjPtr retVal;

    retVal = NewObject(sound, 0);
    SetVar(retVal, SOUNDNOTEHANDLE, NewReal(noteHandle));
    SetMethod(retVal, SOUNDNOTEOFF, EndObjNote);
    return retVal;
}

ObjPtr NewSound(soundName)
char *soundName;
{
    return NewSampleSound(soundName);
}

ObjPtr NewSampleSound(soundName)
char *soundName;
{
    ObjPtr retVal, soundType = NULLOBJ;
    ThingListPtr runner;

#ifdef TESTSOUND
return NewFmToneSound();
#endif

    /* search for existing sound type with name soundName */
    if (!soundList)
    {
#ifdef DEBUGSOUND
fprintf(stderr, "remade soundlist!\n");
#endif
	soundList = NewList();
	AddToReferenceList(soundList);
    }

    runner = LISTOF(soundList);
    while(runner)
    {
	char *foo;

	foo = GetString(GetStringVar("NewSound", runner->thing, FILENAME));
	if (foo)
	/* if (!strcmp(soundName, GetString(GetStringVar("NewSound", runner->thing, FILENAME)))) */
	{
	if (!strcmp(soundName, foo))
	{
	    /* found it, use it as class */
	    soundType = runner -> thing;
	    break;
	}
	}
	else
	{
	    char str[100];
	    sprintf(str, "bad object? runner = %x, runner -> thing = %x", runner, runner -> thing);
	    ReportError ("NewFmToneSound", str);
	}
	runner = runner -> next;
    }

    if (!soundType)
    {
	soundType = NewSampleSoundType(soundName);
#ifdef DEBUGSOUND
fprintf(stderr, "had to recreate sound type %s\n", soundName);
#endif
    }
    retVal = NewObject(soundType, 0);
    SetVar(retVal, SOUNDON, ObjFalse);
    return retVal;
}

ObjPtr NewFmToneSound()
{
    ObjPtr retVal;
    float pitch;

#ifdef DEBUGSOUND
fprintf(stderr, "creating FM tone sound\n");
#endif

    retVal = NewObject(soundFmToneClass, 0);
    SetVar(retVal, SOUNDON, ObjFalse);


    return retVal;
}

ObjPtr NewWoodblockSound()
{
    ObjPtr retVal;

fprintf(stderr, "don't use Woodblock, Bwana! (it's off-key)\n");

    retVal = NewObject(soundWoodblockClass, 0);
    SetVar(retVal, SOUNDON, ObjFalse);
    return retVal;
}

ObjPtr NewFmBellSound()
{
    ObjPtr retVal;

    retVal = NewObject(soundFmBellClass, 0);
    SetVar(retVal, SOUNDON, ObjFalse);
    return retVal;
}

ObjPtr NewAddBellSound()
{
    ObjPtr retVal;

    retVal = NewObject(soundAddBellClass, 0);
    SetVar(retVal, SOUNDON, ObjFalse);
    return retVal;
}

ObjPtr NewBellSound()
{
    return NewAddBellSound();
}

ObjPtr PlayObjSampleSound(sound)
ObjPtr sound;
{
    char *name;

#ifdef DEBUGSOUND
fprintf(stderr, "got to play method. soundVetoed = %d\n", soundVetoed);
#endif
    if (soundVetoed)
    {
	return ObjFalse;
    }
    name = GetString(GetStringVar ("PlayObjSampleSound", sound, FILENAME));
    SetVar(sound, SOUNDON, ObjTrue);
    PlaySampleSound(name);
    return ObjTrue;
}

ObjPtr PlayObjFmBellSound(sound)
ObjPtr sound;
{
    char *name;
    ObjPtr ampObj, pitchObj, durObj;
    float amp, pitch, dur;

#ifdef DEBUGSOUND
fprintf(stderr, "got to Bell play method. soundVetoed = %d\n", soundVetoed);
#endif
    if (soundVetoed)
    {
	return ObjFalse;
    }

    amp = GetReal(GetRealVar("PlayObjFmBellSound", sound, SOUNDAMPLITUDE));
    pitch = GetReal(GetRealVar("PlayObjFmBellSound", sound, SOUNDPITCH));
    dur = GetReal(GetRealVar("PlayObjFmBellSound", sound, SOUNDDURATION));

    SetVar(sound, SOUNDON, ObjTrue);

    /* reality check */
    if (amp > 1 || amp < 0 || dur > 1000 || dur < 0)
    {
	ReportError("PlayObjFmBellSound", "bad FmBell parameters!");
	return ObjFalse;
    }
    PlayFmBellSound(amp, pitch, dur);
    return ObjTrue;
}

ObjPtr PlayObjAddBellSound(sound)
ObjPtr sound;
{
    char *name;
    ObjPtr ampObj, pitchObj, durObj;
    float amp, pitch, dur;

#ifdef DEBUGSOUND
fprintf(stderr, "got to AddBell play method. soundVetoed = %d\n", soundVetoed);
#endif
    if (soundVetoed)
    {
	return ObjFalse;
    }

/* HAK here is where I need to get/set parameters */
    amp = GetReal(GetRealVar("PlayObjAddBellSound", sound, SOUNDAMPLITUDE));
    pitch = GetReal(GetRealVar("PlayObjAddBellSound", sound, SOUNDPITCH));
    dur = GetReal(GetRealVar("PlayObjAddBellSound", sound, SOUNDDURATION));

    SetVar(sound, SOUNDON, ObjTrue);

    /* reality check */
    if (amp > 1 || amp < 0 || dur > 1000 || dur < 0)
    {
	ReportError("PlayObjAddBellSound", "bad AddBell parameters!\n");
	return ObjFalse;
    }
    PlayAddBellSound(amp, pitch, dur);
    return ObjTrue;
}

ObjPtr PlayObjWoodblockSound(sound)
ObjPtr sound;
{
    char *name;
    ObjPtr ampObj, pitchObj, durObj;
    float amp, pitch, dur;

#ifdef DEBUGSOUND
fprintf(stderr, "got to Woodblock play method. soundVetoed = %d\n", soundVetoed);
#endif
    if (soundVetoed)
    {
	return ObjFalse;
    }

/* HAK here is where I need to get/set parameters */
    amp = GetReal(GetRealVar("PlayObjWoodblockSound", sound, SOUNDAMPLITUDE));
    pitch = GetReal(GetRealVar("PlayObjWoodblockSound", sound, SOUNDPITCH));
    dur = GetReal(GetRealVar("PlayObjWoodblockSound", sound, SOUNDDURATION));

    SetVar(sound, SOUNDON, ObjTrue);

    /* reality check */
    if (amp > 1 || amp < 0 || dur > 1000 || dur < 0)
    {
	ReportError("PlayObjWoodblockSound", "bad Woodblock parameters!");
	return ObjFalse;
    }

#ifdef DEBUGSOUND
fprintf(stderr, "amp = %f, pitch = %f, dur = %f\n", amp, pitch, dur);
#endif
    PlayWoodblockSound(amp, pitch, dur);
    return ObjTrue;
}

ObjPtr PlayObjFmToneSound(sound)
ObjPtr sound;
{
    ObjPtr noteObj;
    float amp, pitch, dur, modIndex, cmRatio, handle, oldhandle, noteHandle;
#ifdef DEBUGSOUND
fprintf(stderr, "got to FmTone play method. soundVetoed = %d\n", soundVetoed);
#endif
    if (soundVetoed)
    {
	return ObjFalse;
    }

    amp = GetReal(GetRealVar("PlayObjFmToneSound", sound, SOUNDAMPLITUDE));
    pitch = GetReal(GetRealVar("PlayObjFmToneSound", sound, SOUNDPITCH));
    dur = GetReal(GetRealVar("PlayObjFmToneSound", sound, SOUNDDURATION));
    modIndex = GetReal(GetRealVar("PlayObjFmToneSound", sound, SOUNDFMMODULATION));
    cmRatio = GetReal(GetRealVar("PlayObjFmToneSound", sound, SOUNDFMCMRATIO));
    handle = GetReal(GetRealVar("PlayObjFmToneSound", sound, SOUNDACTORHANDLE));
    oldhandle = handle;

    SetVar(sound, SOUNDON, ObjTrue);

#if TESTSOUND
#if 1
    pitch = pitch + (random() % 16) + (random() % 16) - 15;
#endif
#endif

    noteHandle = PlayFmToneSound(&handle, amp, pitch, dur, modIndex, cmRatio);
    if (oldhandle != handle)
    {
	SetVar (sound, SOUNDACTORHANDLE, NewReal(handle));
    }
    noteObj = NewNote(sound, noteHandle);
#ifdef DEBUGSOUND
fprintf(stderr, "setting WakeMe, noteObj = %x, dur = %f\n", noteObj, dur);
#endif
    WakeMe(noteObj, SOUNDNOTEOFF, Clock() + dur);
}


/* endif CAVESOUND from way up there */
#endif


/*should go into ScianMethods? */

ObjPtr Play(sound)
ObjPtr sound;
{
#ifdef CAVESOUND
    FuncTyp playSound;

    playSound = GetMethod(sound, PLAYSOUND);
    if (playSound)
    {
	(*playSound) (sound);
    }
#ifdef DEBUGSOUND
else fprintf(stderr, "no method!\n");
#endif

    return ObjTrue;

#endif
}

/* PlayPitch plays a sound object at the given pitch. Note
 * side effect of resetting SOUNDPITCH var..
 */
ObjPtr PlayPitch(sound, pitch)
ObjPtr sound;
real pitch;
{
#ifdef CAVESOUND
    SetVar (sound, SOUNDPITCH, NewReal(pitch));
    return Play(sound);
#endif
}

ObjPtr TestCleaner(obj)
ObjPtr obj;
{
#ifdef DEBUGSOUND
    fprintf(stderr, "Cleanup!\n");
#endif
    return ObjTrue;
}

void InitSound()
{
#ifdef CAVESOUND
    ObjPtr hak;

    if (getenv("SCIAN_VETO_SOUND"))
    {
	if (showConfig)
	{
	    printf("Sound vetoed.\n");
	}
	soundVetoed = 1;
    }
#ifdef DEBUGSOUND
fprintf(stderr, "initializing sound\n");
#endif
    soundClass = NewObject(0,0);
    SetMethod (soundClass, PLAYSOUND, PlayObjSampleSound);
    
    SetVar(soundClass, SOUNDPITCH, NewReal(SNDPITCH));
    SetVar(soundClass, SOUNDAMPLITUDE, NewReal(SNDAMPLITUDE));
    SetVar(soundClass, SOUNDDURATION, NewReal(SNDDURATION));
    AddToReferenceList(soundClass);

    soundBellClass = NewObject(soundClass, 0);
    AddToReferenceList(soundBellClass);
    SetVar(soundBellClass, SOUNDDURATION, NewReal(SNDBELLDURATION));

    soundFmBellClass = NewObject(soundBellClass, 0);
    SetMethod (soundFmBellClass, PLAYSOUND, PlayObjFmBellSound);
    AddToReferenceList(soundFmBellClass);

    soundAddBellClass = NewObject(soundBellClass, 0);
    SetMethod (soundAddBellClass, PLAYSOUND, PlayObjAddBellSound);
    AddToReferenceList(soundAddBellClass);

    soundWoodblockClass = NewObject(soundClass, 0);
    SetMethod (soundWoodblockClass, PLAYSOUND, PlayObjWoodblockSound);
    SetVar(soundWoodblockClass, SOUNDAMPLITUDE, NewReal(SNDAMPLITUDE));
    SetVar(soundWoodblockClass, SOUNDPITCH, NewReal(500));
    SetVar(soundWoodblockClass, SOUNDDURATION, NewReal(5));
    AddToReferenceList(soundWoodblockClass);

    soundFmToneClass = NewObject(soundClass, 0);
    SetMethod (soundFmToneClass, PLAYSOUND, PlayObjFmToneSound);
    SetVar (soundFmToneClass, SOUNDAMPLITUDE, NewReal(0.1));
    SetVar (soundFmToneClass, SOUNDFMMODULATION, NewReal(0.75));
/* NewReal(SNDFMMODULATION)); */
    SetVar (soundFmToneClass, SOUNDFMCMRATIO, NewReal(SNDFMCMRATIO));
    SetVar (soundFmToneClass, SOUNDACTORHANDLE, NewReal(-1.0));
    SetVar (soundFmToneClass, REPRIEVE, ObjTrue);
    SetMethod(soundFmToneClass, CLEANUP, TestCleaner);
    AddToReferenceList(soundFmToneClass);

    soundList = NewList();
    AddToReferenceList(soundList);

    if (!(sampleDir = getenv("SCIAN_SOUND_SAMPLE_DIR")))
    {
#ifdef NCSA
	sampleDir = "/usr/local/CAVE/sound/SAMPLES";
#else
	sampleDir = ".";
#endif
    }
#ifdef DEBUGSOUND
fprintf(stderr, "sampleDir set to %s.\n", sampleDir);
#endif

#endif

}

void KillSound()
{
#ifdef CAVESOUND
    if (soundClass)
    {
	RemoveFromReferenceList(soundClass);
    }
    SoundClose();
#endif
}

