#include "Scian.h"
#include "ScianTypes.h"
#include "ScianIDs.h"
#include "ScianGarbageMan.h"
#include "ScianErrors.h"
#include "ScianLists.h"
#include "ScianSound.h"

#ifdef CAVESOUND

/* include file for CAVE sound */
#include "client.h"


int first_time = 1;
INSTR *instPtr = 0;
float noteHandle = hNil;
ObjPtr soundClass;
ObjPtr soundList;

/* NOTE should be split into ScianObjSound and ScianSound? */

void SoundOpen()
{
    instPtr = Alloc(sizeof(INSTR));
    if (!FReadFile("samp.in", instPtr))
    {
fprintf(stderr, "couldn't read instrument file samp.in\n");
	return;
    }

    if (!FBgnSoundServer())
    {
fprintf(stderr, "connection to sound server failed\n");
	return;
    }
}

void SoundClose()
{
    EndSoundServer();
    return;
}

int bad_inst = 0;

void PlaySound(sName)
char *sName;
{
    mm mmT;
    if (first_time)
    {
	SoundOpen();
	first_time = 0;
    }

    if (!instPtr && !bad_inst)
    {
fprintf(stderr, "no instrument pointer!\n");
	bad_inst = 1;
	return;
    }

    /* NOTE: may have to keep track of note handles */
    noteHandle = HNoteFromPinstr(instPtr);
fprintf(stderr, "noteHandle = %f\n", noteHandle);


    sprintf(mmT.name, "_%s", szFileSamp);
    mmT.param[0] = noteHandle;
    strcpy((char*)(mmT.param+1), sName);
    Msgsend(NULL, &mmT, strlen(sName)/sizeof(float) + 5);
	/* + 5 is random */

    UnmuteNote(noteHandle);
}

/* ===== ABOVE - sound interface functions. BELOW - Sound Obj functions ===== */

/* returns a new sound object */
ObjPtr NewSoundType(soundName)
char *soundName;
{
    ObjPtr retVal;

fprintf (stderr, "soundClass = %lx\n", soundClass);

    retVal = NewObject(soundClass, 0);
    SetVar(retVal, FILENAME, NewString(soundName));
    if (!soundList)
    {
	soundList = NewList();
	AddToReferenceList(soundList);
    }
    PrefixList (soundList, retVal);
    return retVal;
}

ObjPtr NewSound(soundName)
char *soundName;
{
    ObjPtr retVal, soundType = NULLOBJ;
    ThingListPtr runner;

    /* search for existing sound type with name soundName */
    if (!soundList)
    {
	soundList = NewList();
	AddToReferenceList(soundList);
    }

    runner = LISTOF(soundList);
    while(runner)
    {
	if (!strcmp(soundName, GetString(GetStringVar("NewSound", runner->thing, FILENAME))))
	{
	    /* found it, use it as class */
	    soundType = runner -> thing;
	    break;
	}
	runner = runner -> next;
    }

    if (!soundType)
    {
	soundType = NewSoundType(soundName);
fprintf(stderr, "had to recreate sound type %s\n", soundName);
    }
    retVal = NewObject(soundType, 0);
    SetVar(retVal, SOUNDON, ObjFalse);
    return retVal;
}

ObjPtr PlayObjSound(sound)
ObjPtr sound;
{
    char *name;

    name = GetString(GetStringVar ("PlayObjSound", sound, FILENAME));
    SetVar(sound, SOUNDON, ObjTrue);
    PlaySound(name);
}

#endif

void InitSound()
{
#ifdef CAVESOUND
    ObjPtr hak;
fprintf(stderr, "initializing sound\n");
    soundClass = NewObject(0,0);
    SetMethod (soundClass, PLAYSOUND, PlayObjSound);
    AddToReferenceList(soundClass);

    soundList = NewList();
    AddToReferenceList(soundList);

/*
    hak = NewSound("bang.aiff");
    PlayObjSound(hak);
    hak = NewSound("hit.aiff");
    PlayObjSound(hak);
    hak = NewSound("trumpet.aiff");
    PlayObjSound(hak);
*/
#endif

}

void KillSound()
{
#ifdef CAVESOUND
    RemoveFromReferenceList(soundClass);
    if (!first_time)
    {
	SoundClose();
    }
#endif
}

#ifdef CAVESOUND
/*should go into ScianMethods? */

ObjPtr Play(sound)
ObjPtr sound;
{
    FuncTyp playSound;

fprintf(stderr,"and here..\n");
    playSound = GetMethod(sound, PLAYSOUND);
    if (playSound)
    {
	(*playSound) (sound);
    }
else fprintf (stderr,"no method!\n");
    return ObjTrue;
}


#endif

