/* ScianVisStreams.c */
/* Tzong-Yow Hwu */
/* 6/16/1994 */
/* 
1. change the control panels for streamlines,
   get rid of mark buckets and replace
   steplengthfactor with steplength (default to 0.05 of maxBound)
   sampleparameter with steps (default to 200)
now working on
2. add picking methods for emitters class
*/

#include "Scian.h"
#include "ScianTypes.h"
#include "ScianArrays.h"
#include "ScianWindows.h"
#include "ScianTextBoxes.h"
#include "ScianButtons.h"
#include "ScianTitleBoxes.h"
#include "ScianObjWindows.h"
#include "ScianIcons.h"
#include "ScianColors.h"
#include "ScianControls.h"
#include "ScianLists.h"
#include "ScianSliders.h"
#include "ScianIDs.h"
#include "ScianDatasets.h"
#include "ScianPictures.h"
#include "ScianDialogs.h"
#include "ScianErrors.h"
#include "ScianEvents.h"
#include "ScianComplexControls.h"
#include "ScianMethods.h"
#include "ScianStyle.h"
#include "ScianVisObjects.h"
#include "ScianVisStreams.h"
#include "ScianDraw.h"
#include "ScianTemplates.h"
#include "ScianTemplateHelper.h"
#include "ScianSymbols.h"
#include "ScianSnap.h"
#include "ScianFixedClasses.h"

#define HWUTEST

#ifdef HWUTEST
#define IDSTART                    727 
#define EMITTERX                   IDSTART + 1 
#define EMITTERY                   IDSTART + 2
#define EMITTERZ                   IDSTART + 3 
#endif

/* should be Scian.h, maybe */
#include <float.h>

/* 
for SIGGRAPH purpose 
#define SIGGRAPH
*/

/* emitters controlling way */
#define USEINTERACTIVE 0
#define USEDATASET  1
#define USEBUCKET   2

#ifdef SIGGRAPH

#define DEFAULTSAMPLEPARAMETER 2.0 
#define MINSAMPLEPARAMETER 0
#define MAXSAMPLEPARAMETER plusInf 

#define DEFAULTSTEPLENGTHFACTOR    0.005     /* c: ds = c * max(bounds) */
#define MINSTEPLENGTHFACTOR 0 
#define MAXSTEPLENGTHFACTOR plusInf 

#define DEFAULTMINSTREAMPOINTS 20
#define MINMINSTREAMPOINTS  1 
#define MAXMINSTREAMPOINTS plusInf 

#else

#define DEFAULTSAMPLEPARAMETER 3.0  /* used to multiply 1/c to get total number of sample points */
#define MINSAMPLEPARAMETER 0
#define MAXSAMPLEPARAMETER plusInf 

#define DEFAULTSTEPLENGTHFACTOR    0.1     /* c: ds = c * max(bounds) */
#define MINSTEPLENGTHFACTOR 0 
#define MAXSTEPLENGTHFACTOR plusInf 

#define DEFAULTMINSTREAMPOINTS 2
#define MINMINSTREAMPOINTS  1 
#define MAXMINSTREAMPOINTS plusInf 

#endif

/* Integration Methods */
#define EULER           0     /* first order */
#define RUNGEKUTTA       1    /* second order */

#define STR_MAXCYLSIDES 512
#define STR_I_THRESHOLD 0.8

/* width of flow style control box */ 
#define PICSTICKBOXWID 185 

/* chord extension factor */
extern double cefs_2[];

#ifdef HAVE_PROTOTYPES
static int rkuttaCalcTrace(real *trace, int maxNTracePoints, int formField, int dataField, int nDataComponents, int nSpatComponents, real start[], real bounds[], real ds);
static int eulerCalcTrace(real *trace, int maxNTracePoints, int formField, int dataField, int nDataComponents, int nSpatComponents, real start[], real bounds[], real ds);
#else
static int rkuttaCalcTrace();
static int eulerCalcTrace();
#endif

ObjPtr visLineEmitters;                 /* Class for line emitters */
ObjPtr visHoseEmitters;                 /* Class for hose emitters */
ObjPtr visStreamLines;      		/* Class for streamlines objects*/
ObjPtr visStreamHoses;      		/* Class for streamhoses objects*/

/******************************************************************************************/
/******************************** stream line object **************************************/
/******************************************************************************************/
/* 
 Both visStreamLines and visStreamHoses use this as ADDCONTROL method.
 Adds controls to a stream object: (streamline or streamhose)
   1. Dataset icon corral,
   2. Sample parameter, (to determine maximum number of points) 
   3. Step length factor,  (c: dS = c * MAX(bounds)),
   4. Mark bucket: a check box for marking unused bucket passed through by stream lines,
   5. Integration methods,
   6. Min points: Only streams consists of at least this many tracing points are visualized.
*/ 
static ObjPtr AddStreamsControls(streamVisObj, panelContents)
ObjPtr streamVisObj, panelContents;
{
  ObjPtr var, titleBox, checkBox, button, radio;
  ObjPtr corral, textBox, icon, defaultIcon, name;
  ObjPtr streamData, mainDataset;
  int left, top, bottom, right, mid;
  Bool markBucket;

  /* get the MAINDATASET object var */
  streamData = GetObjectVar("AddStreamsControls", streamVisObj, MAINDATASET);
  if (!streamData) return ObjFalse;
  while (NULLOBJ != (mainDataset = GetVar(streamData, MAINDATASET)))
    {
      streamData = mainDataset;
    }

  /* Set up the visualization icon corral at the top left of the panel */
  left = MAJORBORDER;
  right = left + ONECORRALWIDTH;
  top = CWINHEIGHT - MAJORBORDER;
  bottom = top - ONECORRALHEIGHT;
  corral = NewIconCorral(NULLOBJ, left, right, bottom, top, 0); 
  SetVar(corral, SINGLECORRAL, ObjTrue);
  SetVar(corral, TOPDOWN, ObjTrue);
  SetVar(corral, NAME, NewString("Stream Field"));
  SetVar(corral, HELPSTRING, NewString("This corral shows the dataset that is \
being used to make the streams visualization object."));  
  PrefixList(panelContents, corral);
  SetVar(corral, PARENT, panelContents);
  SetVar(corral, REPOBJ, streamVisObj);
  SetMethod(corral, DROPINCONTENTS, DropInMainDatasetCorral);

  /* Create a text box for the corral */
  top = bottom - TEXTBOXSEP;
  bottom = top - TEXTBOXHEIGHT;
  textBox = NewTextBox(left, right,bottom, top, 0, "Stream Field Text", "Stream Field");
  PrefixList(panelContents, textBox);
  SetVar(textBox, PARENT, panelContents);
  SetTextAlign(textBox, CENTERALIGN);

  /* Put a dataset representation icon into the corral */
  MakeVar(streamData, NAME);
  name = GetVar(streamData, NAME);
  MakeVar(streamData, DEFAULTICON);
  defaultIcon = GetVar(streamData, DEFAULTICON);
  if (defaultIcon)
    {
      icon = NewObject(defaultIcon, 0);
      SetVar(icon, NAME, name);
    }
  else
    {
      icon = NewIcon(0, 0, ICONQUESTION, GetString(name));
    }
  SetVar(icon, ICONLOC, NULLOBJ);
  SetVar(icon, REPOBJ, streamData);
  DropIconInCorral(corral, icon);

  /* add a check box control for marking buckets passed through by particle traces */ 
  top = CWINHEIGHT - MAJORBORDER;
  bottom = top - CHECKBOXHEIGHT;
  left = right + 6 * MINORBORDER;
  right = left + PICSTICKBOXWID + 30;
  if (!GetVar(streamVisObj, MARKBUCKET))
    {
      SetVar(streamVisObj, MARKBUCKET, NewInt(1));
    }
  markBucket = GetPredicate(streamVisObj, MARKBUCKET);
  checkBox = NewCheckBox(left, right, bottom, top, "Mark Bucket", markBucket);
  PrefixList(panelContents, checkBox);
  SetVar(checkBox, HELPSTRING, 
	NewString("This check box affects the selection of emitters in generating \
	streams.  The emitters are traversed sequentially to initiate the \
	integration of each stream.  If this box is checked then \
	those of the remaining emitters passed by generated streams are marked \
	so that they are not used to initiate streams.")); 
  SetVar(checkBox, PARENT, panelContents);
  AssocDirectControlWithVar(checkBox, streamVisObj, MARKBUCKET);

  /* Set up the integration method control box */
  top = bottom - 2 * MINORBORDER; 
  bottom = top - (TITLEBOXTOP + 2 * MINORBORDER + 1 * CHECKBOXSPACING + 2 * CHECKBOXHEIGHT);

  /* title box for integration method */
  titleBox = NewTitleBox(left, right, bottom, top,  "Integration Method");
  PrefixList(panelContents, titleBox);
  SetVar(titleBox, PARENT, panelContents);

  /* add radio */
  radio = NewRadioButtonGroup("Integration Method Group");
  PrefixList(panelContents, radio);
  SetVar(radio, PARENT, panelContents);
  SetVar(radio, HELPSTRING, 
	NewString("These radio buttons control the integration method used \
for stream tracing.")); 

  /* Make the buttons */
  top -= TITLEBOXTOP + MINORBORDER;
  bottom = top - CHECKBOXHEIGHT;
  button = NewRadioButton(left + MINORBORDER, (left + right) / 2,
			  bottom, top, "Euler");

  AddRadioButton(radio, button);
  top -= CHECKBOXHEIGHT + CHECKBOXSPACING;
  bottom = top - CHECKBOXHEIGHT;
  button = NewRadioButton(left + MINORBORDER, (left + right) / 2,
			  bottom, top, "Second RungeKutta");
  AddRadioButton(radio, button);

  bottom -= MINORBORDER;

  var = GetIntVar("AddStreamsControls", streamVisObj, INTEGRATIONMETHOD);
  if (!var)
    {
      SetVar(streamVisObj, INTEGRATIONMETHOD, NewInt(EULER));
    }
  AssocDirectControlWithVar(radio, streamVisObj, INTEGRATIONMETHOD);

  /* create a real text box for sample number parameter control */
  top = bottom - 4 * MINORBORDER;
  bottom = top - EDITBOXHEIGHT;
  mid = bottom + EDITBOXHEIGHT / 2 ;
  left = MAJORBORDER + 3 * MINORBORDER;
  right = left + ACNAMEWIDTH;
  textBox = NewTextBox(left, right, 
		       mid - TEXTBOXHEIGHT / 2 + EDITBOXDOWN, 
		       mid + TEXTBOXHEIGHT / 2 + EDITBOXDOWN, 
		       0, "Sample Parameter Text", "Sample Parameter:");
  PrefixList(panelContents, textBox);
  SetVar(textBox, PARENT, panelContents);
  var = GetRealVar("AddStreamsControls", streamVisObj, SAMPLEPARAMETER);
  if (!var)
    {
      SetVar(streamVisObj, SAMPLEPARAMETER, NewReal(DEFAULTSAMPLEPARAMETER));
    }

  textBox = NewTextBox(right + MINORBORDER,
			 right + MINORBORDER + ACTEXTWIDTH,
			 mid - EDITBOXHEIGHT / 2, mid + EDITBOXHEIGHT / 2,
			 EDITABLE + WITH_PIT + ONE_LINE, "Sample Parameter", "");
  SetVar(textBox, PARENT, panelContents);
  PrefixList(panelContents, textBox);
  SetTextAlign(textBox, RIGHTALIGN);
  SetVar(textBox, HELPSTRING, 
     NewString("This text box controls the parameter used to calculate the \
maximum number of points to be used for tracing stream.  The number obtained \
by multiplying this parameter with the reciprocal of STEPLENGTHFACTOR will be the \
maximum number of sample points the integration trace will traverse. \
A trace stops when the trace goes out of grid domain, \
when the trace stagnates (the magnitude of the velocity is 0), \
or when this maxinum number of sample points is reached."));

  AssocTextRealControlWithVar(textBox, streamVisObj, SAMPLEPARAMETER, MINSAMPLEPARAMETER, MAXSAMPLEPARAMETER, TR_NE_TOP | TR_NE_BOTTOM);

  /* create a real text control for step length factor */
  bottom -= EDITBOXHEIGHT + MINORBORDER;
  mid = bottom + EDITBOXHEIGHT / 2;
  textBox = NewTextBox(left, right, 
		       mid - TEXTBOXHEIGHT / 2 + EDITBOXDOWN, 
		       mid + TEXTBOXHEIGHT / 2 + EDITBOXDOWN, 
		       0, "Step Length Factor Text", "Step Length Factor:");
  PrefixList(panelContents, textBox);
  SetVar(textBox, PARENT, panelContents);

  var = GetRealVar("AddStreamsControls", streamVisObj, STEPLENGTHFACTOR);
  if (!var)
    {
      SetVar(streamVisObj, STEPLENGTHFACTOR, NewReal(DEFAULTSTEPLENGTHFACTOR));
    }

  textBox = NewTextBox(right + MINORBORDER,
			 right + MINORBORDER + ACTEXTWIDTH,
			 mid - EDITBOXHEIGHT / 2, mid + EDITBOXHEIGHT / 2,
			 EDITABLE + WITH_PIT + ONE_LINE, "Step Length Factor", "");
  SetVar(textBox, PARENT, panelContents);
  PrefixList(panelContents, textBox);
  SetTextAlign(textBox, RIGHTALIGN);
  SetVar(textBox, HELPSTRING, NewString("This text box controls the factor used to \
calculate the step length of a particle trace.  The value obtained by multiplying the \
factor with the maximum of the spatial bounds will be used as the first estimate of each \
tracing step.  The length is then used to calculate the estimate of \
the advanced time for each step of tracing integration."));
  AssocTextRealControlWithVar(textBox, streamVisObj, STEPLENGTHFACTOR, MINSTEPLENGTHFACTOR, MAXSTEPLENGTHFACTOR, TR_NE_TOP | TR_NE_BOTTOM);

  /* create an int text control for the lower bounds of points traced for a stream */
  bottom -= EDITBOXHEIGHT + MINORBORDER;
  mid = bottom + EDITBOXHEIGHT / 2;
  textBox = NewTextBox(left, right, 
		       mid - TEXTBOXHEIGHT / 2 + EDITBOXDOWN, 
		       mid + TEXTBOXHEIGHT / 2 + EDITBOXDOWN, 
		       0, "Minimum Points Text", "Minimum Points:");
  PrefixList(panelContents, textBox);
  SetVar(textBox, PARENT, panelContents);
  var = GetIntVar("AddStreamsControls", streamVisObj, MINSTREAMPOINTS);
  if (!var)
    {
      SetVar(streamVisObj, MINSTREAMPOINTS, NewInt(DEFAULTMINSTREAMPOINTS));
    }
  textBox = NewTextBox(right + MINORBORDER,
			 right + MINORBORDER + ACTEXTWIDTH,
			 mid - EDITBOXHEIGHT / 2, mid + EDITBOXHEIGHT / 2,
			 EDITABLE + WITH_PIT + ONE_LINE, "Minimum Points", "");
  SetVar(textBox, PARENT, panelContents);
  PrefixList(panelContents, textBox);
  SetTextAlign(textBox, RIGHTALIGN);
  SetVar(textBox, HELPSTRING, NewString("This text box controls the lower bound of the length, in terms of points traced, of the stream lines visualized. Only stream with its number of points traced greater than or equal to this number will be displayed."));
  AssocTextIntControlWithVar(textBox, streamVisObj, MINSTREAMPOINTS, MINMINSTREAMPOINTS, MAXMINSTREAMPOINTS, TR_NE_TOP | TR_NE_BOTTOM);

  return ObjTrue;
}

/* keep a copy of stream path so that we do not have to make every time */
static ObjPtr MakeStreamPath(streamVisObj)
ObjPtr streamVisObj;
{
  int nSpatComponents;    /* # components of data form */ /* is the rank of buckets */
  int nVecComponents;     /* # components of vector data */
  int nEmitterComponents, nEmitterDims;
  long whichEmitter, nEmitters, emitterDims[3], emitterIndices[3];
  int whichComponent, whichIndex;
  NameTyp whichName;
  real bucketBounds[6], bucketDivisionUnits[6];
  real *elements, *emitterBucketsElements;
  ObjPtr *objElements;
  ObjPtr var, streamData;      /* maindataset of stream */
  ObjPtr emitterDataset, emitterBuckets; 
  ObjPtr streamPath;
  long maxNTracePoints;   /* upper limit of the trace points */
  int integrationMethod;
  int emitterControl;
  real bounds[6], maxBoundLength;
  real ds;            /* integration length and time step */
  real *trace, *forwardTrace, *backwardTrace;   /* a trace */
  int nForwardTracePoints, nBackwardTracePoints;      /* the number of points traced */
  real stepLengthFactor;  /* c */
  real sampleFactor;      /* times 1/c to get maximum number of trace points */
  Bool markBucket;
  int i;

  ObjPtr emitterSet;
  ObjPtr *emitterSetElements;
  real *emitterElements;

  fprintf(stderr, "Making new stream path.\n");

  if (!(streamData = GetVar(streamVisObj, MAINDATASET)))
    {
      return ObjFalse;
    }
  /* set cur form */
  if (false == SetCurForm(FORMFIELD, streamData) || 
      false == SetCurField(FIELD1, streamData))
    {
      /* unable to set cur form or set cur field */
      return ObjFalse;
    }
  /* Get number of components */
  nSpatComponents = GetNComponents(FORMFIELD);
  nVecComponents = GetNComponents(FIELD1);

  /* Get integration method */
  /* INTEGRATIONMETHOD has a default value */
  var = GetVar(streamVisObj, INTEGRATIONMETHOD);
  integrationMethod = GetInt(var);

  /* Get step length factor */
  /* has a default value */
  var = GetVar(streamVisObj, STEPLENGTHFACTOR);
  stepLengthFactor = GetReal(var);

  /* Get sample parameter */
  /* has a default value */
  var = GetVar(streamVisObj, SAMPLEPARAMETER);
  sampleFactor = GetReal(var);
  /* This number is a upper limit for both forward and backward traces. */
  maxNTracePoints = (long) (sampleFactor / stepLengthFactor);

  /* Get mark bucket bool var */
  /* has a default value */
  markBucket = GetPredicate(streamVisObj, MARKBUCKET);

  /* get the EMITTERCONTROL var */
  /* has a default value */
  emitterControl = GetInt(GetIntVar("MakeStreamPath", streamVisObj, EMITTERCONTROL));

  /* Get dataform bounds */
  if (NULLOBJ == (var = GetDatasetFormBounds(streamData)))
    {
      return ObjFalse;
    }
  elements = ELEMENTS(var);
  for (i = 0; i < 2 * nSpatComponents; i++)
     {
       bounds[i] = elements[i];
     }
  /* get the maximum of the bound length */
  maxBoundLength = 0.0;
  for (i = 0; i < nSpatComponents; i++)
     {
       real temp;

       temp = bounds[2 * i + 1] - bounds[2 * i];
       if (temp > maxBoundLength) maxBoundLength = temp;
     }
  /* compute ds */
  ds = maxBoundLength * stepLengthFactor;

  if (emitterControl == USEBUCKET)
    {
      /* make emitterBuckets */
      MakeVar(streamVisObj, EMITTERBUCKETS);
      emitterBuckets = GetVar(streamVisObj, EMITTERBUCKETS);
      emitterBucketsElements = ELEMENTS(emitterBuckets);

      /* get XYZ BUCKETDIM */
      nEmitterDims = nEmitterComponents = nSpatComponents;
      for (whichComponent = 0; whichComponent < nSpatComponents; whichComponent++)
         {
           whichName = whichComponent == 2 ? ZLOWERBOUND: 
		       (whichComponent ? YLOWERBOUND: XLOWERBOUND);
           MakeVar(streamVisObj, whichName);
           var = GetVar(streamVisObj, whichName);
           bucketBounds[2 * whichComponent] = GetReal(var);
	       
           whichName = whichComponent == 2 ? ZUPPERBOUND: 
		       (whichComponent ? YUPPERBOUND: XUPPERBOUND);
           MakeVar(streamVisObj, whichName);
           var = GetVar(streamVisObj, whichName);
           bucketBounds[2 * whichComponent + 1] = GetReal(var);
	   
           if (bucketBounds[2 * whichComponent] == bucketBounds[2 * whichComponent + 1])
	     {
	       emitterDims[whichComponent] = 1;
	       bucketDivisionUnits[whichComponent] = 0; 
	     }
           else
             {
               whichName = whichComponent == 2 ? ZBUCKETDIM : 
		          (whichComponent ? YBUCKETDIM : XBUCKETDIM);
               MakeVar(streamVisObj, whichName);
               var = GetIntVar("MakeStreamPath", streamVisObj, whichName); 
               emitterDims[whichComponent] = GetInt(var);

               if (bucketBounds[2 * whichComponent] > bucketBounds[2 * whichComponent + 1])
                 {
	           /* reverse order from big to small */
	           bucketDivisionUnits[whichComponent] = 
	            ((bounds[2 * whichComponent + 1] - bounds[2 * whichComponent]) - 
		     (bucketBounds[2 * whichComponent] - bucketBounds[2 * whichComponent + 1])) / 
		      emitterDims[whichComponent];
	         }
               else
	         {
	           bucketDivisionUnits[whichComponent] = 
	            ((bucketBounds[2 * whichComponent + 1] - bucketBounds[2 * whichComponent])) / 
	            emitterDims[whichComponent];
	         }
             }
         }
      nEmitters = 1;
      for (whichComponent = 0; whichComponent < nEmitterComponents; whichComponent++)
         {
           nEmitters *= emitterDims[whichComponent];
         }

      /* reset the mark flags for emitter buckets every time the 
	 streampath is made */
      for (whichEmitter = 0; whichEmitter < nEmitters; whichEmitter++)
	 {
	   emitterBucketsElements[whichEmitter * (nEmitterComponents + 1) +
	    nEmitterComponents] = 1.0;
	 }
    }
  else if (emitterControl == USEDATASET)
    {
      /* dataset must has form */
      emitterDataset = GetVar(streamVisObj, EMITTERDATASET);
      if (NULLOBJ == emitterDataset)
	{
	  /* no dataset in emitter control corral */
	  fprintf(stderr, "The dataset is a null obj.\n");
          SetVar(streamVisObj, STREAMPATH, NULLOBJ);
          return ObjTrue;
	}

      SetCurForm(FIELD2, emitterDataset);
      nEmitterComponents = GetNComponents(FIELD2);
      if (nEmitterComponents != nSpatComponents) return ObjFalse;

      /* get form dimensions */
      nEmitterDims = CountTraversalDims(FIELD2);
      GetTraversalDims(FIELD2, emitterDims);
      nEmitters = 1;
      for (whichComponent = 0; whichComponent < nEmitterComponents; whichComponent++)
         {
           nEmitters *= emitterDims[whichComponent];
         }
    }
  else if (emitterControl == USEINTERACTIVE)
    {
      if (NULLOBJ != (emitterSet = GetVar(streamVisObj, EMITTERSET)))
	{
          emitterSetElements = (ObjPtr *) ELEMENTS(emitterSet);
	  nEmitters = DIMS(emitterSet)[0];
        }
      else
	{
          SetVar(streamVisObj, STREAMPATH, NULLOBJ);
          return ObjTrue;
	}
    }
  else
    {
      /* dummy case */
    }
    
  /* use array of AT_OBJECT for storing array objects */
  if (NULLOBJ == (streamPath = NewArray(AT_OBJECT, 1, &nEmitters)))
    {
      return ObjFalse;
    }
  objElements = (ObjPtr *) ELEMENTS(streamPath);
  for (whichEmitter = 0; whichEmitter < nEmitters; whichEmitter++)
     {
       objElements[whichEmitter] = NULLOBJ;
     }

  if (!(trace = (real *) Alloc(2 * maxNTracePoints * nSpatComponents * sizeof(real))))
    {
      OMErr();
      return ObjFalse;
    }
  forwardTrace = trace + maxNTracePoints * nSpatComponents;
  backwardTrace = trace;

  emitterIndices[0] = emitterIndices[1] = emitterIndices[2] = 0;

  /* do the trace starting from every emitter */
  for (whichEmitter = 0; whichEmitter < nEmitters; whichEmitter++)
     {
       real emitter[3];
       ObjPtr thisPath;
       int nPoints;

       /* get an emitter according to the type of EMITTERCONTROL */
       if (emitterControl == USEBUCKET)
	 {
	   long index;

           /* advance emitterIndices */
	   for (whichIndex = 0; whichIndex < nEmitterDims; whichIndex++)
	      {
	        if (++emitterIndices[whichIndex] >= emitterDims[whichIndex])
	          {
		    emitterIndices[whichIndex] = 0;
	          }
                else
	          {
		    break;
	          }
	      }

	   index = whichEmitter * (nEmitterComponents + 1);

	   if ( (true == markBucket) &&
		(emitterBucketsElements[index + nEmitterComponents] < 0.5))
	     {
	       /* the bucket is marked, i.e., passed by other stream paths */
	       continue;
	     }
           else
	     {
	       for (whichComponent = 0; whichComponent < nEmitterComponents; whichComponent++)
		  {
		    emitter[whichComponent] = emitterBucketsElements[index + whichComponent];
		  }
	     }

	 }
       else if (emitterControl == USEDATASET)
	 {
	   for (whichComponent = 0; whichComponent < nEmitterComponents; whichComponent++)
	      {
                 emitter[whichComponent] = 
		      SelectFieldComponent(FIELD2, whichComponent, emitterIndices);
	      }

           /* advance emitterIndices */
	   for (whichIndex = 0; whichIndex < nEmitterDims; whichIndex++)
	      {
	        if (++emitterIndices[whichIndex] >= emitterDims[whichIndex])
	          {
		    emitterIndices[whichIndex] = 0;
	          }
                else
	          {
		    break;
	          }
	      }
	 }
       else if (emitterControl == USEINTERACTIVE)
	 {
	   ObjPtr thisEmitterObj;

	   thisEmitterObj = emitterSetElements[whichEmitter];
	   emitterElements = (real *) ELEMENTS(thisEmitterObj); 
           emitter[0] = emitterElements[0];
           emitter[1] = emitterElements[1];
           emitter[2] = emitterElements[2];
	 }
       else
	 {
	   /* dummy case */
	   continue;
	 }

       if (integrationMethod == RUNGEKUTTA)
	 {
           /* trace forward */
           nForwardTracePoints = rkuttaCalcTrace(forwardTrace, maxNTracePoints, FORMFIELD, FIELD1, 
				      nVecComponents, nSpatComponents, emitter, bounds, ds);
           /* trace backward */
           nBackwardTracePoints = rkuttaCalcTrace(backwardTrace, maxNTracePoints, FORMFIELD, FIELD1, 
				      nVecComponents, nSpatComponents, emitter, bounds, -ds);
	 }
       else if (integrationMethod == EULER)
	 {
           nForwardTracePoints = eulerCalcTrace(forwardTrace, maxNTracePoints, FORMFIELD, FIELD1, 
				      nVecComponents, nSpatComponents, emitter, bounds, ds);
           nBackwardTracePoints = eulerCalcTrace(backwardTrace, maxNTracePoints, FORMFIELD, FIELD1, 
				      nVecComponents, nSpatComponents, emitter, bounds, -ds);

           /* HWUCHEAT 
	   nBackwardTracePoints = 1;
	   */
	 }
       else
	 {
	   /* nothings */
	 }

#ifdef DEBUD 
       fprintf(stderr, "%ld _th forward trace has %d points\n", whichEmitter, nForwardTracePoints);
       fprintf(stderr, "%ld _th backward trace has %d points\n", whichEmitter, nBackwardTracePoints);
#endif
       /* figure out number of points in the path */
       if (nBackwardTracePoints <= 1 && nForwardTracePoints <= 1 )
	 {
	   /* no stream path for this emitter at all */
	   continue;
	 }

       /* The emitter point is duplicated in the forward and backward path traces */
       nPoints = nBackwardTracePoints + nForwardTracePoints - 1;

       if (NULLOBJ == (thisPath = NewRealArray(1, nPoints * nSpatComponents)))
         {
	   Free(trace);
	   return ObjFalse;
         }
       elements = ELEMENTS(thisPath);
       objElements[whichEmitter] = thisPath;
       
       /* fill the elements */
       if (nBackwardTracePoints > 1)
	 {
	   int f, b, fIndex, bIndex;

	   for (f = 0, b = nBackwardTracePoints - 1; f < nBackwardTracePoints; f++, b--)
	      {
		fIndex = nSpatComponents * f;
		bIndex = nSpatComponents * b;

		for (whichComponent = 0; whichComponent < nSpatComponents; whichComponent++)
		   {
		     elements[fIndex + whichComponent] = backwardTrace[bIndex + whichComponent];
		   }
	      }
	 }

       if (nForwardTracePoints > 1)
	 {
	   int f, fIndex, fStart, b, bIndex, bStart;

           if (nBackwardTracePoints > 1)
	     {
               bStart = nBackwardTracePoints;
	       fStart = 1;
	     }
           else
	     {
               bStart = fStart =  0;
	     }

	   for (f = fStart, b = bStart ; f < nForwardTracePoints; f++, b++)
	      {
		fIndex = nSpatComponents * f;
		bIndex = nSpatComponents * b;

		for (whichComponent = 0; whichComponent < nSpatComponents; whichComponent++)
		   {
		     elements[bIndex + whichComponent] = forwardTrace[fIndex + whichComponent];
		   }
	      }
	 }

       if ((USEBUCKET == emitterControl) && (true == markBucket))
	 {
            /* check if the stream line passes other unused bucket */
            long bucketOffset, whichOffset;
	    long j;
            float offset;
	    real *sample;
      
	    for (j = 0; j < nPoints; j++)
	       {
		 /* go through the elements of the trace and exmamine each point */
		 sample = elements + nSpatComponents * j;

                 bucketOffset = 0;
                 for (whichComponent = nSpatComponents - 1; whichComponent >= 0; whichComponent--)
                    {
                      if (bucketBounds[2 * whichComponent] == bucketBounds[2 * whichComponent + 1])
                        {
	                  whichOffset = 0;
	                }
                      else if (bucketBounds[2 * whichComponent] > bucketBounds[2 * whichComponent + 1])
                        {
	                  if (sample[whichComponent] < bucketBounds[whichComponent * 2] &&
		              sample[whichComponent] >= bucketBounds[whichComponent * 2 + 1])
		            {
			      /* out side of the bounds */
			      break;
		            }
                          else
		            {
		              if ((offset = sample[whichComponent] - bucketBounds[whichComponent * 2]) < 0)
		                {
		                  offset += bounds[2 * whichComponent + 1] - bounds[2 * whichComponent];
		                }
                              whichOffset = (long) (offset / bucketDivisionUnits[whichComponent]);
		           }
	                }
                      else
	                {
	                  if (sample[whichComponent] < bucketBounds[whichComponent * 2] ||
		              sample[whichComponent] >= bucketBounds[whichComponent * 2 + 1])
		            {
			      break;
		            }
                          else
		            {
		              offset = sample[whichComponent] - bucketBounds[whichComponent * 2]; 
                              whichOffset = (long) (offset / bucketDivisionUnits[whichComponent]);
		            }
	                }

		      /* disregard buckets with indices preceding the cur emitter bucket */
                      if (whichOffset < emitterIndices[whichComponent])
			{
			  break;
			}
       
                      if (whichComponent > 0)
	                {
                          bucketOffset = (bucketOffset + whichOffset) * emitterDims[whichComponent - 1];
	                }
                      else
	                {
	                  bucketOffset += whichOffset;
	                }
                    }

		 /* mark it only if all components were traversed and
		    bucketOffset is after current emitter */
		 if (whichComponent == -1)
		   {
	             emitterBucketsElements[bucketOffset * (nSpatComponents + 1) + nSpatComponents] = 0.0;
		   }
	      }
	   }
     }

  Free(trace);
  SetVar(streamVisObj, STREAMPATH, streamPath);
  return ObjTrue;
}

static ObjPtr MakeStreamLinesSurface(streamLineVisObj)
ObjPtr streamLineVisObj;
/*Makes the surface of stream vis object*/
{
  ObjPtr streamData;      /* maindataset of stream */
  ObjPtr var;             /* an object pointer */
  ObjPtr surface;         /* surface picture */
  ObjPtr streamPath, thisPath;
  ObjPtr *objElements;
  real *elements;
  int nSpatComponents;    /* # components of data form */ /* is the rank of buckets */
  VertexPtr *vertices;
  long nVertices;
  int minPoints;
  long nPaths, n;

  fprintf(stderr, "Making new stream line surface.\n");

  if (!(streamData = GetVar(streamLineVisObj, MAINDATASET)))
    {
      return ObjFalse;
    }

  /* Get number of spatial components */
  if (false == SetCurForm(FORMFIELD, streamData))
    {
      return ObjFalse;
    }
  nSpatComponents = GetNComponents(FORMFIELD);
  
  /* Get minimum stream points */
  if (!(var = GetVar(streamLineVisObj, MINSTREAMPOINTS)))
    {
      minPoints = DEFAULTMINSTREAMPOINTS; 
    }
  else
    {
      minPoints = GetInt(var);
    }

  /* Create a new surface picture */
  if (!(surface = NewPicture()))
    {
      return ObjFalse;
    }
  
  /* Get STREAMPATH */
  MakeVar(streamLineVisObj, STREAMPATH);
  if (streamPath = GetVar(streamLineVisObj, STREAMPATH))
    {
      if (RANK(streamPath) != 1)
        {
          return ObjFalse;
        }
      nPaths = DIMS(streamPath)[0];
      objElements = (ObjPtr *) ELEMENTS(streamPath);

      for (n = 0; n < nPaths; n++)
         {
           long i;

           if (NULLOBJ == (thisPath = objElements[n]))
	     {
	       /* see if there is an real array there */
	       continue;
	     }
           if (1 != RANK(thisPath))
	     {
	       return ObjFalse;
	     }
           nVertices = DIMS(thisPath)[0] / nSpatComponents;

           if (nVertices < minPoints)
	     {
	       continue;
	     }

           elements = (real *) ELEMENTS(thisPath);
           if (NULL == (vertices = (VertexPtr *)Alloc(nVertices * sizeof(VertexPtr))))
	     {
	       OMErr();
	       return ObjFalse;
	     }
           for (i = 0; i < nVertices; i++)
	      {
	        long index;
    
	        if (!(vertices[i] = NewVertex(surface, 0)))
	          {
		    Free(vertices);
		    return ObjFalse;
	          }

                index = i * nSpatComponents;
	        vertices[i]->position[0] = elements[index];
	        vertices[i]->position[1] = elements[index + 1];
	        vertices[i]->position[2] = (nSpatComponents > 2 ? elements[index + 2] : 0.0);
	      }
           AppendSPolylineToPicture(surface, 1, 0, nVertices, vertices);
           Free(vertices);
         }
    }

  /* set the pictures's MAINDATASET for facilitating ColorPictureByObject */
  SetVar(surface, MAINDATASET, streamData);
  SetVar(streamLineVisObj, SURFACE, surface);
  SetVar(surface, REPOBJ, streamLineVisObj);
  return ObjTrue;
}

/* Initializes stream lines visualization class */
#ifdef HAVE_PROTOTYPES
void InitStreamlines(void)
#else
void InitStreamlines()
#endif
{
    ObjPtr icon;
#ifdef GRAPHICS
    Object o;
    float n[3], v[3];
    short c[3];
#endif

    /*Class for streams */
    visStreamLines = NewObject(visLineEmitters, 0);
    AddToReferenceList(visStreamLines);
    DefineFixedClass(visStreamLines, CLASS_STREAMLINES, "visStreamlines");

    /* A. Variables setup: essential variables for all new vis classes */
    /* 1. a NAME */
    SetVar(visStreamLines, NAME, NewString("Streamlines"));

    /* 2. a DEFAULTICON */
    SetVar(visStreamLines, DEFAULTICON, icon = NewObject(visIcon, 0));
    SetVar(icon, WHICHICON, NewInt(ICONQUESTION));
    SetVar(icon, NAME, NewString("Streamlines"));
    SetVar(icon, HELPSTRING, NewString("This icon represents a streamline visualization object."));  
#ifdef GRAPHICS
    o = genobj();
    makeobj(o);
    bgnpolygon();
    c[0] = 128;
    c[1] = 128;
    c[2] = 128;
    c3s(c);
    n[0] = 0.270598;
    n[1] = 0.923880;
    n[2] = -0.270598;
    n3f(n);
    v[0] = 0.405412;
    v[1] = 0.418478;
    v[2] = 0.394588;
    v3f(v);
    v[0] = -0.394588;
    v[1] = 0.418478;
    v[2] = -0.405412;
    v3f(v);
    n[0] = 0.00000;
    n[1] = 1.00000;
    n[2] = 0.00000;
    n3f(n);
    v[0] = -0.400000;
    v[1] = 0.420000;
    v[2] = -0.400000;
    v3f(v);
    v[0] = 0.400000;
    v[1] = 0.420000;
    v[2] = 0.400000;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.400000;
    v[1] = 0.420000;
    v[2] = 0.400000;
    v3f(v);
    v[0] = -0.400000;
    v[1] = 0.420000;
    v[2] = -0.400000;
    v3f(v);
    n[0] = -0.270598;
    n[1] = 0.923880;
    n[2] = 0.270598;
    n3f(n);
    v[0] = -0.405412;
    v[1] = 0.418478;
    v[2] = -0.394588;
    v3f(v);
    v[0] = 0.394588;
    v[1] = 0.418478;
    v[2] = 0.405412;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.394588;
    v[1] = 0.418478;
    v[2] = 0.405412;
    v3f(v);
    v[0] = -0.405412;
    v[1] = 0.418478;
    v[2] = -0.394588;
    v3f(v);
    n[0] = -0.500000;
    n[1] = 0.707107;
    n[2] = 0.500000;
    n3f(n);
    v[0] = -0.410000;
    v[1] = 0.414142;
    v[2] = -0.390000;
    v3f(v);
    v[0] = 0.390000;
    v[1] = 0.414142;
    v[2] = 0.410000;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.390000;
    v[1] = 0.414142;
    v[2] = 0.410000;
    v3f(v);
    v[0] = -0.410000;
    v[1] = 0.414142;
    v[2] = -0.390000;
    v3f(v);
    n[0] = -0.653282;
    n[1] = 0.382683;
    n[2] = 0.653282;
    n3f(n);
    v[0] = -0.413066;
    v[1] = 0.407654;
    v[2] = -0.386934;
    v3f(v);
    v[0] = 0.386934;
    v[1] = 0.407654;
    v[2] = 0.413066;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.386934;
    v[1] = 0.407654;
    v[2] = 0.413066;
    v3f(v);
    v[0] = -0.413066;
    v[1] = 0.407654;
    v[2] = -0.386934;
    v3f(v);
    n[0] = -0.707107;
    n[1] = 0.00000;
    n[2] = 0.707107;
    n3f(n);
    v[0] = -0.414142;
    v[1] = 0.400000;
    v[2] = -0.385858;
    v3f(v);
    v[0] = 0.385858;
    v[1] = 0.400000;
    v[2] = 0.414142;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.385858;
    v[1] = 0.400000;
    v[2] = 0.414142;
    v3f(v);
    v[0] = -0.414142;
    v[1] = 0.400000;
    v[2] = -0.385858;
    v3f(v);
    n[0] = -0.653282;
    n[1] = -0.382683;
    n[2] = 0.653282;
    n3f(n);
    v[0] = -0.413066;
    v[1] = 0.392346;
    v[2] = -0.386934;
    v3f(v);
    v[0] = 0.386934;
    v[1] = 0.392346;
    v[2] = 0.413066;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.386934;
    v[1] = 0.392346;
    v[2] = 0.413066;
    v3f(v);
    v[0] = -0.413066;
    v[1] = 0.392346;
    v[2] = -0.386934;
    v3f(v);
    n[0] = -0.500000;
    n[1] = -0.707107;
    n[2] = 0.500000;
    n3f(n);
    v[0] = -0.410000;
    v[1] = 0.385858;
    v[2] = -0.390000;
    v3f(v);
    v[0] = 0.390000;
    v[1] = 0.385858;
    v[2] = 0.410000;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.390000;
    v[1] = 0.385858;
    v[2] = 0.410000;
    v3f(v);
    v[0] = -0.410000;
    v[1] = 0.385858;
    v[2] = -0.390000;
    v3f(v);
    n[0] = -0.270598;
    n[1] = -0.923880;
    n[2] = 0.270598;
    n3f(n);
    v[0] = -0.405412;
    v[1] = 0.381522;
    v[2] = -0.394588;
    v3f(v);
    v[0] = 0.394588;
    v[1] = 0.381522;
    v[2] = 0.405412;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.394588;
    v[1] = 0.381522;
    v[2] = 0.405412;
    v3f(v);
    v[0] = -0.405412;
    v[1] = 0.381522;
    v[2] = -0.394588;
    v3f(v);
    n[0] = 0.00000;
    n[1] = -1.00000;
    n[2] = 0.00000;
    n3f(n);
    v[0] = -0.400000;
    v[1] = 0.380000;
    v[2] = -0.400000;
    v3f(v);
    v[0] = 0.400000;
    v[1] = 0.380000;
    v[2] = 0.400000;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.400000;
    v[1] = 0.380000;
    v[2] = 0.400000;
    v3f(v);
    v[0] = -0.400000;
    v[1] = 0.380000;
    v[2] = -0.400000;
    v3f(v);
    n[0] = 0.270598;
    n[1] = -0.923880;
    n[2] = -0.270598;
    n3f(n);
    v[0] = -0.394588;
    v[1] = 0.381522;
    v[2] = -0.405412;
    v3f(v);
    v[0] = 0.405412;
    v[1] = 0.381522;
    v[2] = 0.394588;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.405412;
    v[1] = 0.381522;
    v[2] = 0.394588;
    v3f(v);
    v[0] = -0.394588;
    v[1] = 0.381522;
    v[2] = -0.405412;
    v3f(v);
    n[0] = 0.500000;
    n[1] = -0.707107;
    n[2] = -0.500000;
    n3f(n);
    v[0] = -0.390000;
    v[1] = 0.385858;
    v[2] = -0.410000;
    v3f(v);
    v[0] = 0.410000;
    v[1] = 0.385858;
    v[2] = 0.390000;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.410000;
    v[1] = 0.385858;
    v[2] = 0.390000;
    v3f(v);
    v[0] = -0.390000;
    v[1] = 0.385858;
    v[2] = -0.410000;
    v3f(v);
    n[0] = 0.653282;
    n[1] = -0.382683;
    n[2] = -0.653282;
    n3f(n);
    v[0] = -0.386934;
    v[1] = 0.392346;
    v[2] = -0.413066;
    v3f(v);
    v[0] = 0.413066;
    v[1] = 0.392346;
    v[2] = 0.386934;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.413066;
    v[1] = 0.392346;
    v[2] = 0.386934;
    v3f(v);
    v[0] = -0.386934;
    v[1] = 0.392346;
    v[2] = -0.413066;
    v3f(v);
    n[0] = 0.707107;
    n[1] = 0.00000;
    n[2] = -0.707107;
    n3f(n);
    v[0] = -0.385858;
    v[1] = 0.400000;
    v[2] = -0.414142;
    v3f(v);
    v[0] = 0.414142;
    v[1] = 0.400000;
    v[2] = 0.385858;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.414142;
    v[1] = 0.400000;
    v[2] = 0.385858;
    v3f(v);
    v[0] = -0.385858;
    v[1] = 0.400000;
    v[2] = -0.414142;
    v3f(v);
    n[0] = 0.653282;
    n[1] = 0.382683;
    n[2] = -0.653282;
    n3f(n);
    v[0] = -0.386934;
    v[1] = 0.407654;
    v[2] = -0.413066;
    v3f(v);
    v[0] = 0.413066;
    v[1] = 0.407654;
    v[2] = 0.386934;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.413066;
    v[1] = 0.407654;
    v[2] = 0.386934;
    v3f(v);
    v[0] = -0.386934;
    v[1] = 0.407654;
    v[2] = -0.413066;
    v3f(v);
    n[0] = 0.500000;
    n[1] = 0.707107;
    n[2] = -0.500000;
    n3f(n);
    v[0] = -0.390000;
    v[1] = 0.414142;
    v[2] = -0.410000;
    v3f(v);
    v[0] = 0.410000;
    v[1] = 0.414142;
    v[2] = 0.390000;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.410000;
    v[1] = 0.414142;
    v[2] = 0.390000;
    v3f(v);
    v[0] = -0.390000;
    v[1] = 0.414142;
    v[2] = -0.410000;
    v3f(v);
    n[0] = 0.270598;
    n[1] = 0.923880;
    n[2] = -0.270598;
    n3f(n);
    v[0] = -0.394588;
    v[1] = 0.418478;
    v[2] = -0.405412;
    v3f(v);
    v[0] = 0.405412;
    v[1] = 0.418478;
    v[2] = 0.394588;
    v3f(v);
    endpolygon();

    bgnpolygon();
    n[0] = 0.707107;
    n[1] = 0.00000;
    n[2] = 0.707107;
    n3f(n);
    v[0] = 0.400000;
    v[1] = 0.420000;
    v[2] = 0.400000;
    v3f(v);
    v[0] = 0.394588;
    v[1] = 0.418478;
    v[2] = 0.405412;
    v3f(v);
    v[0] = 0.390000;
    v[1] = 0.414142;
    v[2] = 0.410000;
    v3f(v);
    v[0] = 0.386934;
    v[1] = 0.407654;
    v[2] = 0.413066;
    v3f(v);
    v[0] = 0.385858;
    v[1] = 0.400000;
    v[2] = 0.414142;
    v3f(v);
    v[0] = 0.386934;
    v[1] = 0.392346;
    v[2] = 0.413066;
    v3f(v);
    v[0] = 0.390000;
    v[1] = 0.385858;
    v[2] = 0.410000;
    v3f(v);
    v[0] = 0.394588;
    v[1] = 0.381522;
    v[2] = 0.405412;
    v3f(v);
    v[0] = 0.400000;
    v[1] = 0.380000;
    v[2] = 0.400000;
    v3f(v);
    v[0] = 0.405412;
    v[1] = 0.381522;
    v[2] = 0.394588;
    v3f(v);
    v[0] = 0.410000;
    v[1] = 0.385858;
    v[2] = 0.390000;
    v3f(v);
    v[0] = 0.413066;
    v[1] = 0.392346;
    v[2] = 0.386934;
    v3f(v);
    v[0] = 0.414142;
    v[1] = 0.400000;
    v[2] = 0.385858;
    v3f(v);
    v[0] = 0.413066;
    v[1] = 0.407654;
    v[2] = 0.386934;
    v3f(v);
    v[0] = 0.410000;
    v[1] = 0.414142;
    v[2] = 0.390000;
    v3f(v);
    v[0] = 0.405412;
    v[1] = 0.418478;
    v[2] = 0.394588;
    v3f(v);
    endpolygon();

    bgnpolygon();
    n[0] = -0.707107;
    n[1] = 0.00000;
    n[2] = -0.707107;
    n3f(n);
    v[0] = -0.400000;
    v[1] = 0.420000;
    v[2] = -0.400000;
    v3f(v);
    v[0] = -0.405412;
    v[1] = 0.418478;
    v[2] = -0.394588;
    v3f(v);
    v[0] = -0.410000;
    v[1] = 0.414142;
    v[2] = -0.390000;
    v3f(v);
    v[0] = -0.413066;
    v[1] = 0.407654;
    v[2] = -0.386934;
    v3f(v);
    v[0] = -0.414142;
    v[1] = 0.400000;
    v[2] = -0.385858;
    v3f(v);
    v[0] = -0.413066;
    v[1] = 0.392346;
    v[2] = -0.386934;
    v3f(v);
    v[0] = -0.410000;
    v[1] = 0.385858;
    v[2] = -0.390000;
    v3f(v);
    v[0] = -0.405412;
    v[1] = 0.381522;
    v[2] = -0.394588;
    v3f(v);
    v[0] = -0.400000;
    v[1] = 0.380000;
    v[2] = -0.400000;
    v3f(v);
    v[0] = -0.394588;
    v[1] = 0.381522;
    v[2] = -0.405412;
    v3f(v);
    v[0] = -0.390000;
    v[1] = 0.385858;
    v[2] = -0.410000;
    v3f(v);
    v[0] = -0.386934;
    v[1] = 0.392346;
    v[2] = -0.413066;
    v3f(v);
    v[0] = -0.385858;
    v[1] = 0.400000;
    v[2] = -0.414142;
    v3f(v);
    v[0] = -0.386934;
    v[1] = 0.407654;
    v[2] = -0.413066;
    v3f(v);
    v[0] = -0.390000;
    v[1] = 0.414142;
    v[2] = -0.410000;
    v3f(v);
    v[0] = -0.394588;
    v[1] = 0.418478;
    v[2] = -0.405412;
    v3f(v);
    endpolygon();

    bgnpolygon();
    n[0] = 0.306147;
    n[1] = 0.923880;
    n[2] = -0.229610;
    n3f(n);
    v[0] = 0.306123;
    v[1] = 0.218478;
    v[2] = 0.395408;
    v3f(v);
    v[0] = -0.293877;
    v[1] = 0.218478;
    v[2] = -0.404592;
    v3f(v);
    n[0] = 0.00000;
    n[1] = 1.00000;
    n[2] = 0.00000;
    n3f(n);
    v[0] = -0.300000;
    v[1] = 0.220000;
    v[2] = -0.400000;
    v3f(v);
    v[0] = 0.300000;
    v[1] = 0.220000;
    v[2] = 0.400000;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.300000;
    v[1] = 0.220000;
    v[2] = 0.400000;
    v3f(v);
    v[0] = -0.300000;
    v[1] = 0.220000;
    v[2] = -0.400000;
    v3f(v);
    n[0] = -0.306147;
    n[1] = 0.923880;
    n[2] = 0.229610;
    n3f(n);
    v[0] = -0.306123;
    v[1] = 0.218478;
    v[2] = -0.395408;
    v3f(v);
    v[0] = 0.293877;
    v[1] = 0.218478;
    v[2] = 0.404592;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.293877;
    v[1] = 0.218478;
    v[2] = 0.404592;
    v3f(v);
    v[0] = -0.306123;
    v[1] = 0.218478;
    v[2] = -0.395408;
    v3f(v);
    n[0] = -0.565685;
    n[1] = 0.707107;
    n[2] = 0.424264;
    n3f(n);
    v[0] = -0.311314;
    v[1] = 0.214142;
    v[2] = -0.391515;
    v3f(v);
    v[0] = 0.288686;
    v[1] = 0.214142;
    v[2] = 0.408485;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.288686;
    v[1] = 0.214142;
    v[2] = 0.408485;
    v3f(v);
    v[0] = -0.311314;
    v[1] = 0.214142;
    v[2] = -0.391515;
    v3f(v);
    n[0] = -0.739104;
    n[1] = 0.382683;
    n[2] = 0.554328;
    n3f(n);
    v[0] = -0.314782;
    v[1] = 0.207654;
    v[2] = -0.388913;
    v3f(v);
    v[0] = 0.285218;
    v[1] = 0.207654;
    v[2] = 0.411087;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.285218;
    v[1] = 0.207654;
    v[2] = 0.411087;
    v3f(v);
    v[0] = -0.314782;
    v[1] = 0.207654;
    v[2] = -0.388913;
    v3f(v);
    n[0] = -0.800000;
    n[1] = 0.00000;
    n[2] = 0.600000;
    n3f(n);
    v[0] = -0.316000;
    v[1] = 0.200000;
    v[2] = -0.388000;
    v3f(v);
    v[0] = 0.284000;
    v[1] = 0.200000;
    v[2] = 0.412000;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.284000;
    v[1] = 0.200000;
    v[2] = 0.412000;
    v3f(v);
    v[0] = -0.316000;
    v[1] = 0.200000;
    v[2] = -0.388000;
    v3f(v);
    n[0] = -0.739104;
    n[1] = -0.382683;
    n[2] = 0.554328;
    n3f(n);
    v[0] = -0.314782;
    v[1] = 0.192346;
    v[2] = -0.388913;
    v3f(v);
    v[0] = 0.285218;
    v[1] = 0.192346;
    v[2] = 0.411087;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.285218;
    v[1] = 0.192346;
    v[2] = 0.411087;
    v3f(v);
    v[0] = -0.314782;
    v[1] = 0.192346;
    v[2] = -0.388913;
    v3f(v);
    n[0] = -0.565685;
    n[1] = -0.707107;
    n[2] = 0.424264;
    n3f(n);
    v[0] = -0.311314;
    v[1] = 0.185858;
    v[2] = -0.391515;
    v3f(v);
    v[0] = 0.288686;
    v[1] = 0.185858;
    v[2] = 0.408485;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.288686;
    v[1] = 0.185858;
    v[2] = 0.408485;
    v3f(v);
    v[0] = -0.311314;
    v[1] = 0.185858;
    v[2] = -0.391515;
    v3f(v);
    n[0] = -0.306147;
    n[1] = -0.923880;
    n[2] = 0.229610;
    n3f(n);
    v[0] = -0.306123;
    v[1] = 0.181522;
    v[2] = -0.395408;
    v3f(v);
    v[0] = 0.293877;
    v[1] = 0.181522;
    v[2] = 0.404592;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.293877;
    v[1] = 0.181522;
    v[2] = 0.404592;
    v3f(v);
    v[0] = -0.306123;
    v[1] = 0.181522;
    v[2] = -0.395408;
    v3f(v);
    n[0] = 0.00000;
    n[1] = -1.00000;
    n[2] = 0.00000;
    n3f(n);
    v[0] = -0.300000;
    v[1] = 0.180000;
    v[2] = -0.400000;
    v3f(v);
    v[0] = 0.300000;
    v[1] = 0.180000;
    v[2] = 0.400000;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.300000;
    v[1] = 0.180000;
    v[2] = 0.400000;
    v3f(v);
    v[0] = -0.300000;
    v[1] = 0.180000;
    v[2] = -0.400000;
    v3f(v);
    n[0] = 0.306147;
    n[1] = -0.923880;
    n[2] = -0.229610;
    n3f(n);
    v[0] = -0.293877;
    v[1] = 0.181522;
    v[2] = -0.404592;
    v3f(v);
    v[0] = 0.306123;
    v[1] = 0.181522;
    v[2] = 0.395408;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.306123;
    v[1] = 0.181522;
    v[2] = 0.395408;
    v3f(v);
    v[0] = -0.293877;
    v[1] = 0.181522;
    v[2] = -0.404592;
    v3f(v);
    n[0] = 0.565685;
    n[1] = -0.707107;
    n[2] = -0.424264;
    n3f(n);
    v[0] = -0.288686;
    v[1] = 0.185858;
    v[2] = -0.408485;
    v3f(v);
    v[0] = 0.311314;
    v[1] = 0.185858;
    v[2] = 0.391515;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.311314;
    v[1] = 0.185858;
    v[2] = 0.391515;
    v3f(v);
    v[0] = -0.288686;
    v[1] = 0.185858;
    v[2] = -0.408485;
    v3f(v);
    n[0] = 0.739104;
    n[1] = -0.382683;
    n[2] = -0.554328;
    n3f(n);
    v[0] = -0.285218;
    v[1] = 0.192346;
    v[2] = -0.411087;
    v3f(v);
    v[0] = 0.314782;
    v[1] = 0.192346;
    v[2] = 0.388913;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.314782;
    v[1] = 0.192346;
    v[2] = 0.388913;
    v3f(v);
    v[0] = -0.285218;
    v[1] = 0.192346;
    v[2] = -0.411087;
    v3f(v);
    n[0] = 0.800000;
    n[1] = 0.00000;
    n[2] = -0.600000;
    n3f(n);
    v[0] = -0.284000;
    v[1] = 0.200000;
    v[2] = -0.412000;
    v3f(v);
    v[0] = 0.316000;
    v[1] = 0.200000;
    v[2] = 0.388000;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.316000;
    v[1] = 0.200000;
    v[2] = 0.388000;
    v3f(v);
    v[0] = -0.284000;
    v[1] = 0.200000;
    v[2] = -0.412000;
    v3f(v);
    n[0] = 0.739104;
    n[1] = 0.382683;
    n[2] = -0.554328;
    n3f(n);
    v[0] = -0.285218;
    v[1] = 0.207654;
    v[2] = -0.411087;
    v3f(v);
    v[0] = 0.314782;
    v[1] = 0.207654;
    v[2] = 0.388913;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.314782;
    v[1] = 0.207654;
    v[2] = 0.388913;
    v3f(v);
    v[0] = -0.285218;
    v[1] = 0.207654;
    v[2] = -0.411087;
    v3f(v);
    n[0] = 0.565685;
    n[1] = 0.707107;
    n[2] = -0.424264;
    n3f(n);
    v[0] = -0.288686;
    v[1] = 0.214142;
    v[2] = -0.408485;
    v3f(v);
    v[0] = 0.311314;
    v[1] = 0.214142;
    v[2] = 0.391515;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.311314;
    v[1] = 0.214142;
    v[2] = 0.391515;
    v3f(v);
    v[0] = -0.288686;
    v[1] = 0.214142;
    v[2] = -0.408485;
    v3f(v);
    n[0] = 0.306147;
    n[1] = 0.923880;
    n[2] = -0.229610;
    n3f(n);
    v[0] = -0.293877;
    v[1] = 0.218478;
    v[2] = -0.404592;
    v3f(v);
    v[0] = 0.306123;
    v[1] = 0.218478;
    v[2] = 0.395408;
    v3f(v);
    endpolygon();

    bgnpolygon();
    n[0] = 0.600000;
    n[1] = 0.00000;
    n[2] = 0.800000;
    n3f(n);
    v[0] = 0.300000;
    v[1] = 0.220000;
    v[2] = 0.400000;
    v3f(v);
    v[0] = 0.293877;
    v[1] = 0.218478;
    v[2] = 0.404592;
    v3f(v);
    v[0] = 0.288686;
    v[1] = 0.214142;
    v[2] = 0.408485;
    v3f(v);
    v[0] = 0.285218;
    v[1] = 0.207654;
    v[2] = 0.411087;
    v3f(v);
    v[0] = 0.284000;
    v[1] = 0.200000;
    v[2] = 0.412000;
    v3f(v);
    v[0] = 0.285218;
    v[1] = 0.192346;
    v[2] = 0.411087;
    v3f(v);
    v[0] = 0.288686;
    v[1] = 0.185858;
    v[2] = 0.408485;
    v3f(v);
    v[0] = 0.293877;
    v[1] = 0.181522;
    v[2] = 0.404592;
    v3f(v);
    v[0] = 0.300000;
    v[1] = 0.180000;
    v[2] = 0.400000;
    v3f(v);
    v[0] = 0.306123;
    v[1] = 0.181522;
    v[2] = 0.395408;
    v3f(v);
    v[0] = 0.311314;
    v[1] = 0.185858;
    v[2] = 0.391515;
    v3f(v);
    v[0] = 0.314782;
    v[1] = 0.192346;
    v[2] = 0.388913;
    v3f(v);
    v[0] = 0.316000;
    v[1] = 0.200000;
    v[2] = 0.388000;
    v3f(v);
    v[0] = 0.314782;
    v[1] = 0.207654;
    v[2] = 0.388913;
    v3f(v);
    v[0] = 0.311314;
    v[1] = 0.214142;
    v[2] = 0.391515;
    v3f(v);
    v[0] = 0.306123;
    v[1] = 0.218478;
    v[2] = 0.395408;
    v3f(v);
    endpolygon();

    bgnpolygon();
    n[0] = -0.600000;
    n[1] = 0.00000;
    n[2] = -0.800000;
    n3f(n);
    v[0] = -0.300000;
    v[1] = 0.220000;
    v[2] = -0.400000;
    v3f(v);
    v[0] = -0.306123;
    v[1] = 0.218478;
    v[2] = -0.395408;
    v3f(v);
    v[0] = -0.311314;
    v[1] = 0.214142;
    v[2] = -0.391515;
    v3f(v);
    v[0] = -0.314782;
    v[1] = 0.207654;
    v[2] = -0.388913;
    v3f(v);
    v[0] = -0.316000;
    v[1] = 0.200000;
    v[2] = -0.388000;
    v3f(v);
    v[0] = -0.314782;
    v[1] = 0.192346;
    v[2] = -0.388913;
    v3f(v);
    v[0] = -0.311314;
    v[1] = 0.185858;
    v[2] = -0.391515;
    v3f(v);
    v[0] = -0.306123;
    v[1] = 0.181522;
    v[2] = -0.395408;
    v3f(v);
    v[0] = -0.300000;
    v[1] = 0.180000;
    v[2] = -0.400000;
    v3f(v);
    v[0] = -0.293877;
    v[1] = 0.181522;
    v[2] = -0.404592;
    v3f(v);
    v[0] = -0.288686;
    v[1] = 0.185858;
    v[2] = -0.408485;
    v3f(v);
    v[0] = -0.285218;
    v[1] = 0.192346;
    v[2] = -0.411087;
    v3f(v);
    v[0] = -0.284000;
    v[1] = 0.200000;
    v[2] = -0.412000;
    v3f(v);
    v[0] = -0.285218;
    v[1] = 0.207654;
    v[2] = -0.411087;
    v3f(v);
    v[0] = -0.288686;
    v[1] = 0.214142;
    v[2] = -0.408485;
    v3f(v);
    v[0] = -0.293877;
    v[1] = 0.218478;
    v[2] = -0.404592;
    v3f(v);
    endpolygon();

    bgnpolygon();
    n[0] = 0.342282;
    n[1] = 0.923880;
    n[2] = -0.171141;
    n3f(n);
    v[0] = 0.206846;
    v[1] = 0.0184776;
    v[2] = 0.396577;
    v3f(v);
    v[0] = -0.193154;
    v[1] = 0.0184776;
    v[2] = -0.403423;
    v3f(v);
    n[0] = 0.00000;
    n[1] = 1.00000;
    n[2] = 0.00000;
    n3f(n);
    v[0] = -0.200000;
    v[1] = 0.0200000;
    v[2] = -0.400000;
    v3f(v);
    v[0] = 0.200000;
    v[1] = 0.0200000;
    v[2] = 0.400000;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.200000;
    v[1] = 0.0200000;
    v[2] = 0.400000;
    v3f(v);
    v[0] = -0.200000;
    v[1] = 0.0200000;
    v[2] = -0.400000;
    v3f(v);
    n[0] = -0.342282;
    n[1] = 0.923880;
    n[2] = 0.171141;
    n3f(n);
    v[0] = -0.206846;
    v[1] = 0.0184776;
    v[2] = -0.396577;
    v3f(v);
    v[0] = 0.193154;
    v[1] = 0.0184776;
    v[2] = 0.403423;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.193154;
    v[1] = 0.0184776;
    v[2] = 0.403423;
    v3f(v);
    v[0] = -0.206846;
    v[1] = 0.0184776;
    v[2] = -0.396577;
    v3f(v);
    n[0] = -0.632456;
    n[1] = 0.707107;
    n[2] = 0.316228;
    n3f(n);
    v[0] = -0.212649;
    v[1] = 0.0141421;
    v[2] = -0.393675;
    v3f(v);
    v[0] = 0.187351;
    v[1] = 0.0141421;
    v[2] = 0.406325;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.187351;
    v[1] = 0.0141421;
    v[2] = 0.406325;
    v3f(v);
    v[0] = -0.212649;
    v[1] = 0.0141421;
    v[2] = -0.393675;
    v3f(v);
    n[0] = -0.826343;
    n[1] = 0.382683;
    n[2] = 0.413171;
    n3f(n);
    v[0] = -0.216527;
    v[1] = 0.00765367;
    v[2] = -0.391737;
    v3f(v);
    v[0] = 0.183473;
    v[1] = 0.00765367;
    v[2] = 0.408263;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.183473;
    v[1] = 0.00765367;
    v[2] = 0.408263;
    v3f(v);
    v[0] = -0.216527;
    v[1] = 0.00765367;
    v[2] = -0.391737;
    v3f(v);
    n[0] = -0.894427;
    n[1] = 0.00000;
    n[2] = 0.447214;
    n3f(n);
    v[0] = -0.217889;
    v[1] = 0.00000;
    v[2] = -0.391056;
    v3f(v);
    v[0] = 0.182111;
    v[1] = 0.00000;
    v[2] = 0.408944;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.182111;
    v[1] = 0.00000;
    v[2] = 0.408944;
    v3f(v);
    v[0] = -0.217889;
    v[1] = 0.00000;
    v[2] = -0.391056;
    v3f(v);
    n[0] = -0.826343;
    n[1] = -0.382683;
    n[2] = 0.413171;
    n3f(n);
    v[0] = -0.216527;
    v[1] = -0.00765367;
    v[2] = -0.391737;
    v3f(v);
    v[0] = 0.183473;
    v[1] = -0.00765367;
    v[2] = 0.408263;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.183473;
    v[1] = -0.00765367;
    v[2] = 0.408263;
    v3f(v);
    v[0] = -0.216527;
    v[1] = -0.00765367;
    v[2] = -0.391737;
    v3f(v);
    n[0] = -0.632456;
    n[1] = -0.707107;
    n[2] = 0.316228;
    n3f(n);
    v[0] = -0.212649;
    v[1] = -0.0141421;
    v[2] = -0.393675;
    v3f(v);
    v[0] = 0.187351;
    v[1] = -0.0141421;
    v[2] = 0.406325;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.187351;
    v[1] = -0.0141421;
    v[2] = 0.406325;
    v3f(v);
    v[0] = -0.212649;
    v[1] = -0.0141421;
    v[2] = -0.393675;
    v3f(v);
    n[0] = -0.342282;
    n[1] = -0.923880;
    n[2] = 0.171141;
    n3f(n);
    v[0] = -0.206846;
    v[1] = -0.0184776;
    v[2] = -0.396577;
    v3f(v);
    v[0] = 0.193154;
    v[1] = -0.0184776;
    v[2] = 0.403423;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.193154;
    v[1] = -0.0184776;
    v[2] = 0.403423;
    v3f(v);
    v[0] = -0.206846;
    v[1] = -0.0184776;
    v[2] = -0.396577;
    v3f(v);
    n[0] = 0.00000;
    n[1] = -1.00000;
    n[2] = 0.00000;
    n3f(n);
    v[0] = -0.200000;
    v[1] = -0.0200000;
    v[2] = -0.400000;
    v3f(v);
    v[0] = 0.200000;
    v[1] = -0.0200000;
    v[2] = 0.400000;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.200000;
    v[1] = -0.0200000;
    v[2] = 0.400000;
    v3f(v);
    v[0] = -0.200000;
    v[1] = -0.0200000;
    v[2] = -0.400000;
    v3f(v);
    n[0] = 0.342282;
    n[1] = -0.923880;
    n[2] = -0.171141;
    n3f(n);
    v[0] = -0.193154;
    v[1] = -0.0184776;
    v[2] = -0.403423;
    v3f(v);
    v[0] = 0.206846;
    v[1] = -0.0184776;
    v[2] = 0.396577;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.206846;
    v[1] = -0.0184776;
    v[2] = 0.396577;
    v3f(v);
    v[0] = -0.193154;
    v[1] = -0.0184776;
    v[2] = -0.403423;
    v3f(v);
    n[0] = 0.632456;
    n[1] = -0.707107;
    n[2] = -0.316228;
    n3f(n);
    v[0] = -0.187351;
    v[1] = -0.0141421;
    v[2] = -0.406325;
    v3f(v);
    v[0] = 0.212649;
    v[1] = -0.0141421;
    v[2] = 0.393675;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.212649;
    v[1] = -0.0141421;
    v[2] = 0.393675;
    v3f(v);
    v[0] = -0.187351;
    v[1] = -0.0141421;
    v[2] = -0.406325;
    v3f(v);
    n[0] = 0.826343;
    n[1] = -0.382683;
    n[2] = -0.413171;
    n3f(n);
    v[0] = -0.183473;
    v[1] = -0.00765367;
    v[2] = -0.408263;
    v3f(v);
    v[0] = 0.216527;
    v[1] = -0.00765367;
    v[2] = 0.391737;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.216527;
    v[1] = -0.00765367;
    v[2] = 0.391737;
    v3f(v);
    v[0] = -0.183473;
    v[1] = -0.00765367;
    v[2] = -0.408263;
    v3f(v);
    n[0] = 0.894427;
    n[1] = 0.00000;
    n[2] = -0.447214;
    n3f(n);
    v[0] = -0.182111;
    v[1] = 0.00000;
    v[2] = -0.408944;
    v3f(v);
    v[0] = 0.217889;
    v[1] = 0.00000;
    v[2] = 0.391056;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.217889;
    v[1] = 0.00000;
    v[2] = 0.391056;
    v3f(v);
    v[0] = -0.182111;
    v[1] = 0.00000;
    v[2] = -0.408944;
    v3f(v);
    n[0] = 0.826343;
    n[1] = 0.382683;
    n[2] = -0.413171;
    n3f(n);
    v[0] = -0.183473;
    v[1] = 0.00765367;
    v[2] = -0.408263;
    v3f(v);
    v[0] = 0.216527;
    v[1] = 0.00765367;
    v[2] = 0.391737;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.216527;
    v[1] = 0.00765367;
    v[2] = 0.391737;
    v3f(v);
    v[0] = -0.183473;
    v[1] = 0.00765367;
    v[2] = -0.408263;
    v3f(v);
    n[0] = 0.632456;
    n[1] = 0.707107;
    n[2] = -0.316228;
    n3f(n);
    v[0] = -0.187351;
    v[1] = 0.0141421;
    v[2] = -0.406325;
    v3f(v);
    v[0] = 0.212649;
    v[1] = 0.0141421;
    v[2] = 0.393675;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.212649;
    v[1] = 0.0141421;
    v[2] = 0.393675;
    v3f(v);
    v[0] = -0.187351;
    v[1] = 0.0141421;
    v[2] = -0.406325;
    v3f(v);
    n[0] = 0.342282;
    n[1] = 0.923880;
    n[2] = -0.171141;
    n3f(n);
    v[0] = -0.193154;
    v[1] = 0.0184776;
    v[2] = -0.403423;
    v3f(v);
    v[0] = 0.206846;
    v[1] = 0.0184776;
    v[2] = 0.396577;
    v3f(v);
    endpolygon();

    bgnpolygon();
    n[0] = 0.447214;
    n[1] = 0.00000;
    n[2] = 0.894427;
    n3f(n);
    v[0] = 0.200000;
    v[1] = 0.0200000;
    v[2] = 0.400000;
    v3f(v);
    v[0] = 0.193154;
    v[1] = 0.0184776;
    v[2] = 0.403423;
    v3f(v);
    v[0] = 0.187351;
    v[1] = 0.0141421;
    v[2] = 0.406325;
    v3f(v);
    v[0] = 0.183473;
    v[1] = 0.00765367;
    v[2] = 0.408263;
    v3f(v);
    v[0] = 0.182111;
    v[1] = 0.00000;
    v[2] = 0.408944;
    v3f(v);
    v[0] = 0.183473;
    v[1] = -0.00765367;
    v[2] = 0.408263;
    v3f(v);
    v[0] = 0.187351;
    v[1] = -0.0141421;
    v[2] = 0.406325;
    v3f(v);
    v[0] = 0.193154;
    v[1] = -0.0184776;
    v[2] = 0.403423;
    v3f(v);
    v[0] = 0.200000;
    v[1] = -0.0200000;
    v[2] = 0.400000;
    v3f(v);
    v[0] = 0.206846;
    v[1] = -0.0184776;
    v[2] = 0.396577;
    v3f(v);
    v[0] = 0.212649;
    v[1] = -0.0141421;
    v[2] = 0.393675;
    v3f(v);
    v[0] = 0.216527;
    v[1] = -0.00765367;
    v[2] = 0.391737;
    v3f(v);
    v[0] = 0.217889;
    v[1] = 0.00000;
    v[2] = 0.391056;
    v3f(v);
    v[0] = 0.216527;
    v[1] = 0.00765367;
    v[2] = 0.391737;
    v3f(v);
    v[0] = 0.212649;
    v[1] = 0.0141421;
    v[2] = 0.393675;
    v3f(v);
    v[0] = 0.206846;
    v[1] = 0.0184776;
    v[2] = 0.396577;
    v3f(v);
    endpolygon();

    bgnpolygon();
    n[0] = -0.447214;
    n[1] = 0.00000;
    n[2] = -0.894427;
    n3f(n);
    v[0] = -0.200000;
    v[1] = 0.0200000;
    v[2] = -0.400000;
    v3f(v);
    v[0] = -0.206846;
    v[1] = 0.0184776;
    v[2] = -0.396577;
    v3f(v);
    v[0] = -0.212649;
    v[1] = 0.0141421;
    v[2] = -0.393675;
    v3f(v);
    v[0] = -0.216527;
    v[1] = 0.00765367;
    v[2] = -0.391737;
    v3f(v);
    v[0] = -0.217889;
    v[1] = 0.00000;
    v[2] = -0.391056;
    v3f(v);
    v[0] = -0.216527;
    v[1] = -0.00765367;
    v[2] = -0.391737;
    v3f(v);
    v[0] = -0.212649;
    v[1] = -0.0141421;
    v[2] = -0.393675;
    v3f(v);
    v[0] = -0.206846;
    v[1] = -0.0184776;
    v[2] = -0.396577;
    v3f(v);
    v[0] = -0.200000;
    v[1] = -0.0200000;
    v[2] = -0.400000;
    v3f(v);
    v[0] = -0.193154;
    v[1] = -0.0184776;
    v[2] = -0.403423;
    v3f(v);
    v[0] = -0.187351;
    v[1] = -0.0141421;
    v[2] = -0.406325;
    v3f(v);
    v[0] = -0.183473;
    v[1] = -0.00765367;
    v[2] = -0.408263;
    v3f(v);
    v[0] = -0.182111;
    v[1] = 0.00000;
    v[2] = -0.408944;
    v3f(v);
    v[0] = -0.183473;
    v[1] = 0.00765367;
    v[2] = -0.408263;
    v3f(v);
    v[0] = -0.187351;
    v[1] = 0.0141421;
    v[2] = -0.406325;
    v3f(v);
    v[0] = -0.193154;
    v[1] = 0.0184776;
    v[2] = -0.403423;
    v3f(v);
    endpolygon();

    bgnpolygon();
    n[0] = 0.371257;
    n[1] = 0.923880;
    n[2] = -0.0928144;
    n3f(n);
    v[0] = 0.107425;
    v[1] = -0.181522;
    v[2] = 0.398144;
    v3f(v);
    v[0] = -0.0925748;
    v[1] = -0.181522;
    v[2] = -0.401856;
    v3f(v);
    n[0] = 0.00000;
    n[1] = 1.00000;
    n[2] = 0.00000;
    n3f(n);
    v[0] = -0.100000;
    v[1] = -0.180000;
    v[2] = -0.400000;
    v3f(v);
    v[0] = 0.100000;
    v[1] = -0.180000;
    v[2] = 0.400000;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.100000;
    v[1] = -0.180000;
    v[2] = 0.400000;
    v3f(v);
    v[0] = -0.100000;
    v[1] = -0.180000;
    v[2] = -0.400000;
    v3f(v);
    n[0] = -0.371257;
    n[1] = 0.923880;
    n[2] = 0.0928144;
    n3f(n);
    v[0] = -0.107425;
    v[1] = -0.181522;
    v[2] = -0.398144;
    v3f(v);
    v[0] = 0.0925748;
    v[1] = -0.181522;
    v[2] = 0.401856;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.0925748;
    v[1] = -0.181522;
    v[2] = 0.401856;
    v3f(v);
    v[0] = -0.107425;
    v[1] = -0.181522;
    v[2] = -0.398144;
    v3f(v);
    n[0] = -0.685994;
    n[1] = 0.707107;
    n[2] = 0.171499;
    n3f(n);
    v[0] = -0.113720;
    v[1] = -0.185858;
    v[2] = -0.396570;
    v3f(v);
    v[0] = 0.0862801;
    v[1] = -0.185858;
    v[2] = 0.403430;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.0862801;
    v[1] = -0.185858;
    v[2] = 0.403430;
    v3f(v);
    v[0] = -0.113720;
    v[1] = -0.185858;
    v[2] = -0.396570;
    v3f(v);
    n[0] = -0.896295;
    n[1] = 0.382683;
    n[2] = 0.224074;
    n3f(n);
    v[0] = -0.117926;
    v[1] = -0.192346;
    v[2] = -0.395519;
    v3f(v);
    v[0] = 0.0820741;
    v[1] = -0.192346;
    v[2] = 0.404481;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.0820741;
    v[1] = -0.192346;
    v[2] = 0.404481;
    v3f(v);
    v[0] = -0.117926;
    v[1] = -0.192346;
    v[2] = -0.395519;
    v3f(v);
    n[0] = -0.970142;
    n[1] = 0.00000;
    n[2] = 0.242536;
    n3f(n);
    v[0] = -0.119403;
    v[1] = -0.200000;
    v[2] = -0.395149;
    v3f(v);
    v[0] = 0.0805972;
    v[1] = -0.200000;
    v[2] = 0.404851;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.0805972;
    v[1] = -0.200000;
    v[2] = 0.404851;
    v3f(v);
    v[0] = -0.119403;
    v[1] = -0.200000;
    v[2] = -0.395149;
    v3f(v);
    n[0] = -0.896295;
    n[1] = -0.382683;
    n[2] = 0.224074;
    n3f(n);
    v[0] = -0.117926;
    v[1] = -0.207654;
    v[2] = -0.395519;
    v3f(v);
    v[0] = 0.0820741;
    v[1] = -0.207654;
    v[2] = 0.404481;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.0820741;
    v[1] = -0.207654;
    v[2] = 0.404481;
    v3f(v);
    v[0] = -0.117926;
    v[1] = -0.207654;
    v[2] = -0.395519;
    v3f(v);
    n[0] = -0.685994;
    n[1] = -0.707107;
    n[2] = 0.171499;
    n3f(n);
    v[0] = -0.113720;
    v[1] = -0.214142;
    v[2] = -0.396570;
    v3f(v);
    v[0] = 0.0862801;
    v[1] = -0.214142;
    v[2] = 0.403430;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.0862801;
    v[1] = -0.214142;
    v[2] = 0.403430;
    v3f(v);
    v[0] = -0.113720;
    v[1] = -0.214142;
    v[2] = -0.396570;
    v3f(v);
    n[0] = -0.371257;
    n[1] = -0.923880;
    n[2] = 0.0928144;
    n3f(n);
    v[0] = -0.107425;
    v[1] = -0.218478;
    v[2] = -0.398144;
    v3f(v);
    v[0] = 0.0925748;
    v[1] = -0.218478;
    v[2] = 0.401856;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.0925748;
    v[1] = -0.218478;
    v[2] = 0.401856;
    v3f(v);
    v[0] = -0.107425;
    v[1] = -0.218478;
    v[2] = -0.398144;
    v3f(v);
    n[0] = 0.00000;
    n[1] = -1.00000;
    n[2] = 0.00000;
    n3f(n);
    v[0] = -0.100000;
    v[1] = -0.220000;
    v[2] = -0.400000;
    v3f(v);
    v[0] = 0.100000;
    v[1] = -0.220000;
    v[2] = 0.400000;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.100000;
    v[1] = -0.220000;
    v[2] = 0.400000;
    v3f(v);
    v[0] = -0.100000;
    v[1] = -0.220000;
    v[2] = -0.400000;
    v3f(v);
    n[0] = 0.371257;
    n[1] = -0.923880;
    n[2] = -0.0928144;
    n3f(n);
    v[0] = -0.0925748;
    v[1] = -0.218478;
    v[2] = -0.401856;
    v3f(v);
    v[0] = 0.107425;
    v[1] = -0.218478;
    v[2] = 0.398144;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.107425;
    v[1] = -0.218478;
    v[2] = 0.398144;
    v3f(v);
    v[0] = -0.0925748;
    v[1] = -0.218478;
    v[2] = -0.401856;
    v3f(v);
    n[0] = 0.685994;
    n[1] = -0.707107;
    n[2] = -0.171499;
    n3f(n);
    v[0] = -0.0862801;
    v[1] = -0.214142;
    v[2] = -0.403430;
    v3f(v);
    v[0] = 0.113720;
    v[1] = -0.214142;
    v[2] = 0.396570;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.113720;
    v[1] = -0.214142;
    v[2] = 0.396570;
    v3f(v);
    v[0] = -0.0862801;
    v[1] = -0.214142;
    v[2] = -0.403430;
    v3f(v);
    n[0] = 0.896295;
    n[1] = -0.382683;
    n[2] = -0.224074;
    n3f(n);
    v[0] = -0.0820741;
    v[1] = -0.207654;
    v[2] = -0.404481;
    v3f(v);
    v[0] = 0.117926;
    v[1] = -0.207654;
    v[2] = 0.395519;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.117926;
    v[1] = -0.207654;
    v[2] = 0.395519;
    v3f(v);
    v[0] = -0.0820741;
    v[1] = -0.207654;
    v[2] = -0.404481;
    v3f(v);
    n[0] = 0.970142;
    n[1] = 0.00000;
    n[2] = -0.242536;
    n3f(n);
    v[0] = -0.0805972;
    v[1] = -0.200000;
    v[2] = -0.404851;
    v3f(v);
    v[0] = 0.119403;
    v[1] = -0.200000;
    v[2] = 0.395149;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.119403;
    v[1] = -0.200000;
    v[2] = 0.395149;
    v3f(v);
    v[0] = -0.0805972;
    v[1] = -0.200000;
    v[2] = -0.404851;
    v3f(v);
    n[0] = 0.896295;
    n[1] = 0.382683;
    n[2] = -0.224074;
    n3f(n);
    v[0] = -0.0820741;
    v[1] = -0.192346;
    v[2] = -0.404481;
    v3f(v);
    v[0] = 0.117926;
    v[1] = -0.192346;
    v[2] = 0.395519;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.117926;
    v[1] = -0.192346;
    v[2] = 0.395519;
    v3f(v);
    v[0] = -0.0820741;
    v[1] = -0.192346;
    v[2] = -0.404481;
    v3f(v);
    n[0] = 0.685994;
    n[1] = 0.707107;
    n[2] = -0.171499;
    n3f(n);
    v[0] = -0.0862801;
    v[1] = -0.185858;
    v[2] = -0.403430;
    v3f(v);
    v[0] = 0.113720;
    v[1] = -0.185858;
    v[2] = 0.396570;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.113720;
    v[1] = -0.185858;
    v[2] = 0.396570;
    v3f(v);
    v[0] = -0.0862801;
    v[1] = -0.185858;
    v[2] = -0.403430;
    v3f(v);
    n[0] = 0.371257;
    n[1] = 0.923880;
    n[2] = -0.0928144;
    n3f(n);
    v[0] = -0.0925748;
    v[1] = -0.181522;
    v[2] = -0.401856;
    v3f(v);
    v[0] = 0.107425;
    v[1] = -0.181522;
    v[2] = 0.398144;
    v3f(v);
    endpolygon();

    bgnpolygon();
    n[0] = 0.242536;
    n[1] = 0.00000;
    n[2] = 0.970142;
    n3f(n);
    v[0] = 0.100000;
    v[1] = -0.180000;
    v[2] = 0.400000;
    v3f(v);
    v[0] = 0.0925748;
    v[1] = -0.181522;
    v[2] = 0.401856;
    v3f(v);
    v[0] = 0.0862801;
    v[1] = -0.185858;
    v[2] = 0.403430;
    v3f(v);
    v[0] = 0.0820741;
    v[1] = -0.192346;
    v[2] = 0.404481;
    v3f(v);
    v[0] = 0.0805972;
    v[1] = -0.200000;
    v[2] = 0.404851;
    v3f(v);
    v[0] = 0.0820741;
    v[1] = -0.207654;
    v[2] = 0.404481;
    v3f(v);
    v[0] = 0.0862801;
    v[1] = -0.214142;
    v[2] = 0.403430;
    v3f(v);
    v[0] = 0.0925748;
    v[1] = -0.218478;
    v[2] = 0.401856;
    v3f(v);
    v[0] = 0.100000;
    v[1] = -0.220000;
    v[2] = 0.400000;
    v3f(v);
    v[0] = 0.107425;
    v[1] = -0.218478;
    v[2] = 0.398144;
    v3f(v);
    v[0] = 0.113720;
    v[1] = -0.214142;
    v[2] = 0.396570;
    v3f(v);
    v[0] = 0.117926;
    v[1] = -0.207654;
    v[2] = 0.395519;
    v3f(v);
    v[0] = 0.119403;
    v[1] = -0.200000;
    v[2] = 0.395149;
    v3f(v);
    v[0] = 0.117926;
    v[1] = -0.192346;
    v[2] = 0.395519;
    v3f(v);
    v[0] = 0.113720;
    v[1] = -0.185858;
    v[2] = 0.396570;
    v3f(v);
    v[0] = 0.107425;
    v[1] = -0.181522;
    v[2] = 0.398144;
    v3f(v);
    endpolygon();

    bgnpolygon();
    n[0] = -0.242536;
    n[1] = 0.00000;
    n[2] = -0.970142;
    n3f(n);
    v[0] = -0.100000;
    v[1] = -0.180000;
    v[2] = -0.400000;
    v3f(v);
    v[0] = -0.107425;
    v[1] = -0.181522;
    v[2] = -0.398144;
    v3f(v);
    v[0] = -0.113720;
    v[1] = -0.185858;
    v[2] = -0.396570;
    v3f(v);
    v[0] = -0.117926;
    v[1] = -0.192346;
    v[2] = -0.395519;
    v3f(v);
    v[0] = -0.119403;
    v[1] = -0.200000;
    v[2] = -0.395149;
    v3f(v);
    v[0] = -0.117926;
    v[1] = -0.207654;
    v[2] = -0.395519;
    v3f(v);
    v[0] = -0.113720;
    v[1] = -0.214142;
    v[2] = -0.396570;
    v3f(v);
    v[0] = -0.107425;
    v[1] = -0.218478;
    v[2] = -0.398144;
    v3f(v);
    v[0] = -0.100000;
    v[1] = -0.220000;
    v[2] = -0.400000;
    v3f(v);
    v[0] = -0.0925748;
    v[1] = -0.218478;
    v[2] = -0.401856;
    v3f(v);
    v[0] = -0.0862801;
    v[1] = -0.214142;
    v[2] = -0.403430;
    v3f(v);
    v[0] = -0.0820741;
    v[1] = -0.207654;
    v[2] = -0.404481;
    v3f(v);
    v[0] = -0.0805972;
    v[1] = -0.200000;
    v[2] = -0.404851;
    v3f(v);
    v[0] = -0.0820741;
    v[1] = -0.192346;
    v[2] = -0.404481;
    v3f(v);
    v[0] = -0.0862801;
    v[1] = -0.185858;
    v[2] = -0.403430;
    v3f(v);
    v[0] = -0.0925748;
    v[1] = -0.181522;
    v[2] = -0.401856;
    v3f(v);
    endpolygon();

    bgnpolygon();
    n[0] = 0.382683;
    n[1] = 0.923880;
    n[2] = 0.00000;
    n3f(n);
    v[0] = 0.00765367;
    v[1] = -0.381522;
    v[2] = 0.400000;
    v3f(v);
    v[0] = 0.00765367;
    v[1] = -0.381522;
    v[2] = -0.400000;
    v3f(v);
    n[0] = 0.00000;
    n[1] = 1.00000;
    n[2] = 0.00000;
    n3f(n);
    v[0] = 0.00000;
    v[1] = -0.380000;
    v[2] = -0.400000;
    v3f(v);
    v[0] = 0.00000;
    v[1] = -0.380000;
    v[2] = 0.400000;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.00000;
    v[1] = -0.380000;
    v[2] = 0.400000;
    v3f(v);
    v[0] = 0.00000;
    v[1] = -0.380000;
    v[2] = -0.400000;
    v3f(v);
    n[0] = -0.382683;
    n[1] = 0.923880;
    n[2] = 0.00000;
    n3f(n);
    v[0] = -0.00765367;
    v[1] = -0.381522;
    v[2] = -0.400000;
    v3f(v);
    v[0] = -0.00765367;
    v[1] = -0.381522;
    v[2] = 0.400000;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = -0.00765367;
    v[1] = -0.381522;
    v[2] = 0.400000;
    v3f(v);
    v[0] = -0.00765367;
    v[1] = -0.381522;
    v[2] = -0.400000;
    v3f(v);
    n[0] = -0.707107;
    n[1] = 0.707107;
    n[2] = 0.00000;
    n3f(n);
    v[0] = -0.0141421;
    v[1] = -0.385858;
    v[2] = -0.400000;
    v3f(v);
    v[0] = -0.0141421;
    v[1] = -0.385858;
    v[2] = 0.400000;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = -0.0141421;
    v[1] = -0.385858;
    v[2] = 0.400000;
    v3f(v);
    v[0] = -0.0141421;
    v[1] = -0.385858;
    v[2] = -0.400000;
    v3f(v);
    n[0] = -0.923880;
    n[1] = 0.382683;
    n[2] = 0.00000;
    n3f(n);
    v[0] = -0.0184776;
    v[1] = -0.392346;
    v[2] = -0.400000;
    v3f(v);
    v[0] = -0.0184776;
    v[1] = -0.392346;
    v[2] = 0.400000;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = -0.0184776;
    v[1] = -0.392346;
    v[2] = 0.400000;
    v3f(v);
    v[0] = -0.0184776;
    v[1] = -0.392346;
    v[2] = -0.400000;
    v3f(v);
    n[0] = -1.00000;
    n[1] = 0.00000;
    n[2] = 0.00000;
    n3f(n);
    v[0] = -0.0200000;
    v[1] = -0.400000;
    v[2] = -0.400000;
    v3f(v);
    v[0] = -0.0200000;
    v[1] = -0.400000;
    v[2] = 0.400000;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = -0.0200000;
    v[1] = -0.400000;
    v[2] = 0.400000;
    v3f(v);
    v[0] = -0.0200000;
    v[1] = -0.400000;
    v[2] = -0.400000;
    v3f(v);
    n[0] = -0.923880;
    n[1] = -0.382683;
    n[2] = 0.00000;
    n3f(n);
    v[0] = -0.0184776;
    v[1] = -0.407654;
    v[2] = -0.400000;
    v3f(v);
    v[0] = -0.0184776;
    v[1] = -0.407654;
    v[2] = 0.400000;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = -0.0184776;
    v[1] = -0.407654;
    v[2] = 0.400000;
    v3f(v);
    v[0] = -0.0184776;
    v[1] = -0.407654;
    v[2] = -0.400000;
    v3f(v);
    n[0] = -0.707107;
    n[1] = -0.707107;
    n[2] = 0.00000;
    n3f(n);
    v[0] = -0.0141421;
    v[1] = -0.414142;
    v[2] = -0.400000;
    v3f(v);
    v[0] = -0.0141421;
    v[1] = -0.414142;
    v[2] = 0.400000;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = -0.0141421;
    v[1] = -0.414142;
    v[2] = 0.400000;
    v3f(v);
    v[0] = -0.0141421;
    v[1] = -0.414142;
    v[2] = -0.400000;
    v3f(v);
    n[0] = -0.382683;
    n[1] = -0.923880;
    n[2] = 0.00000;
    n3f(n);
    v[0] = -0.00765367;
    v[1] = -0.418478;
    v[2] = -0.400000;
    v3f(v);
    v[0] = -0.00765367;
    v[1] = -0.418478;
    v[2] = 0.400000;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = -0.00765367;
    v[1] = -0.418478;
    v[2] = 0.400000;
    v3f(v);
    v[0] = -0.00765367;
    v[1] = -0.418478;
    v[2] = -0.400000;
    v3f(v);
    n[0] = 0.00000;
    n[1] = -1.00000;
    n[2] = 0.00000;
    n3f(n);
    v[0] = 0.00000;
    v[1] = -0.420000;
    v[2] = -0.400000;
    v3f(v);
    v[0] = 0.00000;
    v[1] = -0.420000;
    v[2] = 0.400000;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.00000;
    v[1] = -0.420000;
    v[2] = 0.400000;
    v3f(v);
    v[0] = 0.00000;
    v[1] = -0.420000;
    v[2] = -0.400000;
    v3f(v);
    n[0] = 0.382683;
    n[1] = -0.923880;
    n[2] = 0.00000;
    n3f(n);
    v[0] = 0.00765367;
    v[1] = -0.418478;
    v[2] = -0.400000;
    v3f(v);
    v[0] = 0.00765367;
    v[1] = -0.418478;
    v[2] = 0.400000;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.00765367;
    v[1] = -0.418478;
    v[2] = 0.400000;
    v3f(v);
    v[0] = 0.00765367;
    v[1] = -0.418478;
    v[2] = -0.400000;
    v3f(v);
    n[0] = 0.707107;
    n[1] = -0.707107;
    n[2] = 0.00000;
    n3f(n);
    v[0] = 0.0141421;
    v[1] = -0.414142;
    v[2] = -0.400000;
    v3f(v);
    v[0] = 0.0141421;
    v[1] = -0.414142;
    v[2] = 0.400000;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.0141421;
    v[1] = -0.414142;
    v[2] = 0.400000;
    v3f(v);
    v[0] = 0.0141421;
    v[1] = -0.414142;
    v[2] = -0.400000;
    v3f(v);
    n[0] = 0.923880;
    n[1] = -0.382683;
    n[2] = 0.00000;
    n3f(n);
    v[0] = 0.0184776;
    v[1] = -0.407654;
    v[2] = -0.400000;
    v3f(v);
    v[0] = 0.0184776;
    v[1] = -0.407654;
    v[2] = 0.400000;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.0184776;
    v[1] = -0.407654;
    v[2] = 0.400000;
    v3f(v);
    v[0] = 0.0184776;
    v[1] = -0.407654;
    v[2] = -0.400000;
    v3f(v);
    n[0] = 1.00000;
    n[1] = 0.00000;
    n[2] = 0.00000;
    n3f(n);
    v[0] = 0.0200000;
    v[1] = -0.400000;
    v[2] = -0.400000;
    v3f(v);
    v[0] = 0.0200000;
    v[1] = -0.400000;
    v[2] = 0.400000;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.0200000;
    v[1] = -0.400000;
    v[2] = 0.400000;
    v3f(v);
    v[0] = 0.0200000;
    v[1] = -0.400000;
    v[2] = -0.400000;
    v3f(v);
    n[0] = 0.923880;
    n[1] = 0.382683;
    n[2] = 0.00000;
    n3f(n);
    v[0] = 0.0184776;
    v[1] = -0.392346;
    v[2] = -0.400000;
    v3f(v);
    v[0] = 0.0184776;
    v[1] = -0.392346;
    v[2] = 0.400000;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.0184776;
    v[1] = -0.392346;
    v[2] = 0.400000;
    v3f(v);
    v[0] = 0.0184776;
    v[1] = -0.392346;
    v[2] = -0.400000;
    v3f(v);
    n[0] = 0.707107;
    n[1] = 0.707107;
    n[2] = 0.00000;
    n3f(n);
    v[0] = 0.0141421;
    v[1] = -0.385858;
    v[2] = -0.400000;
    v3f(v);
    v[0] = 0.0141421;
    v[1] = -0.385858;
    v[2] = 0.400000;
    v3f(v);
    endpolygon();

    bgnpolygon();
    v[0] = 0.0141421;
    v[1] = -0.385858;
    v[2] = 0.400000;
    v3f(v);
    v[0] = 0.0141421;
    v[1] = -0.385858;
    v[2] = -0.400000;
    v3f(v);
    n[0] = 0.382683;
    n[1] = 0.923880;
    n[2] = 0.00000;
    n3f(n);
    v[0] = 0.00765367;
    v[1] = -0.381522;
    v[2] = -0.400000;
    v3f(v);
    v[0] = 0.00765367;
    v[1] = -0.381522;
    v[2] = 0.400000;
    v3f(v);
    endpolygon();

    bgnpolygon();
    n[0] = 0.00000;
    n[1] = 0.00000;
    n[2] = 1.00000;
    n3f(n);
    v[0] = 0.00000;
    v[1] = -0.380000;
    v[2] = 0.400000;
    v3f(v);
    v[0] = -0.00765367;
    v[1] = -0.381522;
    v[2] = 0.400000;
    v3f(v);
    v[0] = -0.0141421;
    v[1] = -0.385858;
    v[2] = 0.400000;
    v3f(v);
    v[0] = -0.0184776;
    v[1] = -0.392346;
    v[2] = 0.400000;
    v3f(v);
    v[0] = -0.0200000;
    v[1] = -0.400000;
    v[2] = 0.400000;
    v3f(v);
    v[0] = -0.0184776;
    v[1] = -0.407654;
    v[2] = 0.400000;
    v3f(v);
    v[0] = -0.0141421;
    v[1] = -0.414142;
    v[2] = 0.400000;
    v3f(v);
    v[0] = -0.00765367;
    v[1] = -0.418478;
    v[2] = 0.400000;
    v3f(v);
    v[0] = 0.00000;
    v[1] = -0.420000;
    v[2] = 0.400000;
    v3f(v);
    v[0] = 0.00765367;
    v[1] = -0.418478;
    v[2] = 0.400000;
    v3f(v);
    v[0] = 0.0141421;
    v[1] = -0.414142;
    v[2] = 0.400000;
    v3f(v);
    v[0] = 0.0184776;
    v[1] = -0.407654;
    v[2] = 0.400000;
    v3f(v);
    v[0] = 0.0200000;
    v[1] = -0.400000;
    v[2] = 0.400000;
    v3f(v);
    v[0] = 0.0184776;
    v[1] = -0.392346;
    v[2] = 0.400000;
    v3f(v);
    v[0] = 0.0141421;
    v[1] = -0.385858;
    v[2] = 0.400000;
    v3f(v);
    v[0] = 0.00765367;
    v[1] = -0.381522;
    v[2] = 0.400000;
    v3f(v);
    endpolygon();

    bgnpolygon();
    n[0] = 0.00000;
    n[1] = 0.00000;
    n[2] = -1.00000;
    n3f(n);
    v[0] = 0.00000;
    v[1] = -0.380000;
    v[2] = -0.400000;
    v3f(v);
    v[0] = -0.00765367;
    v[1] = -0.381522;
    v[2] = -0.400000;
    v3f(v);
    v[0] = -0.0141421;
    v[1] = -0.385858;
    v[2] = -0.400000;
    v3f(v);
    v[0] = -0.0184776;
    v[1] = -0.392346;
    v[2] = -0.400000;
    v3f(v);
    v[0] = -0.0200000;
    v[1] = -0.400000;
    v[2] = -0.400000;
    v3f(v);
    v[0] = -0.0184776;
    v[1] = -0.407654;
    v[2] = -0.400000;
    v3f(v);
    v[0] = -0.0141421;
    v[1] = -0.414142;
    v[2] = -0.400000;
    v3f(v);
    v[0] = -0.00765367;
    v[1] = -0.418478;
    v[2] = -0.400000;
    v3f(v);
    v[0] = 0.00000;
    v[1] = -0.420000;
    v[2] = -0.400000;
    v3f(v);
    v[0] = 0.00765367;
    v[1] = -0.418478;
    v[2] = -0.400000;
    v3f(v);
    v[0] = 0.0141421;
    v[1] = -0.414142;
    v[2] = -0.400000;
    v3f(v);
    v[0] = 0.0184776;
    v[1] = -0.407654;
    v[2] = -0.400000;
    v3f(v);
    v[0] = 0.0200000;
    v[1] = -0.400000;
    v[2] = -0.400000;
    v3f(v);
    v[0] = 0.0184776;
    v[1] = -0.392346;
    v[2] = -0.400000;
    v3f(v);
    v[0] = 0.0141421;
    v[1] = -0.385858;
    v[2] = -0.400000;
    v3f(v);
    v[0] = 0.00765367;
    v[1] = -0.381522;
    v[2] = -0.400000;
    v3f(v);
    endpolygon();

    closeobj();
    SetVar(icon, CACHEGRAPHICS, NewInt(o));
#endif

    /* 3. a CONTROLICON: click on this to show control panel */
    icon = NewIcon(0, 0, ICONQUESTION, "Streamlines");
    SetVar(icon, HELPSTRING,
	NewString("Click on this icon to see a panel of controls for the streamline object."));
    SetVar(visStreamLines, CONTROLICON, icon);
    SetVar(icon, PANELHELP, NewString("The visualization object is intended to show stream lines."));

    /* 3.5 add snap shot variables */
    AddSnapVar(visStreamLines, INTEGRATIONMETHOD);
    AddSnapVar(visStreamLines, SAMPLEPARAMETER);
    AddSnapVar(visStreamLines, STEPLENGTHFACTOR);
    AddSnapVar(visStreamLines, MINSTREAMPOINTS);
    AddSnapVar(visStreamLines, MARKBUCKET);
    AddSnapVar(visStreamLines, STREAMPATH);


    /* 4. variables specific to the new vis class: define same default values for vis
          objects of this class */
    SetVar(visStreamLines, INTEGRATIONMETHOD, NewInt(EULER)); 
    SetVar(visStreamLines, SAMPLEPARAMETER, NewReal(DEFAULTSAMPLEPARAMETER));
    SetVar(visStreamLines, STEPLENGTHFACTOR, NewReal(DEFAULTSTEPLENGTHFACTOR));
    SetVar(visStreamLines, MINSTREAMPOINTS, NewInt(DEFAULTMINSTREAMPOINTS));
#ifdef SIGGRAPH
    SetVar(visStreamLines, MARKBUCKET, NewInt(0));
#else
    SetVar(visStreamLines, MARKBUCKET, NewInt(1));
#endif
    

    /* B. Dependecy set up */
    DeclareDependency(visStreamLines, SURFACE, STREAMPATH);
    DeclareDependency(visStreamLines, SURFACE, MINSTREAMPOINTS);
    DeclareDependency(visStreamLines, STREAMPATH, INTEGRATIONMETHOD);
    DeclareDependency(visStreamLines, STREAMPATH, SAMPLEPARAMETER);
    DeclareDependency(visStreamLines, STREAMPATH, STEPLENGTHFACTOR);
    DeclareDependency(visStreamLines, STREAMPATH, MARKBUCKET);
    DeclareDependency(visStreamLines, STREAMPATH, EMITTERCONTROL);
    DeclareDependency(visStreamLines, STREAMPATH, EMITTERBUCKETS);
    DeclareDependency(visStreamLines, STREAMPATH, EMITTERDATASET);
    DeclareDependency(visStreamLines, STREAMPATH, EMITTERSET);
    DeclareIndirectDependency(visStreamLines, STREAMPATH, MAINDATASET, CHANGED);
    DeclareIndirectDependency(visStreamLines, STREAMPATH, MAINDATASET, DATA);

    SetMethod(visStreamLines, ADDCONTROLS, AddStreamsControls);
    SetMethod(visStreamLines, STREAMPATH, MakeStreamPath);
    SetMethod(visStreamLines, SURFACE, MakeStreamLinesSurface);


    /* D. Vismapping set up */
    DefineVisMapping(DS_HASFORM | DS_HASFIELD | DS_VECTOR, -1, 3, 3, visStreamLines);
    DefineVisMapping(DS_HASFORM | DS_HASFIELD | DS_VECTOR, -1, 2, 3, visStreamLines);
    DefineVisMapping(DS_HASFORM | DS_HASFIELD | DS_VECTOR, -1, 3, 2, visStreamLines);
    DefineVisMapping(DS_HASFORM | DS_HASFIELD | DS_VECTOR, -1, 2, 2, visStreamLines);
}

#ifdef HAVE_PROTOTYPES
void KillStreamlines(void)
#else
void KillStreamlines()
#endif
{
    DeleteThing(visStreamLines);
}


/******************************************************************************************/
/******************************** stream hose object **************************************/
/******************************************************************************************/
static ObjPtr MakeStreamHosesSurface(streamHoseVisObj)
ObjPtr streamHoseVisObj;
/*Makes the surface of stream vis object*/
{
  ObjPtr streamData;      /* maindataset of stream */
  ObjPtr var;             /* an object pointer */
  ObjPtr surface;         /* surface picture */
  int nSpatComponents;    /* # components of data form */ /* is the rank of buckets */
  int whichComponent;
  real *elements;             
  ObjPtr *objElements, thisPath, streamPath;
  VertexPtr *vertices;
  int minPoints;
  float sizeFactor, sizeOffset, sizeConstant;
  ObjPtr sizeObj;
  Bool capEnds;
  real size;
  PolysPtr polys;
  int curNSides, nSubDivisions;
  long nPaths, n, nVertices;

  fprintf(stderr, "Making new stream hose sruface.\n");

  if (!(streamData = GetVar(streamHoseVisObj, MAINDATASET)))
    {
      return ObjFalse;
    }

  /* set cur form */
  if (false == SetCurForm(FORMFIELD, streamData))
    {
      /* unable to set cur form or set cur field */
      return ObjFalse;
    }
  /* Get number of components */
  nSpatComponents = GetNComponents(FORMFIELD);

  /* Create a new surface picture */
  if (!(surface = NewPicture()))
    {
      return ObjFalse;
    }

  /* Get STREAMPATH */
  MakeVar(streamHoseVisObj, STREAMPATH);
  if (NULLOBJ == (streamPath = GetVar(streamHoseVisObj, STREAMPATH)))
    {
      SetVar(surface, MAINDATASET, streamData);
      SetVar(streamHoseVisObj, SURFACE, surface);
      SetVar(surface, REPOBJ, streamHoseVisObj);
      return ObjTrue;
    }
  if (RANK(streamPath) != 1)
    {
      return ObjFalse;
    }
  nPaths = DIMS(streamPath)[0];
  objElements = (ObjPtr *) ELEMENTS(streamPath);

  /* Get minimum stream points */
  if (!(var = GetVar(streamHoseVisObj, MINSTREAMPOINTS)))
    {
      minPoints = DEFAULTMINSTREAMPOINTS; 
    }
  else
    {
      minPoints = GetInt(var);
    }
  

  /* figure out the size switch */
  var = GetRealVar("MakeStreamsSurface", streamHoseVisObj, SIZEFACTOR);
  if (var)
    {
     sizeFactor = GetReal(var);
    }
  else
    {
      sizeFactor = 1.0;
    }

  var = GetRealVar("MakeStreamsSurface", streamHoseVisObj, SIZEOFFSET);
  if (var)
    {
      sizeOffset = GetReal(var);
    }
  else
    {
      sizeOffset = 0.0;
    }

  sizeObj = NULLOBJ;
  MakeVar(streamHoseVisObj, SIZESWITCH);
  if (GetPredicate(streamHoseVisObj, SIZESWITCH))
    {
      sizeObj = GetVar(streamHoseVisObj, SIZEOBJ);
    }
  if (sizeObj)
    {
       SetCurField(FIELD2, sizeObj);
       SetCurForm(FIELD3, sizeObj);
    }
  else
    {
      var = GetRealVar("MakeStreamsSurface", streamHoseVisObj, SIZECONSTANT);
      if (var)
        {
          sizeConstant = GetReal(var);
        }
      else
        {
          sizeConstant = 0.1;
        }
      size = sizeConstant * sizeFactor + sizeOffset;
    }

  /* get nSubDivisions, now assumes it's 2 */
#ifdef SIGGRAPH
  nSubDivisions = 0;
  curNSides = 4;
  capEnds = false;
#else
  var = GetIntVar("MakeStreamsSurface", streamHoseVisObj, FRUSTUMSUBDIV);
  if (var)
    {
      nSubDivisions = GetInt(var);
    }
  else
    {
      nSubDivisions = 2;
    }
  curNSides = 4;
  while(nSubDivisions--)
    {
      curNSides *= 2;
    }

  /* see if cap is needed by users */
  capEnds = GetPredicate(streamHoseVisObj, CAPENDSP);
#endif


  /* create polys for use */
  polys = AppendPolysToPicture(surface);

  for (n = 0; n < nPaths; n++)
     {
       long i;

       if (NULLOBJ == (thisPath = objElements[n]))
         {
           /* see if there is an real array there */
           continue;
         }
       if (1 != RANK(thisPath))
         {
           return ObjFalse;
         }
       nVertices = DIMS(thisPath)[0] / nSpatComponents;

       if (nVertices < minPoints)
	 {
	   continue;
	 }

       elements = (real *) ELEMENTS(thisPath);
       if (NULL == (vertices = (VertexPtr *)Alloc(nVertices * sizeof(VertexPtr))))
         {
           OMErr();
           return ObjFalse;
         }
       for (i = 0; i < nVertices; i++)
          {
            long index;
       
            if (!(vertices[i] = NewVertex(surface, 0)))
              {
	        Free(vertices);
	        return ObjFalse;
              }

            index = i * nSpatComponents;
            vertices[i]->position[0] = elements[index];
            vertices[i]->position[1] = elements[index + 1];
            vertices[i]->position[2] = (nSpatComponents > 2 ? elements[index + 2] : 0.0);
          }
       {
#define ISZEROVEC(a)  ((ABS((a)[0]) + ABS((a)[1]) + ABS((a)[2])) == 0.0)
         VertexPtr hoseEnd[2][STR_MAXCYLSIDES], v[4];
         float normal[3], sum[3], cross[3], axial[2][3], radial[4][3]; 
	 float denom, nom, unSizedHoseEnd[2][STR_MAXCYLSIDES][3];
	 int m, curVert, nextVert, prevSec, curSec, sidesInc;
	 int beg, end, mid, inc, isZeroVec;
	 float sizeVal;

	 /* Use the vector pt0 - pt1 as surface normals on the first point. */ 
	 curVert = nextVert  = 0;
	 curSec = 0;

	 /* makes sure next vertex is actually with a different coordinate */
         do {
		nextVert++;
	        axial[curSec][0] = vertices[curVert]->position[0] - vertices[nextVert]->position[0];
	        axial[curSec][1] = vertices[curVert]->position[1] - vertices[nextVert]->position[1];
	        axial[curSec][2] = vertices[curVert]->position[2] - vertices[nextVert]->position[2];
	    } while((isZeroVec = ISZEROVEC(axial[curSec])) && nextVert < nVertices - 1);

         if (isZeroVec)
	   {
	     continue;  /* skip to next line */
	   }

	 /* if the axial vector is a 0 vector then we are in trouble when normalizing it. */
         NORMALIZE(axial[curSec]);

	 if (ABS(axial[curSec][0]) >= STR_I_THRESHOLD)
	   {
	       /* cross with base j (0, 1, 0) to get first radial vector */
	       radial[0][0] = -axial[curSec][2]; 
	       radial[0][1] = 0.0;
	       radial[0][2] = axial[curSec][0];
	   }
         else
	   {
	       /* cross with base i (1, 0, 0) to get first radial vector */
	       radial[0][0] = 0.0;
	       radial[0][1] = axial[curSec][2];
	       radial[0][2] = -axial[curSec][1];
	   }
         NORMALIZE(radial[0]);

	   /* cross axial[curSec] with radial[0] to get radial[1] */
	   /* and radial[1] is automatically normalized since axial[curSec] 
	      is perpendicular to radial[0] and both are normalized */
	 CROSS(axial[curSec], radial[0], radial[1]);

	   /* radial[2] and radial[3] are inverse of radial[0] and radial[1] respectively */
         for (i = 0; i < 3; i++)
	    {
	      radial[2][i] = -radial[0][i];
	      radial[3][i] = -radial[1][i];
            }

         sidesInc = curNSides / 4;
         for (m = 0; m < curNSides; m++ )
            {
	      hoseEnd[curSec][m] = InsertVertex(surface, vertices[curVert]);
	    }

	 /* normal of hoses on vertices of curSec */
         for (m = 0; m < 4; m++)
	    {
	      for (i = 0; i < 3; i++)
	         {
	           hoseEnd[curSec][m * sidesInc]->normal[i] = radial[m][i];
	         }
	    }

         for (inc = sidesInc, m = 0; inc > 1; inc /= 2, m++)
	    {
	      for (beg = 0; beg < curNSides; beg += inc)
	         {
	           mid = beg + inc / 2;
                   /* circular clipping, works only curNSides is the power of 2 */
	           end = (beg + inc) & (curNSides - 1);  
                   for (i = 0; i < 3; i++)
	              {
                        hoseEnd[curSec][mid]->normal[i] = cefs_2[m] * 
                         (hoseEnd[curSec][beg]->normal[i] + hoseEnd[curSec][end]->normal[i]);
	              }
	          }
            }

         if (sizeObj)
	   {
	     sizeVal = SampleSpatScalar(FIELD2, FIELD3, 3, vertices[curVert]->position, true);
	     if (sizeVal == missingData)
	       {
	         sizeVal = 1.0;
               }
	     size = sizeVal * sizeFactor + sizeOffset;
	     /* otherwise, a constant size is set up previously */
           }

         /* positions of hoseEnd on vertices of curSec */
         for (m = 0; m < curNSides; m++)
	    {
	      for (i = 0; i < 3; i++)
	         {
	           unSizedHoseEnd[curSec][m][i] = 
	            vertices[curVert]->position[i] + hoseEnd[curSec][m]->normal[i];
                   hoseEnd[curSec][m]->position[i] = 
	            vertices[curVert]->position[i] + size * hoseEnd[curSec][m]->normal[i];
                 }
	    }

	 if (true == capEnds)
	   {
	     VertexPtr cap[STR_MAXCYLSIDES];

	     /* cap vertices */
             for (m = 0; m < curNSides; m++)
	        {
	          cap[m] = InsertVertex(surface, vertices[curVert]);
	          for (i = 0; i < 3; i++)
	             {
	               cap[m]->position[i] = hoseEnd[curSec][m]->position[i];
	               cap[m]->normal[i] = axial[curSec][i];
                     }
	        }

             /* append cap to polys */
	     AppendSPolyToPolys(polys, curNSides, cap);
	   }

	 while(!isZeroVec)
	    {
		curVert = nextVert;
		prevSec = curSec;
		curSec ^= 1;      /* flip between 1 and 0 */

		/* insert some vertices onto the surface */
                for (m = 0; m < curNSides; m++)
		   {
		     hoseEnd[curSec][m] = InsertVertex(surface, vertices[curVert]);
		   }

		/* get axial vector */
		for (i = 0; i < 3; i++)
		   {
		     axial[prevSec][i] = -axial[prevSec][i];  /* inverse previous axial */
		   }
                if (curVert == nVertices - 1)
		  {
		    /* last vertex, trick it to think that isZeroVec is true */
		    isZeroVec = 1;
		  }
                else
		  {
                    do{
	                nextVert++;
	                for (i = 0; i < 3; i++)
	                   {
	                     axial[curSec][i] = 
	                     vertices[curVert]->position[i] - vertices[nextVert]->position[i];
	                   }
	              } while((isZeroVec = ISZEROVEC(axial[curSec])) && nextVert < nVertices - 1);
		  }

                if (isZeroVec)
		  {
		    /* then all the vertices lelf are of the same position as curVert */
		    /* All we have to do is to construct the last section and cap if any */
	            /* get axial vector, inverse the previous one */
	            for (i = 0; i < 3; i++)
	               {
		         /* use inversed previous axial */
	                 normal[i] = axial[curSec][i] = axial[prevSec][i];  
	               }
		  }
                else 
		  {
                    NORMALIZE(axial[curSec]);

		    for (i = 0; i < 3; i++)
		       {
		         sum[i] = axial[prevSec][i] + axial[curSec][i];
		       }
		    CROSS(axial[prevSec], axial[curSec], cross);
    
		    if ( cross[0] == 0.0 && cross[1] == 0.0 && cross[2] == 0.0)
		      {
		        /* the three vertices are colinear */
		        for (i = 0; i < 3; i++)
		           {
			     normal[i] = axial[curSec][i];
		           }
		      }
                    else
		      {
		        CROSS(cross, sum, normal);
		      }
#ifdef DEBUG
		    /* bad thing */
                    if (ISZEROVEC(normal))
		      {
		        fprintf(stderr, "In MakeStreamsSruface bisection normal is 0 vector"); 
			fprintf(stderr, "for stream line number %ld\n", n);
		      }
#endif
		  }

                /* find the parameter t at the intersection of the parametric line function 
		   of the line from sizedHoseEnd[i] along axial[prevSec] (point + t * axial)
		   and the bisection surface of curSec */
                /* t = sum(nk * (mk - pk))/sum(nk * ak), where nk is normal vector,
		   mk is the current vertex and pk is the point on the previous surface, and
		   ak is the axial vector of prevSec */
                for (m = 0; m < curNSides; m++)
		   {
		     denom = nom = 0.0;
		     for (i = 0; i < 3; i++)
			{
	                  nom += normal[i] * (vertices[curVert]->position[i] - 
					      unSizedHoseEnd[prevSec][m][i]);
                          denom += normal[i] * axial[prevSec][i];
			}
		     for (i = 0; i < 3; i++)
			{
	                  unSizedHoseEnd[curSec][m][i] = unSizedHoseEnd[prevSec][m][i] + 
			       nom * (axial[prevSec][i] / denom);
			  /* this is the offset from center to vertices */ 
		          hoseEnd[curSec][m]->normal[i] = 
			   (unSizedHoseEnd[curSec][m][i] - vertices[curVert]->position[i]);
                        }
#ifdef DEBUG
		     if ( hoseEnd[curSec][m]->normal[0] == 0.0 && hoseEnd[curSec][m]->normal[1]  == 0.0 && hoseEnd[curSec][m]->normal[2] == 0.0 )
		       {
                         fprintf(stderr, "Error, The octgonal side normal is 0 vector\n");
		       }
                     if (denom == 0.0 || nom >= FLT_MAX || denom >= FLT_MAX)
		       {
		         fprintf(stderr, "nom is %g and denom is %g\n", nom, denom);
			 fprintf(stderr, "current axial is <%g, %g, %g>\n", 
			 axial[curSec][0], axial[curSec][1], axial[curSec][2]);
		       }
#endif
		   }

                if (sizeObj)
	          {
	            sizeVal = SampleSpatScalar(FIELD2, FIELD3, 3, 
				  vertices[curVert]->position, true);
	            if (sizeVal == missingData)
	              {
		        sizeVal = 1.0;
                      }
		    size = sizeVal * sizeFactor + sizeOffset;
                  }

	        /* scale according to the size  */
	        for (m = 0; m < curNSides; m++)
	           {
                     for (i = 0; i < 3; i++)
		        {
	                  hoseEnd[curSec][m]->position[i] = vertices[curVert]->position[i]
	                    + size * hoseEnd[curSec][m]->normal[i];
                        }
		     /* normalize the normal after the scaling */
                     NORMALIZE(hoseEnd[curSec][m]->normal);
	           }

		/* appends spolys to polys */
		for (m = 0; m < curNSides - 1; m++)
		   {
		     v[0] = hoseEnd[prevSec][m];
		     v[1] = hoseEnd[curSec][m];
		     v[2] = hoseEnd[curSec][m + 1];
		     v[3] = hoseEnd[prevSec][m + 1];
		     AppendSPolyToPolys(polys, 4, v);
		   }
                v[0] = hoseEnd[prevSec][curNSides - 1];
		v[1] = hoseEnd[curSec][curNSides - 1];
		v[2] = hoseEnd[curSec][0];
		v[3] = hoseEnd[prevSec][0];
		AppendSPolyToPolys(polys, 4, v);

		/* for last section, appends a cap if wanted */
	        if (isZeroVec && true == capEnds)
	          {
	            VertexPtr cap[STR_MAXCYLSIDES];
     
	            /* cap vertices */
                    for (m = 0; m < curNSides; m++)
	               {
		         cap[m] = InsertVertex(surface, vertices[curVert]);
	                 for (i = 0; i < 3; i++)
	                    {
	                      cap[m]->position[i] = hoseEnd[curSec][m]->position[i];
		              cap[m]->normal[i] = axial[curSec][i];
                            }
	               }

                    /* append cap to polys */
	            AppendSPolyToPolys(polys, curNSides, cap);
	          }
	    } /* while loop */
       }
       Free(vertices);
     }

  /* set the pictures's MAINDATASET for facilitating ColorPictureByObject */
  SetVar(surface, MAINDATASET, streamData);
  SetVar(streamHoseVisObj, SURFACE, surface);
  SetVar(surface, REPOBJ, streamHoseVisObj);
  return ObjTrue;
}

static ObjPtr MakeStreamHoseSized(streamVisObj)
ObjPtr streamVisObj;
{
#ifdef SIGGRAPH
  SetVar(streamVisObj, SIZESWITCH, NewInt(0));
  SetVar(streamVisObj, SIZECONSTANT, NewReal(1.0));
  SetVar(streamVisObj, SIZEFACTOR, NewReal(75.0));
  SetVar(streamVisObj, SIZEOFFSET, NewReal(0.0));
#else
  ObjPtr sizeObj;
  ObjPtr mainDataset;
  ObjPtr var;

  /* Traverse the class hierarchy of the visualization object to get 
     it's real main dataset object. */ 
  mainDataset = GetVar(streamVisObj, MAINDATASET);
  while (var = GetVar(mainDataset, MAINDATASET))
    {
      mainDataset = var;
    }

  /* See if it has any sizeObj attatched to it. */
  /* Now, get the SIZEOBJ of the main dataset. */
  sizeObj = GetVar(mainDataset, SIZEOBJ);

  if (sizeObj)
    {
      SetVar(streamVisObj, SIZEOBJ, sizeObj);
    }
  else
    {
      SetVar(streamVisObj, SIZEOBJ, mainDataset);
    }
  SetVar(streamVisObj, SIZESWITCH, NewInt(1));
#endif
  return ObjTrue;
}

#ifdef HAVE_PROTOTYPES
void InitStreamhoses(void)
#else
void InitStreamhoses()
#endif
{
    ObjPtr icon;

    visStreamHoses = NewObject(visHoseEmitters, 0);
    AddToReferenceList(visStreamHoses);
    DefineFixedClass(visStreamHoses, CLASS_STREAMLINES, "visStreamhoses");

    /* A. Variables setup: essential variables for all new vis classes */
    /* 1. a NAME */
    SetVar(visStreamHoses, NAME, NewString("Streamhoses"));

    /* 2. a DEFAULTICON */
    SetVar(visStreamHoses, DEFAULTICON, icon = NewObject(visIcon, 0));
    SetVar(icon, WHICHICON, NewInt(ICONQUESTION));
    SetVar(icon, NAME, NewString("Streamhoses"));
    SetVar(icon, HELPSTRING,
	NewString("This icon represents a stream visualization object.  \
This object is used to display streamhoses."));

    /* 3. a CONTROLICON: click on this to show control panel */
    icon = NewIcon(0, 0, ICONQUESTION, "Streamhoses");
    SetVar(icon, HELPSTRING,
	NewString("Click on this icon to see a panel of controls for the streamhoses object."));
    SetVar(visStreamHoses, CONTROLICON, icon);
    SetVar(icon, PANELHELP, NewString("The streamhoses visualization object is intended \
to show visualization objects of stream hoses."));


    /* 3.5 add snap shot variables */
    AddSnapVar(visStreamHoses, INTEGRATIONMETHOD);
    AddSnapVar(visStreamHoses, SAMPLEPARAMETER);
    AddSnapVar(visStreamHoses, STEPLENGTHFACTOR);
    AddSnapVar(visStreamHoses, MINSTREAMPOINTS);
    AddSnapVar(visStreamHoses, MARKBUCKET);
    AddSnapVar(visStreamHoses, STREAMPATH);


    /* 4. variables specific to the new vis class: define same default values for vis
          objects of this class */
    SetVar(visStreamHoses, INTEGRATIONMETHOD, NewInt(EULER)); 
    SetVar(visStreamHoses, SAMPLEPARAMETER, NewReal(DEFAULTSAMPLEPARAMETER));
    SetVar(visStreamHoses, STEPLENGTHFACTOR, NewReal(DEFAULTSTEPLENGTHFACTOR));
    SetVar(visStreamHoses, MINSTREAMPOINTS, NewInt(DEFAULTMINSTREAMPOINTS));
#ifdef SIGGRAPH
    SetVar(visStreamHoses, MARKBUCKET, NewInt(0));
#else
    SetVar(visStreamHoses, MARKBUCKET, NewInt(1));
#endif

    /* B. Dependecy set up */
    DeclareDependency(visStreamHoses, SURFACE, STREAMPATH);
    DeclareDependency(visStreamHoses, SURFACE, MINSTREAMPOINTS);
    DeclareDependency(visStreamHoses, STREAMPATH, INTEGRATIONMETHOD);
    DeclareDependency(visStreamHoses, STREAMPATH, SAMPLEPARAMETER);
    DeclareDependency(visStreamHoses, STREAMPATH, STEPLENGTHFACTOR);
    DeclareDependency(visStreamHoses, STREAMPATH, MARKBUCKET);
    DeclareDependency(visStreamHoses, STREAMPATH, EMITTERCONTROL);
    DeclareDependency(visStreamHoses, STREAMPATH, EMITTERBUCKETS);
    DeclareDependency(visStreamHoses, STREAMPATH, EMITTERDATASET);
    DeclareDependency(visStreamHoses, STREAMPATH, EMITTERSET);
    DeclareIndirectDependency(visStreamHoses, STREAMPATH, MAINDATASET, CHANGED);
    DeclareIndirectDependency(visStreamHoses, STREAMPATH, MAINDATASET, DATA);

    /* C. Methods set up */
    SetMethod(visStreamHoses, STREAMPATH, MakeStreamPath);
    SetMethod(visStreamHoses, ADDCONTROLS, AddStreamsControls);
    SetMethod(visStreamHoses, SURFACE, MakeStreamHosesSurface);
    SetMethod(visStreamHoses, SIZESWITCH, MakeStreamHoseSized);

    /* D. Vismapping set up */
    DefineVisMapping(DS_HASFORM | DS_HASFIELD | DS_VECTOR, -1, 3, 3, visStreamHoses);
    DefineVisMapping(DS_HASFORM | DS_HASFIELD | DS_VECTOR, -1, 2, 3, visStreamHoses);
    DefineVisMapping(DS_HASFORM | DS_HASFIELD | DS_VECTOR, -1, 3, 2, visStreamHoses);
    DefineVisMapping(DS_HASFORM | DS_HASFIELD | DS_VECTOR, -1, 2, 2, visStreamHoses);
}

#ifdef HAVE_PROTOTYPES
void KillStreamhoses(void)
#else
void KillStreamhoses()
#endif
/*Kills streams visualization*/
{
    DeleteThing(visStreamHoses);
}

/*****************************************************************************************/
/**************************** visLine(Hose)Emitters Class Begin **************************/
/*****************************************************************************************/

#define DEFAULTBUCKETDIM  3
#define MINBUCKETDIM      1
#define MAXBUCKETDIM      plusInf

/* DROPINCONTENTS method for emitter corral */
static ObjPtr DropInEmitterCorral(corral, object, x, y)
ObjPtr corral, object;
int x, y;
{
 ObjPtr emitterVisObj;
 ObjPtr fieldObj;
 ObjPtr icon;
 ObjPtr name;
 ObjPtr defaultIcon;
 ObjPtr contents;
 long info;

 /*Find the visualization object*/
 emitterVisObj = GetObjectVar("DropInEmitterCorral", corral, REPOBJ);
 if (!emitterVisObj)
    {
	return ObjFalse;
    }

    /*Get the field object*/
    fieldObj = GetObjectVar("DropInEmitterCorral", object, REPOBJ);
    if (!fieldObj)
    {
	return ObjFalse;
    }

    if (!IsDataset(fieldObj))
    {
	fieldObj = GetVar(fieldObj, MAINDATASET);
    }

    if (!IsDataset(fieldObj))
    {
	WarnUser(CW_CANNOTDROPNONDATASET);
	return ObjFalse;
    }

    /* must use only non geometry dataset */
    info = GetDatasetInfo(fieldObj);
    if ((info & DS_HASGEOMETRY) && (fieldObj != GetVar(emitterVisObj, MAINDATASET)))
    {
	WarnUser(CW_NOGEOMETRYERROR);
	return ObjFalse;
    }
    /* needs more tests here, has form, ... ?*/
    if (!(info & DS_HASFORM))
      {
	fprintf(stderr, "Dataset must have form.\n");
	return ObjFalse;
      }

    /*Create an icon for it*/
    MakeVar(fieldObj, NAME);
    name = GetStringVar("DropInEmitterCorral", fieldObj, NAME);
    if (!name)
    {
	return ObjFalse;
    }

    MakeVar(fieldObj, DEFAULTICON);
    defaultIcon = GetVar(fieldObj, DEFAULTICON);
    if (defaultIcon)
    {
	ObjPtr locArray;
	real loc[2];
	icon = NewObject(defaultIcon, 0);
	SetVar(icon, NAME, name);
	loc[0] = x;
	loc[1] = y;
	locArray = NewRealArray(1, 2L);
	CArray2Array(locArray, loc);
	SetVar(icon, ICONLOC, locArray);
    }
    else
    {
	icon = NewIcon(x, y, ICONQUESTION, GetString(name));
    }
    
    /*Make the icon point to the field*/
    SetVar(icon, REPOBJ, fieldObj);

    /*Make it the only icon in the corral*/
    contents = NewList();
    PrefixList(contents, icon);
    SetVar(corral, CONTENTS, contents);
    SetVar(contents, PARENT, corral);
    SetVar(icon, PARENT, corral);
    RecalcScroll(corral);

    ImInvalid(corral);

    /* Make this the EMITTERDATASET obj */
    SetVar(emitterVisObj, EMITTERDATASET, fieldObj);

    return ObjTrue;
}

/* EMITTERBUCKETS making method  */
static ObjPtr MakeEmitterBuckets(emitterVisObj)
ObjPtr emitterVisObj;
{
  ObjPtr emitterData, var, emitterBuckets;
  int nComponents, whichComponent;
  real bounds[6], *elements;
  long nBuckets, whichBucket, bucketIndices[3], index;
  NameTyp whichName;
  long bucketDims[3];
  real bucketBounds[6], divisionUnits[3];

  fprintf(stderr, "Making new emitter buckets.\n");

  /* get the MAINDATASET of the emitterVisObj */
  if (!(emitterData = GetVar(emitterVisObj, MAINDATASET)))
    {
      return ObjFalse;
    }

  /* get the rank of the form of the MAINDATASET */
  if (false == SetCurForm(FIELD3, emitterData))
    {
      return ObjFalse;
    }
  nComponents = GetNComponents(FIELD3);

  /* get bounds of the form of the MAINDATASET */
  if (NULLOBJ == (var = GetDatasetFormBounds(emitterData)))
    {
      return ObjFalse;
    }
  elements = ELEMENTS(var);
  for (whichComponent = 0; whichComponent < nComponents; whichComponent++)
     {
       bounds[2 * whichComponent] = elements[2 * whichComponent];
       bounds[2 * whichComponent + 1] = elements[2 * whichComponent + 1];
     }

  /* get XYZ BUCKETDIM, LOWERBOUND, UPPERBOUND */
  for (whichComponent = 0; whichComponent < nComponents; whichComponent++)
     {
       whichName = whichComponent == 2 ? ZLOWERBOUND: (whichComponent ? YLOWERBOUND: XLOWERBOUND);
       MakeVar(emitterVisObj, whichName); 
       var = GetVar(emitterVisObj, whichName); 
       if (var)
         {
           bucketBounds[2 * whichComponent] = GetReal(var);
         }
       else
         {
           bucketBounds[2 * whichComponent] = bounds[2 * whichComponent];
	   SetVar(emitterVisObj, whichName, NewReal(bounds[2 * whichComponent]));
         }
	   
       whichName = whichComponent == 2 ? ZUPPERBOUND: (whichComponent ? YUPPERBOUND: XUPPERBOUND);
       MakeVar(emitterVisObj, whichName); 
       var = GetVar(emitterVisObj, whichName); 
       if (var)
         {
           bucketBounds[2 * whichComponent + 1] = GetReal(var);
         }
       else
         {
           bucketBounds[2 * whichComponent + 1] = bounds[2 * whichComponent + 1];
	   SetVar(emitterVisObj, whichName, NewReal(bounds[2 * whichComponent + 1]));
         }
	   
       if (bucketBounds[2 * whichComponent] == bucketBounds[2 * whichComponent + 1])
	 {
	   bucketDims[whichComponent] = 1;
	   divisionUnits[whichComponent] = 0; 
	 }
       else
         {
           whichName = whichComponent == 2 ? ZBUCKETDIM : 
			      (whichComponent ? YBUCKETDIM : XBUCKETDIM);
           MakeVar(emitterVisObj, whichName); 
           var = GetVar(emitterVisObj, whichName); 
           bucketDims[whichComponent] = GetInt(var);

           if (bucketBounds[2 * whichComponent] > bucketBounds[2 * whichComponent + 1])
             {
	       /* reverse order from big to small */
	       divisionUnits[whichComponent] = 
	        ((bounds[2 * whichComponent + 1] - bounds[2 * whichComponent]) - 
		 (bucketBounds[2 * whichComponent] - bucketBounds[2 * whichComponent + 1])) / 
		 bucketDims[whichComponent];
	     }
           else
	     {
	       divisionUnits[whichComponent] = 
	        ((bucketBounds[2 * whichComponent + 1] - bucketBounds[2 * whichComponent])) / 
	        bucketDims[whichComponent];
	     }
         }
     }

  /* look for the bucket's position and set it as the contents of the EMITTERBUCKET var */
  /* use an array of size (nComponents + 1)*nBuckets to represent the bucket position and marked */
  nBuckets = 1;
  for (whichComponent = 0; whichComponent < nComponents; whichComponent++)
     {
       nBuckets *= bucketDims[whichComponent];
     }
  if (NULLOBJ == (emitterBuckets = NewRealArray(1, (nComponents + 1) * nBuckets)))
    {
      return ObjFalse;
    }

  elements = ELEMENTS(emitterBuckets);
  /* fill in the emitterBuckets */
  bucketIndices[0] = bucketIndices[1] = bucketIndices[2] = 0;
  for(whichBucket = 0; whichBucket < nBuckets; whichBucket++)
     {
       index = (nComponents + 1) * whichBucket;
       for (whichComponent = 0; whichComponent < nComponents; whichComponent++) 
	  {
            elements[index + whichComponent] = bucketBounds[2 * whichComponent] + 
	      (bucketIndices[whichComponent] + 0.5) * divisionUnits[whichComponent]; 
            if (elements[index + whichComponent] > bounds[2 * whichComponent + 1])
	      {
                elements[index + whichComponent] -= bounds[2 * whichComponent + 1];
	      }
	  }
       /* uncheck the marked flag */
       elements[index + whichComponent] = 1.0;

       /* advance the bucket indices */
       for (whichComponent = 0; whichComponent < nComponents; whichComponent++)
	  {
	    if (++bucketIndices[whichComponent] >= bucketDims[whichComponent])
	      {
		bucketIndices[whichComponent] = 0;
	      }
            else
	      {
		break;
	      }
	  }
     }

  SetVar(emitterVisObj, EMITTERBUCKETS, emitterBuckets);

  return ObjTrue;
}

#ifdef HAVE_PROTOTYPES
ObjPtr NewEmitter(real *vector)
#else
ObjPtr NewEmitter(vector)
real *vector;
#endif
{
  ObjPtr retVal;
  real *elements;

  if (NULLOBJ == (retVal = NewRealArray(1, 3L)))
    {
      fprintf(stderr, "Error in NewEmitter, unable to get a new real array.\n");
      return NULLOBJ;
    }
  elements = ELEMENTS(retVal);
  elements[0] = vector[0];
  elements[1] = vector[1];
  elements[2] = vector[2];

  return retVal;
}

/* ADDEMITTERS method  */
static ObjPtr AddEmitters(emitterVisObj, emitter)
ObjPtr emitterVisObj;
ObjPtr emitter;
{
  ObjPtr emitterSet, oldEmitterSet;
  ObjPtr *emitterSetElements, *oldEmitterSetElements;
  long nEmitters, oldNEmitters;
  long i;

  fprintf(stderr, "Adding new emitters.\n");

  if (NULLOBJ != (oldEmitterSet = GetVar(emitterVisObj, EMITTERSET)))
    {
      oldNEmitters = DIMS(oldEmitterSet)[0];
      oldEmitterSetElements = ELEMENTS(oldEmitterSet);
    }
  else
    {
      oldNEmitters = 0;
    }

  nEmitters = oldNEmitters + 1;
  if (NULLOBJ == (emitterSet = NewArray(AT_OBJECT, 1, &nEmitters)))
    {
      return ObjFalse;
    }

  emitterSetElements = ELEMENTS(emitterSet);
  for (i = 0 ; i < oldNEmitters; i++)
     {
       emitterSetElements[i] = oldEmitterSetElements[i];
     }
  emitterSetElements[i] = emitter;

  SetVar(emitterVisObj, EMITTERSET, emitterSet);
  ImInvalid(emitterVisObj);

  return ObjTrue;
}

/* DELETEEMITTERS method  */
static ObjPtr DeleteEmitters(emitterVisObj)
ObjPtr emitterVisObj;
{
  fprintf(stderr, "Deleting emitters.\n");

  if (NULLOBJ != GetVar(emitterVisObj, EMITTERSET))
    {
      SetVar(emitterVisObj, EMITTERSET, NULLOBJ);
      ImInvalid(emitterVisObj);
    }
  return ObjTrue;
}

/* method of Add An Emitter Button */
static ObjPtr AddAnEmitter(button)
ObjPtr button;
{
  ObjPtr emitterVisObj;
  ObjPtr emitterX, emitterY, emitterZ, emitter;
  real vector[3];
  FuncTyp method;

  emitterVisObj = GetObjectVar("AddAnEmitter", button, REPOBJ);
  vector[2] = 0.0;
  if (NULLOBJ == (emitterX = GetVar(emitterVisObj, EMITTERX)))
    {
      /* no emitter control set up yet */
      return ObjTrue;
    }

  vector[0] = GetReal(emitterX);
  vector[1] = GetReal(GetRealVar("AddAnEmitter", emitterVisObj, EMITTERY));
  if (NULLOBJ != (emitterZ = GetVar(emitterVisObj, EMITTERZ)))
    {
      vector[2] = GetReal(emitterZ);
    }

  if (NULLOBJ != (emitter = NewEmitter(vector)))
    {
      method = GetMethodSurely("AddAnEmitter", emitterVisObj, ADDEMITTERS);
      if (method)
	{
          (*method)(emitterVisObj, emitter);
	}
      return ObjTrue;
    }
  else 
    {
      return ObjFalse;
    }
}

/* method of Remove All Emitters Button */
static ObjPtr DeleteAllEmitters(button)
ObjPtr button;
{
  ObjPtr emitterVisObj;
  FuncTyp method;

  emitterVisObj = GetObjectVar("DeleteAllEmitters", button, REPOBJ);
  method = GetMethodSurely("DeleteAllEmitters", emitterVisObj, DELETEEMITTERS);
  if (method)
    {
      (*method)(emitterVisObj);
    }
  return ObjTrue;
}

/* Emitter control for visLineEmitters and visHoseEmitters */
static ObjPtr AddEmittersControls(emitterVisObj, panelContents)
ObjPtr emitterVisObj, panelContents;
{
  ObjPtr emitterData, mainDataset;
  int left, top, bottom, right, radioLeft, radioRight;
  int width, singleWidth;
  int nComponents, whichComponent;
  NameTyp whichName;
  ObjPtr var, titleBox, checkBox, radio, button, textBox, corral;
  real *elements;
  float bounds[6];
  Bool deleteEmitters;

  /* get the MAINDATASET object var */
  if (NULLOBJ == (emitterData = GetObjectVar("AddEmittersControls", emitterVisObj, MAINDATASET)) )
    {
      return ObjFalse;
    }
  while (NULLOBJ != (mainDataset = GetVar(emitterData, MAINDATASET)))
    {
      emitterData = mainDataset;
    }

  if (false == SetCurForm(FIELD2, emitterData))
    {
      return ObjFalse;
    }
  nComponents = GetNComponents(FIELD2);

  /* Get dataform bounds, for use with bounds of buckets */
  if (NULLOBJ == (var = GetDatasetFormBounds(emitterData)))
    {
      return ObjFalse;
    }
  elements = ELEMENTS(var);
  for (whichComponent = 0; whichComponent < 2 * nComponents; whichComponent++)
     {
       bounds[whichComponent] = (float) elements[whichComponent];
     }

  /* add EMITTERCONTROL radio */
  radio = NewRadioButtonGroup("Emitter Control Group");
  PrefixList(panelContents, radio);
  SetVar(radio, PARENT, panelContents);
  SetVar(radio, HELPSTRING, 
	NewString("These radio buttons control the way the emitters are \
specified.  Either buckets, datasets or Wand devices is to be used."));

  /* Add an USEINTERACTIVE radio button */
  right = CWINWIDTH - CORRALBORDER - CWINCORRALWIDTH - MAJORBORDER;
  left = right - PICSTICKBOXWID;
  top = CWINHEIGHT - MAJORBORDER;
  bottom = top - CHECKBOXHEIGHT;
  button = NewRadioButton(left, (left + right) / 2, bottom, top, "Interactive");
  AddRadioButton(radio, button);

#ifdef HWUTEST
  width = right - left - 2 * MINORBORDER;
  singleWidth = width / 3;
  top = bottom - MINORBORDER;
  bottom = top - EDITBOXHEIGHT;
  for (whichComponent = 0; whichComponent < nComponents; ++whichComponent)
     {
       right = left +  singleWidth;
       /* lower bound text boxes */
       whichName = whichComponent == 2 ? EMITTERZ: (whichComponent ? EMITTERY : EMITTERX);
       MakeVar(emitterVisObj, whichName);
       var = GetVar(emitterVisObj, whichName); 
       if (!var)
         {
           SetVar(emitterVisObj, whichName, 
		   NewReal((bounds[2 * whichComponent] + bounds[2 * whichComponent + 1]) / 2.0));
         }
       textBox = NewTextBox(left, right, bottom, top, EDITABLE + WITH_PIT + ONE_LINE, 
	         whichComponent == 2 ? "Emitter Z" : 
	         (whichComponent ? "Emitter Y" : "Emitter X"), ""); 
       PrefixList(panelContents, textBox);
       SetVar(textBox, PARENT, panelContents);
       SetTextAlign(textBox, RIGHTALIGN);
       SetVar(textBox, HELPSTRING, NewString("This text box contains the lower bound of the range for use with bucket divisions of the indicated axis."));
       AssocTextRealControlWithVar(textBox, emitterVisObj, whichName, 
	  bounds[2 * whichComponent], bounds[2 * whichComponent + 1], 0);

       left = right + MINORBORDER;
     }

  right = CWINWIDTH - CORRALBORDER - CWINCORRALWIDTH - MAJORBORDER;
  left = right - PICSTICKBOXWID;
  top = bottom - MINORBORDER;
  bottom = top - CHECKBOXHEIGHT;
  button = NewButton(left, right, bottom, top, "Add An Emitters");
  PrefixList(panelContents, button);
  SetVar(button, PARENT, panelContents);
  SetVar(button, REPOBJ, emitterVisObj);
  SetMethod(button, CHANGEDVALUE, AddAnEmitter);
  SetVar(button, HELPSTRING, NewString(
        "Pressing this button adds an emitter to the current set of emitters."));

  top = bottom - MINORBORDER;
  bottom = top - CHECKBOXHEIGHT;
  button = NewButton(left, right, bottom, top, "Remove All Emitters");
  PrefixList(panelContents, button);
  SetVar(button, PARENT, panelContents);
  SetVar(button, REPOBJ, emitterVisObj);
  SetMethod(button, CHANGEDVALUE, DeleteAllEmitters);
  SetVar(button, HELPSTRING, NewString(
        "Pressing this button removes all current emitters."));
#endif

  /* Add a Use Dataset radio button */
  radioLeft = MAJORBORDER;
  radioRight = radioLeft + PICSTICKBOXWID;
  top = CWINHEIGHT - MAJORBORDER;
  bottom = top - CHECKBOXHEIGHT;
  button = NewRadioButton(radioLeft, (radioLeft + radioRight) / 2, bottom, top, "Use Dataset");
  AddRadioButton(radio, button);

  /* The icon corral contains a dataset icon whose REPOBJ is the dataset used to 
     generate the emitter emitters. */
  left = radioLeft;
  right = left + ONECORRALWIDTH;
  top = bottom - MINORBORDER;
  bottom = top - ONECORRALHEIGHT;

  corral = NewIconCorral(NULLOBJ, left, right, bottom, top, 0); 
  SetVar(corral, SINGLECORRAL, ObjTrue);
  SetVar(corral, TOPDOWN, ObjTrue);
  SetVar(corral, NAME, NewString("Emitter Field"));
  SetVar(corral, HELPSTRING, NewString("This corral shows the dataset whose dataform is \
being used to make the emitters.  To replace it with another dataset, \
drag the icon of the other dataset into this corral."));
  PrefixList(panelContents, corral);
  SetVar(corral, PARENT, panelContents);
  SetVar(corral, REPOBJ, emitterVisObj);
  SetMethod(corral, DROPINCONTENTS, DropInEmitterCorral);

  /* Create a text box for the corral */
  top = bottom - TEXTBOXSEP;
  bottom = top - TEXTBOXHEIGHT;
  textBox = NewTextBox(left, right, bottom, top, 0, "Emitter Field Text", "Emitter Field");
  PrefixList(panelContents, textBox);
  SetVar(textBox, PARENT, panelContents);
  SetTextAlign(textBox, CENTERALIGN);

  /* Put a dataset representation icon into the corral if there's a EMITTERDATASET obj */
  if (var = GetVar(emitterVisObj, EMITTERDATASET))
    {
      ObjPtr name, defaultIcon, icon;

      MakeVar(var, NAME);
      name = GetVar(var, NAME);
      MakeVar(var, DEFAULTICON);
      defaultIcon = GetVar(var, DEFAULTICON);
      if (defaultIcon)
        {
          icon = NewObject(defaultIcon, 0);
          SetVar(icon, NAME, name);
        }
      else
        {
          icon = NewIcon(0, 0, ICONQUESTION, GetString(name));
        }
      SetVar(icon, ICONLOC, NULLOBJ);
      SetVar(icon, REPOBJ, var);
      DropIconInCorral(corral, icon);
    }

  /* Make a use Bucket radio button */
  top = bottom - MINORBORDER;
  bottom = top - CHECKBOXHEIGHT;
  button = NewRadioButton(radioLeft, (radioLeft + radioRight) / 2, bottom, top, "Use Bucket");
  AddRadioButton(radio, button);

  /* title box for Bucket Dimensions */
  top = bottom - MINORBORDER;
  bottom = top - 3 * EDITBOXHEIGHT - 2 * CHECKBOXSPACING  - 2 * MINORBORDER - TITLEBOXTOP;
  /*
  left = radioLeft + MINORBORDER;
  */
  left = MAJORBORDER;
  right = CWINWIDTH - CORRALBORDER - CWINCORRALWIDTH - MAJORBORDER;
  titleBox = NewTitleBox(left, right, bottom, top,  "Bucket Dimensions");
  PrefixList(panelContents, titleBox);
  SetVar(titleBox, PARENT, panelContents);

  left += MINORBORDER;
  right -= MINORBORDER;

  textBox = NewTextBox(left, left + SBFUNCTIONNAMEWIDTH, 
		       top - MINORBORDER - TITLEBOXTOP - EDITBOXHEIGHT + 2 * EDITBOXDOWN, 
		       top - MINORBORDER - TITLEBOXTOP + 2 * EDITBOXDOWN, 
		       0, "Minimum Text", "Minimum:");
  PrefixList(panelContents, textBox);
  SetVar(textBox, PARENT, panelContents);
  SetTextAlign(textBox, LEFTALIGN);

  textBox = NewTextBox(left, left + SBFUNCTIONNAMEWIDTH, 
       top - MINORBORDER - TITLEBOXTOP - 2 * EDITBOXHEIGHT - CHECKBOXSPACING + 2 * EDITBOXDOWN, 
       top - MINORBORDER - TITLEBOXTOP - EDITBOXHEIGHT - CHECKBOXSPACING + 2 * EDITBOXDOWN, 
		       0, "Maximum Text", "Maximum:");
  PrefixList(panelContents, textBox);
  SetVar(textBox, PARENT, panelContents);
  SetTextAlign(textBox, LEFTALIGN);

  textBox = NewTextBox(left, left + SBFUNCTIONNAMEWIDTH, 
       top - MINORBORDER - TITLEBOXTOP - 3 * EDITBOXHEIGHT - 2 * CHECKBOXSPACING + 2 * EDITBOXDOWN, 
       top - MINORBORDER - TITLEBOXTOP - 2 * EDITBOXHEIGHT - 2 * CHECKBOXSPACING + 2 * EDITBOXDOWN, 
		       0, "Divisions Text", "Divisions:");
  PrefixList(panelContents, textBox);
  SetVar(textBox, PARENT, panelContents);
  SetTextAlign(textBox, LEFTALIGN);

  left += SBFUNCTIONNAMEWIDTH + MINORBORDER;
  width = right - left - 2 * MINORBORDER;
  singleWidth = width / 3;

  for (whichComponent = 0; whichComponent < nComponents; ++whichComponent)
     {
       int tempTop, tempBottom;

       right = left +  singleWidth;

       tempTop = top - MINORBORDER - TITLEBOXTOP;
       tempBottom = tempTop - EDITBOXHEIGHT;
       /* lower bound text boxes */
       whichName = whichComponent == 2 ? ZLOWERBOUND: (whichComponent ? YLOWERBOUND : XLOWERBOUND);
       MakeVar(emitterVisObj, whichName);
       var = GetVar(emitterVisObj, whichName); 
       if (!var)
         {
           SetVar(emitterVisObj, whichName, NewReal(bounds[2 * whichComponent]));
         }
       textBox = NewTextBox(left, right, tempBottom, tempTop, EDITABLE + WITH_PIT + ONE_LINE, 
	         whichComponent == 2 ? "Z Lower Bound" : 
	         (whichComponent ? "Y Lower Bound" : "X Lower Bound"), ""); 
       PrefixList(panelContents, textBox);
       SetVar(textBox, PARENT, panelContents);
       SetTextAlign(textBox, RIGHTALIGN);
       SetVar(textBox, HELPSTRING, NewString("This text box contains the lower bound of the range for use with bucket divisions of the indicated axis."));
       AssocTextRealControlWithVar(textBox, emitterVisObj, whichName, 
	  bounds[2 * whichComponent], bounds[2 * whichComponent + 1], 0);

       tempTop = tempBottom - CHECKBOXSPACING;
       tempBottom = tempTop - EDITBOXHEIGHT;
       /* upper bound text boxes */
       whichName = whichComponent == 2 ? ZUPPERBOUND : (whichComponent ? YUPPERBOUND : XUPPERBOUND);
       MakeVar(emitterVisObj, whichName);
       var = GetVar(emitterVisObj, whichName); 
       if (!var)
         {
           SetVar(emitterVisObj, whichName, NewReal(bounds[2 * whichComponent + 1]));
         }
       textBox = NewTextBox(left, right, tempBottom, tempTop, EDITABLE + WITH_PIT + ONE_LINE, 
	         whichComponent == 2 ? "Z Upper Bound" : 
	         (whichComponent ? "Y Upper Bound" : "X Upper Bound"), ""); 
       PrefixList(panelContents, textBox);
       SetVar(textBox, PARENT, panelContents);
       SetTextAlign(textBox, RIGHTALIGN);
       SetVar(textBox, HELPSTRING, NewString("This text box contains the upper bound of the range for use with bucket divisions of the indicated axis."));
       AssocTextRealControlWithVar(textBox, emitterVisObj, whichName, 
	  bounds[2 * whichComponent], bounds[2 * whichComponent + 1], 0);

       /* bucket dimension divisions */
       tempTop = tempBottom - CHECKBOXSPACING;
       tempBottom = tempTop - EDITBOXHEIGHT;
       whichName = whichComponent == 2 ? ZBUCKETDIM : (whichComponent ? YBUCKETDIM : XBUCKETDIM);
       MakeVar(emitterVisObj, whichName);
       var = GetVar(emitterVisObj, whichName); 
       if (!var)
         {
           SetVar(emitterVisObj, whichName, NewInt(DEFAULTBUCKETDIM));
         }
       textBox = NewTextBox(left, right, tempBottom, tempTop, EDITABLE + WITH_PIT + ONE_LINE, 
	         whichComponent == 2 ? "Z Bucket Dims" : 
	         (whichComponent ? "Y Bucket Dims" : "X Bucket Dims"), ""); 
       PrefixList(panelContents, textBox);
       SetVar(textBox, PARENT, panelContents);
       SetTextAlign(textBox, RIGHTALIGN);
       SetVar(textBox, HELPSTRING, NewString("This text box contains the bucket dimensions of the indicated axis."));
       AssocTextIntControlWithVar(textBox, emitterVisObj, whichName, 
			  MINBUCKETDIM, MAXBUCKETDIM, TR_NE_TOP | TR_NE_BOTTOM);
       
       left = right + MINORBORDER;
     }

  /* figure out the emitters control way */
  var = GetVar(emitterVisObj, EMITTERCONTROL);
  if (!var)
    {
      SetVar(emitterVisObj, EMITTERCONTROL, NewInt(USEBUCKET));
    }
  AssocDirectControlWithVar(radio, emitterVisObj, EMITTERCONTROL);

  return ObjTrue;
}

/* Initializes visHoseEmitters visualization class */
#ifdef HAVE_PROTOTYPES
void InitHoseEmitters(void)
#else
void InitHoseEmitters()
#endif
{
    ObjPtr icon;

    visHoseEmitters = NewObject(visSized, 0);
    AddToReferenceList(visHoseEmitters);

    SetVar(visHoseEmitters, NAME, NewString("visHoseEmitters"));

    SetVar(visHoseEmitters, DEFAULTICON, icon = NewObject(visIcon, 0));
    SetVar(icon, WHICHICON, NewInt(ICONQUESTION));
    SetVar(icon, NAME, NewString("Emitters"));
    SetVar(icon, HELPSTRING,
	NewString("This icon represents a emitters visualization object.  \
This object is used to get emitters for vector field."));

    icon = NewIcon(0, 0, ICONQUESTION, "Emitters");
    SetVar(icon, HELPSTRING,
	NewString("Click on this icon to see a panel of controls for the \
Emitters object."));
    SetVar(visHoseEmitters, CONTROLICON, icon);
    SetVar(icon, PANELHELP, NewString("The emitterss visualization object is intended \
to control the emitters of a vector field."));

    AddSnapVar(visHoseEmitters, XBUCKETDIM);
    AddSnapVar(visHoseEmitters, YBUCKETDIM);
    AddSnapVar(visHoseEmitters, ZBUCKETDIM);
    AddSnapVar(visHoseEmitters, XUPPERBOUND);
    AddSnapVar(visHoseEmitters, YUPPERBOUND);
    AddSnapVar(visHoseEmitters, ZUPPERBOUND);
    AddSnapVar(visHoseEmitters, XLOWERBOUND);
    AddSnapVar(visHoseEmitters, YLOWERBOUND);
    AddSnapVar(visHoseEmitters, ZLOWERBOUND);
    AddSnapVar(visHoseEmitters, EMITTERCONTROL);
    AddSnapVar(visHoseEmitters, EMITTERBUCKETS);
    AddSnapVar(visHoseEmitters, EMITTERDATASET);

#ifdef SIGGRAPH
    SetVar(visHoseEmitters, XBUCKETDIM, NewInt(7));
    SetVar(visHoseEmitters, YBUCKETDIM, NewInt(10));
    SetVar(visHoseEmitters, ZBUCKETDIM, NewInt(3));
    SetVar(visHoseEmitters, XLOWERBOUND, NewReal(-42000.0));
    SetVar(visHoseEmitters, XUPPERBOUND, NewReal(-15000.0));
    SetVar(visHoseEmitters, YLOWERBOUND, NewReal(-34000.0));
    SetVar(visHoseEmitters, YUPPERBOUND, NewReal(5000.0));
    SetVar(visHoseEmitters, ZLOWERBOUND, NewReal(0.0));
    SetVar(visHoseEmitters, ZUPPERBOUND, NewReal(6000.0));
    SetVar(visHoseEmitters, EMITTERCONTROL, NewInt(USEBUCKET));
#else
    SetVar(visHoseEmitters, XBUCKETDIM, NewInt(DEFAULTBUCKETDIM));
    SetVar(visHoseEmitters, YBUCKETDIM, NewInt(DEFAULTBUCKETDIM));
    SetVar(visHoseEmitters, ZBUCKETDIM, NewInt(DEFAULTBUCKETDIM));
    SetVar(visHoseEmitters, EMITTERCONTROL, NewInt(USEINTERACTIVE));
#endif

    DeclareDependency(visHoseEmitters, EMITTERBUCKETS, XBUCKETDIM);
    DeclareDependency(visHoseEmitters, EMITTERBUCKETS, YBUCKETDIM);
    DeclareDependency(visHoseEmitters, EMITTERBUCKETS, ZBUCKETDIM);
    DeclareDependency(visHoseEmitters, EMITTERBUCKETS, XUPPERBOUND);
    DeclareDependency(visHoseEmitters, EMITTERBUCKETS, YUPPERBOUND);
    DeclareDependency(visHoseEmitters, EMITTERBUCKETS, ZUPPERBOUND);
    DeclareDependency(visHoseEmitters, EMITTERBUCKETS, XLOWERBOUND);
    DeclareDependency(visHoseEmitters, EMITTERBUCKETS, YLOWERBOUND);
    DeclareDependency(visHoseEmitters, EMITTERBUCKETS, ZLOWERBOUND);

    SetMethod(visHoseEmitters, EMITTERBUCKETS, MakeEmitterBuckets);
    SetMethod(visHoseEmitters, ADDCONTROLS, AddEmittersControls);

    /* for adding and deleting emitters */
    SetMethod(visHoseEmitters, ADDEMITTERS, AddEmitters);
    SetMethod(visHoseEmitters, DELETEEMITTERS, DeleteEmitters);

    return;
}

/* Initializes visLineEmitters visualization class */
#ifdef HAVE_PROTOTYPES
void InitLineEmitters(void)
#else
void InitLineEmitters()
#endif
{
    ObjPtr icon;

    visLineEmitters = NewObject(visSurface, 0);
    AddToReferenceList(visLineEmitters);

    SetVar(visLineEmitters, NAME, NewString("visLineEmitters"));

    SetVar(visLineEmitters, DEFAULTICON, icon = NewObject(visIcon, 0));
    SetVar(icon, WHICHICON, NewInt(ICONQUESTION));
    SetVar(icon, NAME, NewString("Emitters"));
    SetVar(icon, HELPSTRING,
	NewString("This icon represents a emitterss visualization object.  \
This object is used to get emitters for vector field."));

    icon = NewIcon(0, 0, ICONQUESTION, "Emitters");
    SetVar(icon, HELPSTRING,
	NewString("Click on this icon to see a panel of controls for the \
Emitters object."));
    SetVar(visLineEmitters, CONTROLICON, icon);
    SetVar(icon, PANELHELP, NewString("The emitterss visualization object is intended \
to control the emitters of a vector field."));

    AddSnapVar(visLineEmitters, XBUCKETDIM);
    AddSnapVar(visLineEmitters, YBUCKETDIM);
    AddSnapVar(visLineEmitters, ZBUCKETDIM);
    AddSnapVar(visLineEmitters, XUPPERBOUND);
    AddSnapVar(visLineEmitters, YUPPERBOUND);
    AddSnapVar(visLineEmitters, ZUPPERBOUND);
    AddSnapVar(visLineEmitters, XLOWERBOUND);
    AddSnapVar(visLineEmitters, YLOWERBOUND);
    AddSnapVar(visLineEmitters, ZLOWERBOUND);
    AddSnapVar(visLineEmitters, EMITTERCONTROL);
    AddSnapVar(visLineEmitters, EMITTERDATASET);
    AddSnapVar(visLineEmitters, EMITTERBUCKETS);

#ifdef SIGGRAPH
    SetVar(visLineEmitters, XBUCKETDIM, NewInt(7));
    SetVar(visLineEmitters, YBUCKETDIM, NewInt(10));
    SetVar(visLineEmitters, ZBUCKETDIM, NewInt(3));
    SetVar(visLineEmitters, XLOWERBOUND, NewReal(-42000.0));
    SetVar(visLineEmitters, XUPPERBOUND, NewReal(-15000.0));
    SetVar(visLineEmitters, YLOWERBOUND, NewReal(-34000.0));
    SetVar(visLineEmitters, YUPPERBOUND, NewReal(5000.0));
    SetVar(visLineEmitters, ZLOWERBOUND, NewReal(0.0));
    SetVar(visLineEmitters, ZUPPERBOUND, NewReal(6000.0));
    SetVar(visLineEmitters, EMITTERCONTROL, NewInt(USEINTERACTIVE));
#else
    SetVar(visLineEmitters, XBUCKETDIM, NewInt(DEFAULTBUCKETDIM));
    SetVar(visLineEmitters, YBUCKETDIM, NewInt(DEFAULTBUCKETDIM));
    SetVar(visLineEmitters, ZBUCKETDIM, NewInt(DEFAULTBUCKETDIM));
    SetVar(visLineEmitters, EMITTERCONTROL, NewInt(USEINTERACTIVE));
#endif

    DeclareDependency(visLineEmitters, EMITTERBUCKETS, XBUCKETDIM);
    DeclareDependency(visLineEmitters, EMITTERBUCKETS, YBUCKETDIM);
    DeclareDependency(visLineEmitters, EMITTERBUCKETS, ZBUCKETDIM);
    DeclareDependency(visLineEmitters, EMITTERBUCKETS, XUPPERBOUND);
    DeclareDependency(visLineEmitters, EMITTERBUCKETS, YUPPERBOUND);
    DeclareDependency(visLineEmitters, EMITTERBUCKETS, ZUPPERBOUND);
    DeclareDependency(visLineEmitters, EMITTERBUCKETS, XLOWERBOUND);
    DeclareDependency(visLineEmitters, EMITTERBUCKETS, YLOWERBOUND);
    DeclareDependency(visLineEmitters, EMITTERBUCKETS, ZLOWERBOUND);

    SetMethod(visLineEmitters, EMITTERBUCKETS, MakeEmitterBuckets);
    SetMethod(visLineEmitters, ADDCONTROLS, AddEmittersControls);

    /* for adding and deleting emitters */
    SetMethod(visLineEmitters, ADDEMITTERS, AddEmitters);
    SetMethod(visLineEmitters, DELETEEMITTERS, DeleteEmitters);

    return;
}

#ifdef HAVE_PROTOTYPES
void KillHoseEmitters(void)
#else
void KillHoseEmitters()
#endif
{
    DeleteThing(visHoseEmitters);
    return;
}

#ifdef HAVE_PROTOTYPES
void KillLineEmitters(void)
#else
void KillLineEmitters()
#endif
{
    DeleteThing(visLineEmitters);
    return;
}
/***************************** visEmitters Class End *************************************/

/****************************** integration ops  *************************************/

#ifdef HAVE_PROTOTYPES
static int eulerCalcTrace(real *trace, int maxNTracePoints, int formField, int dataField, int nDataComponents, int nSpatComponents, real start[], real bounds[], real ds)
#else
static int eulerCalcTrace(trace, maxNTracePoints, formField, dataField, nDataComponents, nSpatComponents, start, bounds, ds)
real *trace;
int maxNTracePoints;
int formField;
int dataField;
int nDataComponents;
int nSpatComponents;
real start[];
real bounds[];
real ds;
#endif
{
  real *dataSample;
  real *curSpatSample, *nextSpatSample;
  int pointsTraced;
  int whichComponent;
  Bool outOfBounds;  /* if the trace goes out of bound */
  Bool isDataMissing; /* if any missing data */
  float dt, stepLength, advanced, sqrDistance, vecMagnitude;
  float epsilonDistance;

#ifdef SIGGRAPH
  float sqrtcoef, estimatedDistance;

  if (nDataComponents == 2)
    {
      sqrtcoef = SQ22;
    }
  else
    /* nDataComponents must be 3 */
    {
      sqrtcoef = SQ33;
    }
#endif
  epsilonDistance = ABS(ds) * FLT_EPSILON;


  if (!(dataSample = (real *) Alloc(sizeof(real) * nDataComponents)))
    {
      OMErr();
      return 0;
    }
  if (!(curSpatSample = (real *) Alloc(sizeof(real) * nSpatComponents)))
    {
      OMErr();
      Free(dataSample);
      return 0;
    }
  if (!(nextSpatSample = (real *) Alloc(sizeof(real) * nSpatComponents)))
    {
      OMErr();
      Free(curSpatSample);
      Free(dataSample);
      return 0;
    }

  /* copy the start point to the current point */
  for (whichComponent = 0; whichComponent < nSpatComponents; whichComponent++)
     {
       curSpatSample[whichComponent] = start[whichComponent];
     }

  /* if ds < 0 then trace backward, where dt should be negative */
  stepLength = ABS(ds);

  /* needs to know what happens if nDataComponents != nSpatComponents */
  for (pointsTraced = 0; pointsTraced < maxNTracePoints; )
     {
       /* put the current sample's poisition to trace */
       for (whichComponent = 0; whichComponent < nSpatComponents; whichComponent++)
	  {
            trace[pointsTraced * nSpatComponents + whichComponent] = curSpatSample[whichComponent];
	  }
       pointsTraced++;

       /* get the vector data of the current point */
       SampleSpatComponent(dataField, formField, nDataComponents, dataSample, nSpatComponents, curSpatSample, true);

       /* if any missing data, then stop */
       isDataMissing = false;
       for (whichComponent = 0; whichComponent < nDataComponents && false == isDataMissing; whichComponent++)
	  {
	    if (dataSample[whichComponent] == missingData)
	      {
		isDataMissing = true;
	      }
	  }
       if (true == isDataMissing) 
	 {
#ifdef DEBUG
	   fprintf(stderr, "Stop integration, running into missing data.\n");
#endif
	   break;
	 }

#ifdef SIGGRAPH
       vecMagnitude = 0;
       for (whichComponent = 0; whichComponent < nDataComponents; whichComponent++)
	  {
	    vecMagnitude += ABS(dataSample[whichComponent]);
	  }
       if (0.0 == vecMagnitude)
	 {
	   break;
	 }
       vecMagnitude = vecMagnitude * sqrtcoef;
       /* estimate the integration time */
       dt = ds / vecMagnitude;

       estimatedDistance = 0.0;
       for (whichComponent = 0; whichComponent < nSpatComponents; whichComponent++)
	  {
	    if (whichComponent >= nDataComponents)
	      {
		break;
	      }
            nextSpatSample[whichComponent] = curSpatSample[whichComponent] + 
					     dt * dataSample[whichComponent];
            advanced = nextSpatSample[whichComponent] - curSpatSample[whichComponent];
	    estimatedDistance += ABS(advanced);
	  }

       estimatedDistance *= sqrtcoef;
       if (epsilonDistance > estimatedDistance)
	 {
	   /* consider it as stagnation */
	   break;
	 }
#else
       vecMagnitude = 0;
       for (whichComponent = 0; whichComponent < nDataComponents; whichComponent++)
	  {
	    vecMagnitude += dataSample[whichComponent] * dataSample[whichComponent];
	  }
       if (0.0 == vecMagnitude)
	 {
	   /* stagnation */
#ifdef DEBUG 
	   fprintf(stderr, "Stop integration, stagnation (0 velocity).\n"); 
#endif
	   break;
	 }
       vecMagnitude = (real) sqrt((double) vecMagnitude);

       /* estimate the integration time */
       dt = ds / vecMagnitude;

       sqrDistance = 0.0;
       for (whichComponent = 0; whichComponent < nSpatComponents; whichComponent++)
	  {
	    /* what happenes if nDataComponents > nSpatComponents */
	    /* not now, since I define vismapping to the constraint of them equal */
	    if (whichComponent >= nDataComponents)
	      {
		break;
	      }
	    /* if dt * nextDataSample[whichComponent] is too small (<< curSpatSample[]), then
	       nextDataSample is going to be the same as curSpatSample */
            nextSpatSample[whichComponent] = curSpatSample[whichComponent] + 
					     dt * dataSample[whichComponent];
            advanced = nextSpatSample[whichComponent] - curSpatSample[whichComponent];
	    sqrDistance += advanced * advanced;
	  }

       /* because later on, we need to use square of the distance in calculating
	  some radial vector, we need an restriction here, and note that the 
	  distance is calculatd using subtraction not using dt * vel directory,
	  this is because of the way the axial is obtained later on */
       if (epsilonDistance > sqrDistance)
	 {
#ifdef DEBUG
	   fprintf(stderr, "Stagnation, less than epsilon advaced distance.\n"); 
#endif
	   /* stagnation */
	   break;
	 }
#endif

       outOfBounds = false;
       for (whichComponent = 0; whichComponent < nSpatComponents && false == outOfBounds; whichComponent++)
	  {
	    if (nextSpatSample[whichComponent] > bounds[2 * whichComponent + 1] ||
		nextSpatSample[whichComponent] < bounds[2 * whichComponent])
              {
		outOfBounds = true;
	      }
	  }
       if (true == outOfBounds) 
	 {
#ifdef DEBUG
	   fprintf(stderr, "Stop integration, going out of bounds.\n");
#endif
	   break;
         }

       for (whichComponent = 0; whichComponent < nSpatComponents; whichComponent++)
	  {
	    curSpatSample[whichComponent] = nextSpatSample[whichComponent];
	  }
     }

  Free(dataSample);
  Free(curSpatSample);
  Free(nextSpatSample);
  return pointsTraced;
}


#ifdef HAVE_PROTOTYPES
static int rkuttaCalcTrace(real *trace, int maxNTracePoints, int formField, int dataField, int nDataComponents, int nSpatComponents, real start[], real bounds[], real ds)
#else
static int rkuttaCalcTrace(trace, maxNTracePoints, formField, dataField, nDataComponents, nSpatComponents, start, bounds, ds)
real *trace;
int maxNTracePoints;
int formField;
int dataField;
int nDataComponents;
int nSpatComponents;
real start[];
real bounds[];
real ds;
#endif
{
  real *curDataSample, *nextDataSample;
  real *curSpatSample, *nextSpatSample;
  int pointsTraced;
  int whichComponent;
  real vecMagnitude;
  Bool outOfBounds;  /* if the trace goes out of bound */
  Bool isDataMissing; /* if any missing data */
  real dt;  
  real stepLength;
  real advanced, sqrDistance;
  float epsilonDistance;

#ifdef SIGGRAPH
  float sqrtcoef, estimatedDistance;

  if (nDataComponents == 2)
    {
      sqrtcoef = SQ22;
    }
  else
    /* nDataComponents must be 3 */
    {
      sqrtcoef = SQ33;
    }
#endif
  epsilonDistance = ABS(ds) * FLT_EPSILON;


  if (!(curDataSample = (real *) Alloc(sizeof(real) * nDataComponents)))
    {
      OMErr();
      return 0;
    }
  if (!(nextDataSample = (real *) Alloc(sizeof(real) * nDataComponents)))
    {
      OMErr();
      Free(curDataSample);
      return 0;
    }
  if (!(curSpatSample = (real *) Alloc(sizeof(real) * nSpatComponents)))
    {
      OMErr();
      Free(curDataSample);
      Free(nextDataSample);
      return 0;
    }
  if (!(nextSpatSample = (real *) Alloc(sizeof(real) * nSpatComponents)))
    {
      OMErr();
      Free(curDataSample);
      Free(nextDataSample);
      Free(curSpatSample);
      return 0;
    }

  /* copy the start point to the current point */
  for (whichComponent = 0; whichComponent < nSpatComponents; whichComponent++)
     {
       curSpatSample[whichComponent] = start[whichComponent];
     }

  /* if ds < 0 then trace backward, where dt should be negative */
  stepLength = ABS(ds);

  /* needs to know what happens if nDataComponents != nSpatComponents */
  for (pointsTraced = 0; pointsTraced < maxNTracePoints; )
     {
       /* put the current sample's poisition to trace */
       for (whichComponent = 0; whichComponent < nSpatComponents; whichComponent++)
	  {
            trace[pointsTraced * nSpatComponents + whichComponent] = curSpatSample[whichComponent];
	  }
       pointsTraced++;

       /* get the vector data of the current point */
       SampleSpatComponent(dataField, formField, nDataComponents, curDataSample, nSpatComponents, curSpatSample, true);

       /* if any missing data, then stop */
       isDataMissing = false;
       for (whichComponent = 0; whichComponent < nDataComponents && false == isDataMissing; whichComponent++)
	  {
	    if (curDataSample[whichComponent] == missingData)
	      {
		isDataMissing = true;
	      }
	  }
       if (true == isDataMissing)
	 {
#ifdef DEBUG
	   fprintf(stderr, "Stop integration, running into missing data.\n");
#endif
	   break;
	 }

       vecMagnitude = 0;
       for (whichComponent = 0; whichComponent < nDataComponents; whichComponent++)
	  {
	    vecMagnitude += curDataSample[whichComponent] * curDataSample[whichComponent];
	  }
       if (0.0 == vecMagnitude)
	 {
#ifdef DEBUG 
	   fprintf(stderr, "Stagnated because of 0 velocity.\n"); 
#endif
	   /* stagnation */
	   break;
	 }
       vecMagnitude = (real) sqrt((double) vecMagnitude);

       /* estimate the integration time */
       dt = ds / vecMagnitude;

       /* first order approximation */
       for (whichComponent = 0; whichComponent < nSpatComponents; whichComponent++)
	  {
	    /* what happenes if nDataComponents > nSpatComponents */
	    if (whichComponent >= nDataComponents)
	      {
		break;
	      }
            nextSpatSample[whichComponent] = curSpatSample[whichComponent] + dt * curDataSample[whichComponent];
	  }

       outOfBounds = false;
       for (whichComponent = 0; whichComponent < nSpatComponents && false == outOfBounds; whichComponent++)
	  {
	    if (nextSpatSample[whichComponent] > bounds[2 * whichComponent + 1] ||
		nextSpatSample[whichComponent] < bounds[2 * whichComponent])
              {
		outOfBounds = true;
	      }
	  }
       if (true == outOfBounds)
	 {
#ifdef DEBUG
	   fprintf(stderr, "Stop integration, going out of bounds.\n");
#endif
	   break;
         }

/* needed for 2nd order rung-kutta begin */
       /* get vector data samples on this approximated point */
       SampleSpatComponent(dataField, formField, nDataComponents, nextDataSample, nSpatComponents, nextSpatSample, true);

       for (whichComponent = 0; whichComponent < nDataComponents && false == isDataMissing; whichComponent++)
	  {
	    if (nextDataSample[whichComponent] == missingData)
	      {
		isDataMissing = true;
	      }
	  }
       if (true == isDataMissing)
	 {
#ifdef DEBUG
	   fprintf(stderr, "Stop integration, running into missing data.\n");
#endif
	   break;
	 }

       /* Runge-Kutta second order adjustment on the vector value */
       for (whichComponent = 0; whichComponent < nDataComponents && false == isDataMissing; whichComponent++)
	  {
	    nextDataSample[whichComponent] = 0.5 * (curDataSample[whichComponent] + 
						    nextDataSample[whichComponent]);
	  }

       sqrDistance = 0.0;
       /* use this vector value to calculate next spatial sample */
       for (whichComponent = 0; whichComponent < nSpatComponents; whichComponent++)
	  {
	    /* what happens if nDataComponents > nSpatComponents */
	    if (whichComponent >= nDataComponents)
	      {
		break;
	      }
	    /* if dt * nextDataSample[whichComponent] is too small (< FLT_EPSILON), then
	       nextDataSample is going to be the same as nextSpatSample */
            nextSpatSample[whichComponent] = curSpatSample[whichComponent] + dt * nextDataSample[whichComponent];
            advanced = nextSpatSample[whichComponent] - curSpatSample[whichComponent];
	    sqrDistance += advanced * advanced;
	  }
       if (epsilonDistance > sqrDistance)
	 {
#ifdef DEBUG
	   fprintf(stderr, "Stagnation, less than epsilon advaced distance.\n"); 
#endif
	   /* stagnation */
	   break;
	 }

       for (whichComponent = 0; whichComponent < nSpatComponents && false == outOfBounds; whichComponent++)
	  {
	    if (nextSpatSample[whichComponent] > bounds[2 * whichComponent + 1] ||
		nextSpatSample[whichComponent] < bounds[2 * whichComponent])
              {
		outOfBounds = true;
	      }
	  }
       if (true == outOfBounds)
	 {
#ifdef DEBUG
	   fprintf(stderr, "Stop integration, going out of bounds.\n");
#endif
	   break;
         }

       for (whichComponent = 0; whichComponent < nSpatComponents; whichComponent++)
	  {
	    curSpatSample[whichComponent] = nextSpatSample[whichComponent];
	  }
     }

  Free(curDataSample);
  Free(nextDataSample);
  Free(curSpatSample);
  Free(nextSpatSample);
  return pointsTraced;
}
/************************** end integration ops  *************************************/

/*
  ScianVisStreams.c
  Tzong-Yow Hwu
  Streams visualization object in SciAn.  
  Should mapped to any datasets with vector field.
  1/08/1993

Outline Particle Trace

Design Specification
Given: y'(x) = f(y(x)), and y(x0) = y0
Want:  particle trace

 Particle Trace Initial Points Determination:
   1. Divide the formfield into field of buckets uniformly 
      ditributed over the spatial bounds. Distribute the grid 
      points of formfield into buckets according to saptial 
      coordinates.  Set a certain number of buckets on each 
      spatial dimension.  The number of buckets along each 
      spatial axe may be either a fixed number(say 5) or 
      varied according to the magnitude of axial bounds (maybe 
      by ratio.)
   2. For a point from within every bucket, do a particle trace
      (forward and backward.)
      Three different ways to do this: 
      a.  Choose the center point of every bucket (the mean of 
	  the spatial coordinates of all points within a bucket) 
	  to be the initial point of a trace and mark every 
	  bucket the trace passes through.  Choose another center 
	  point from within a bucket among those unmarked to do 
	  another trace.  Until there is no unmarked buckets to 
	  initiate another trace. 
      b.  Just simply initiate a trace from the center of every 
	  bucket.
      c.  Same as a. except instead of using the center point to 
          initiate a trace, use the point with a maximum divergence 
	  within the bucket.  Starting from the bucket with the 
	  maximum average divergence until all buckets are marked.
 Particle Trace Integration Proceeding:  
   1. Approximate the advanced distance of each step of integration 
      by ds = c * MAX(bounds), where c is a parameter (with a default 
      value and can be reset by the user.)  
   2. Then the time interval dt = ds / |v|, where v is the velocity 
      vector at the initiate point, x0, of each step of integration.
   3. Advance the particle to next point x1 by 1st order Euler or 2nd order 
      Runge-Kutta method.
 Particle Trace Termination:
   An integration stops if 
    1.  the magnitude of the velocity at a point is close to 0 (stagnated), 
    2.  the number of points stepped through exceeds a certain upper
	limit ( factor/c ), or
    3.  the trace goes out of domain of the formfield.

Code Backbone:

0.  Make and fill buckets. 

1.  Compute ds.

2.  Start by getting the indices of the initial point y0
    from within each bucket.

3.  Find the velocity at the point by SampleSpatComponent. 

4.  Calculate the magnitude of the velocity. (|v|)

5.  If the magnitude of the velocity is 0, then choose another 
    initial point, if all points are stagnated then go to
    another bucket, otherwise continue.

6.  Estimate the advanced time for this step. (dt = ds/|v|)

7.  Advance the integration using 2nd order Runge-Kutta method 
    to the next point.

8.  Get the position of next point and mark the bucket within
    which the point in.

9.  Calculate the distance advanced. 
     (the length from last point to next point)

10. If the distance is greater than 1.2 * ds, 
     then adjust the advanced time by a factor(say, ds/dl) and do 
     the integration again(go to step 7). Otherwise continue.

11. If the new point is outside of the bounds of the formfield, stop.

12. Get the velocity v of the new point by using its position and 
    SampleSpatComponent.

13. If any component returned is missing then stop.

14. If |v| is 0, then stagnated here and stop the integration.

15. See if the total number of points traced execeeds a certain
    upper limit, then stop.  

16. If stop, then goes to next unmarked bucket for another trace, 
    otherwise, continue to step 7.

17. Continue until every bucket is traversed.
*/
